/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2018 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Moritz Gelb, Felix Metzner                               *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#include <analysis/modules/FSRCorrection/ConservativeFSRCorrectionModule.h>

#include <framework/gearbox/Const.h>
#include <framework/logging/Logger.h>
#include <framework/datastore/RelationArray.h>

#include <analysis/dataobjects/Particle.h>
#include <mdst/dataobjects/MCParticle.h>

#include <analysis/DecayDescriptor/ParticleListName.h>
#include <analysis/utility/PCmsLabTransform.h>

#include <cmath>
#include <algorithm>
#include <TMatrixFSym.h>

namespace Belle2 {

  REG_MODULE(ConservativeFSRCorrection)

  ConservativeFSRCorrectionModule::ConservativeFSRCorrectionModule() :
    Module()

  {
    setDescription("Takes the particles from the given lepton list copies them to the output list "
                   "and adds the 4-vector of the closest photon (considered as radiative) to the "
                   "lepton, if the given criteria for maximum angle and energy are fulfilled.");
    setPropertyFlags(c_ParallelProcessingCertified);

    addParam("leptonListName", m_leptonListName,
             "The lepton list containing the leptons to correct, should already exists.");
    addParam("gammaListName", m_gammaListName,
             "The gammas list containing possibly radiative gammas, should already exist.");
    addParam("angleThreshold", m_angleThreshold,
             "The maximum angle (in degrees) between the lepton and the (radiative) gamma to be accepted.",
             m_angleThreshold);
    addParam("thetaThreshold", m_thetaThreshold,
             "The maximum theta (in degrees) between the lepton and the (radiative) gamma to be accepted."
             "Used only for arc method.",
             m_thetaThreshold);
    addParam("energyThreshold", m_energyThreshold,
             "The maximum energy of the (radiative) gamma to be accepted.",
             m_energyThreshold);
    addParam("momentumFactor", m_momentumFactor,
             "The maximum momentum of the (radiative) gamma with respect to the lepton momentum: "
             "p_gamma < momentumFactor * p_lepton.",
             m_momentumFactor);
    addParam("method", m_method_param,
             "The method which is used to look for potential photons. Can be either centeredCone, "
             "offsetCone or arc.",
             m_method_param);
  }


  void ConservativeFSRCorrectionModule::initialize()
  {
    if (m_method_param == "centeredCone") {
      m_method = c_CenteredCone;
    } else if (m_method_param == "offsetCone") {
      m_method = c_OffsetCone;
    } else if (m_method_param == "arc") {
      m_method = c_Arc;
    } else {
      B2FATAL("The provided method " << m_method_param << " is not valid. "
              "Must be one of 'centeredCone', 'offsetCone' or 'arc'!");
    }

    StoreObjPtr<ParticleList>().isRequired(m_leptonListName);
    StoreObjPtr<ParticleList>().isRequired(m_gammaListName);

    m_maxAngle = cos(m_angleThreshold * M_PI / 180.0);
    m_maxTheta = cos(m_thetaThreshold * M_PI / 180.0);
  }


  void ConservativeFSRCorrectionModule::event()
  {
    StoreArray<Particle> particles;

    const StoreObjPtr<ParticleList> leptonList(m_leptonListName);
    const StoreObjPtr<ParticleList> gammaList(m_gammaListName);

    std::vector<int> usedGammas;

    // loop over leptons, and find suitable photons as candidates for FSR
    for (unsigned i = 0; i < leptonList->getListSize(); i++) {
      Particle* lepton = leptonList->getParticle(i);

      TLorentzVector lepton4Vector = lepton->get4Vector();
      TLorentzVector correctedLepton4Vec = lepton->get4Vector();
      TVector3 pLepton = lepton->getMomentum();

      Particle* fsrGamma = nullptr;
      bool fsrGammaFound = false;
      double cosMax = -1.0;

      // look for a possible FSR gamma
      for (unsigned j = 0; j < gammaList->getListSize(); j++) {
        Particle* gamma = gammaList->getParticle(j);

        // check if gamma energy is below threshold
        if (gamma->getEnergy() > m_energyThreshold) {
          continue;
        }

        // check if gamma momentum is valid:
        if (gamma->getMomentumMagnitude() > m_momentumFactor * lepton->getMomentumMagnitude()) {
          continue;
        }

        bool candidateFound = false;
        if (m_method == c_CenteredCone) {
          candidateFound = checkGammaWithCenteredCone(lepton, gamma, cosMax);
        } else if (m_method == c_OffsetCone) {
          candidateFound = checkGammaWithOffsetCone(lepton, gamma, cosMax);
        } else if (m_method == c_Arc) {
          candidateFound = checkGammaWithArc(lepton, gamma, cosMax);
        }

        if (candidateFound) {
          // check if gamma was already used
          if (std::find(usedGammas.begin(), usedGammas.end(), gamma->getMdstArrayIndex()) != usedGammas.end()) {
            continue;
          }
          fsrGammaFound = true;
          fsrGamma = gamma;
        }
      }

      if (fsrGammaFound) {
        correctedLepton4Vec = lepton4Vector + fsrGamma->get4Vector();
        usedGammas.push_back(fsrGamma->getMdstArrayIndex());

        fsrGamma->addExtraInfo("fsrCorrected", float(fsrGammaFound));

        lepton->addExtraInfo("fsrCorrected", float(fsrGammaFound));
        lepton->addExtraInfo("fsrCorrectedE", float(correctedLepton4Vec.E()));
        lepton->addExtraInfo("fsrCorrectedPx", float(correctedLepton4Vec.Px()));
        lepton->addExtraInfo("fsrCorrectedPy", float(correctedLepton4Vec.Py()));
        lepton->addExtraInfo("fsrCorrectedPz", float(correctedLepton4Vec.Pz()));
        lepton->addExtraInfo("fsrCorrectedP", float(correctedLepton4Vec.P()));
        lepton->addExtraInfo("fsrGammaE", float(fsrGamma->get4Vector().E()));
        lepton->addExtraInfo("fsrGammaPx", float(fsrGamma->get4Vector().Px()));
        lepton->addExtraInfo("fsrGammaPy", float(fsrGamma->get4Vector().Py()));
        lepton->addExtraInfo("fsrGammaPz", float(fsrGamma->get4Vector().Pz()));
        lepton->addExtraInfo("fsrGammaP", float(fsrGamma->get4Vector().P()));

        if (fsrGamma->getRelated<MCParticle>()) {
          lepton->addExtraInfo("fsrMCGammaE", fsrGamma->getRelated<MCParticle>()->get4Vector().E());
          lepton->addExtraInfo("fsrMCGammaPx", fsrGamma->getRelated<MCParticle>()->get4Vector().Px());
          lepton->addExtraInfo("fsrMCGammaPy", fsrGamma->getRelated<MCParticle>()->get4Vector().Py());
          lepton->addExtraInfo("fsrMCGammaPz", fsrGamma->getRelated<MCParticle>()->get4Vector().Pz());
          lepton->addExtraInfo("fsrMCGammaP", fsrGamma->getRelated<MCParticle>()->get4Vector().P());
        } else {
          lepton->addExtraInfo("fsrMCGammaE", float(-999));
          lepton->addExtraInfo("fsrMCGammaPx", float(-999));
          lepton->addExtraInfo("fsrMCGammaPy", float(-999));
          lepton->addExtraInfo("fsrMCGammaPz", float(-999));
          lepton->addExtraInfo("fsrMCGammaP", float(-999));
        }

        // // update error matrix
        // const TMatrixFSym& lepErrorMatrix = lepton->getMomentumVertexErrorMatrix();
        // const TMatrixFSym& fsrErrorMatrix = fsrGamma->getMomentumVertexErrorMatrix();
        // // todo add check

        // for (int irow = 0; irow < c_DimMatrix; irow++) {
        //   for (int icol = irow; icol < c_DimMatrix; icol++) {
        //     if (irow > 3 || icol > 3) {
        //        lepton->addExtraInfo("fsrCorrectedMomentumVertexMatrix" + std::to_string(irow) + std::to_string(icol),
        //                            float(lepErrorMatrix(irow, icol)));
        //     } else {
        //       lepton->addExtraInfo("fsrCorrectedMomentumVertexMatrix" + std::to_string(irow) + std::to_string(icol),
        //                           float(lepErrorMatrix(irow, icol) + fsrErrorMatrix(irow, icol)));
        //     }
        //   }
        // }
      }
    }
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithCenteredCone(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();
    TVector3 coneAxis = pLepton;
    double cosAngle = cos(coneAxis.Angle(pGamma));

    // check if angle is current best
    if (cosAngle > currentCosMax) {
      currentCosMax = cosAngle;
      // check if this angle fulfills the maxAngle bound
      if (cosAngle > m_maxAngle) {
        return true;
      }
    }
    return false;
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithOffsetCone(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();
    TVector3 coneAxis = pLepton;
    coneAxis.SetPhi(pLepton.Phi() + lepton->getCharge() * m_angleThreshold * M_PI / 180.0);
    double cosAngle = cos(coneAxis.Angle(pGamma));

    // check if angle is current best
    if (cosAngle > currentCosMax) {
      currentCosMax = cosAngle;
      // check if this angle fulfills the maxAngle bound
      if (cosAngle > m_maxAngle) {
        return true;
      }
    }
    return false;
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithArc(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();

    TVector3 arcAxis = pLepton;
    arcAxis.SetPhi(pLepton.Phi() + lepton->getCharge() * 1./3. * m_angleThreshold * M_PI / 180.0);

    double cosAngle = cos(arcAxis.Angle(pGamma));

    TVector3 arcAxisInPhiPlane = arcAxis;
    arcAxisInPhiPlane.SetTheta(0.);
    TVector3 arcAxisInThetaPlane = arcAxis;
    arcAxisInThetaPlane.SetPhi(0.);

    TVector3 pGammaInPhiPlane = pGamma;
    pGammaInPhiPlane.SetTheta(0.);
    TVector3 pGammaInThetaPlane = pGamma;
    pGammaInThetaPlane.SetPhi(0.);

    double cosPhi = cos(arcAxisInPhiPlane.Angle(pGammaInPhiPlane));
    double cosTheta = cos(arcAxisInThetaPlane.Angle(pGammaInThetaPlane));

    // check if this angle fulfills the maxAngle and maxTheta bound
    if (cosPhi > m_maxAngle and cosTheta > m_maxTheta) {
      // check if angle is current best
      if (cosAngle > currentCosMax) {
        currentCosMax = cosAngle;
          return true;
      }
    }
    return false;
  }
} // end Belle2 namespace

