/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2018 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Moritz Gelb, Felix Metzner                               *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#pragma once

#include <framework/core/Module.h>

#include <analysis/VariableManager/Utility.h>
#include <analysis/dataobjects/ParticleList.h>
#include <analysis/dataobjects/Particle.h>

#include <framework/datastore/StoreArray.h>
#include <framework/datastore/StoreObjPtr.h>

#include <analysis/DecayDescriptor/DecayDescriptor.h>

#include <string>


namespace Belle2 {
  class Particle;

  /**
   * Final state radiation correction module (conservative version)
   * This module finds the (closest) radiative gamma related to a lepton which fulfills the specified criteria.
   * The potential radiative gamma is marked as such and its 4-momentum is added as additional information to the lepton.
   * The module is intended to be used for electrons.
   */
  class ConservativeFSRCorrectionModule : public Module {

  public:

    /// Constructor
    ConservativeFSRCorrectionModule();

    /// Module initialization.
    virtual void initialize() override;

    /// Event processor.
    virtual void event() override;

    enum {c_DimMatrix = 7};

  private:
    /**
     * Check if fsr Gamma candidate is valid by checking if it is in a cone with an opening angle of
     * 2*m_angleThreshold and for which the symmetry axis is the momentum vector of the lepton.
     * @param lepton Particle to which the fsr gamma shall be added.
     * @param gamma Particle to consider as fsr gamma candidate.
     * @param currentCosMax The current best value for the angle between the lepton and the gamma.
     * @return Boolean indicating if the gamma is the current best match.
     */
    bool checkGammaWithCenteredCone(const Particle* lepton, const Particle* gamma, double& currentCosMax);

    /**
     * Check if fsr Gamma candidate is valid by checking if it is in a cone with an opening angle of
     * 2*m_angleThreshold and on whose surface the lepton momentum vector lies and for which the symmetry
     * axis points away from the magnetic force acting upon the charged lepton.
     * @param lepton Particle to which the fsr gamma shall be added.
     * @param gamma Particle to consider as fsr gamma candidate.
     * @param currentCosMax The current best value for the angle between the cone axis and the gamma.
     * @return Boolean indicating if the gamma is the current best match.
     */
    bool checkGammaWithOffsetCone(const Particle* lepton, const Particle* gamma, double& currentCosMax);

    /**
     * Check if fsr Gamma candidate is valid by checking the distance in phi (=m_angleThreshold) and
     * theta (=m_thetaThreshold) between the photon and the lepton momentum vector.
     * @param lepton Particle to which the fsr gamma shall be added.
     * @param gamma Particle to consider as fsr gamma candidate.
     * @param currentCosMax The current best value for the angle between the lepton and the gamma.
     * @return Boolean indicating if the gamma is the current best match.
     */
    bool checkGammaWithArc(const Particle* lepton, const Particle* gamma, double& currentCosMax);


    /// Module parameters
    std::string m_leptonListName;  /**< input lepton ParticleList */
    std::string m_gammaListName;  /**< input gamma ParticleList */

    double m_angleThreshold = 1.5;  /**< max angle to be accepted (in degree) (cone opening angle is 2 times this!) */
    double m_thetaThreshold = 1.0;  /**< max theta angle to be accepted (in degree) Only used in the arc method */
    double m_energyThreshold = 1.0;  /**< max energy of gamma to be accepted */
    double m_momentumFactor = 1.0;  /**< factor for maximal gamma momentum compared to the lepton momentum */

    std::string m_method_param = "offset_cone";  /**< parameter defining which method will be used. */

    /// Member variables
    double m_maxAngle;  /**< max angle */
    double m_maxTheta;  /**< max theta angle used in arc method */

    /// Definition of possible methods with which to look for fsr photons
    enum EMethods {
      c_CenteredCone,  /**< A cone centered around the lepton momentum */
      c_OffsetCone,  /**< A cone on which the lepton momentum vector sits and whose symmetry axis
                       * is pointing into the opposite direction of magnetic force acting upon the charged particle */
      c_Arc  /**< An arc defined by two angles; one for the phi- and one for the theta direction.
               * The arc will also be offset. */
    };

    EMethods m_method;
  };
}
