/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2010 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Max Welsch, Felix Metzner                                *
 *                                                                        *
 * File for Untracked Privat Variables for B -> D(*(*)) l nu              *
 * Created on October 17th 2018                                           *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#include <analysis/VariableManager/OwnVariables.h>
#include <analysis/utility/ReferenceFrame.h>
#include <analysis/VariableManager/Manager.h>

#include <analysis/dataobjects/Particle.h>

#include <mdst/dataobjects/MCParticle.h>

#include <framework/datastore/StoreArray.h>
#include <framework/utilities/Conversion.h>

#include <TLorentzVector.h>
#include <TVector3.h>

#include <boost/lexical_cast.hpp>
#include <cmath>

using namespace std;

namespace Belle2 {
  namespace Variable {

    bool isValidBMeson(const Particle* part, unsigned int min_daughters)
    {
      if (part == nullptr) {
        return false;
      }
      int abs_pdg = abs(part->getPDGCode());

      if (abs_pdg != 511 and abs_pdg != 521) {
        return false;
      }

      unsigned int n_daughters = part->getNDaughters();
      if (n_daughters < min_daughters) {
        return false;
      }

      int d1_pdg = abs(part->getDaughter(1)->getPDGCode());
      if ((d1_pdg < 11) or (d1_pdg > 15)) {
        return false;
      }

      if (min_daughters < 3) {
        return true;
      }

      int d2_pdg = abs(part->getDaughter(2)->getPDGCode());
      if ((d2_pdg < 12) or (d2_pdg > 16)) {
        return false;
      }

      return true;
    }

    bool isValidBMesonMC(const MCParticle* part, unsigned int min_daughters)
    {
      if (part == nullptr) {
        return false;
      }
      int abs_pdg = abs(part->getPDG());

      if (abs_pdg != 511 and abs_pdg != 521) {
        return false;
      }

      unsigned int n_daughters = part->getNDaughters();
      if (n_daughters < min_daughters) {
        return false;
      }

      auto daughters = part->getDaughters();

      int d1_pdg = abs(daughters.at(1)->getPDG());
      if ((d1_pdg < 11) or (d1_pdg > 15)) {
        return false;
      }

      if (min_daughters < 3) {
        return true;
      }

      int d2_pdg = abs(daughters.at(2)->getPDG());
      if ((d2_pdg < 12) or (d2_pdg > 16)) {
        return false;
      }

      return true;
    }

    bool isValidUpsilon4S(const Particle* part)
    {
      if (part == nullptr) {
        return false;
      }
      int abs_pdg = abs(part->getPDGCode());

      if (abs_pdg != 300553) {
        return false;
      }

      unsigned int n_daughters = part->getNDaughters();
      if (n_daughters != 2) {
        return false;
      }

      int d0_pdg = abs(part->getDaughter(0)->getPDGCode());
      int d1_pdg = abs(part->getDaughter(1)->getPDGCode());
      if ((d0_pdg != 511) and (d0_pdg != 521)) {
        return false;
      }
      if ((d1_pdg != 511) and (d1_pdg != 521)) {
        return false;
      }

      return true;
    }

    double pStarL(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
        tagVec.SetE(-beamEnergy);


        TLorentzRotation lab2BsigRestFrame = TLorentzRotation(-(T.rotateCmsToLab() * -tagVec).BoostVector());

        TLorentzVector lVecInLab = part->getDaughter(1)->getDaughter(1)->get4Vector();

        return (lab2BsigRestFrame * lVecInLab).Vect().Mag();
      }
    }

    double qSquared(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D(*) meson
        const Particle* daughter0 = part->getDaughter(0);

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pD = frame.getMomentum(daughter0);

        auto q = pB - pD;

        return q.Mag2();
      }
    }

    double qSquaredMC(const Particle* part)
    {
      auto mc_part = part->getMCParticle();

      if (not isValidBMesonMC(mc_part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mc_daughters = mc_part->getDaughters();

        // get D(*) meson
        const Particle daughter0(mc_daughters.at(0));

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pD = frame.getMomentum(&daughter0);

        auto q = pB - pD;

        return q.Mag2();
      }
    }

    double qSquaredReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms()
                                * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms()
                              * part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(-beamEnergy);

        return (-tagVec - DVec).M2();
      }
    }

    double qSquaredReco2(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector tagVec = T.rotateLabToCms()
                                * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms()
                              * part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(-tagVec.E());

        return (-tagVec - DVec).M2();
      }
    }

    double qSquaredReco3(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector upsilon4SBoost = T.getBoostVector();

        TLorentzVector tagVec = part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        return (upsilon4SBoost - tagVec - DVec).M2();
      }
    }

    double qSquaredReco4(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        TLorentzVector upsilon4SBoost = T.getBoostVector();
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = part->getDaughter(1)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (part->getDaughter(1)->getDaughter(2)) {
          TLorentzVector pPi = part->getDaughter(1)->getDaughter(2)->get4Vector();
          DVec += pPi;
        }

        tagVec.SetE(beamEnergy);

        return (upsilon4SBoost - tagVec - DVec).M2();
      }
    }

    double qSquaredRecoDStarStar(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(0) == nullptr) {
        return -999;
      } else if (part->getDaughter(1)->getDaughter(2) == nullptr) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        TLorentzVector tagVec = T.rotateLabToCms()
                                * part->getDaughter(0)->get4Vector();
        TLorentzVector DVec = T.rotateLabToCms()
                              * part->getDaughter(1)->getDaughter(0)->get4Vector();
        TLorentzVector Pi0Vec = T.rotateLabToCms()
                                * part->getDaughter(1)->getDaughter(2)->get4Vector();

        tagVec.SetE(-beamEnergy);

        return (-tagVec - DVec - Pi0Vec).M2();
      }
    }

    double w(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D** meson
        const Particle* daughter0 = part->getDaughter(0);

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pDstst = frame.getMomentum(daughter0);

        auto q = pB - pDstst;

        return (pB.M2() + pDstst.M2() - q.M2()) / (2 * pB.M() * pDstst.M());
      }
    }

    double wMC(const Particle* part)
    {
      auto mc_part = part->getMCParticle();
      if (not isValidBMesonMC(mc_part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mc_daughters = mc_part->getDaughters();
        // get D** meson
        Particle dstst{mc_daughters.at(0)};
        const Particle* daughter0 = &dstst;

        TLorentzVector pB = frame.getMomentum(part);
        TLorentzVector pDstst = frame.getMomentum(daughter0);

        auto q = pB - pDstst;

        return (pB.M2() + pDstst.M2() - q.M2()) / (2 * pB.M() * pDstst.M());
      }
    }

    double wReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pD = T.rotateLabToCms() * sigB->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * sigB->getDaughter(2)->get4Vector();
          pD += pPi;
        }

        pTag.SetE(-beamEnergy);

        TLorentzVector pBsig = -pTag;
        TLorentzVector q = -pTag - pD;

        return (pBsig.M2() + pD.M2() - q.M2()) / (2 * pBsig.M() * pD.M());
      }
    }

    double costhetal(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get lepton and neutrino
        const Particle* daughter1 = part->getDaughter(1);
        const Particle* daughter2 = part->getDaughter(2);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;

        pl.Boost(motherBoost);
        pW.Boost(motherBoost);
        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double costhetalMC(const Particle* part)
    {

      auto mc_part = part->getMCParticle();
      if (not isValidBMesonMC(mc_part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mc_daughters = mc_part->getDaughters();
        // get lepton and neutrino
        Particle d1{mc_daughters.at(1)};
        Particle d2{mc_daughters.at(2)};
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;

        pl.Boost(motherBoost);
        pW.Boost(motherBoost);
        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double costhetalReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pD = T.rotateLabToCms() * sigB->getDaughter(0)->get4Vector();
        TLorentzVector pl = T.rotateLabToCms() * sigB->getDaughter(1)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = T.rotateLabToCms() * sigB->getDaughter(2)->get4Vector();
          pD += pPi;
        }

        pTag.SetE(-beamEnergy);

        TLorentzVector pW = -pTag - pD;

        pl.Boost(-pW.BoostVector());

        return cos(pW.Angle(pl.Vect()));
      }
    }

    double costhetav(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        // get D** and daugther(D**, 0)
        const Particle* daughter0 = part->getDaughter(0);
        const Particle* granddaughter00 = daughter0->getDaughter(0);

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pD = frame.getMomentum(granddaughter00);

        pDstst.Boost(motherBoost);
        pD.Boost(motherBoost);
        pD.Boost(-pDstst.BoostVector());

        return cos(pDstst.Angle(pD.Vect()));
      }
    }

    double costhetavMC(const Particle* part)
    {
      auto mc_part = part->getMCParticle();
      if (not isValidBMesonMC(mc_part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mc_daughters = mc_part->getDaughters();
        auto mc_granddaughters = mc_daughters.at(0)->getDaughters();

        // get D** and daugther(D**, 0)
        Particle d0{mc_daughters.at(0)};
        Particle gd00{mc_granddaughters.at(0)};

        const Particle* daughter0 = &d0;
        const Particle* granddaughter00 = &gd00;

        TVector3 motherBoost = -frame.getMomentum(part).BoostVector();

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pD = frame.getMomentum(granddaughter00);

        pDstst.Boost(motherBoost);
        pD.Boost(motherBoost);
        pD.Boost(-pDstst.BoostVector());

        return cos(pDstst.Angle(pD.Vect()));
      }
    }

    double costhetavReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pDst = sigB->getDaughter(0)->get4Vector();
        TLorentzVector pD = sigB->getDaughter(0)->getDaughter(0)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = sigB->getDaughter(2)->get4Vector();
          pDst += pPi;
        }

        pTag.SetE(-beamEnergy);
        TLorentzVector pBsig_Lab = T.rotateCmsToLab() * (-pTag);

        pDst.Boost(-pBsig_Lab.BoostVector());
        pD.Boost(-pBsig_Lab.BoostVector());
        pD.Boost(-pDst.BoostVector());

        return cos(pDst.Angle(pD.Vect()));
      }
    }

    double chiprime(const Particle* part)
    {
      if (not isValidBMeson(part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        const Particle* daughter0 = part->getDaughter(0);
        const Particle* daughter1 = part->getDaughter(1);
        const Particle* daughter2 = part->getDaughter(2);
        const Particle* granddaughter00 = daughter0->getDaughter(0);

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(granddaughter00);

        TVector3 WBoost = -pW.BoostVector();
        TVector3 DststBoost = -pDstst.BoostVector();

        pD.Boost(DststBoost);
        double costhetav =
          (-pD.Px() * pDstst.Px() - pD.Py() * pDstst.Py() - pD.Pz() * pDstst.Pz()) / (pD.P() * pDstst.P());

        pl.Boost(WBoost);
        double costhetawl = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 ltrans = pl.Vect() - costhetawl * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vtrans = pD.Vect() - costhetav * pD.Vect().Perp() * pDstst.Vect().Unit();

        double chi = atan2(ltrans.Cross(vtrans) * pW.Vect().Unit(), ltrans * vtrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiprimeMC(const Particle* part)
    {
      auto mc_part = part->getMCParticle();
      if (not isValidBMesonMC(mc_part)) {
        return -999;
      } else {
        const auto& frame = ReferenceFrame::GetCurrent();

        auto mc_daughters = mc_part->getDaughters();
        // get D** and daugther(D**, 0)
        Particle d0{mc_daughters.at(0)};
        Particle d1{mc_daughters.at(1)};
        Particle d2{mc_daughters.at(2)};

        const Particle* daughter0 = &d0;
        const Particle* daughter1 = &d1;
        const Particle* daughter2 = &d2;

        auto mc_granddaughters = mc_daughters.at(0)->getDaughters();
        Particle gd00{mc_granddaughters.at(0)};
        const Particle* granddaughter00 = &gd00;

        TLorentzVector pDstst = frame.getMomentum(daughter0);
        TLorentzVector pl = frame.getMomentum(daughter1);
        TLorentzVector pnu = frame.getMomentum(daughter2);
        TLorentzVector pW = pl + pnu;
        TLorentzVector pD = frame.getMomentum(granddaughter00);

        TVector3 WBoost = -pW.BoostVector();
        TVector3 DststBoost = -pDstst.BoostVector();

        pD.Boost(DststBoost);
        double costhetav =
          (-pD.Px() * pDstst.Px() - pD.Py() * pDstst.Py() - pD.Pz() * pDstst.Pz()) / (pD.P() * pDstst.P());

        pl.Boost(WBoost);
        double costhetawl = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 ltrans = pl.Vect() - costhetawl * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vtrans = pD.Vect() - costhetav * pD.Vect().Perp() * pDstst.Vect().Unit();

        double chi = atan2(ltrans.Cross(vtrans) * pW.Vect().Unit(), ltrans * vtrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    double chiprimeReco(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      } else {
        PCmsLabTransform T;
        double beamEnergy = T.getCMSEnergy() / 2.;

        // get B mesons. The Second daughter is expected to be the signal B meson.
        const Particle* tagB = part->getDaughter(0);
        const Particle* sigB = part->getDaughter(1);

        TLorentzVector pTag = T.rotateLabToCms() * tagB->get4Vector();
        TLorentzVector pDst = sigB->getDaughter(0)->get4Vector();
        TLorentzVector pD = sigB->getDaughter(0)->getDaughter(0)->get4Vector();
        TLorentzVector pl = sigB->getDaughter(1)->get4Vector();

        // Handling of D** cases by adding the pion of the D** decay, which is reconstructed
        // as third daughter of the B_sig, to the D meson 4-momentum.
        if (sigB->getDaughter(2)) {
          TLorentzVector pPi = sigB->getDaughter(2)->get4Vector();
          pDst += pPi;
        }

        pTag.SetE(-beamEnergy);
        TLorentzVector pBsig_Lab = T.rotateCmsToLab() * (-pTag);

        TLorentzVector pW = pBsig_Lab - pDst;

        pl.Boost(-pW.BoostVector());
        pD.Boost(-pDst.BoostVector());

        double costhetav = (-pD.Px() * pDst.Px() - pD.Py() * pDst.Py() - pD.Pz() * pDst.Pz()) / (pD.P() * pDst.P());
        double costhetawl = (-pl.Px() * pW.Px() - pl.Py() * pW.Py() - pl.Pz() * pW.Pz()) / (pl.P() * pW.P());

        TVector3 ltrans = pl.Vect() - costhetawl * pl.Vect().Perp() * pW.Vect().Unit();
        TVector3 vtrans = pD.Vect() - costhetav * pD.Vect().Perp() * pDst.Vect().Unit();

        double chi = atan2(ltrans.Cross(vtrans) * pW.Vect().Unit(), ltrans * vtrans);
        if (chi < 0) chi += 2 * TMath::Pi();

        return chi;
      }
    }

    template<typename Iterator>
    TLorentzVector sumMomentumOfAllDescendantsWithPDDG(
      const MCParticle* mcPart,
      Iterator begin,
      Iterator end
    )
    {
      TLorentzVector sumOf4Momenta(0., 0., 0., 0.);
      for (const MCParticle* mcDaughter : mcPart->getDaughters()) {
        if (std::find(begin, end, abs(mcDaughter->getPDG())) != end) {
          sumOf4Momenta += Particle(mcDaughter).get4Vector();
        } else if (mcDaughter->getNDaughters() > 0) {
          sumOf4Momenta += sumMomentumOfAllDescendantsWithPDDG(mcDaughter, begin, end);
        }
      }

      return sumOf4Momenta;
    };

    double mcMissingMassSquared(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        return secondNuMomentumSum.M2();
      } else if (secondNuMomentumSum.P() == 0.) {
        return firstNuMomentumSum.M2();
      } else {
        return -1;
      }
    }

    double missingMomentum(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms()
                              * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms()
                              * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

      return missingMomentum.Vect().Mag();
    }

    double mcMissingMomentum(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        return secondNuMomentumSum.Vect().Mag();
      } else if (secondNuMomentumSum.P() == 0.) {
        return firstNuMomentumSum.Vect().Mag();
      } else {
        return -1;
      }
    }

    double missingMomentumInCMSFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms()
                              * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms()
                              * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = tagVec + sigVec;

      return missingMomentum.Vect().Mag();
    }

    double missingMomentumInBSigRestFrame(const Particle* part)
    {
      if (not isValidUpsilon4S(part)) {
        return -999;
      }
      PCmsLabTransform T;
      double beamEnergy = T.getCMSEnergy() / 2.;
      TLorentzVector tagVec = T.rotateLabToCms()
                              * part->getDaughter(0)->get4Vector();
      TLorentzVector sigVec = T.rotateLabToCms()
                              * part->getDaughter(1)->get4Vector();
      tagVec.SetE(beamEnergy);
      TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

      TVector3 boost = part->getDaughter(1)->get4Vector().BoostVector();
      TLorentzRotation lab2BsigRestFrame = TLorentzRotation(-boost);

      return (lab2BsigRestFrame * missingMomentum).Vect().Mag();
    }

    double mcMissingMomentumInBSigRestFrame(const Particle*)
    {
      StoreArray <MCParticle> mcParticles("MCParticles");
      if (mcParticles.getEntries() == 0) {
        return -999;
      }

      MCParticle* mcUpsilon4S = mcParticles[0];
      if (mcUpsilon4S->getPDG() != 300553) {
        return -999;
      }

      std::vector<MCParticle*> mcU4SDaughters = mcUpsilon4S->getDaughters();
      if (mcU4SDaughters.size() != 2) {
        return -999;
      }

      const MCParticle* firstB = mcU4SDaughters.at(0);
      const MCParticle* secondB = mcU4SDaughters.at(1);

      if ((abs(firstB->getPDG()) != 511 && abs(firstB->getPDG()) != 521)
          || (abs(secondB->getPDG()) != 511 && abs(secondB->getPDG()) != 521)) {
        return -999;
      }

      std::vector<int> nuPDGs {12, 14, 16};

      TLorentzVector firstNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(firstB, nuPDGs.begin(), nuPDGs.end());
      TLorentzVector secondNuMomentumSum = sumMomentumOfAllDescendantsWithPDDG(secondB, nuPDGs.begin(), nuPDGs.end());

      if (firstNuMomentumSum.P() == 0. && secondNuMomentumSum.P() == 0) {
        return -2.;
      } else if (firstNuMomentumSum.P() == 0.) {
        TVector3 boost = Particle(secondB).get4Vector().BoostVector();
        TLorentzRotation lab2SecondBRestFrame = TLorentzRotation(-boost);
        return (lab2SecondBRestFrame * secondNuMomentumSum).Vect().Mag();
      } else if (secondNuMomentumSum.P() == 0.) {
        TVector3 boost = Particle(firstB).get4Vector().BoostVector();
        TLorentzRotation lab2FirstBRestFrame = TLorentzRotation(-boost);
        return (lab2FirstBRestFrame * firstNuMomentumSum).Vect().Mag();
      } else {
        return -1;
      }
    }

    Manager::FunctionPtr angleWithMissingMomentum(const std::vector<std::string>& arguments)
    {
      if (arguments.size() == 2) {
        std::string particle = arguments[0];
        int daughterNumber = 0;
        int daughterOfDaughterNumber = 0;
        try {
          daughterNumber = Belle2::convertString<int>(arguments[0]);
        } catch (boost::bad_lexical_cast&) {
          B2WARNING("First argument of angleWithMissingMass meta function must be integer!");
          return nullptr;
        }
        try {
          daughterOfDaughterNumber = Belle2::convertString<int>(arguments[1]);
        } catch (boost::bad_lexical_cast&) {
          B2WARNING("Second argument of angleWithMissingMass meta function must be integer!");
          return nullptr;
        }

        auto func = [daughterNumber, daughterOfDaughterNumber](const Particle * part) -> double {
          if (not isValidUpsilon4S(part))
          {
            return -999;
          }
          int nDaughters = static_cast<int>(part->getNDaughters());
          if (daughterNumber >= nDaughters)
          {
            return -999;
          }

          int nDaughtersOfDaughter = static_cast<int>(part->getDaughter(daughterNumber)->getNDaughters());
          if (daughterOfDaughterNumber >= nDaughtersOfDaughter)
          {
            return -999;
          }

          PCmsLabTransform T;
          double beamEnergy = T.getCMSEnergy() / 2.;
          TLorentzVector tagVec = T.rotateLabToCms() * part->getDaughter(0)->get4Vector();
          TLorentzVector sigVec = T.rotateLabToCms() * part->getDaughter(1)->get4Vector();
          tagVec.SetE(beamEnergy);
          TLorentzVector missingMomentum = T.rotateCmsToLab() * (tagVec + sigVec);

          const Particle* Bsig = part->getDaughter(daughterNumber);
          TVector3 boost = Bsig->get4Vector().BoostVector();
          TLorentzRotation lab2BsigRestFrame = TLorentzRotation(-boost);

          TVector3 missingMom_inRestFrame = (lab2BsigRestFrame * missingMomentum).Vect();

          TVector3 particleV = (lab2BsigRestFrame * part->getDaughter(daughterNumber)->getDaughter(
            daughterOfDaughterNumber)->get4Vector()).Vect();

          return std::cos(particleV.Angle(missingMom_inRestFrame));
        };
        return func;
      } else {
        B2FATAL("Wrong number of arguments for meta function angleWithMissingMomentum");
      }
    }


    VARIABLE_GROUP("MYVARIABLES");
    REGISTER_VARIABLE("pStarL", pStarL,
                      "Magnitude of the lepton momentum in the rest frame of the "
                      "signal B meson extracted via the tag side B's 4-momentum "
                      "for which the energy component is replaced by 1/2 of the "
                      "beam energy.\n"
                      "Must be a calculated from the Upsilon(4S) for which the"
                      "second daughter must be the signal B meson.");
    REGISTER_VARIABLE("qSquared", qSquared,
                      "Squared difference of the four-momenta of the B meson "
                      "and the first daughter of the B meson in the B rest frame. "
                      "The B meson is required to have 3 daughter particles!");
    REGISTER_VARIABLE("qSquaredMC", qSquaredMC,
                      "Squared difference of the four-momenta of the B meson "
                      "and the first daughter of the B meson in the B rest frame. "
                      "The B meson is required to have 3 daughter particles!");
    REGISTER_VARIABLE("qSquaredReco", qSquaredReco,
                      "To be applied on Upsilon(4S) particles to retrieve the "
                      "squared difference of the four-momenta of the signal B meson "
                      "and its first daughter. The signal B meson must be the second "
                      "daughter of the Upsilon(4S) meson! The Upsilon(4S) meson must "
                      "have 2 B mesons as daughters to be valid.");
    REGISTER_VARIABLE("qSquaredReco2", qSquaredReco2, "");
    REGISTER_VARIABLE("qSquaredReco3", qSquaredReco3, "");
    REGISTER_VARIABLE("qSquaredReco4", qSquaredReco4, "");
    REGISTER_VARIABLE("qSquaredRecoDStarStar", qSquaredRecoDStarStar,
                      "To be applied on Upsilon(4S) particles to retrieve the "
                      "squared difference of the four-momenta of the signal B meson "
                      "and its first and third daughter which together form a D** candidate. "
                      "The signal B meson must be the second daughter of the Upsilon(4S) meson! "
                      "The Upsilon(4S) meson must have 2 B mesons as daughters to be valid.");
    REGISTER_VARIABLE("w", w,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("wMC", wMC,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("wReco", wReco,
                      "product of four velocities of B meson and D** daugther");
    REGISTER_VARIABLE("costhetal", costhetal,
                      "cosine of angle between virtual W boson and lepton in W rest frame");
    REGISTER_VARIABLE("costhetalMC", costhetalMC,
                      "cosine of angle between virtual W boson and lepton in W rest frame");
    REGISTER_VARIABLE("costhetalReco", costhetalReco,
                      "cosine of angle between virtual W boson and lepton in W rest frame calculated "
                      "from reconstructed Upsilon(4S) meson.");
    REGISTER_VARIABLE("costhetav", costhetav,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("costhetavMC", costhetavMC,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("costhetavReco", costhetavReco,
                      "cosine of angle between D** meson and D(*) meson in D(*) rest frame");
    REGISTER_VARIABLE("chi_prime", chiprime,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("chi_primeMC", chiprimeMC,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("chi_primeReco", chiprimeReco,
                      "angle between decay planes of virtual W boson and D** boson");
    REGISTER_VARIABLE("mcMissingMassSquared", mcMissingMassSquared,
                      "Returns the true squared Missing Mass of the event.");
    REGISTER_VARIABLE("missingMomentum", missingMomentum,
                      "Returns the Missing Momentum of the event. "
                      "The Signal B meson must be the second daughter of the Upsilon(4s)!\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("mcMissingMomentum", mcMissingMomentum,
                      "Returns the true Missing Momentum of the event in the lab frame.");
    REGISTER_VARIABLE("missingMomentumInCMSFrame", missingMomentumInCMSFrame,
                      "Returns the Missing Momentum of the event in the center of mass frame, "
                      "The signal B meson must be the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("missingMomentumInBSigRestFrame", missingMomentumInBSigRestFrame,
                      "Returns the Missing Momentum of the event in the rest frame of the signal side, "
                      "if signal B is the second daughter of the Upsilon(4s).\n"
                      "The Missing Momentum is calculated using the beam energy as energy of the tag-side B.");
    REGISTER_VARIABLE("mcMissingMomentumInBSigRestFrame", mcMissingMomentumInBSigRestFrame,
                      "Returns the true Missing Momentum of the event in the signal B meson's rest frame.");
    REGISTER_VARIABLE("angleWithMissingMomentum(daughterNumber, daughterOfDaughterNumber)", angleWithMissingMomentum,
                      "Returns the cosin of the angle between the missing momentum (hence the function must "
                      "be applied on the Upsilon(4S)!) and the momentum vector of the daughterOfDaughterNumber-th "
                      "daughter of the daughterNumber-th daughter of the particle it is applied to in the rest frame "
                      "of the signal B meson's rest frame.");

  } // OwnVariables namespace
} // Belle2 namespace