import os
import pickle
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from PyFastBDT import FastBDT
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score

from typing import List, Tuple, Dict, Union, Optional

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.plotting.mva_plots import plot_roc_curve
from analysistools.dataframe_tools.hdf5_utils import load_hdf5


cs_variable_dict = {
    'daughter__bo1__cm__spR2__bc': HistVariable('daughter__bo1__cm__spR2__bc', n_bins=25, scope=(0., 1.),
                                                var_name=r"$\mathrm{R}_2$"),
    'daughter__bo1__cm__spcosTBTO__bc': HistVariable('daughter__bo1__cm__spcosTBTO__bc', n_bins=25, scope=(0., 1.),
                                                     var_name=r"$\cos\theta_\mathrm{TBTO}$"),
    'daughter__bo1__cm__spcosTBz__bc': HistVariable('daughter__bo1__cm__spcosTBz__bc', n_bins=25, scope=(0., 1.),
                                                    var_name=r"$\cos\theta_\mathrm{TBz}$"),
    'daughter__bo1__cm__spthrustBm__bc': HistVariable('daughter__bo1__cm__spthrustBm__bc', n_bins=25, scope=(0.5, 1.),
                                                      var_name=r"$\mathrm{T}_\mathrm{B}$"),
    'daughter__bo1__cm__spthrustOm__bc': HistVariable('daughter__bo1__cm__spthrustOm__bc', n_bins=25, scope=(0.5, 1.),
                                                      var_name=r"$\mathrm{T}_\mathrm{O}$"),
    'daughter__bo1__cm__spKSFWVariables__boet__bc__bc': HistVariable('daughter__bo1__cm__spKSFWVariables__boet__bc__bc',
                                                                     n_bins=25, scope=(2., 12.),
                                                                     var_name=r"$\mathrm{KSFW}\;E_t$"),
    'daughter__bo1__cm__spKSFWVariables__bomm2__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bomm2__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;M_m^2$",
        scope=(-10, 10)),
    'daughter__bo1__cm__spKSFWVariables__bohso00__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso00__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{0,0}$",
        scope=(0, 1.5)),
    'daughter__bo1__cm__spKSFWVariables__bohso02__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso02__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{0,2}$",
        scope=(-0.6, 1.1)),
    'daughter__bo1__cm__spKSFWVariables__bohso04__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso04__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{0,4}$",
        scope=(-0.4, 0.6)),
    'daughter__bo1__cm__spKSFWVariables__bohso10__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso10__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{1,0}$",
        scope=(0, 1.25)),
    'daughter__bo1__cm__spKSFWVariables__bohso12__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso12__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{1,2}$",
        scope=(-0.2, 0.55)),
    'daughter__bo1__cm__spKSFWVariables__bohso14__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso14__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{1,4}$",
        scope=(-0.2, 0.3)),
    'daughter__bo1__cm__spKSFWVariables__bohso20__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso20__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{2,0}$",
        scope=(0, 0.7)),
    'daughter__bo1__cm__spKSFWVariables__bohso22__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso22__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{2,2}$",
        scope=(-0.2, 0.5)),
    'daughter__bo1__cm__spKSFWVariables__bohso24__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohso24__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{so}_{2,4}$",
        scope=(-0.2, 0.3)),
    'daughter__bo1__cm__spKSFWVariables__bohoo0__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohoo0__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{oo}_0$",
        scope=(0.0, 0.3)),
    'daughter__bo1__cm__spKSFWVariables__bohoo1__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohoo1__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{oo}_1$",
        scope=(-0.04, 0.04)),
    'daughter__bo1__cm__spKSFWVariables__bohoo2__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohoo2__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{oo}_2$",
        scope=(-0.02, 0.12)),
    'daughter__bo1__cm__spKSFWVariables__bohoo3__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohoo3__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{oo}_3$",
        scope=(-0.04, 0.04)),
    'daughter__bo1__cm__spKSFWVariables__bohoo4__bc__bc': HistVariable(
        'daughter__bo1__cm__spKSFWVariables__bohoo4__bc__bc', n_bins=25, var_name=r"$\mathrm{KSFW}\;H^{oo}_4$",
        scope=(-0.025, 0.08)),
    'daughter__bo1__cm__spCleoCone__bo1__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo1__bc__bc', n_bins=30,
                                                               scope=(0, 5),
                                                               var_name=r"$\mathrm{CleoCone}(1)$"),
    'daughter__bo1__cm__spCleoCone__bo2__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo2__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(2)$"),
    'daughter__bo1__cm__spCleoCone__bo3__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo3__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(3)$"),
    'daughter__bo1__cm__spCleoCone__bo4__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo4__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(4)$"),
    'daughter__bo1__cm__spCleoCone__bo5__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo5__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(5)$"),
    'daughter__bo1__cm__spCleoCone__bo6__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo6__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(6)$"),
    'daughter__bo1__cm__spCleoCone__bo7__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo7__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(7)$"),
    'daughter__bo1__cm__spCleoCone__bo8__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo8__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(8)$"),
    'daughter__bo1__cm__spCleoCone__bo9__bc__bc': HistVariable('daughter__bo1__cm__spCleoCone__bo9__bc__bc', n_bins=30,
                                                               scope=(0, 6),
                                                               var_name=r"$\mathrm{CleoCone}(9)$")
}

# List of additional variables needed for training process but not as input for the cs mva.
cs_additional_variables = {
    "isNotContinuumEvent",  # Target variable for the mva training.
    "m2RecoilSignalSide",  # Signal variable used to check correlation with input variables for the mva.
    "isSignalAcceptMissingNeutrino",  # Can be used to extract only correctly reconstructed signal events.
    "daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc",  # Can be used to extract corr. reconstructed sig. events.
}


class ContinuumSuppression:

    def __init__(self,
                 cs_variables: List[str],
                 target_variable: str,
                 weight: Optional[str] = None):
        """
        Interface for the training and application of the
        continuum suppression using a mva method.

        :param cs_variables: List of variables to use for the
        continuum suppression.
        :param target_variable: Name of the variable for the
        'signal' when training the classifier.
        :param weight: Name of the weight to use in the DataFrame.
        """

        self._cs_variables = cs_variables
        self._target_variable = target_variable
        self._weight = weight
        self._has_trained_classifier = False

        self._classifier = None

    def add_classifier(self,
                       classifier=FastBDT.Classifier(),
                       is_trained: bool = False):
        """
        Adds a mva classifier to the continuum suppression.
        Can be trained or untrained. The classifier should
        have a sklearn interface.
        :param classifier: A mva classifier with sklearn interface.
        :param is_trained: Flag if the classifier is already trained.
        :return:
        """
        assert self._classifier is None
        self._classifier = classifier
        self._has_trained_classifier = is_trained

    def load_classifier_weights(self, weight_file_path: Union[str, os.fspath]) -> None:
        """
        Loads the classifier weights from an already prepared weight file.
        :param weight_file_path: Path to the weight file.
        :return: None
        """
        if not os.path.exists(weight_file_path):
            raise ValueError(f"The provided path to the weight file does not exist:\n{weight_file_path}")

        self._classifier.load(weightfile=weight_file_path)
        self._has_trained_classifier = True

    def train(self, data: pd.DataFrame) -> None:
        """
        Trains the added classifier with data present
        in the DataFrame.

        :param data: Training data. DataFrame has to hold the
        continuum suppression as columns.
        :return: None.
        """
        x_train = data[self._cs_variables].values
        y_train = data[self._target_variable].values.reshape(-1)

        if self._weight is None:
            w_train = np.ones(y_train.shape)
        else:
            w_train = data[self._weight].values

        w_train = self.normalize_weights(y_train, w_train)

        self._classifier.fit(x_train, y_train, w_train)
        self._has_trained_classifier = True

    @staticmethod
    def normalize_weights(target: np.ndarray,
                          weights: np.ndarray) -> np.ndarray:
        """
        Normalizes event weights so that the sum of
        signal and background weights is equal.

        :param target: Target labels. Expected to be in {0, 1}.
        :param weights: Array of event weights.
        :return: Normalized event weights
        """
        n_sig = np.sum(target == 1)
        n_bkg = np.sum(target == 0)

        weights[np.where(target == 1)] *= 1 / n_sig
        weights[np.where(target == 0)] *= 1 / n_bkg

        return weights

    def apply(self,
              data: pd.DataFrame,
              output_column: str = 'cs_output') -> None:
        """
        Applies a train continuum suppression classifier to a
        pd.DataFrame and adds a new column with the classifier
        output to the pd.DataFrame.

        :param data: DataFrame where the classifier should
        be applied on.
        :param output_column: Name of the column containing the
        classifier output which will be appended to the given
        DataFrame.
        :return: None
        """
        if not self._has_trained_classifier:
            raise RuntimeError("Trying to apply an untrained classifier.")

        x_predict = data[self._cs_variables].values

        pcs = self._classifier.predict(x_predict)

        data.loc[:, output_column] = pcs

    def save_classifier_weights(self, weight_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Saves the weight file for a trained classifier.
        :param weight_file_path: Path under which the weight file shall be saved.
        :param overwrite: Whether to overwrite an existing weight file.
        :return: None
        """
        if os.path.isfile(weight_file_path) and not overwrite:
            raise RuntimeError("The specified weight file already exists and the overwrite option is set to False.")
        self._classifier.save(weightfile=weight_file_path)

        var_file_path = self.get_var_file_from_weight_file(weight_file_path=weight_file_path)
        self.dump_variables(var_file_path, overwrite=overwrite)

    def dump_variables(self, pickle_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Dump variables used by the cs mva as pickled object.
        :param pickle_file_path: Path under which the variables shall be dumped as pickled object.
        :param overwrite:  Whether to overwrite an existing file.
        :return: None
        """
        ContinuumSuppression.dump_list_of_variables(
            variable_list=self._cs_variables,
            pickle_file_path=pickle_file_path,
            overwrite=overwrite
        )

    def dump_classifier(self, pickle_file_path: Union[str, os.fspath], overwrite: bool = False) -> None:
        """
        Dump classifier object as pickled object.
        :param pickle_file_path: Path under which the classifier shall be dumped as pickled object.
        :param overwrite: Whether to overwrite an existing file.
        :return: None
        """
        if os.path.isfile(pickle_file_path) and not overwrite:
            raise RuntimeError("The specified pickle file already exists and the overwrite option is set to False.")
        with open(pickle_file_path, "wb") as pkl_file:
            pickle.dump(self, pkl_file)

    @staticmethod
    def get_var_file_from_weight_file(weight_file_path: Union[str, os.fspath]) -> str:
        """
        Convenience function to get the file name of the variable file to a weight file.
        :param weight_file_path: Path to the weight file.
        :return: Path to respective variable file.
        """
        return os.fspath(f"{os.path.splitext(weight_file_path)[0]}_vars.pkl")

    @staticmethod
    def load_classifier(weight_file_path: Union[str, os.fspath], var_file_path: Union[None, str, os.fspath] = None):
        """
        Returns a classifier with the weights from the specified weight file.
        :param weight_file_path: Path to the weight file for the classifier.
        :param var_file_path: Path to the variable file for the classifier.
                              Default: None -> Trying to get it from weight file name.
        :return: A ContinuumSuppression classifier.
        """
        if not os.path.isfile(weight_file_path):
            raise ValueError(f"The specified weight file\n {weight_file_path} \ndoes not exists.")

        if var_file_path is None:
            var_file_path = ContinuumSuppression.get_var_file_from_weight_file(weight_file_path=weight_file_path)
            if not os.path.isfile(var_file_path):
                raise RuntimeError(f"Automatic extraction of variables file path from weight file failed!"
                                   f"Got variables file path\n{var_file_path}\nfrom weight file\n{weight_file_path}")
        variables = ContinuumSuppression.load_pickled_variables(pickle_file_path=var_file_path)

        cs_classifier = ContinuumSuppression(variables, "", "")
        cs_classifier.add_classifier()
        cs_classifier.load_classifier_weights(weight_file_path=weight_file_path)
        return cs_classifier

    @staticmethod
    def load_pickled_classifier(pickle_file_path: Union[str, os.fspath]):
        """
        Loads a pickled classifier from the specified pickle file.
        :param pickle_file_path: Path to the pickle file containing a classifier object.
        :return: A ContinuumSuppression classifier.
        """
        if not os.path.isfile(pickle_file_path):
            raise ValueError(f"The specified pickle file\n {pickle_file_path} \ndoes not exists.")
        with open(pickle_file_path, "rb") as pkl_file:
            cs_classifier = pickle.load(pkl_file)
        return cs_classifier

    @staticmethod
    def load_pickled_variables(pickle_file_path: Union[str, os.fspath]) -> List[str]:
        """
        Loads a pickled variable list for a classifier from the specified pickle file.
        :param pickle_file_path: Path to the pickle file containing the variables.
        :return: A list of variables.
        """
        if not os.path.isfile(pickle_file_path):
            raise ValueError(f"The specified pickle file\n {pickle_file_path} \ndoes not exists.")
        with open(pickle_file_path, "rb") as pkl_file:
            cs_variables = pickle.load(pkl_file)
        return cs_variables

    @staticmethod
    def dump_list_of_variables(
            variable_list: List[str],
            pickle_file_path: Union[str, os.fspath],
            overwrite: bool = False
    ) -> None:
        """
        Dump given list of variables.

        :param variable_list: List of variables to be saved in pickle file.
        :param pickle_file_path: Path under which the variables shall be dumped as pickled object.
        :param overwrite: Whether to overwrite an existing file.
        :return: None
        """
        if os.path.isfile(pickle_file_path) and not overwrite:
            raise RuntimeError("The specified pickle file already exists and the overwrite option is set to False.")
        with open(pickle_file_path, "wb") as pkl_file:
            pickle.dump(variable_list, pkl_file)

    @staticmethod
    def prepare_data_samples(
            class1_dfs: List[pd.DataFrame],
            class2_dfs: List[pd.DataFrame],
            train_test_split_ratio: float = 0.5
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Splits data frames for category 1 and 2 into subsets for training and evaluation of the
        CS mva and returns a tuple of four data frames: Training set category 1, training set category 2,
        test set category 1 and test set category 2.

        :param class1_dfs: Data frame containing data of the first class.
        :param class2_dfs: Data frame containing data of the second class.
        :param train_test_split_ratio: Ratio by which the data frames will be split into train
        and test sample. Default = 0.5.
        :return: Tuple (class 1 train df, class 2 train df, class 1 test df, class 2 test df)
        """
        train1 = []
        test1 = []
        for df in class1_dfs:
            train, test = train_test_split(df, train_size=train_test_split_ratio)
            train1.append(train)
            test1.append(test)

        train2 = []
        test2 = []
        for df in class2_dfs:
            train, test = train_test_split(df, train_size=train_test_split_ratio)
            train2.append(train)
            test2.append(test)

        return (
            pd.concat(train1, ignore_index=True, sort=False),
            pd.concat(train2, ignore_index=True, sort=False),
            pd.concat(test1, ignore_index=True, sort=False),
            pd.concat(test2, ignore_index=True, sort=False)
        )

    @staticmethod
    def get_uncorrelated_variables(
            data_sample: pd.DataFrame,
            variables: List[str],
            max_correlation: float,
            signal_var: str
    ) -> List[str]:
        """
        Method to get a set of variables from the provided list "variables" which are
        uncorrelated to a specified other variable "signal_var".

        The given data_sample is used to evaluate the correlation. As correlated to
        signal_var are considered such variables for which abs(correlation to signal_var)
        is larger than max_correlation.

        The reduced list of variable strings is returned.

        :param data_sample: Set used to evaluate the correlation.
        :param variables: Variables to be checked for their correlation to signal_var.
        :param max_correlation: The maximal correlation allowed. Variables with values above are removed.
        :param signal_var: The signal variable against which the correlation is checked.
        :return: List of variable strings of uncorrelated variables.
        """
        corr = data_sample[variables + [signal_var]].corr()
        final_var_list = list(corr.query(f'abs({signal_var}) < {max_correlation}').index)

        print(f"The following variables have been omitted due to their correlation to the signal variable {signal_var}")
        print("\n".join([var for var in variables if var not in final_var_list]))

        return final_var_list

    def test_evaluation(self, test_data: pd.DataFrame, output_dir: Union[str, os.fspath]) -> float:
        """
        A test application of this Continuum Suppression MVA, yielding some performance values and plots.

        The CS is applied to the provided test data sample provided via the parameter test_data
        based on which the performance of the CS MVA is evaluated.
        The results are written into the directory specified via output_dir.
        The return value is the area under the ROC curve for this CS MVA.

        :param test_data: Pandas DataFrame containing a labeled test data set.
        :param output_dir: The path to a directory into which the results will be written.
        :return: Returns the area under the roc curve.
        """
        self.apply(test_data, output_column="cs_mva_output")

        # TODO: Add distibution plots of input pdfs and correlation plots using test_cs_input...
        # test_cs_input = test_data[self._cs_variables].values
        test_cs_truth = test_data[self._target_variable].values
        test_cs_output = test_data["cs_mva_output"].values

        roc_auc = roc_auc_score(test_cs_truth, test_cs_output)

        fig = plt.figure(num=None, figsize=(12, 9), dpi=100, facecolor='w', edgecolor='k')
        ax = plt.gca()
        fig, ax = plot_roc_curve(y_true=test_cs_truth, y_predict=test_cs_output, fig_ax_tuple=(fig, ax))
        export(fig=fig, target_dir=output_dir, filename="cs_mva_roc_plot")

        return roc_auc


def run_cs_training(
        train_data_file_path: Union[str, os.fspath],
        variable_file_path: Union[str, os.fspath],
        weight_file_path: Union[str, os.fspath],
        target_variable: str = "isNotContinuumEvent",
        fast_bdt_settings: Optional[Dict[str, str]] = None,
) -> None:
    """
    Trains a FastBDT for Continuum Suppression using a provided, prepared data set,
    and storing the resulting weight files as well as the list of used variables.

    The truth target used for the supervised training process is specified via the
    parameter target_variable.
    The FastBDT settings can be set via the optional parameter fast_bdt_settings,
    which expects a dictionary mapping the setting names to their respective values.

    The variables used as input for the MVA are provided as pickled list loaded from
    a pickle file specified by variable_file_path.

    The resulting classifier weights and the variable list required to apply the
    CS MVA are stored to a path defined by the argument weight_file_path.

    :param train_data_file_path: Path to the training data sample hdf5 file.
    :param variable_file_path: Path to the pickle file containing the variable to be used.
    :param weight_file_path: Path under which the trained weights and variables will be stored.
    :param target_variable: The variable used as truth target in the training.
    :param fast_bdt_settings: Settings used for the FastBDT MVA.
    :return: None
    """
    final_vars = ContinuumSuppression.load_pickled_variables(variable_file_path)
    cs_train_data = load_hdf5(hdf5_file_path=train_data_file_path, hdf5_key="train_sample")

    cs_train_data["cs_weights"] = np.ones(len(cs_train_data.index))

    cs = ContinuumSuppression(
        cs_variables=final_vars,
        target_variable=target_variable,
        weight='cs_weights'
    )

    if fast_bdt_settings is None:
        fast_bdt_settings = {
            "nTrees": 200,
        }

    cs.add_classifier(classifier=FastBDT.Classifier(**fast_bdt_settings))
    cs.train(cs_train_data)

    cs.save_classifier_weights(weight_file_path=weight_file_path, overwrite=True)


def evaluate_cs_training(
        test_data_file_path: Union[str, os.fspath],
        weight_file_path: Union[str, os.fspath],
        output_dir_path: Union[str, os.fspath]
) -> float:
    """
    Evaluates a given trained Continuum Suppression classifier.

    :param test_data_file_path: Path to the hdf5 file containing the test data set.
    :param weight_file_path: Path to the weights of the CS MVA to be tested.
    :param output_dir_path: Path to the directory under which the output of the
                            evaluation is stored.
    :return: Area under the ROC curve.
    """
    cs = ContinuumSuppression.load_classifier(weight_file_path=weight_file_path)
    cs_test_data = load_hdf5(hdf5_file_path=test_data_file_path, hdf5_key="train_sample")

    roc_auc_value = cs.test_evaluation(cs_test_data, output_dir=output_dir_path)

    return roc_auc_value
