#!/usr/bin/python
from __future__ import absolute_import

import argparse
import subprocess
import os
import sys
import shutil
import socket


script_dir = sys.path[0]

mc_types = [
    "Dlnu",
    "Dstlnu", 
    "Dststlnu", 
    "SignalDststlnupiz", 
    "SignalDststlnupim"
]

dec_file_dirs = {
    "Dlnu": "{0}/dec_files/b_to_dlnu".format(script_dir),
    "Dstlnu": "{0}/dec_files/b_to_dstlnu".format(script_dir),
    "Dststlnu": "{0}/dec_files/b_to_dststlnu".format(script_dir),
    "SignalDststlnupiz": "{0}/dec_files/b_to_dststlnu_piz_signal".format(script_dir),
    "SignalDststlnupim": "{0}/dec_files/b_to_dststlnu_pim_signal".format(script_dir),
}

def check_host():
    """
    Checks if scripts are executed on the KEKCC. If not, the
    execution of the script is terminated.
    """
    if "cc.kek.jp" not in socket.gethostname():
        sys.exit("This script can only be executet on kekcc. Terminating!")

class EvtGenerationManager:
    """
    Manages event generation with EvtGen:
      * copies and removes decay file to the runEvtgen directory
      * executes the runEvtgen script
    """
    def __init__(self, dec_file_dir):
        """
        Constructor
        :param dec_file_dir: Path to decay file directory.
        :type dec_file_dir: str
        """
        self._dec_file_dir = dec_file_dir
        self._evtgen_dir = "{0}/evtgen".format(sys.path[0])
        self._dec_files = os.listdir(dec_file_dir)

    def copy_dec_files_to_evtgen_dir(self):
        """
        Copies all decay file in the give directory to the evtgen
        directory.
        """
        for file in self._dec_files:
            shutil.copy("{0}/{1}".format(self._dec_file_dir, file),
                        "{0}/{1}".format(self._evtgen_dir, file))

    def tidy_up_evtgen_dir(self):
        """
        Deletes all decay files and job scripts in the given directory
        from the evtgen directory.
        """
        for file in self._dec_files:
            try:
                os.remove("{0}/{1}".format(self._evtgen_dir, file))
            except OSError:
                print "File {0}/{1} does not exist!".format(
                    self._evtgen_dir, file)

        subprocess.call(["rm -f {0}/evtgen/*evtgen_exp*".format(sys.path[0])], shell=True)
        subprocess.call(["rm -f {0}/evtgen/*fpda_pid*".format(sys.path[0])], shell=True)

    def generate_events(self, total_num, job_num, output_path):
        """
        Calls the ./evtgen/runEvtgen script with the given
        parameters. This creates basf scripts which are submitted
        to the LSF batch system. The generated events and log
        files are stored in the given output directory

        :param total_num: Total number of events to generate.
        :type total_num: int
        :param job_num: Maximum number of events submitted per
        LSF job.
        :type job_num: int
        :param output_path: Path to the output directory. Has to
        be of the for /abs/path/to/dir/. The last forward slash
        cannot be omitted.
        :type output_path: str
        :return:
        """
        os.chdir("{0}/evtgen".format(sys.path[0]))
        output_path = "{0}evtgen/".format(output_path)

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        for dec_file in self._dec_files:
            print "\n\n"
            print "\tGenerating {0} events for decay file {1}/{2}".format(
                total_num, dec_file_dir, dec_file)
            print "\tOutput will be put in {0}".format(output_path)
            print "\n\n"
            subprocess.check_call(["./runEvtgen",
                                   "nBB-Y4S.txt",
                                   dec_file,
                                   "Y4S.conf",
                                   str(total_num),
                                   str(job_num),
                                   output_path])


def parse_command_line_args():
    """
    Parses command line arguments.

    :return: Command line arguments.
    """

    parser = argparse.ArgumentParser(description="This script starts the event generation with EvtGen for the given"
                                                 " Decay type.")

    parser.add_argument("-t", action="store", dest="dec_type", type=str, required=True,
                        help="Decay type: Currently available is 'Dlnu', 'Dstlnu', 'Dststlnu'.")
    parser.add_argument("-o", action="store", dest="out_path", type=str, required=True,
                        help="Absolute output path for the generated events. "
                             "Has to end with a /. "
                             "Example: '/group/belle2/users/welsch/test_mc/evtgen'")
    parser.add_argument("-n", action="store", dest="total_num", type=int, required=True,
                        help="Total number of events to generate per decay file of the specified decay type.")
    parser.add_argument("-j", action="store", dest="job_num", type=int, required=True,
                        help="Maximum number of events per job.")

    parser.add_argument("--clean", action="store_true", required=False,
                        help="Tidy up the evtgen directory after all jobs ran successfully.")

    parsed_args = parser.parse_args()

    if parsed_args.out_path[-1] != "/":
        raise ValueError("Output path does not end with '/'!")

    return parsed_args

if __name__ == "__main__":
    check_host()

    args = parse_command_line_args()

    if args.dec_type not in mc_types:
        raise ValueError("Specified decay type not known!")

    dec_file_dir = dec_file_dirs[args.dec_type]

    evtgen_manager = EvtGenerationManager(dec_file_dir)

    if not args.clean:
        evtgen_manager.copy_dec_files_to_evtgen_dir()

        evtgen_manager.generate_events(args.total_num,
                                       args.job_num,
                                       args.out_path)
    else:
        evtgen_manager.tidy_up_evtgen_dir()



