#!/usr/bin/python
from __future__ import absolute_import

import argparse
import subprocess
import os
import socket
import sys

script_dir = sys.path[0]

mc_types = ["Dlnu", "Dstlnu", "Dststlnu", "SignalDststlnu"]

dec_file_dirs = {
    "Dlnu": "{0}/dec_files/b_to_dlnu".format(script_dir),
    "Dstlnu": "{0}/dec_files/b_to_dstlnu".format(script_dir),
    "Dststlnu": "{0}/dec_files/b_to_dststlnu".format(script_dir),
    "SignalDststlnu": "{0}/dec_files/b_to_dststlnu_signal".format(script_dir),
}

def check_host():
    """
    Checks if scripts are executed on the KEKCC. If not, the
    execution of the script is terminated.
    """
    if "cc.kek.jp" not in socket.gethostname():
        sys.exit("This script can only be executed on kekcc. Terminating!")

def get_dec_types(dec_file_dir):
    """
    Creates a list of decay types from the .dec files in the
    given directory by removing the file extension.

    :param dec_file_dir: Path to the .dec files.
    :type dec_file_dir: str

    :return: List of decay types.
    """
    return [os.path.splitext(file)[0] for file in os.listdir(dec_file_dir)]


def parse_command_line_args():
    """
    Parses command line arguments.

    :return: Command line arguments.
    """
    parser = argparse.ArgumentParser(description="This script starts the event simulation with GSim for the given"
                                                 " Decay type.")

    parser.add_argument("-t", action="store", dest="dec_type", type=str, required=True,
                        help="Decay type: Currently available is 'Dlnu', 'Dstlnu', 'Dststlnu'.")
    parser.add_argument("-i", action="store", dest="in_path", type=str, required=True,
                        help="Absolute path to the generated events. If the events are stored in "
                             "/some/path/evtgen/gen/DECAYS, you have to provide /some/path.")

    parser.add_argument("--clean", action="store_true", required=False,
                        help="Tidy up the gsim directory after all jobs ran successfully.")

    args = parser.parse_args()

    return args


class EvtSimulationManager:

    def __init__(self, dec_file_dir, in_path):
        self._dec_types = get_dec_types(dec_file_dir)
        self._path_to_gen = "{0}evtgen/gen".format(in_path)
        self._outpath = str(in_path)

    def simulate(self):

        os.chdir("{0}/gsim".format(sys.path[0]))

        for dec_type in self._dec_types:
            abs_path = "{0}/{1}/".format(self._path_to_gen, dec_type)
            print "\n\n"
            print "\tExecuting detector simulation for decay {0}".format(dec_type)
            print "\tInput from {0}".format(abs_path)
            print "\tOutput will be put in {0}/gsim/mdst/{1}/".format(self._outpath,
                                                               dec_type)
            print "\n\n"


            subprocess.check_call(["./runGsimReco.csh",
                                   self._outpath,
                                   abs_path,
                                   dec_type])


if __name__ == "__main__":

    check_host()

    args = parse_command_line_args()

    if args.clean:
        os.chdir("{0}/gsim".format(sys.path[0]))
        subprocess.check_call(["./clean.csh"], shell=True)
    else:
        print "Starting detector simulation for decay type: {0}".format(args.dec_type)
        dec_file_dir = dec_file_dirs[args.dec_type]
        gsim_manager = EvtSimulationManager(dec_file_dir, args.in_path)
        gsim_manager.simulate()
