"""
This file calculates 1 sigma variations of the form factor parameters of the
LLSW_nominal model for D^*{**} decays and generates new decay files accordingly.

Maximilian Welsch 2018
"""
import logging
import numpy as np
import os
from typing import List

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


class ParameterVariation:
    """
    This class performs the parameter variation for the form factor
    parameters of the LLSW_nominal model.
    """

    def __init__(self, nom_vals: np.array, cov_mat: np.array = None):
        """
        Constructor.
        :param nom_vals: Nominal parameter values. Shape (n,).
        :param cov_mat: Parameter covariance matrix. Shape (n, n).
        """
        self._nom_vals = nom_vals
        self._n_params = nom_vals.shape[0]
        self._cov_mat = cov_mat
        self._stds = None
        self._corr_mat = None
        self._eig_vals = None
        self._eig_vecs = None
        self._variation = None

    def calculate_covariance_matrix(self,  stds: np.array, corr_mat: np.array):
        """
        If no covariance matrix was given in the constructor, the covariance
        matrix can be calculated with this function using the correlation
        matrix and the standard deviations of the parameters.
        :param stds: Parameter standard deviations. Shape (n,).
        :param corr_mat: Parameter correlation matrix. Shape (n, n).
        """
        self._stds = stds

        self._corr_mat = corr_mat
        self._cov_mat = np.matmul(np.diag(stds),
                                  np.matmul(corr_mat, np.diag(stds)))

    def _calc_eigvals_eigvecs(self):
        """
        Calculates eigenvalues and vectors of the covariance matrix
        """
        if self._cov_mat is None:
            raise ValueError('Covariance matrix is not initialized yet.')

        self._eig_vals, self._eig_vecs = np.linalg.eig(self._cov_mat)

    def _calc_variation(self):
        """
        Calculates the 1 sigma variation for each parameter.
        """
        if self._eig_vecs is None or self._eig_vals is None:
            self._calc_eigvals_eigvecs()

        # note that the columns of this array corerspond to the variation of
        #  the parameters
        self._variation = np.matmul(self._eig_vecs,
                                    np.sqrt(np.diag(self._eig_vals)))

    def calculate_variation(self):
        """
        Helper function if variations should be calculated manually.
        """
        self._calc_variation()

    @property
    def up_variation(self):
        """
        Calculates 1 sigma upwards variation of the parameters.
        :return: Upwards variation.
        """
        if self._variation is None:
            self._calc_variation()

        nom_val = np.repeat(np.reshape(self._nom_vals,
                                       (self._n_params, 1)),
                            repeats=self._n_params,
                            axis=1)
        return np.around(nom_val + self._variation, decimals=2)

    @property
    def up_variation_as_list(self) -> List[List[float]]:
        """
        :return: Up variations as list.
        """
        up_variation = self.up_variation
        up_variation_list = up_variation.T.tolist()
        return up_variation_list

    @property
    def down_variation(self):
        """
         Calculates 1 sigma downwards variation of the parameters.
        :return: Downwards variation.
        """
        if self._variation is None:
            self._calc_variation()

        nom_val = np.repeat(np.reshape(self._nom_vals,
                                       (self._n_params, 1)),
                            repeats=self._n_params,
                            axis=1)
        return np.around(nom_val - self._variation, decimals=2)

    @property
    def down_variation_as_list(self) -> List[List[float]]:
        """
        :return: Down variations as list.
        """
        down_variation = self.down_variation
        down_variation_list = down_variation.T.tolist()
        return down_variation_list

    def print_summary(self):
        """

        :return:
        """
        print(50*'-')
        print(f'Nominal Values:\n{self._nom_vals}\n')
        print(f'Standard Deviations:\n{self._stds}\n')
        print(f'Correlation Matrix:\n{self._corr_mat}\n')
        print(f'Covariance Matrix:\n{self._cov_mat}\n')
        print(f'Eigen Values:\n{self._eig_vals}\n')
        print(f'Eigen Vectors:\n{self._eig_vecs}\n')
        print(f'1 Sigma Variation up:\n{self.up_variation}\n')
        print(f'1 Sigma Variation down:\n{self.down_variation}\n')


class DecFileManipulator:
    """
    This class creates new decay file with LLSW_nominal parameters exchanged for
    the D^{**} decay.
    """

    def __init__(self, original_file: str, keyword: str):
        self._keyword = keyword
        self._original_file = original_file
        self._original_file_name = original_file.split('/')[-1]
        with open(original_file, 'r') as of:
            self._original_lines = [line for line in of]
        self._new_lines = None

    def replace_parameters(self, params: List[float]):
        nparams = len(params)
        self._new_lines = list()
        for line in self._original_lines:
            if self._keyword in line:
                line_wo_params = self._cut_off_old_params(line, nparams)
                line_w_new_params = self._add_new_params(line_wo_params, params)
                self._new_lines.append(line_w_new_params)
                continue
            self._new_lines.append(line)

    @staticmethod
    def _cut_off_old_params(line: str, nparams: int) -> str:
        """
        Cuts off the parameters in the D^{**} decay command.
        :param str line: Decay line with old parameters.
        :param int nparams: Number of parameters.
        :return str: Decay line w/o parameters
        """
        return ' '.join(line.split(' ')[:-nparams])

    @staticmethod
    def _add_new_params(line: str, params: List[float]) -> str:
        """
        Adds the new paramters at the end of the line.
        :param str line: Old line w/o parameters.
        :param List[float] params: New parameters.
        :return str: New decay line.
        """
        param_string = ' '.join([str(param) for param in params])
        # add end of command character for EvtGen decay files
        param_string += ';\n'
        return ' '.join([line, param_string])

    def print_old_file(self):
        """
        Prints the old decay file.
        """
        for line in self._original_lines:
            print(line)

    def print_new_file(self):
        """
        Prints the new decay file
        """
        for line in self._new_lines:
            print(line)

    def save_as(self, extension: str, new_dir: str):
        org_id = self._original_file_name.split('.')[0]
        new_file = '_'.join([org_id, extension]) + '.dec'
        with open(f'{new_dir}/{new_file}', 'w') as nf:
            for line in self._new_lines:
                nf.write(line)


def create_llsw_variations():
    logging.info('Start calculation.')
    nominal = {
        'narrow': np.array([0.71, -1.6, -0.5, 2.9]),
        'broad': np.array([0.68, -0.2, 0.3])
    }

    standard_deviation = {
        'narrow': np.array([0.07, 0.2, 0.3, 1.6]),
        'broad': np.array([0.2, 1.2, 0.3])
    }

    correlation_matrix = {
        'narrow': np.array([[1, -0.83, 0.66, -0.63],
                            [-0.83, 1, -0.27, 0.20],
                            [0.66, -0.27, 1, -0.93],
                            [-0.63, 0.20, -0.93, 1]]),
        'broad': np.array([[1, -0.95, -0.35],
                           [-0.95, 1, 0.51],
                           [-0.35, 0.51, 1]])
    }

    dec_file_dir = 'decay_files'
    dec_type = {
        'narrow': ('Done', 'Dtwostar'),
        'broad': ('Dprimeone', 'Dzerostar')
    }

    for type in ['narrow', 'broad']:
        logging.info(f'Processing type {type}')
        variation = ParameterVariation(nominal[type])
        variation.calculate_covariance_matrix(standard_deviation[type],
                                              correlation_matrix[type])
        variation.calculate_variation()
        variation.print_summary()

        for i, var in enumerate(variation.up_variation_as_list):
            logging.info(f'Variation up {i}: {var}')
            decs = dec_type[type]
            for dec in decs:
                logging.info(f'Processing Channel: {dec}')
                for dec_file in os.listdir(f'{dec_file_dir}/LLSW_nominal/{dec}'):
                    variation_dir = f'{dec_file_dir}/LLSW_variation_up_{i}/{dec}'
                    if not os.path.exists(variation_dir):
                        os.makedirs(variation_dir)
                    if any(dec in dec_file for dec in decs):
                        dfmanip = DecFileManipulator(
                            f'{dec_file_dir}/LLSW_nominal/{dec}/{dec_file}', 'LLSW')
                        dfmanip.replace_parameters(var)
                        dfmanip.save_as(f'variation_up_{i}', variation_dir)
                    else:
                        continue

        for i, var in enumerate(variation.down_variation_as_list):
            logging.info(f'Variation down {i}: {var}')
            decs = dec_type[type]
            for dec in decs:
                logging.info(f'Processing Channel: {dec}')

                for dec_file in os.listdir(f'{dec_file_dir}/LLSW_nominal/{dec}'):
                    variation_dir = f'{dec_file_dir}/LLSW_variation_down_{i}/{dec}'
                    if not os.path.exists(variation_dir):
                        os.makedirs(variation_dir)
                    if any(dec in dec_file for dec in decs):
                        dfmanip = DecFileManipulator(
                            f'{dec_file_dir}/LLSW_nominal/{dec}/{dec_file}', 'LLSW')
                        dfmanip.replace_parameters(var)
                        dfmanip.save_as(f'variation_down_{i}', variation_dir)
                    else:
                        continue


if __name__ == '__main__':
    create_llsw_variations()