"""
Definition of all tasks used for creating the data used for the mc reweighting
for the B --> D^{**} l nu decays from the ISGW2 to the LLSW decay model.

Maximilian Welsch 2018.
"""

import os
from subprocess import check_call

import time

import b2luigi
from b2luigi.basf2_helper import Basf2PathTask

import pandas as pd
import root_pandas

from rdstar.mc_reweighting.BtoDststlnu.generate_events import generate_events
from rdstar.mc_reweighting.BtoDststlnu.reweighting_variables import calculate_reweighting_variables

class EventGenerationTask(Basf2PathTask):
    """
    This task is responsible for the actual event generation using Basf2.
    """
    n_events = b2luigi.IntParameter()
    model_variation = b2luigi.Parameter()
    decay_channel = b2luigi.Parameter()
    decay_file = b2luigi.Parameter()

    def output(self):
        yield self.add_to_output('events.root')

    def create_path(self):
        path_to_dec_file = f'decay_files/{self.model_variation}/{self.decay_channel}/{self.decay_file}'
        return generate_events(self.n_events,
                               path_to_dec_file,
                               self.get_output_file_name('events.root'))


class MergeIntoCommonDataFrameTask(b2luigi.Task):
    """
    This task merges all root files for each decay channel and parameter
    variation into one dataframe.
    """
    n_events = b2luigi.IntParameter()
    model_variation = b2luigi.Parameter()
    decay_channel = b2luigi.Parameter()

    def requires(self):
        for decay_file in os.listdir(f'decay_files/{self.model_variation}/{self.decay_channel}'):
            yield self.clone(EventGenerationTask,
                             n_events=self.n_events,
                             model_variation=self.model_variation,
                             decay_channel=self.decay_channel,
                             decay_file=decay_file)

    def output(self):
        yield self.add_to_output('merged.pkl')

    def run(self):
        inputs = self.get_input_file_names('events.root')
        df = root_pandas.read_root(inputs)
        df.to_pickle(self.get_output_file_name(f'merged.pkl'))


class ReweightingVariablesCalculationTask(b2luigi.Task):
    """
    This task takes the combined dataframes and calculates the required variables
    for the reweighting defined in `reweighting_variables.py`.
    """

    n_events = b2luigi.IntParameter()
    model_variation = b2luigi.Parameter()
    decay_channel = b2luigi.Parameter()

    def requires(self):
        yield self.clone(MergeIntoCommonDataFrameTask,
                         n_events=self.n_events,
                         model_variation=self.model_variation,
                         decay_channel=self.decay_channel)

    def output(self):
        yield self.add_to_output(f'reweighting_vars.h5')

    def run(self):
        # there is only one input file so take only the 0th component
        input = self.get_input_file_names(f'merged.pkl')[0]
        df = pd.read_pickle(input)
        new_df = calculate_reweighting_variables(df)
        new_df.to_hdf(
            self.get_output_file_name(f'reweighting_vars.h5'),
            key='df',
            mode='w')


class MainTask(b2luigi.Task):
    """
    Main Tasks for the monte carlo production. Requires all other tasks.
    You have to generate the decay files for the varied parameters by executing
    the `calculate_llsw_variations.py` script.
    """
    n_events = b2luigi.IntParameter()
    dec_model = b2luigi.Parameter()

    def requires(self):
        dec_model_variation_dict = {
            'ISGW2': [
                'ISGW2_nominal'
            ],
            'LLSW': [
                'LLSW_nominal',
                'LLSW_variation_down_0',
                'LLSW_variation_down_1',
                'LLSW_variation_down_2',
                'LLSW_variation_down_3',
                'LLSW_variation_up_0',
                'LLSW_variation_up_1',
                'LLSW_variation_up_2',
                'LLSW_variation_up_3'
            ],
            'MyISGW2': [
                'MyISGW2_nominal'
            ]
        }
        dec_channels = ['Done',
                        'Dprimeone',
                        'Dtwostar',
                        'Dzerostar']

        dec_model_variations = dec_model_variation_dict[self.dec_model]
        for dec_model_variation in dec_model_variations:
            for dec_channel in dec_channels:

                # skip if this combination is used since there are only three
                # parameters for Dprimeone and Dzerostar
                if dec_channel in ['Dprimeone', 'Dzerostar'] and \
                        dec_model_variation in ['LLSW_variation_up_3',
                                                'LLSW_variation_down_3']:
                    continue

                yield self.clone(ReweightingVariablesCalculationTask,
                                 n_events=self.n_events,
                                 model_variation=dec_model_variation,
                                 decay_channel=dec_channel)


if __name__ == '__main__':
    # Important: dec_model has to be adjusted prior to the event generation.
    # Only then the correct particle widths for each model are used.
    dec_model = 'LLSW'

    if dec_model == 'ISGW2':
        check_call("./evt_pdl/replace_evt_pdl.sh evt_pdl/belle_evt.pdl.backup", shell=True)
    elif dec_model == 'MyISGW2':
        check_call("./evt_pdl/replace_evt_pdl.sh evt_pdl/belle_evt.pdl.backup", shell=True)
    elif dec_model == 'LLSW':
        check_call("./evt_pdl/replace_evt_pdl.sh evt_pdl/belleII_evt.pdl.backup", shell=True)

    time.sleep(1)

    b2luigi.process(MainTask(1_000_000, dec_model), workers=8)
