from abc import ABC, abstractmethod
from functools import lru_cache

import numpy as np
from scipy.integrate import quad


class ISGW2toLLSWDssWeight(ABC):
    def __init__(self,
                 tau_w1: float,
                 tau_wp: float):

        self._tau_w1 = tau_w1
        self._tau_wp = tau_wp

        self._mb = 4.2
        self._mc = 1.4

        self._L = 0.40
        self._Lp = 0.80
        self._Ls = 0.76

        self._eta_1 = 0
        self._eta_2 = 0.
        self._eta_3 = 0.
        self._eta_b = 0.

        self._chi_1 = 0.
        self._chi_2 = 0.
        self._chi_b = 0.

    @abstractmethod
    def dgamma_isgw2(self, w: float, mB: float, mD: float) -> float:
        """
        Differential decay rate with respect to w using the ISGW2
        form factor parameterization. Terms that cancel in a ratio
        with another differential decay rate are neglected.

        :param w: Product of 4-velocities of B meson and D** meson.
        :param mB: B mass.
        :param mD: D** mass.
        :return: dGamma/dw
        """
        pass

    def to_dgamma(self, mB, mD):
        """
        Multiplicative factor to get the real branching fraction.
        """
        r = mD/mB
        return self._gf**2*self._vcb**2*mB**5/48/np.pi**3*r**3

    def _gamma_isgw2(self, mB, mD):
        """
        Total decay rate using the the ISGW2 form factor
        parameterization. Terms that cancel in a ratio with
        another total decay rate are neglected.

        :param mB: B mass.
        :param mD: D** mass.
        :return: Gamma
        """
        wmin, wmax = 1.0, self._wmax(mB, mD)
        return quad(self.dgamma_isgw2, wmin, wmax, (mB, mD))

    def _gamma_llsw(self, mB, mD):
        """
        Total decay rate using the the LLSW form factor
        parameterization. Terms that cancel in a ratio with
        another total decay rate are neglected.

        :param mB: B mass.
        :param mD: D** mass.
        :return: Gamma
        """
        wmin, wmax = 1.0, self._wmax(mB, mD)
        return quad(self.dgamma_llsw, wmin, wmax, (mB, mD))

    def _get_normalization(self, mB, mD):
        """
        Calculates the correct normalization of each weight as the
        ratio of the total decay rates.

        :param mB: B mass.
        :param mD: D** mass.
        :return:
        """
        gamma_isgw2 = self._gamma_isgw2(mB, mD)
        gamma_llsw = self._gamma_llsw(mB, mD)

        if gamma_llsw[0] != 0.:
            return gamma_isgw2[0]/gamma_llsw[0]
        else:
            return 1.0

    @abstractmethod
    def dgamma_llsw(self, w: float, mB: float, mD: float) -> float:
        """
        Differential decay rate with respect to w using the LLSW
        form factor parameterization. Terms that cancel in a ratio
        another differential decay rates are neglected.

        :param w: Product of 4-velocities of B meson and D** meson.
        :param mB: B mass.
        :param mD: D** mass.
        :return: dGamma/dw
        """
        pass

    @staticmethod
    def _wmax(mB: float, mD: float) -> float:
        """
        Helper function calculating the maximum w value.
        """
        return (mB**2 + mD**2) / (2*mB*mD)

    def _get_etas(self, *args) -> float:
        """
        Helper function for a expression in _dgamma_isgw.
        """
        lqcd2 = 0.04
        nflav = 4
        temp = 0.6

        if len(args) == 1:

            mass = args[0]

            if mass > 0.6:
                if mass < 1.85:
                    nflav = 3.0
                temp = 12.0 * np.pi / (33.0 - 2.0 * nflav) / np.log(mass * mass / lqcd2)

            return temp

        elif len(args) == 2:

            massq = args[0]
            massx = args[1]

            if massx > 0.6:
                if massq < 1.85:
                    nflav = 3.0
                temp = 12.0 * np.pi / (33.0 - 2.0 * nflav) / np.log(massx * massx / lqcd2)

            return temp

        else:
            raise RuntimeError("Called evt_get_etas with the wrong number of arguments.")

    def _isgur_wise_function(self, w):
        """
        Implementation of the Isgur-Wise function.
        :return: Leading Isgur-Wise function values.
        """
        return self._tau_w1 * (1 + (w - 1.) * self._tau_wp)

    @lru_cache(maxsize=10000)
    def get_weight(self, w, mB, mD):
        """
        Calculates the event form factor weight.

        :param w: Product of 4-velocities of B meson and D** meson.
        :param mB: B mass.
        :param mD: D** mass.
        :return: FF weight.
        """
        d_w = self.dgamma_isgw2(w, mB, mD)
        u_w = self.dgamma_llsw(w, mB, mD)

        weight = u_w / d_w
        weight *= self._get_normalization(mB, mD)

        return weight


class ISGW2toLLSWDprimeoneWeight(ISGW2toLLSWDssWeight):

    def __init__(self,
                 tau_w1: float = 0.68,
                 tau_wp: float = -0.2,
                 zeta_1: float = 0.3,):

        super().__init__(tau_w1, tau_wp)
        self._zeta_1 = zeta_1

    def dgamma_isgw2(self,  w: float, mB: float, mD: float):
        """
        Differential decay rate for B->D'1*lnu with the ISGW2 model.
        :return:
        """
        q2 = mB ** 2 + mD ** 2 - 2 * mB * mD * w
        trueW = w

        # get R
        mcR = 2.0 * np.sqrt(mB * mD) / (mB + mD)

        # get r
        r = mD / mB

        # get FF

        t = q2
        mass = mD

        msb = 5.2
        msd = 0.33
        bb2 = 0.431 * 0.431
        mbb = 5.31
        nf = 4.0

        msq = 1.82
        bx2 = 0.33 * 0.33
        mbx = (3.0 * 2.49 + 2.40) / 4.0  # check here
        nfp = 3.0

        mtb = msb + msd
        mtx = msq + msd
        mb = mB
        mx = mass

        mup = 1.0 / (1.0 / msq + 1.0 / msb)
        mum = 1.0 / (1.0 / msq - 1.0 / msb)
        bbx2 = 0.5 * (bb2 + bx2)
        tm = (mb - mx) * (mb - mx)
        if t > tm:
            t = 0.99 * tm
        wt = 1.0 + (tm - t) / (2.0 * mbb * mbx)

        mqm = 0.1
        r2 = 3.0 / (4.0 * msb * msq) + 3 * msd * msd / (2 * mbb * mbx * bbx2) + (
                16.0 / (mbb * mbx * (33.0 - 2.0 * nfp))) * np.log(
            self._get_etas(mqm) / self._get_etas(msq))

        f5 = np.sqrt(mtx / mtb) * (np.sqrt(bx2 * bb2) / bbx2) ** (5.0 / 2.0) / ((1.0 + r2 * (tm - t) / 18.0) ** 3.0)

        f5q = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** (-0.5)
        f5l = f5 * (mbb / mtb) ** 0.5 * (mbx / mtx) ** 0.5
        f5cppcm = f5 * (mbb / mtb) ** (-1.5) * (mbx / mtx) ** 0.5
        f5cpmcm = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** (-0.5)

        if msq == msd:
            ql = -1.0 * (msd * (5.0 + wt) * f5q / (2.0 * mtx * np.sqrt(bb2) * 6.0))

            ll = -1.0 * mtb * np.sqrt(bb2) * f5l * (1 / mum + ((msd * mtx * (wt - 1) / bb2) * (
                    (5.0 + wt) / (6.0 * msq) - (msd * bb2) / (2.0 * mum * mtx * bbx2))))

            cppcm = (-1.0 * (msd * mtx * f5cppcm / (2.0 * msq * mtb * np.sqrt(bb2))) * (
                    1 - (msd * msq * bb2) / (2.0 * mtx * mum * bbx2)))

            cpmcm = 1.0 * (msd * mtx * f5cpmcm / (2.0 * msq * mtb * np.sqrt(bb2))) * (
                    ((wt + 2.0) / 3.0) - (msd * msq * bb2) / (2.0 * mtx * mum * bbx2)) * (mtb / mtx)
        else:

            ql = f5q * np.sqrt(1.0 / 6.0) * msd / (np.sqrt(bb2) * mtx) * (1.0 - bb2 * mtb / (4.0 * msd * msq * msb))
            ll = f5l * np.sqrt(2.0 / 3.0) * mtb * np.sqrt(bb2) * (
                    1.0 / (2.0 * msq) - 3.0 / (2.0 * msb) + msd * mtx * (wt - 1) / bb2 * (
                    1.0 / msq - msd * bb2 / (2.0 * mum * mtx * bbx2)))
            cppcm = msd * msd * bx2 * f5cppcm / (np.sqrt(6.0) * mtb * msq * np.sqrt(bb2) * bbx2)
            cpmcm = -np.sqrt(2.0 / 3.0) * msd * f5cpmcm / (np.sqrt(bb2) * mtx) * (1 + msd * bx2 / (2.0 * msq * bbx2))

        qf = ql
        lf = ll
        cpf = (cppcm + cpmcm) / 2.0
        cmf = (cppcm - cpmcm) / 2.0

        conv = mcR * mB * (1.0 + r)
        gA = -conv * qf
        gV1 = 2.0 * lf / conv
        gV3rgV2 = conv * cpf
        r_tm1 = (trueW - r) * gV1 + (trueW * trueW - 1.0) * gV3rgV2
        tm1 = r_tm1 * r_tm1
        tm2 = 2.0 * (1.0 - 2.0 * r * trueW + r * r) * (gV1 * gV1 + (trueW * trueW - 1.0) * gA * gA)

        oldFF = tm1 + tm2
        oldFF *= np.sqrt(trueW * trueW - 1.0)

        return oldFF

    def dgamma_llsw(self, w: float, mB: float, mD: float):
        r = mD / mB

        gA = self.gAD1p(w)
        gV1 = self.gV1D1p(w)
        gV2 = self.gV2D1p(w)
        gV3 = self.gV3D1p(w)

        gV3rgV2 = gV3 + r * gV2

        r_tm1 = (w - r) * gV1 + (w ** 2 - 1.0) * gV3rgV2
        tm1 = r_tm1 ** 2
        tm2 = 2.0 * (1.0 - 2.0 * r * w + r * r) * (gV1 ** 2 + (w ** 2 - 1.0) * gA ** 2)
        newFF = tm1 + tm2
        newFF *= np.sqrt(w ** 2 - 1.0)
        return newFF

    def gV1D1p(self, w: float):
        ec = 1 / (2 * self._mc)
        eb = 1 / (2 * self._mb)

        z12 = self._isgur_wise_function(w)

        gv1 = 0
        gv1 += (w - 1.) * z12
        gv1 += ec * ((w * self._Ls - self._L) * z12 + (w - 1.) * (-2 * self._chi_1))
        gv1 -= eb * ((self._Ls * (2. * w + 1) - self._L * (w + 2.)) * z12 - 2 * (
                    w ** 2 - 1.) * self._zeta_1 * z12 - (
                             w - 1.) * self._chi_b)

        return gv1

    def gV2D1p(self, w: float):
        ec = 1 / (2 * self._mc)

        z12 = self._isgur_wise_function(w)

        gv2 = 0
        gv2 += 2 * ec * (self._zeta_1 * z12 - self._chi_2)

        return gv2

    def gV3D1p(self, w: float):
        ec = 1 / (2 * self._mc)
        eb = 1 / (2 * self._mb)

        z12 = self._isgur_wise_function(w)

        gv3 = 0
        gv3 += eb * ((self._Ls * (2. * w + 1) - self._L * (w + 2.)) / (w + 1.) * z12 - 2 * (
                    w - 1.) * self._zeta_1 * z12 - self._chi_b)
        gv3 -= z12
        gv3 -= ec * ((w * self._Ls - self._L) / (
                    w + 1) * z12 + 2. * self._zeta_1 * z12 - 2 * self._chi_1 + 2 * self._chi_2)

        return gv3

    def gAD1p(self, w: float):
        ec = 1 / (2 * self._mc)
        eb = 1 / (2 * self._mb)

        z12 = self._isgur_wise_function(w)

        ga = 0
        ga += eb * (2. * (w - 1.) * self._zeta_1 * z12 - (self._Ls * (2. * w + 1.) - self._L * (w + 2.)) / (
                    w + 1.) * z12 + self._chi_b)
        ga += z12
        ga += ec * ((w * self._Ls - self._L) / (w + 1.) * z12 - 2 * self._chi_1)

        return ga

class ISGW2toLLSWDzerostarWeight(ISGW2toLLSWDssWeight):
    """
        Reweighting code for B->D'1lnu from ISGW2 to LLSW FF parameterization.
        """

    def __init__(self,
                 tau_w1: float = 0.68,
                 tau_wp: float = -0.2,
                 zeta_1: float = 0.3, ):
        super().__init__(tau_w1, tau_wp)
        self._zeta_1 = zeta_1

    def dgamma_isgw2(self, w: float, mB: float, mD: float):
        q2 = mB ** 2 + mD ** 2 - 2 * mB * mD * w
        trueW = w

        # get R
        mcR = 2.0 * np.sqrt(mB * mD) / (mB + mD)

        # get r
        r = mD / mB

        # get FF

        t = q2
        mass = mD

        msb = 5.2
        msd = 0.33
        bb2 = 0.431 * 0.431
        mbb = 5.31
        nf = 4.0

        msq = 1.82
        bx2 = 0.33 * 0.33
        mbx = (3.0 * 2.49 + 2.40) / 4.0  # changed
        nfp = 3.0

        mtb = msb + msd
        mtx = msq + msd
        mb = 5.279
        mx = mass

        mup = 1.0 / (1.0 / msq + 1.0 / msb)
        mum = 1.0 / (1.0 / msq - 1.0 / msb)
        bbx2 = 0.5 * (bb2 + bx2)
        tm = (mb - mx) * (mb - mx)
        if t > tm:
            t = 0.99 * tm
        wt = 1.0 + (tm - t) / (2.0 * mbb * mbx)

        mqm = 0.1
        r2 = 3.0 / (4.0 * msb * msq) + 3 * msd * msd / (2 * mbb * mbx * bbx2) + (
                16.0 / (mbb * mbx * (33.0 - 2.0 * nfp))) * np.log(
            self._get_etas(mqm) / self._get_etas(msq))

        f5 = np.sqrt(mtx / mtb) * (np.sqrt(bx2 * bb2) / bbx2) ** (5.0 / 2.0) / ((1.0 + r2 * (tm - t) / 18.0) ** 3.0)

        f5uppum = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** 0.5
        f5upmum = f5 * (mbb / mtb) ** 0.5 * (mbx / mtx) ** (-0.5)

        uppum = -1.0 * f5uppum * np.sqrt(2.0 / (3.0 * bb2)) * msd
        upmum = 1.0 * f5upmum * np.sqrt(2.0 / (3.0 * bb2)) * msd * mtb / mtx

        upf = (uppum + upmum) / 2.0
        umf = (uppum - upmum) / 2.0

        gp = (upf + (1.0 - r) * umf / (1.0 + r)) / mcR
        gm = ((1.0 - r) * upf / (1.0 + r) + umf) / mcR
        RoldFF = (1.0 + r) * gp - (1.0 - r) * gm
        oldFF = RoldFF * RoldFF

        oldFF *= (trueW ** 2 - 1.0) * np.sqrt(trueW ** 2 - 1.0)

        return oldFF

    def dgamma_llsw(self, w: float, mB: float, mD: float):
        r = mD / mB

        gp = self.gpD0(w)
        gm = self.gmD0(w)

        r_newFF = (1.0 + r) * gp - (1.0 - r) * gm
        newFF = r_newFF * r_newFF
        newFF *= (w ** 2 - 1.0) * np.sqrt(w ** 2 - 1.0)

        return newFF

    def gpD0(self, w: float):
        ec = 1 / (2 * self._mc)
        eb = 1 / (2 * self._mb)

        z12 = self._isgur_wise_function(w)

        gp = 0
        gp += ec * (2. * (w - 1.) * self._zeta_1 * z12 - 3. * z12 * (w * self._Ls - self._L) / (w + 1.))
        gp += -eb * ((self._Ls * (2. * w + 1.) - self._L * (w + 2.)) / (w + 1.) * z12 - 2. * (
                w - 1.) * self._zeta_1 * z12)

        return gp

    def gmD0(self, w: float):
        ec = 1 / (2 * self._mc)
        eb = 1 / (2 * self._mb)

        z12 = self._isgur_wise_function(w)

        gm = 0

        gm += z12
        gm += ec * (6 * self._chi_1 - 2 * (w + 1) * self._chi_2)
        gm += eb * self._chi_b

        return gm


class ISGW2toLLSWDtwostarWeight(ISGW2toLLSWDssWeight):
    def __init__(self,
                 tau_w1: float = 0.71,
                 tau_wp: float = -1.6,
                 tau_1: float = -0.5,
                 tau_2: float = 2.9
                 ):
        super().__init__(tau_w1, tau_wp)
        self._tau_1 = tau_1
        self._tau_2 = tau_2

    def dgamma_isgw2(self, w: float, mB: float, mD: float):
        q2 = mB**2 + mD**2 - 2*mB*mD*w
        trueW = w

        # get R
        mcR = 2.0 * np.sqrt(mB * mD) / (mB + mD)

        # get r
        r = mD / mB

        # get FF

        t = q2
        mass = mD

        msb = 5.2
        msd = 0.33
        bb2 = 0.431 * 0.431
        mbb = 5.31
        nf = 4.0

        msq = 1.82
        bx2 = 0.33 * 0.33
        mbx = (5.0 * 2.46 + 3.0 * 2.42) / 8.0
        nfp = 3.0

        mtb = msb + msd
        mtx = msq + msd
        mb = mB
        mx = mass

        mup = 1.0 / (1.0 / msq + 1.0 / msb)
        mum = 1.0 / (1.0 / msq - 1.0 / msb)
        bbx2 = 0.5 * (bb2 + bx2)
        tm = (mb - mx) * (mb - mx)
        if t > tm:
            t = 0.99 * tm
        wt = 1.0 + (tm - t) / (2.0 * mbb * mbx)

        mqm = 0.1
        r2 = 3.0 / (4.0 * msb * msq) + 3 * msd * msd / (2 * mbb * mbx * bbx2) + (
                16.0 / (mbb * mbx * (33.0 - 2.0 * nfp))) * np.log(
            self._get_etas(mqm) / self._get_etas(msq))

        f5 = np.sqrt(mtx / mtb) * (np.sqrt(bx2 * bb2) / bbx2) ** (5.0 / 2.0) / ((1.0 + r2 * (tm - t) / 18.0) ** 3.0)

        f5h = f5 * (mbb / mtb) ** (-1.5) * (mbx / mtx) ** (-0.5)
        f5k = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** 0.5
        f5bppbm = f5 * (mbb / mtb) ** (-2.5) * (mbx / mtx) ** 0.5
        f5bpmbm = f5 * (mbb / mtb) ** (-1.5) * (mbx / mtx) ** (-0.5)

        hf = f5h * (msd / (np.sqrt(8.0 * bb2) * mtb)) * ((1.0 / msq) - (msd * bb2 / (2.0 * mum *
                                                                                     mtx * bbx2)))

        kf = f5k * (msd / (np.sqrt(2.0 * bb2))) * (1.0 + wt)

        bppbm = ((msd * msd * f5bppbm * bx2) / (np.sqrt(32.0 * bb2) * msq * msb * mtb * bbx2)) * (
                1.0 - (msd * bx2 / (2.0 * mtb * bbx2)))

        bpmbm = -1.0 * (msd * f5bpmbm / (np.sqrt(2.0 * bb2) * msb * mtx)) * (
                1.0 - ((msd * msb * bx2) / (2.0 * mup * mtb * bbx2)) + (
                (msd * bx2 * (1.0 - ((msd * bx2) / (2.0 * mtb * bbx2)))) / (4.0 * msq * bbx2)))

        bpf = (bppbm + bpmbm) / 2.0
        bmf = (bppbm - bpmbm) / 2.0

        conv = mcR * mB * (1 + r)
        kV = -conv * mB * hf
        kA1 = 2.0 * mB * kf / conv
        kA3rkA2 = conv * mB * bpf
        r_tm1 = (trueW - r) * kA1 + (trueW * trueW - 1.0) * kA3rkA2
        tm1 = 2.0 * r_tm1 * r_tm1 / 3.0
        tm2 = (1 - 2.0 * r * trueW + r * r) * (kA1 * kA1 + (trueW * trueW - 1.0) * kV * kV)

        oldFF = tm1 + tm2
        oldFF *= (trueW * trueW - 1.0) * np.sqrt(trueW * trueW - 1.0)

        return oldFF

    def dgamma_llsw(self, w: float, mB: float, mD: float):

        r = mB/mD

        ka1 = self.kA1D2(w)
        ka2 = self.kA2D2(w)
        ka3 = self.kA3D2(w)
        kv = self.kVD2(w)

        ka3rka2 = ka3 + r * ka2

        r_tm1 = (w - r) * ka1 + (w**2 - 1.)*ka3rka2
        tm1 = 2.0 * r_tm1 * r_tm1 / 3.0

        tm2 = (1 - 2.0 *r*w + r*r) * (ka1 * ka1 + (w * w - 1.0) * kv * kv)

        newFF = tm1 + tm2
        newFF *= (w**2 - 1.)*np.sqrt(w**2 - 1.)

        return newFF

    def kA1D2(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        ka1 = 0
        ka1 -= (1. + w) * t32
        ka1 -= eb * ((w - 1.) * ((self._Lp + self._L) * t32 - (2. * w + 1.) * self._tau_1 * t32 - self._tau_2 * t32) + (
                    1. + w) * self._eta_b)
        ka1 -= ec * ((w - 1.) * (self._tau_1 * t32 - self._tau_2 * t32) + (w + 1.) * (-2 * self._eta_1 + self._eta_3))

        return ka1

    def kA2D2(self, w):
        ec = 1. / (2. * self._mc)

        t32 = self._isgur_wise_function(w)

        ka2 = 0
        ka2 -= 2.*ec*(self._tau_1 * t32 + self._eta_2)

        return ka2

    def kA3D2(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        ka3 = 0
        ka3 += t32
        ka3 += eb * ((self._Lp + self._L) * t32 - (2.*w + 1.) * self._tau_1 * t32 - self._tau_2 * t32 + self._eta_b)
        ka3 -= ec * (self._tau_1 * t32 + self._tau_2 * t32 + 2 * self._eta_1 - 2 * self._eta_2 - self._eta_3)

        return ka3

    def kVD2(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        kv = 0
        kv -= t32
        kv -= eb * ((self._Lp + self._L) * t32 - (2. * w + 1) * self._tau_1 * t32 - self._tau_2 * t32 + self._eta_b)
        kv -= ec * (self._tau_1 * t32 - self._tau_2 * t32 - 2 * self._eta_1 + self._eta_3)

        return kv

class ISGW2toLLSWDoneWeight(ISGW2toLLSWDssWeight):
    def __init__(self,
                 tau_w1: float = 0.71,
                 tau_wp: float = -1.6,
                 tau_1: float = -0.5,
                 tau_2: float = 2.9
                 ):
        super().__init__(tau_w1, tau_wp)
        self._tau_1 = tau_1
        self._tau_2 = tau_2

    def dgamma_isgw2(self, w: float, mB: float, mD: float):
        """
                Differential decay rate for B->D1lnu with the ISGW2 model.
                :return:
                """
        q2 = mB**2 + mD**2 - 2*mB*mD*w
        trueW = w

        # get R
        mcR = 2.0 * np.sqrt(mB * mD) / (mB + mD)

        # get r
        r = mD / mB

        # get FF

        t = q2
        mass = mD

        msb = 5.2
        msd = 0.33
        bb2 = 0.431 * 0.431
        mbb = 5.31
        nf = 4.0

        msq = 1.82
        bx2 = 0.33 * 0.33
        mbx = (5.0 * 2.46 + 3.0 * 2.42) / 8.0
        nfp = 3.0

        mtb = msb + msd
        mtx = msq + msd
        mb = mB
        mx = mass

        mup = 1.0 / (1.0 / msq + 1.0 / msb)
        mum = 1.0 / (1.0 / msq - 1.0 / msb)
        bbx2 = 0.5 * (bb2 + bx2)
        tm = (mb - mx) * (mb - mx)
        if t > tm:
            t = 0.99 * tm
        wt = 1.0 + (tm - t) / (2.0 * mbb * mbx)

        mqm = 0.1
        r2 = 3.0 / (4.0 * msb * msq) + 3 * msd * msd / (2 * mbb * mbx * bbx2) + (
                16.0 / (mbb * mbx * (33.0 - 2.0 * nfp))) * np.log(
            self._get_etas(mqm, mqm) / self._get_etas(msq, msq))

        f5 = np.sqrt(mtx / mtb) * (np.sqrt(bx2 * bb2) / bbx2) ** (5.0 / 2.0) / ((1.0 + r2 * (tm - t) / 18.0) ** 3.0)

        f5v = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** (-0.5)
        f5r = f5 * (mbb / mtb) ** 0.5 * (mbx / mtx) ** 0.5
        f5sppsm = f5 * (mbb / mtb) ** (-1.5) * (mbx / mtx) ** 0.5
        f5spmsm = f5 * (mbb / mtb) ** (-0.5) * (mbx / mtx) ** (-0.5)

        if msq == msd:
            vv = f5v * (((mtb * np.sqrt(bb2)) / (4.0 * np.sqrt(2.0) * msb * msq * mtx)) + (
                    ((wt - 1) * msd) / (6.0 * np.sqrt(2.0 * bb2) * mtx)))

            rr = f5r * mtb * np.sqrt(bb2 / 2) * ((1.0 / mup) + ((msd * mtx * (wt - 1) * (wt - 1)) / (3.0 * msq * bb2)))

            sppsm = msd * f5sppsm / (np.sqrt(2.0 * bb2) * mtb) * (1.0 - (msd / msq) + ((msd * bb2) /
                                                                                       (2.0 * mup * bbx2)))

            spmsm = msd * f5spmsm / (np.sqrt(2.0 * bb2) * msq) * (((4 - wt) / 3.0) - ((msd * msq * bb2) /
                                                                                      (2.0 * mtx * mup * bbx2)))

        else:
            vv = -1.0 * msd * f5v / (2.0 * np.sqrt(3.0 * bb2) * mtx) * (
                    (wt + 1) / 2.0 + bb2 * mtb / (2.0 * msd * msq * msb))

            rr = -2.0 * mtb * np.sqrt(bb2 / 3.0) * f5r * (1.0 / msq + mtx * msd * (wt - 1) / (2.0 * bb2) *
                                                          ((wt + 1) / (2.0 * msq) - msd * bb2 / (
                                                                  2.0 * mum * mtx * bbx2)))

            sppsm = -1.0 * np.sqrt(3.0) * msd * f5sppsm / (2.0 * np.sqrt(bb2) * mtb) * (1 - msd / (3.0 * msq) -
                                                                                        msd * bb2 / (3.0 * bbx2) * (
                                                                                                1.0 / (
                                                                                                2.0 * mum) - 1.0 / mup))

            spmsm = -1.0 * msd * f5spmsm / (2.0 * np.sqrt(3.0 * bb2) * mtx) * ((2 - wt) * mtx / msq +
                                                                               msd * bb2 / bbx2 * (1.0 / (
                            2.0 * mum) - 1.0 / mup))

        vf = vv

        rf = rr
        spf = (sppsm + spmsm) / 2.0
        smf = (sppsm - spmsm) / 2.0
        conv = mcR * mB * (1.0 + r)

        fA = -conv * vf
        fV1 = 2.0 * rf / conv
        fV3rfV2 = conv * spf

        r_tm1 = (trueW - r) * fV1 + (trueW * trueW - 1.0) * fV3rfV2
        tm1 = r_tm1 * r_tm1
        tm2 = 2.0 * (1.0 - 2.0 * r * trueW + r * r) * (fV1 * fV1 + (trueW * trueW - 1.0) * fA * fA)

        oldFF = tm1 + tm2
        oldFF *= np.sqrt(trueW * trueW - 1.0)

        return oldFF

    def dgamma_llsw(self, w: float, mB: float, mD: float):
        r = mD / mB

        fv1 = self.fV1D1(w)
        fv2 = self.fV2D1(w)
        fv3 = self.fV3D1(w)
        fa = self.fAD1(w)

        fV3rfV2 = fv3 + r * fv2

        r_tm1 = (w - r) * fv1 + (w**2 - 1.0) * fV3rfV2
        tm1 = r_tm1 * r_tm1
        tm2 = 2.0 * (1.0 - 2.0 * r * w + r**2) * (fv1**2 + (w**2 - 1.0) * fa**2)
        newFF = tm1 + tm2

        newFF *= np.sqrt(w**2 - 1.0)

        return newFF

    def fV1D1(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        fv1 = 0

        fv1 += (1. - w ** 2) * t32
        fv1 -= eb * (w ** 2 - 1.) * ((self._Lp + self._L) * t32 - (
                    2. * w + 1) * self._tau_1 * t32 - self._tau_2 * t32 + self._eta_b)
        fv1 -= ec * (4. * (w + 1.) * (w * self._Lp - self._L) * t32 - (w ** 2 - 1) * (
                3. * self._tau_1 * t32 - 3. * self._tau_2 * t32 + 2 * self._eta_1 + 3 * self._eta_3))

        fv1 /= np.sqrt(6)

        return fv1

    def fV2D1(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        fv2 = 0

        fv2 -= 3. * t32
        fv2 -= 3 * eb * ((self._Lp + self._L) * t32 - (2 * w + 1) * self._tau_1 * t32 - self._tau_2 * t32 + self._eta_b)
        fv2 -= ec * ((4. * w - 1) * self._tau_1 * t32 + 5 * self._tau_2 * t32 + 10 * self._eta_1 + 4 * (
                    w - 1.) * self._eta_2 - 5 * self._eta_3)

        fv2 /= np.sqrt(6)

        return fv2

    def fV3D1(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        fv3 = 0

        fv3 += (w - 2.) * t32
        fv3 += eb * ((2. + w) * ((self._Lp + self._L) * t32 - (2. * w + 1.) * self._tau_1 * t32 - self._tau_2 * t32) - (
                    2. - w) * self._eta_b)
        fv3 += ec * (4. * (w * self._Lp - self._L) * t32 + (2. + w) * self._tau_1 * t32 + (
                    2. + 3. * w) * self._tau_2 * t32 - 2. * (
                             6. + w) * self._eta_1 - 4. *
                     (w - 1) * self._eta_2 - (3. * w - 2.) * self._eta_3)

        fv3 /= np.sqrt(6)

        return fv3

    def fAD1(self, w):
        ec = 1. / (2. * self._mc)
        eb = 1. / (2. * self._mb)

        t32 = self._isgur_wise_function(w)

        fa = 0

        fa += -(w + 1.) * t32
        fa -= eb * ((w - 1.) * ((self._Lp + self._L) * t32 - (2. * w + 1.) * self._tau_1 * t32 - self._tau_2 * t32) + (
                w + 1.) * self._eta_b)
        fa -= ec * (4. * (w * self._Lp - self._L) * t32 - 3. * (w - 1.) * (self._tau_1 * t32 - self._tau_2 * t32) + (
                w + 1.) * (
                            -2 * self._eta_1 - 3 * self._eta_3))

        fa /= np.sqrt(6)

        return fa

