"""
Simulate D^{**} events for the reweighting of the ISGW2_nominal to the LLSW_nominal decay model.
"""

import basf2
import b2biiConversion
import generators
import modularAnalysis
import os
from variables import variables


def add_alias(particle, observables):
    """Add alias to variables.variables in order to distinguish the different particle four vectors

    :param particle: String. Name of the particle.
    :param observables: List of strings. Variable names.
    :return: List of strings. Alias names.
    """
    lalias = []

    for observable in observables:
        alias = particle + '_' + observable

        if particle == 'B':
            corrected_observable = f'daughter(0, {observable})'
        elif particle == 'Dstarstar':
            corrected_observable = f'daughter(0, daughter(0, {observable}))'
        elif particle == 'l':
            corrected_observable = f'daughter(0, daughter(1, {observable}))'
        elif particle == 'nu':
            corrected_observable = f'daughter(0, daughter(2, {observable}))'
        elif particle == 'D':
            corrected_observable = f'daughter(0, daughter(0, daughter(0, {observable})))'
        elif particle == 'pi':
            corrected_observable = f'daughter(0, daughter(0, daughter(1, {observable})))'

        variables.addAlias(alias, corrected_observable)
        lalias.append(alias)

    return lalias


def get_four_vectors():
    """"Get the variable names for the four vectors necessary for the reweighting.
    :return: List of strings. """

    four_vectors = ['E', 'px', 'py', 'pz']
    lalias = []

    for particle in ['B', 'Dstarstar', 'l', 'nu', 'D', 'pi']:
        lalias.extend(add_alias(particle, four_vectors))

    return lalias


def generate_events(num_events, dec_file, output_file_name):
    """"Simulates events with given decay file and saves the root file to the output directory

    :param num_events: Int. Number of events to simulate.
    :param dec_file: String. Path to the evtgen decay file.
    :return: None
    """
    main_path = basf2.create_path()
    modularAnalysis.setupEventInfo(num_events, main_path)
    generators.add_evtgen_generator(main_path, 'signal',
                                    signaldecfile=dec_file)
    modularAnalysis.loadGearbox(main_path)
    modularAnalysis.fillParticleListFromMC('Upsilon(4S)', '', addDaughters=True, path=main_path)
    modularAnalysis.matchMCTruth('Upsilon(4S)', path=main_path)

    four_vectors = get_four_vectors()
    modularAnalysis.variablesToNtuple('Upsilon(4S)', four_vectors,
                                      filename=output_file_name, path=main_path)

    return main_path


def get_events_from_b2bii(input_file, output_file_name):
    """
    Get events from Belle MC using the b2bii conversion.
    :param input_file:
    :param output_file_name:
    :return:
    """
    os.environ['BELLE_POSTGRES_SERVER'] = 'can01'
    os.environ['USE_GRAND_REPROCESS_DATA'] = '1'
    os.environ['PGUSER'] = 'g0db'
    os.environ['BELLE2_FILECATALOG'] = 'NONE'

    own_db_path = '/home/belle2/metzner/databases/RDStar_b2bii/mc_beamparams/localdb/'

    basf2.reset_database()
    basf2.use_database_chain()

    basf2.use_local_database(
        os.path.join(own_db_path, 'database.txt'),
        own_db_path,
        False,
        basf2.LogLevel.WARNING
    )

    main_path = basf2.create_path()

    b2biiConversion.convertBelleMdstToBelleIIMdst(
        inputBelleMDSTFile=input_file,
        applyHadronBJSkim=True,
        useBelleDBServer=os.environ['BELLE_POSTGRES_SERVER'],
        path=main_path,
    )

    modularAnalysis.fillParticleListFromMC('Upsilon(4S)', '', addDaughters=True, path=main_path)
    modularAnalysis.matchMCTruth('Upsilon(4S)', path=main_path)

    four_vectors = get_four_vectors()
    modularAnalysis.variablesToNtuple('Upsilon(4S)', four_vectors,
                                      filename=output_file_name, path=main_path)

    return main_path
