"""
This script analyses the results for the B -> D** l nu form factor reweighting
from the ISGW2 to the LLSW decay model.
"""

import os
import re
import rdstar
import inspect
import logging
import numpy as np
import pandas as pd
from multiprocessing import pool

import matplotlib.pyplot as plt
import matplotlib
import b2luigi as luigi

matplotlib.use("Agg")

from analysistools.reweighting.rbf_reweighter import RbfReweighter

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors, xlabel_pos, ylabel_pos


logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

source_path = "/storage/9/welsch/reweighting"
isgw2_dir = os.path.join(source_path, "isgw2_data")
llsw_dir = os.path.join(source_path, "llsw_data")

analysis_dir = os.path.join(source_path, "analysis")
plot_dir = analysis_dir

pkg_dir = os.path.dirname(inspect.getfile(rdstar))
rw_source_path = os.path.join(pkg_dir, "data", "dstst_reweighting_data")

n = 1000000
sample_frac = 1.

w = HistVariable("w", 20, (1., 1.45), "w", "GeV")
costhetal = HistVariable("costhetal", 20, (-1., 1.), r"$\cos (\theta_l)$", "")
costhetav = HistVariable("costhetanu", 20, (-1., 1.), r"$\cos (\theta_V)$", "")
chi_prime = HistVariable("chi_prime", 20, (0, 2 * np.pi), r"$\chi '$", "")
weights = HistVariable("weights", 20, (-0.5, 3.5), "Weights", "")

rbf_arguments = {
    "Done": {
        "variables": ["w", "costhetal", "costhetanu"],
        "num_bins": [10, 10, 10],
        "limits": [(1., 1.42), (-1., 1.), (-1., 1.)]
    },
    "Dtwostar": {
        "variables": ["w", "costhetal"],
        "num_bins": [20, 20],
        "limits": [(1., 1.42), (-1., 1.)]
    },
    "Dprimeone": {
        "variables": ["w", "costhetal"],
        "num_bins": [20, 20],
        "limits": [(1., 1.425), (-1., 1.)]
    },
    "Dzerostar": {
        "variables": ["w"],
        "num_bins": [50],
        "limits": [(1., 1.425)]
    }
}

kit_colors = [
    KITColors.kit_purple,
    KITColors.kit_orange,
    KITColors.kit_maygreen,
    KITColors.kit_green,
    KITColors.kit_cyan,
    KITColors.kit_blue,
    KITColors.kit_yellow,
    KITColors.kit_red,
    KITColors.kit_black
]

set_matplotlibrc_params()

plot_titles = {
    "Done": r"$B\rightarrow D_1 \ell \nu$",
    "Dtwostar": r"$B\rightarrow D_2^* \ell \nu$",
    "Dprimeone": r"$B\rightarrow D_1^\prime \ell \nu$",
    "Dzerostar": r"$B\rightarrow D_0^\ast \ell \nu$",
}


def load_llsw_data_sample(path_to_hdf, sample=True):
    """
    Loads llsw HDF file and splits the dataframe into its components
    specified by the llsw_variations list.

    :param path_to_hdf: Absolute path to the llsw hdf file.

    :return: Dictionary of dataframes. Keys are the variation titles.
    """
    print(f"Loading file: {path_to_hdf}")
    data = pd.read_hdf(path_to_hdf, "df")
    variations = data["variation"].unique()

    if sample:
        return {variation: data.query(f"variation=='{variation}'").sample(n) for variation in variations}
    else:
        return {variation: data.query(f"variation=='{variation}'") for variation in variations}


def load_isgw2_data_sample(path_to_hdf, sample=True) -> pd.DataFrame:
    """
    Loads the specified HDF file

    :param path_to_hdf: Absolute path to the isgw2 hdf file.
    :return: A pandas.DataFrame.
    """
    print(f"Loading file: {path_to_hdf}")
    data = pd.read_hdf(path_to_hdf, "df")

    if sample:
        return data.sample(n)
    else:
        return data


def plot_llsw_variations(hist_variable, sample, ax):
    """
    Plots nominal LLSW mode and its variations on given axis.

    :param hist_variable: HistVariable instance.
    :param sample: Dictionary of pandas.DataFrames.
    :param ax: maptlolib.axis instance.
    :return: None
    """
    hw = SimpleHistogramPlot(hist_variable)

    for variation, color in zip(sample.keys(), kit_colors):
        hw.add_component(
            variation,
            sample[variation],
            color=color)
    hw.plot_on(ax, normed=True)
    ax.legend(frameon=True)


def plot_llsw_uncertainty(hist_variable, sample, ax):
    """
    Plots nominal LLSW mode and its variations on given axis.

    :param hist_variable: HistVariable instance.
    :param sample: Dictionary of pandas.DataFrames.
    :param ax: maptlolib.axis instance.
    :return: None
    """
    hw = SimpleHistogramPlot(hist_variable)
    hw.add_component(
        "LLSW nominal",
        sample["nominal"],
        color=KITColors.kit_red
    )
    hw.plot_on(ax, normed=True, draw_legend=False)
    h_nom, _ = np.histogram(sample[f"nominal"][hist_variable.df_label],
                            bins=hw.bin_edges,
                            density=True)
    errors_sq = np.zeros_like(h_nom)

    for i in range(int((len(sample.keys()) - 1) / 2)):
        h_up, _ = np.histogram(sample[f"up_{i}"][hist_variable.df_label],
                               bins=hw.bin_edges,
                               density=True)
        h_down, _ = np.histogram(sample[f"down_{i}"][hist_variable.df_label],
                                 bins=hw.bin_edges,
                                 density=True)
        errors_sq += get_diff_vec(h_nom, h_up, h_down) ** 2

    errors = np.sqrt(errors_sq)

    ax.bar(x=hw.bin_mids, height=2 * errors, width=hw.bin_width,
           label="sys. uncertainty", bottom=h_nom - errors,
           color=KITColors.kit_red, alpha=0.7, lw=0, fill=True)
    ax.legend(frameon=False, bbox_to_anchor=(1, 1))

    return ax


def get_diff_vec(h_nom, h_up, h_down):
    """
    Calculates the difference between a nominal histogram
    and its up and down variations. If the down variation has
    a higher value than the up variation, a relative minus sign
    is added to the difference. This is used for the construction
    of a covariance matrix as outer product of the difference
    vector, as is sets the correct sign for negatively correlated
    bins.

    :param h_nom: Nominal histogram bin counts.
    :param h_up: Up varied histogram bin counts.
    :param h_down: Down varied histogram bin counts.
    :return: A 'difference' vector.
    """
    sign = np.ones_like(h_up)
    mask = h_up < h_down
    sign[mask] = -1

    diff_up = np.abs(h_up - h_nom)
    diff_down = np.abs(h_down - h_nom)
    diff_sym = (diff_up + diff_down) / 2
    return sign * diff_sym


def compare_isgw2_and_llsw(hist_var, isgw2_df, llsw_df, ax1, ax2):
    """
    Plots histograms and the their ratio of the ISGW2 and LLSW
    model for the given variable. Also displays the systematic
    uncertaitnties of the LLSW model.

    :param hist_var: HistVariable instance.
    :param isgw2_df: A pandas.DataFrame
    :param llsw_df: Dictionary of pandas.DataFrames.
    :param ax1: maptlolib.axis instance.
    :param ax2: maptlolib.axis instance.
    :return: None
    """
    hw = SimpleHistogramPlot(hist_var)
    hw.add_component("LLSW nominal", llsw_df["nominal"], color=KITColors.kit_red)
    hw.add_component("ISGW2", isgw2_df, color=KITColors.kit_blue)
    hw.plot_on(ax1, normed=True, draw_legend=False)
    ax1.set_xlabel("")
    ax2.set_xlabel(hist_var.x_label, xlabel_pos)

    llsw_nom, _ = np.histogram(llsw_df[f"nominal"][hist_var.df_label], bins=hw.bin_edges, density=True)
    errors_sq = np.zeros_like(llsw_nom)

    for i in range(int((len(llsw_df.keys()) - 1) / 2)):
        llsw_up, _ = np.histogram(llsw_df[f"up_{i}"][hist_var.df_label], bins=hw.bin_edges, density=True)
        llsw_down, _ = np.histogram(llsw_df[f"down_{i}"][hist_var.df_label], bins=hw.bin_edges, density=True)
        errors_sq += get_diff_vec(llsw_nom, llsw_up, llsw_down) ** 2

    errors = np.sqrt(errors_sq)

    ax1.bar(x=hw.bin_mids, height=2 * errors, width=hw.bin_width,
            label="sys. uncertainty", bottom=llsw_nom - errors,
            color=KITColors.kit_red, alpha=0.7, lw=0, fill=True)

    if hist_var.df_label == "w":
        ax1.legend(frameon=False, loc="upper left", ncol=2)
        ylims = ax1.get_ylim()
        ax1.set_ylim(ylims[0], 1.25 * ylims[1])

    isgw2_bc, _ = np.histogram(isgw2_df[hist_var.df_label], bins=hw.bin_edges, density=True)

    ratio = (llsw_nom - isgw2_bc) / llsw_nom
    ratio_error = isgw2_bc / llsw_nom ** 2 * errors
    ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
    ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
    ax2.set_ylabel(r"$\frac{\mathrm{LLSW - ISGW2}}{\mathrm{LLSW}}$")
    ax2.set_ylim((-1, 1))


def create_hist_ratio_figure():
    """
    Create a matplotlib.Figure for histogram ratio plots.

    :return: A maptlotlib.Figure instance and a matplotlib.axes instance.
    """
    return plt.subplots(2, 1, figsize=(5, 5), dpi=200, sharex=True, gridspec_kw={"height_ratios": [3.5, 1]})


def compare_reweighed_and_llsw(hist_var, isgw2_df, isgw2_weights, llsw_df, ax1, ax2):
    """
    Plots histograms and the their ratio of the ISGW2 and LLSW
    model for the given variable. Also displays the systematic
    uncertaitnties of the LLSW model.

    :param hist_var: HistVariable instance.
    :param isgw2_df: A pandas.DataFrame
    :param isgw2_weights: Dicitionary of numpy.ndarrays.
    :param llsw_df: Dictionary of pandas.DataFrames.
    :param ax1: maptlolib.axis instance.
    :param ax2: maptlolib.axis instance.
    :return: None
    """
    hw = SimpleHistogramPlot(hist_var)
    hw.add_component("LLSW nominal", llsw_df["nominal"], color=KITColors.kit_red)
    hw.add_component("ISGW2 reweighted\nto LLSW", isgw2_df, weights=isgw2_weights["nominal"], color=KITColors.kit_green)
    hw.plot_on(ax1, normed=False, draw_legend=False)
    ax1.set_xlabel("")
    ax2.set_xlabel(hist_var.x_label, xlabel_pos)

    llsw_nom, _ = np.histogram(llsw_df[f"nominal"][hist_var.df_label], bins=hw.bin_edges)
    llsw_errors_sq = np.zeros_like(llsw_nom, dtype=float)

    for i in range(int((len(llsw_df.keys()) - 1) / 2)):
        llsw_up, _ = np.histogram(llsw_df[f"up_{i}"][hist_var.df_label], bins=hw.bin_edges)
        llsw_down, _ = np.histogram(llsw_df[f"down_{i}"][hist_var.df_label], bins=hw.bin_edges)
        variation_error_sq = get_diff_vec(llsw_nom, llsw_up, llsw_down) ** 2
        llsw_errors_sq += variation_error_sq

    llsw_errors = np.sqrt(llsw_errors_sq)

    isgw2_nom, _ = np.histogram(isgw2_df[hist_var.df_label], weights=isgw2_weights["nominal"],
                                bins=hw.bin_edges)
    isgw2_errors_sq = np.zeros_like(isgw2_nom)

    for i in range(int((len(llsw_df.keys()) - 1) / 2)):
        isgw2_up, _ = np.histogram(isgw2_df[hist_var.df_label],
                                   weights=isgw2_weights[f"up_{i}"],
                                   bins=hw.bin_edges)
        isgw2_down, _ = np.histogram(isgw2_df[hist_var.df_label],
                                     weights=isgw2_weights[f"down_{i}"],
                                     bins=hw.bin_edges)
        isgw2_errors_sq += get_diff_vec(isgw2_nom, isgw2_up, isgw2_down) ** 2
    isgw2_errors = np.sqrt(isgw2_errors_sq)

    ax1.bar(x=hw.bin_mids, height=2 * llsw_errors, width=hw.bin_width,
            label="sys. uncertainty", bottom=llsw_nom - llsw_errors,
            color=KITColors.kit_red, alpha=0.5, lw=0, fill=True)
    ax1.bar(x=hw.bin_mids, height=2 * isgw2_errors, width=hw.bin_width,
            label="sys. uncertainty", bottom=isgw2_nom - isgw2_errors,
            color=KITColors.kit_green, alpha=0.5, lw=0, fill=True)

    if hist_var.df_label == "w":
        ax1.legend(frameon=False, loc="upper left", ncol=2)
        ylims = ax1.get_ylim()
        ax1.set_ylim(ylims[0], 1.5 * ylims[1])

    ratio = (llsw_nom - isgw2_nom) / llsw_nom
    ratio_error = np.sqrt((isgw2_nom / llsw_nom ** 2) ** 2 * llsw_errors_sq + (1 / llsw_nom) ** 2 * isgw2_errors_sq)
    ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
    ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
    ax2.set_ylabel(r"$\frac{\mathrm{LLSW - ISGW2}}{\mathrm{LLSW}}$")
    ax2.set_ylim((-1, 1))


def compare_reweighed_and_isgw2(hist_var, isgw2_df, isgw2_weights, ax1):
    """
    Plots histograms and the their ratio of the ISGW2 and LLSW
    model for the given variable. Also displays the systematic
    uncertaitnties of the LLSW model.

    :param hist_var: HistVariable instance.
    :param isgw2_df: A pandas.DataFrame
    :param isgw2_weights: Dicitionary of numpy.ndarrays.
    :param llsw_df: Dictionary of pandas.DataFrames.
    :param ax1: maptlolib.axis instance.
    :param ax2: maptlolib.axis instance.
    :return: None
    """
    hw = SimpleHistogramPlot(hist_var)
    hw.add_component("ISGW2 reweighted\nto LLSW", isgw2_df, weights=isgw2_weights["nominal"], color=KITColors.kit_green)
    hw.add_component("ISGW2", isgw2_df, color=KITColors.kit_blue)
    hw.plot_on(ax1, normed=False, draw_legend=False)
    # ax1.set_xlabel("")
    ax1.set_xlabel(hist_var.x_label, xlabel_pos)

    isgw2_orig, _ = np.histogram(isgw2_df[hist_var.df_label], bins=hw.bin_edges)
    isgw2_orig_errors_sq = isgw2_orig

    isgw2_nom, _ = np.histogram(isgw2_df[hist_var.df_label], weights=isgw2_weights["nominal"],
                                bins=hw.bin_edges)
    isgw2_errors_sq = np.zeros_like(isgw2_nom)

    for i in range(int((len(isgw2_weights.keys()) - 1) / 2)):
        isgw2_up, _ = np.histogram(isgw2_df[hist_var.df_label],
                                   weights=isgw2_weights[f"up_{i}"],
                                   bins=hw.bin_edges)
        isgw2_down, _ = np.histogram(isgw2_df[hist_var.df_label],
                                     weights=isgw2_weights[f"down_{i}"],
                                     bins=hw.bin_edges)
        isgw2_errors_sq += get_diff_vec(isgw2_nom, isgw2_up, isgw2_down) ** 2
    isgw2_errors = np.sqrt(isgw2_errors_sq)

    ax1.bar(x=hw.bin_mids, height=2 * isgw2_errors, width=hw.bin_width,
            label="sys. uncertainty", bottom=isgw2_nom - isgw2_errors,
            color=KITColors.kit_green, alpha=0.5, lw=0, fill=True)

    if hist_var.df_label == "w":
        ax1.legend(frameon=False, loc="upper left", ncol=2)
        ylims = ax1.get_ylim()
        ax1.set_ylim(ylims[0], 1.25 * ylims[1])

    # ratio = (isgw2_orig - isgw2_nom) / isgw2_orig
    # ratio_error = np.sqrt(
    #     (isgw2_nom / isgw2_orig ** 2) ** 2 * isgw2_orig_errors_sq + (1 / isgw2_orig) ** 2 * isgw2_errors_sq)
    # ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
    # ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
    # ax2.set_ylabel(r"$\frac{\mathrm{ISGW2 - LLSW}}{\mathrm{ISGW2}}$")
    # ax2.set_ylim((-1, 1))


def plot_weights(rwweights, ax):
    """
    Plots a histogram of the given weights.

    :param rwweights: Array of weights.
    :param ax: A matplotlib.axis instance.
    :return: None
    """
    weight_var = HistVariable("weight", 50, (-0.5, 3.), r"$\mathrm{weight}$", "")
    hw = SimpleHistogramPlot(weight_var)
    hw.add_component("weights", rwweights, color=KITColors.kit_blue)
    hw.plot_on(ax, yaxis_scale=1.0, normed=True)


def rbf_helper(dstst_type, variation, isgw2_df, llsw_df):
    print(f"Creating RBF for {variation}")

    rbf_path = os.path.join(rw_source_path, f"rbf_{dstst_type}_{variation}.json")
    print(rbf_path)
    rbf = RbfReweighter.import_from_json(rbf_path)
    # rbf = RbfReweighter(**(rbf_arguments[dstst_type]))
    # rbf.create_interpolation(isgw2_df.sample(n), llsw_df.sample(n))
    # rbf.export_to_json(rbf_path)
    return variation, rbf


def weights_helper(dstst_type, variation, rbf, isgw2_df):
    print(f"Calculating weights for {variation}")
    return variation, rbf(*[isgw2_df[var].values for var in rbf._variables])


def create_rbf(dss_type, variation, isgw2_df, llsw_df, rbf_path):
    print(f"Creating RBF for {variation}")

    rbf = RbfReweighter(**(rbf_arguments[dss_type]))
    rbf.create_interpolation(isgw2_df.sample(n), llsw_df.sample(n))
    rbf.export_to_json(rbf_path)

    return variation, rbf


def apply_rbf(dstst_type, variation, rbf, isgw2_df):
    print(f"Calculating weights for {dstst_type}_{variation}")
    return variation, rbf(*[isgw2_df[var].values for var in rbf._variables])


class ExternalDataTask(luigi.ExternalTask):
    dss_type = luigi.Parameter()
    ff_model = luigi.Parameter()

    def output(self):
        return luigi.LocalTarget(os.path.join(source_path, f"{self.ff_model}_data", f"{self.dss_type}.h5"))


class GetLLSWVatiationsList(luigi.Task):
    dss_type = luigi.Parameter()

    def requires(self):
        return {
            "llsw": ExternalDataTask(dss_type=self.dss_type, ff_model="llsw")
        }

    def output(self):
        return luigi.LocalTarget(os.path.join(llsw_dir, f"{self.dss_type}_variations.txt"))

    def run(self):
        llsw_sample = load_llsw_data_sample(self.input()["llsw"].path)

        with self.output().open("w") as variation_list:
            for key in llsw_sample.keys():
                variation_list.write(key + "\n")


class RbfCreationTask(luigi.Task):
    dss_type = luigi.Parameter()

    def requires(self):
        return {
            "isgw2": ExternalDataTask(dss_type=self.dss_type, ff_model="isgw2"),
            "llsw": ExternalDataTask(dss_type=self.dss_type, ff_model="llsw"),
            "llsw_variation_list": GetLLSWVatiationsList(dss_type=self.dss_type)
        }

    def run(self):
        llsw_sample = load_llsw_data_sample(self.input()["llsw"].path, sample=False)
        isgw2_sample = load_isgw2_data_sample(self.input()["isgw2"].path, sample=False)

        args = [
            (self.dss_type, key, isgw2_sample, df, self.output()[key].path) for
            key, df in llsw_sample.items()
        ]

        pool.Pool(processes=4).starmap(
            rbf_helper,
            args
        )

    def output(self):
        with self.input()["llsw_variation_list"].open("r") as input_file:
            llsw_variation_list = input_file.read().splitlines()

        print(llsw_variation_list)
        return {key: luigi.LocalTarget(os.path.join(rw_source_path, f"rbf_{self.dss_type}_{key}.json")) for
                key in llsw_variation_list}


class WeightCalculationTask(luigi.Task):
    dss_type = luigi.Parameter()

    def requires(self):
        return {
            "isgw2": ExternalDataTask(dss_type=self.dss_type, ff_model="isgw2"),
            "rbfs": RbfCreationTask(dss_type=self.dss_type)
        }

    def output(self):
        return luigi.LocalTarget(os.path.join(isgw2_dir, f"{self.dss_type}_with_weights.h5"))

    def load_rbf(self):
        rbf_dict = {
            key: RbfReweighter.import_from_json(target.path) for key, target in self.input()["rbfs"].items()
        }
        return rbf_dict

    def run(self):
        isgw2_sample = load_isgw2_data_sample(self.input()["isgw2"].path)
        rbf_dict = self.load_rbf()

        weights_res = pool.Pool(processes=4).starmap(
            apply_rbf,
            [(self.dss_type, key, rbf, isgw2_sample) for key, rbf in rbf_dict.items()])

        for var, weight in weights_res:
            isgw2_sample[var] = weight

        isgw2_sample.to_hdf(self.output().path, "with_weights")


class LLSWComparisonPlots(luigi.Task):
    dss_type = luigi.Parameter()
    plot_dir = os.path.join(plot_dir, "llsw_variation_comparison")

    def requires(self):
        return {
            "llsw": ExternalDataTask(dss_type=self.dss_type, ff_model="llsw")
        }

    def output(self):
        plots = {}

        for hist_var in [w, costhetal, costhetav, chi_prime]:
            plots[hist_var.df_label] = luigi.LocalTarget(
                os.path.join(self.plot_dir, f"llsw_variations_{self.dss_type}_{hist_var.df_label}.pdf")
            )

        return plots

    def run(self):
        llsw_sample = load_llsw_data_sample(self.input()["llsw"].path)

        for hist_var in [w, costhetal, costhetav, chi_prime]:
            fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=400)
            plot_llsw_variations(hist_var, llsw_sample, ax)
            ax.set_title("LLSW", loc="left")
            ax.set_title(plot_titles[self.dss_type], loc="right")

            self.output()[hist_var.df_label].makedirs()
            fig.savefig(self.output()[hist_var.df_label].path, bbox_inches="tight")


class PreReweightingComparison(luigi.Task):
    dss_type = luigi.Parameter()
    plot_dir = os.path.join(plot_dir, "pre_reweighting_comparison")

    def requires(self):
        return {
            "isgw2": WeightCalculationTask(dss_type=self.dss_type),
            "llsw": ExternalDataTask(dss_type=self.dss_type, ff_model="llsw"),
            "variation_list": GetLLSWVatiationsList(dss_type=self.dss_type)
        }

    def output(self):

        plots = {}

        for hist_var in [w, costhetal, costhetav, chi_prime]:
            plots[hist_var.df_label] = luigi.LocalTarget(
                os.path.join(self.plot_dir, f"pre_reweighting_{self.dss_type}_{hist_var.df_label}.pdf")
            )

        return plots

    def run(self):
        llsw_sample = load_llsw_data_sample(self.input()["llsw"].path)
        isgw2_sample = pd.read_hdf(self.input()["isgw2"].path)

        # -- LLSW and ISGW2 comparison
        for hist_var in [w, costhetal, costhetav, chi_prime]:
            fig, ax = create_hist_ratio_figure()
            plt.subplots_adjust(hspace=0.1)
            compare_isgw2_and_llsw(hist_var, isgw2_sample, llsw_sample, ax[0], ax[1])
            ax[0].set_title(plot_titles[self.dss_type], loc="right")

            self.output()[hist_var.df_label].makedirs()
            fig.savefig(self.output()[hist_var.df_label].path, bbox_inches="tight")


@luigi.inherits(WeightCalculationTask)
@luigi.inherits(GetLLSWVatiationsList)
class PlotWeights(luigi.Task):
    plot_dir = os.path.join(plot_dir, "weights")

    def requires(self):
        return {
            "isgw2_with_weights": self.clone(WeightCalculationTask),
            "llsw_variation_list": self.clone(GetLLSWVatiationsList)
        }

    def load_llsw_variation_list(self):
        with self.input()["llsw_variation_list"].open("r") as input_file:
            llsw_variation_list = input_file.read().splitlines()
        return llsw_variation_list

    def output(self):
        llsw_variation_list = self.load_llsw_variation_list()

        targets = {}

        for variation in llsw_variation_list:
            targets[variation] = luigi.LocalTarget(
                os.path.join(self.plot_dir, f"isgw2_weights_{self.dss_type}_{variation}.pdf")
            )
        return targets

    def run(self):

        isgw2_sample = pd.read_hdf(self.input()["isgw2_with_weights"].path)
        llsw_variation_list = self.load_llsw_variation_list()

        for variation in llsw_variation_list:
            weights_arr = isgw2_sample[variation].values

            fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=400)
            # plot_weights(weights_arr, ax)
            bc, be, _ = ax.hist(weights_arr, bins=30, color=KITColors.kit_blue,
                                edgecolor=KITColors.kit_black, histtype="stepfilled")
            ax.set_yscale("log")
            bw = be[1] - be[0]
            mean = np.mean(weights_arr)
            median = np.median(weights_arr)
            if variation == "nominal":
                ax.set_xlabel("nominal Weight", **xlabel_pos)
            elif "down" in variation:
                num = int(re.findall(r"\d+", variation)[0])
                ax.set_xlabel(f"Weight Variation Down {num}", **xlabel_pos)
            elif "up" in variation:
                num = int(re.findall(r"\d+", variation)[0])
                ax.set_xlabel(f"Weight Variation Up {num}", **xlabel_pos)

            ax.axvline(mean, label=f"Mean: {mean:.2f}", color=KITColors.kit_orange)
            ax.axvline(median, label=f"Median: {median:.2f}", color=KITColors.kit_maygreen, ls="--")
            # ax.set_ylabel("Normalized in arb. units", **ylabel_pos)
            ax.set_ylabel(f"Events / ({bw:.3f})", **ylabel_pos)
            ax.set_title(plot_titles[self.dss_type], loc="right")
            # ax.set_title(variation, loc="left")
            ax.legend(frameon=False)

            self.output()[variation].makedirs()
            fig.savefig(self.output()[variation].path, bbox_inches="tight")


@luigi.inherits(WeightCalculationTask)
@luigi.inherits(GetLLSWVatiationsList)
class PlotReweightedvsLLSW(luigi.Task):
    plot_dir = os.path.join(plot_dir, "reweighted_vs_llsw")

    def requires(self):
        return {
            "isgw2_with_weights": self.clone(WeightCalculationTask),
            "llsw": ExternalDataTask(dss_type=self.dss_type, ff_model="llsw"),
            "llsw_variation_list": self.clone(GetLLSWVatiationsList)
        }

    def output(self):
        plots = {}
        for hist_var in [w, costhetal, costhetav, chi_prime]:
            plots[hist_var.df_label] = luigi.LocalTarget(
                os.path.join(self.plot_dir, f"reweighted_vs_llsw_{self.dss_type}_{hist_var.df_label}.pdf")
            )
        return plots

    def load_llsw_variation_list(self):
        with self.input()["llsw_variation_list"].open("r") as input_file:
            llsw_variation_list = input_file.read().splitlines()
        return llsw_variation_list

    def get_isgw2_weights(self, isgw2_sample):
        llsw_variation_list = self.load_llsw_variation_list()
        return {variation: isgw2_sample[variation].values for variation in llsw_variation_list}

    def run(self):
        isgw2_sample = pd.read_hdf(self.input()["isgw2_with_weights"].path)
        llsw_sample = load_llsw_data_sample(self.input()["llsw"].path)

        isgw2_weights = self.get_isgw2_weights(isgw2_sample)

        for hist_var in [w, costhetal, costhetav, chi_prime]:
            fig, ax = create_hist_ratio_figure()
            plt.subplots_adjust(hspace=0.1)
            compare_reweighed_and_llsw(hist_var, isgw2_sample,
                                       isgw2_weights, llsw_sample, ax[0], ax[1])
            ax[0].set_title(plot_titles[self.dss_type], loc="right")

            self.output()[hist_var.df_label].makedirs()
            fig.savefig(self.output()[hist_var.df_label].path, bbox_inches="tight")


@luigi.inherits(WeightCalculationTask)
@luigi.inherits(GetLLSWVatiationsList)
class PlotReweightedvsISGW2(luigi.Task):
    plot_dir = os.path.join(plot_dir, "reweighted_vs_isgw2")

    def requires(self):
        return {
            "isgw2_with_weights": self.clone(WeightCalculationTask),
            "llsw_variation_list": self.clone(GetLLSWVatiationsList)
        }

    def output(self):
        plots = {}
        for hist_var in [w, costhetal, costhetav, chi_prime]:
            plots[hist_var.df_label] = luigi.LocalTarget(
                os.path.join(self.plot_dir, f"reweighted_vs_isgw2_{self.dss_type}_{hist_var.df_label}.pdf")
            )
        return plots

    def load_llsw_variation_list(self):
        with self.input()["llsw_variation_list"].open("r") as input_file:
            llsw_variation_list = input_file.read().splitlines()
        return llsw_variation_list

    def get_isgw2_weights(self, isgw2_sample):
        llsw_variation_list = self.load_llsw_variation_list()
        return {variation: isgw2_sample[variation].values for variation in llsw_variation_list}

    def run(self):
        isgw2_sample = pd.read_hdf(self.input()["isgw2_with_weights"].path)
        isgw2_weights = self.get_isgw2_weights(isgw2_sample)

        for hist_var in [w, costhetal, costhetav, chi_prime]:
            # fig, ax = create_hist_ratio_figure()
            # plt.subplots_adjust(hspace=0.1)
            fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=400)
            compare_reweighed_and_isgw2(hist_var, isgw2_sample, isgw2_weights, ax)
            ax.set_title(plot_titles[self.dss_type], loc="right")

            self.output()[hist_var.df_label].makedirs()
            fig.savefig(self.output()[hist_var.df_label].path, bbox_inches="tight")


class ReweightingAnalysis(luigi.WrapperTask):
    dss_type = luigi.Parameter()

    def requires(self):
        yield PlotReweightedvsISGW2(self.dss_type)
        yield PlotReweightedvsLLSW(self.dss_type)
        yield PlotWeights(self.dss_type)
        yield PreReweightingComparison(self.dss_type)
        yield LLSWComparisonPlots(self.dss_type)


class AnalysisMaster(luigi.WrapperTask):

    def requires(self):
        for dstst_type in ["Done", "Dtwostar", "Dprimeone", "Dzerostar"]:
            yield ReweightingAnalysis(dss_type=dstst_type)


if __name__ == "__main__":
    luigi.process(AnalysisMaster(), workers=5)
