"""
This moodule contains utility functions for the calculations of the reweighting factors.

Maximilian Welsch 2018.
"""
import numpy as np
import pandas as pd

from analysistools.physics_vectors.fourvector import FourVector


def var_m2_dstst(df: pd.DataFrame) -> np.array:
    """
    Calculates the invariant mass of the D** meson.

    :param df:
    :return: Inv. Mass of D**
    """

    pDstarstar = FourVector(df[['Dstarstar_E', 'Dstarstar_px',
                                'Dstarstar_py', 'Dstarstar_pz']].values)

    return pDstarstar.M()

def var_mB(df: pd.DataFrame) -> np.array:
    """
    Calculates the invariant mass of the D** meson.

    :param df:
    :return: Inv. Mass of D**
    """

    pDstarstar = FourVector(df[['B_E', 'B_px',
                                'B_py', 'B_pz']].values)

    return pDstarstar.M()


def var_w(df: pd.DataFrame) -> np.array:
    """
    Calculates w, where
    w = (m_B^2 + m_{D^{**}}^2 - q`^2)/(2*m_B*m_{D^{**}}).
    :param df: DataFrame
    :return: w
    """
    pB = FourVector(df[['B_E', 'B_px', 'B_py', 'B_pz']].values)
    pDstarstar = FourVector(df[['Dstarstar_E', 'Dstarstar_px',
                                'Dstarstar_py', 'Dstarstar_pz']].values)

    q = pB - pDstarstar
    return (pB.M2() + pDstarstar.M2() - q.M2()) / (2 * pB.M() * pDstarstar.M())


def var_costhetal(df: pd.DataFrame) -> np.array:
    """
    Calculates \cos(\theta)_l, which is the angle between the lepton and the
    W boson in the W boson rest frame.
    :param df: DataFrame
    :return: \cos(\theta)_l
    """
    pB = FourVector(df[['B_E', 'B_px', 'B_py', 'B_pz']].values)
    pl = FourVector(df[['l_E', 'l_px', 'l_py', 'l_pz']].values)
    pnu = FourVector(df[['nu_E', 'nu_px', 'nu_py', 'nu_pz']].values)

    pW = pl + pnu
    # boost W and lepton in the CMS of the B meson
    bB = -pB.BoostVector()
    pl.Boost(bB)
    pW.Boost(bB)
    # boost lepton in the CMS of the W boson
    bW = -pW.BoostVector()
    pl.Boost(bW)
    return np.cos(pW.Angle(pl.Vec))


def var_costhetanu(df: pd.DataFrame) -> np.array:
    """
    Calculates \cos(\theta)_\nu, which is the angle between the D^{(*)}
    and the D^{**} in the D^{**} rest frame
    :param df: DataFrame
    :return: \cos(\theta)_\nu
    """
    pB = FourVector(df[['B_E', 'B_px', 'B_py', 'B_pz']].values)
    pDstarstar = FourVector(df[['Dstarstar_E', 'Dstarstar_px',
                                'Dstarstar_py', 'Dstarstar_pz']].values)
    pD = FourVector(df[['D_E', 'D_px', 'D_py', 'D_pz']].values)

    bB = -pB.BoostVector()
    pDstarstar.Boost(bB)
    pD.Boost(bB)
    bDstarstar = -pDstarstar.BoostVector()
    pD.Boost(bDstarstar)
    return np.cos(pDstarstar.Angle(pD.Vec))


def var_chi_prime(df: pd.DataFrame) -> np.array:
    """
    Calculates \chi^', which is the angle bewtween
    the decay plane of the D^{**} and the decay plane of the W boson.
    :param df: DataFrame
    :return: \chi^'
    """
    pl = FourVector(df[['l_E', 'l_px', 'l_py', 'l_pz']].values)
    pnu = FourVector(df[['nu_E', 'nu_px', 'nu_py', 'nu_pz']].values)
    pW = pl + pnu
    pDstarstar = FourVector(df[['Dstarstar_E', 'Dstarstar_px',
                                'Dstarstar_py', 'Dstarstar_pz']].values)

    pD = FourVector(df[['D_E', 'D_px', 'D_py', 'D_pz']].values)

    boostWCM = -pW.BoostVector()
    boostDstarstarCM = -pDstarstar.BoostVector()

    pD.Boost(boostDstarstarCM)
    pl.Boost(boostWCM)

    costhetav = (pD.X*pDstarstar.X + pD.Y*pDstarstar.Y +
                 pD.Z*pDstarstar.Z)/(pD.T*pDstarstar.T)

    costhetawl = (-pl.X*pDstarstar.X - pl.Y*pDstarstar.Y -
                  pl.Z*pDstarstar.Z)/(pl.T*pDstarstar.T)

    # actual chi calculation
    ltrans = pl.Vec - pW.Vec.Unit().Scale(costhetawl*pl.Vec.Perp())
    vtrans = pD.Vec - pDstarstar.Vec.Unit().Scale(costhetav*pD.Vec.Perp())

    chi = np.arctan2(ltrans.Cross(vtrans)*pW.Vec.Unit(), ltrans*vtrans)
    chi[chi < 0] += 2*np.pi

    return chi


def calculate_reweighting_variables(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates variables needed for reweighting and creates a new dataframe.
    :param df: Original dataframe
    :return: DataFrame with reweighting variables.
    """
    return pd.DataFrame({'w': var_w(df),
                         'costhetal': var_costhetal(df),
                         'costhetanu': var_costhetanu(df),
                         'chi_prime': var_chi_prime(df),
                         'mDss': var_m2_dstst(df),
                         'mB': var_mB(df),
                         '__weight__': np.ones((len(df.index)))
                         })
