import os
import numpy as np
import pandas as pd

from typing import Optional, List
from collections import defaultdict
from matplotlib import pyplot as plt

from matplotlib.offsetbox import AnchoredText

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors, xlabel_pos, ylabel_pos

from analysistools.batch_tools.file_handler import Belle_list_types

from rdstar.offline_analysis import rdstar_plots
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting

set_matplotlibrc_params()


def get_diff_vec(h_nom, h_up, h_down):
    sign = np.ones_like(h_up)
    mask = h_up < h_down
    sign[mask] = -1

    diff_up = np.abs(h_up - h_nom)
    diff_down = np.abs(h_down - h_nom)
    diff_sym = (diff_up + diff_down) / 2
    return sign * diff_sym


def compare_reweighed_var(hist_var, df, weight_columns, ax1, ax2=None):
    hw = SimpleHistogramPlot(hist_var)
    hw.add_component("after", df, weights=weight_columns["nominal"], color=KITColors.kit_green)
    hw.add_component("before", df, color=KITColors.kit_blue)
    hw.plot_on(ax1, normed=False, draw_legend=False)

    if ax2 is not None:
        ax1.set_xlabel("")
        ax2.set_xlabel(hist_var.x_label, xlabel_pos)

    orig, _ = np.histogram(df[hist_var.df_label], bins=hw.bin_edges)
    orig_errors_sq = orig

    new_nom, _ = np.histogram(df[hist_var.df_label], weights=df[weight_columns["nominal"]], bins=hw.bin_edges)
    new_errors_sq = np.zeros_like(new_nom)

    for i in range(int((len(weight_columns.keys()) - 1) / 2)):
        new_up, _ = np.histogram(df[hist_var.df_label], weights=df[weight_columns[f"up_{i}"]], bins=hw.bin_edges)
        new_down, _ = np.histogram(df[hist_var.df_label], weights=df[weight_columns[f"down_{i}"]], bins=hw.bin_edges)
        new_errors_sq += get_diff_vec(new_nom, new_up, new_down) ** 2
    new_errors = np.sqrt(new_errors_sq)

    ax1.bar(x=hw.bin_mids, height=2 * new_errors, width=hw.bin_width,
            label="reweighted\nsys. uncertainty", bottom=new_nom - new_errors,
            color=KITColors.kit_green, alpha=0.5, lw=0, fill=True)

    ax1.legend(frameon=True, loc='best')

    if ax2 is not None:
        ratio = (orig - new_nom) / orig
        ratio_error = np.sqrt(
            (new_nom / orig ** 2) ** 2 * orig_errors_sq + (1 / orig) ** 2 * new_errors_sq)
        ax2.axhline(y=0, color=KITColors.dark_grey, alpha=0.8)
        ax2.errorbar(hw.bin_mids, ratio, yerr=ratio_error, ls="", marker=".", color=KITColors.kit_black)
        ax2.set_ylabel(r"$\frac{\mathrm{before - after}}{\mathrm{before}}$")
        ax2.set_ylim((-1, 1))


def get_info_from_output_path(output_path):
    dir_path, file_name = os.path.split(output_path)
    base_name = file_name.split("_plot")[0]
    return dir_path, base_name


def map_weight_info_to_key(weight_info):
    if "nom" in weight_info:
        return "nominal"
    var_info = weight_info.split("_")
    if not len(var_info) == 3 or not var_info[1].isdigit():
        raise RuntimeError(f"Encountered unknown weight info: {weight_info}.")
    return f"{var_info[2]}_{var_info[1]}"


def map_decay_key_to_name(decay_key):
    decay_map = {
        "btod": "B to D l nu",
        "btodst": "B to D* l nu",
        "Dzerostar": "B to D*_0 l nu",
        "Dprimeone": "B to D'_1 l nu",
        "Done": "B to D_1 l nu",
        "Dtwostar": "B to D*_2 l nu",
    }
    return decay_map[decay_key]


def get_cols_and_decay(df):
    weight_col_dict = defaultdict(dict)

    for col in df.columns:
        if "weight_ff" in col:
            _, _, decay_key, weight_info = col.split("_", 3)
            weight_info_name = map_weight_info_to_key(weight_info)
            weight_col_dict[decay_key].update({weight_info_name: col})

    return weight_col_dict


def plot_ff_weight_distribution(df, condition, decay_key, nom_weight, weight_cols):
    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=200, facecolor='w', edgecolor='k')
    ax = plt.gca()

    hw = SimpleHistogramPlot(nom_weight)
    hw.add_component("", df.loc[condition], color=KITColors.kit_blue)
    hw.plot_on(ax, normed=False, draw_legend=False)
    ax.set_yscale("log")
    ax.set_title(label=f"Weight Distribution for {map_decay_key_to_name(decay_key)}", loc="center")

    info = f"mean: {df.loc[condition, weight_cols['nominal']].mean():2.3f}\n" \
        f"median: {df.loc[condition, weight_cols['nominal']].median():2.3f}\n" \
        f"min: {df.loc[condition, weight_cols['nominal']].min():2.3f}\n" \
        f"max: {df.loc[condition, weight_cols['nominal']].max():2.3f}"
    anchored_info_text = AnchoredText(info, loc='upper right')
    ax.add_artist(anchored_info_text)

    return fig, ax


def plot_ff_variable_distribution(df, condition, hist_var, decay_key, weight_cols):
    fig, ax = plt.subplots(2, 1, figsize=(6, 4.5), dpi=200, sharex="all", gridspec_kw={"height_ratios": [3.5, 1]})
    plt.subplots_adjust(hspace=0.1)

    compare_reweighed_var(hist_var=hist_var, df=df.loc[condition], weight_columns=weight_cols, ax1=ax[0], ax2=ax[1])

    title = f"Reweighting of {map_decay_key_to_name(decay_key)} in {hist_var.df_label}"
    ax[0].set_title(label=title, loc="center")

    return fig, ax


def plot_ff_var_vs_weight(df, condition, hist_var, nom_weight, decay_key, weight_cols):
    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=200, facecolor='w', edgecolor='k')
    ax = plt.gca()
    plt.hist2d(
        df.loc[condition, hist_var.df_label],
        df.loc[condition, weight_cols["nominal"]],
        bins=(40, 40),
        cmin=1
    )
    ax.set_title(label=fr"Weight vs {hist_var._var_name} for {map_decay_key_to_name(decay_key)}", loc="center")
    ax.set_xlabel(hist_var.x_label, xlabel_pos)
    ax.set_ylabel(nom_weight.x_label, ylabel_pos)
    plt.colorbar()

    return fig, ax


def translate_labels(column_name):
    if column_name.endswith("_nom"):
        return "Nominal"
    elif "_var_" in column_name:
        rest = column_name.split("_var_", 1)[1]
        if rest.count("_") >= 2:
            rest = rest.split("_", 1)[0]
        num, direction = rest.split("_", 1)
        assert num.isdigit(), num
        assert direction == "down" or direction == "up", direction
        return f"{direction.capitalize()} Variation {int(num) + 1}"
    else:
        RuntimeError(f"Unable to parse {column_name}")


def weight_box_plot(
        df,
        decay_key,
        whisker_quantile=0.0,
        plot_fliers=False,
        plot_mean=True,
        cols=None,
        dss_type=None,
        plot_zero_line=True,
):
    assert 0.0 <= whisker_quantile < 0.5, whisker_quantile

    dss_types = ["Dzerostar", "Dprimeone", "Done", "Dtwostar"]
    assert dss_type in dss_types or dss_type is None, dss_type
    if cols is None:
        if dss_type is not None:
            cols = [c for c in df.columns if ("_var_" in c or c.endswith("_nom")) and dss_type in c]
        else:
            cols = [c for c in df.columns if "_var_" in c or c.endswith("_nom")]
    assert len([t for t in dss_types if any(t in c for c in cols)]) <= 1
    assert len([c for c in cols if c.endswith("_nom")]) == 1

    if whisker_quantile == 0.:
        whiskers = "range"
    else:
        whiskers = [0. + 100. * whisker_quantile, 100. - 100. * whisker_quantile]

    fig = plt.figure(num=None, figsize=(8.5, 14), dpi=100, facecolor='w', edgecolor='k')
    ax = plt.gca()

    box_plot_props = {
        "widths": 0.7,
        "notch": False,
        "vert": False,
        "boxprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "whiskerprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "capprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "medianprops": dict(color=KITColors.kit_red, linewidth=1.4, linestyle="-", markerfacecolor=KITColors.kit_red,
                            markeredgecolor=KITColors.kit_red, marker="|"),
        "meanprops": dict(color=KITColors.kit_orange, linewidth=1.2, linestyle="--",
                          markerfacecolor=KITColors.kit_orange, markeredgecolor=KITColors.kit_orange, marker=""),
        "flierprops": dict(markerfacecolor=KITColors.kit_red, marker='+', markersize=3,
                           markeredgecolor=KITColors.kit_red)
    }

    res = plt.boxplot(
        x=[df[col].values for col in cols],
        whis=whiskers,
        showfliers=plot_fliers,
        labels=cols,
        showmeans=plot_mean,
        meanline=True,
        **box_plot_props
    )

    mean_x, median_x = None, None
    mean_objects = res["means"] if plot_mean else res["medians"]
    for data_index, (mean_obj, median_obj) in enumerate(zip(mean_objects, res["medians"])):
        if cols[data_index].endswith("_nom"):
            if plot_mean:
                mean_x = mean_obj._x[0]
            median_x = median_obj._x[0]
    if plot_mean:
        assert mean_x is not None
    assert median_x is not None

    if plot_mean:
        ax.vlines(x=mean_x, ymin=1.3, ymax=len(cols) + 0.5, color=KITColors.kit_orange, linewidth=0.8, linestyle="--",
                  label="Mean of Nominal")
    ax.vlines(x=median_x, ymin=1.3, ymax=len(cols) + 0.5, color=KITColors.kit_red, linewidth=0.8, linestyle="--",
              label="Median of Nominal")

    legend_object1 = ax.legend(frameon=True, loc=3)

    legend_lines = [
        plt.Line2D([0], [0], color=KITColors.kit_red, lw=2, ls="-"),
        plt.Line2D([0], [0], color=KITColors.kit_orange, lw=2, ls="-")
    ]

    if plot_mean:
        ax.legend(legend_lines, ["Medians", "Means"], frameon=True, loc=4)
    else:
        ax.legend(legend_lines[:1], ["Medians"], frameon=True, loc=4)

    ax.add_artist(legend_object1)

    x_lims = ax.get_xlim()
    scaling = (x_lims[1] - x_lims[0]) / 1000.

    if plot_zero_line and x_lims[0] < 0. < x_lims[1]:
        ax.vlines(x=0., ymin=0., ymax=len(cols) + 0.5, color=KITColors.grey, linewidth=0.8, linestyle="-")

    if whisker_quantile > 0. and not plot_fliers:
        min_caps = [c for i, c in enumerate(res["caps"]) if i % 2 == 0]
        max_caps = [c for i, c in enumerate(res["caps"]) if i % 2 == 1]
        for data_index, (min_cap, max_cap) in enumerate(zip(min_caps, max_caps)):
            col = cols[data_index]
            min_x = min_cap._x[0]
            max_x = max_cap._x[0]
            mean_y = np.mean(min_cap._y)
            min_val = df[col].min()
            max_val = df[col].max()

            ax.text(x=min_x - 66 * scaling, y=mean_y, s=r"$\leftarrow$", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 60 * scaling, y=mean_y, s=r"$\rightarrow$", fontsize=10, ha="left", va="center")
            ax.text(x=min_x - 16 * scaling, y=mean_y, s=f"min\n{min_val:.2f}", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 16 * scaling, y=mean_y, s=f"max\n{max_val:.2f}", fontsize=10, ha="left", va="center")

    ax.set_yticklabels([translate_labels(c) for c in cols])

    ax.set_ylim((-0.6 if plot_mean else -0.1, len(cols) + 1.4))
    if whisker_quantile > 0.:
        x_lims = ax.set_xlim(
            (x_lims[0] - 0.10 * abs(x_lims[1] - x_lims[0]), x_lims[1] + 0.10 * abs(x_lims[1] - x_lims[0])))
        info_text = f"Whiskers contain weight values between the {int(whiskers[0])}% and " \
            f"the {int(whiskers[1])}% quantiles."
    else:
        x_lims = ax.set_xlim(
            (x_lims[0] - 0.05 * abs(x_lims[1] - x_lims[0]), x_lims[1] + 0.05 * abs(x_lims[1] - x_lims[0])))
        info_text = f"Whiskers contain all weight values."

    ax.text(x=np.min(x_lims) + 0.5 * abs(x_lims[1] - x_lims[0]), y=len(cols) + 0.8, s=info_text,
            fontsize=12, ha="center", va="baseline")

    ax.set_title(f"Box Plots of {map_decay_key_to_name(decay_key)} Form Factor Weights")
    ax.set_xlabel("Weight Values")

    return fig, ax


def get_data_mask(df, col, quantile=0.05):
    assert 0. <= quantile < 0.5, quantile
    if quantile == 0.:
        return np.full(len(df.index), True, dtype=bool)
    return (df[col] > df[col].quantile(0. + quantile)) & (df[col] < df[col].quantile(1.0 - quantile))


def weight_violin_plot(df, decay_key, mask_quantile=0.05, points=60, cols=None, dss_type=None, plot_zero_line=True):
    dss_types = ["Dzerostar", "Dprimeone", "Done", "Dtwostar"]
    assert dss_type in dss_types or dss_type is None, dss_type
    if cols is None:
        if dss_type is not None:
            cols = [c for c in df.columns if ("_var_" in c or c.endswith("_nom")) and dss_type in c]
        else:
            cols = [c for c in df.columns if "_var_" in c or c.endswith("_nom")]
    assert len([t for t in dss_types if any(t in c for c in cols)]) <= 1
    assert len([c for c in cols if c.endswith("_nom")]) == 1

    fig = plt.figure(num=None, figsize=(8.5, 14), dpi=100, facecolor='w', edgecolor='k')
    ax = plt.gca()

    violin_plot_props = {
        "vert": False,
        "widths": 0.9,
        "points": points,
        "showmeans": True,
        "showextrema": True,
        "showmedians": True,
    }

    res = plt.violinplot(
        dataset=[df.loc[get_data_mask(df, col, quantile=mask_quantile), col].values for col in cols],
        positions=np.arange(0, len(cols)), **violin_plot_props)

    for vp_body in res["bodies"]:
        vp_body.set_color(KITColors.kit_blue)
        vp_body.set_alpha(0.9)

    res["cmedians"].set_color(KITColors.kit_red)
    res["cmeans"].set_color(KITColors.kit_orange)

    res["cbars"].set_color("none")

    res["cmins"].set_color(KITColors.kit_black)
    res["cmaxes"].set_color(KITColors.kit_black)

    x_lims = ax.get_xlim()
    scaling = (x_lims[1] - x_lims[0]) / 1000.

    if plot_zero_line and x_lims[0] < 0. < x_lims[1]:
        ax.vlines(x=0., ymin=-0.4, ymax=len(cols) - 0.5, color=KITColors.grey, linewidth=0.8, linestyle="-")

    mean_x, median_x = None, None
    for data_index, (mean_path, median_path) in enumerate(zip(res["cmeans"].get_paths(), res["cmedians"].get_paths())):
        if cols[data_index].endswith("_nom"):
            mean_x = mean_path.vertices[0][0]
            median_x = median_path.vertices[0][0]
    assert mean_x is not None and median_x is not None

    ax.vlines(x=mean_x, ymin=-0.4, ymax=len(cols) - 0.5, color=KITColors.kit_orange, linewidth=0.8, linestyle="--",
              label="Mean of Nominal")
    ax.vlines(x=median_x, ymin=-0.4, ymax=len(cols) - 0.5, color=KITColors.kit_red, linewidth=0.8, linestyle="--",
              label="Median of Nominal")

    if mask_quantile > 0.:
        for data_index, (min_paths, max_paths) in enumerate(zip(res["cmins"].get_paths(), res["cmaxes"].get_paths())):
            col = cols[data_index]
            min_x = min_paths.vertices[0][0]
            max_x = max_paths.vertices[0][0]
            mean_y = np.mean(min_paths.vertices[:, 1])
            min_val = df[col].min()
            max_val = df[col].max()

            ax.text(x=min_x - 66 * scaling, y=mean_y, s=r"$\leftarrow$", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 60 * scaling, y=mean_y, s=r"$\rightarrow$", fontsize=10, ha="left", va="center")
            ax.text(x=min_x - 16 * scaling, y=mean_y, s=f"min\n{min_val:.2f}", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 16 * scaling, y=mean_y, s=f"max\n{max_val:.2f}", fontsize=10, ha="left", va="center")

    ax.set_yticks(np.arange(0, len(cols)))
    ax.set_yticklabels([translate_labels(c) for c in cols])

    ax.set_ylim((-0.6, len(cols) + 0.6))
    if mask_quantile > 0.:
        x_lims = ax.set_xlim(
            (x_lims[0] - 0.10 * abs(x_lims[1] - x_lims[0]), x_lims[1] + 0.10 * abs(x_lims[1] - x_lims[0]))
        )
        info_text = f"Whiskers contain weight values between the {int(mask_quantile * 100)}% " \
            f"and the {int((1. - mask_quantile) * 100)}% quantiles."
    else:
        x_lims = ax.set_xlim(
            (x_lims[0] - 0.05 * abs(x_lims[1] - x_lims[0]), x_lims[1] + 0.05 * abs(x_lims[1] - x_lims[0]))
        )
        info_text = f"Whiskers contain all weight values."

    ax.text(x=np.min(x_lims) + 0.5 * abs(x_lims[1] - x_lims[0]), y=len(cols), s=info_text, fontsize=12,
            ha="center", va="baseline")
    legend_object1 = ax.legend(frameon=True, loc=3)

    legend_lines = [
        plt.Line2D([0], [0], color=KITColors.kit_red, lw=2, ls="-"),
        plt.Line2D([0], [0], color=KITColors.kit_orange, lw=2, ls="-")
    ]
    ax.legend(legend_lines, ["Medians", "Means"], frameon=True, loc=4)
    ax.add_artist(legend_object1)

    ax.set_title(f"Violin Plots of {map_decay_key_to_name(decay_key)} Form Factor Weights")
    ax.set_xlabel("Weight Values")

    return fig, ax


def plot_reweighted(
        df: pd.DataFrame,
        output_path: str,
        file_name_extension: Optional[str]=None,
        normal_condition=False,
        hist_vars: Optional[List[HistVariable]] = None
):
    """ Generates and saves different FF plots.

    :param df: Dataframe
    :param output_path: Output directory and name prefix
    :param file_name_extension: Prepended to the plot name
    :param normal_condition: If False and the sample is not in "Dzerostar", "Dprimeone", "Done", "Dtwostar",
        a default mask is applied to the dataframe (e.g. experiment number < 30, etc.).
        If True, the mask only selects the rows with non-zero weights.
    :param hist_vars: List of histvariables that are used for the plot_ff_variable_distribution
        plots. If None, a default setting is used.
    :return:
    """
    set_matplotlibrc_params()
    dir_path, base_name = get_info_from_output_path(output_path)
    if file_name_extension is not None:
        base_name = f"{base_name}_{file_name_extension}"

    if hist_vars is None:
        w = HistVariable("w", 20, (1., 1.45), "w", "GeV")
        costhetal = HistVariable("costhetal", 20, (-1., 1.), r"$\cos (\theta_l)$", "")
        costhetav = HistVariable("costhetav", 20, (-1., 1.), r"$\cos (\theta_v)$", "")
        chi_prime = HistVariable("chi_prime", 20, (0, 2 * np.pi), r"$\chi '$", "")

        hist_vars = [costhetal, costhetav, chi_prime, w]

    file_name = ""

    decay_and_weight_type_dict = get_cols_and_decay(df)

    for decay_key, weight_cols in decay_and_weight_type_dict.items():
        if normal_condition or any(c in decay_key for c in ["Dzerostar", "Dprimeone", "Done", "Dtwostar"]):
            main_cond = pd.notnull(df[weight_cols["nominal"]])
        else:
            e_cond = df.unique_event_id.str.startswith("Charged") | df.unique_event_id.str.startswith("Mixed")
            df.loc[~e_cond, "exp"] = df.loc[~e_cond].unique_event_id.str.split("_").str[1].astype("float")
            df.loc[e_cond, "exp"] = df.loc[e_cond].unique_event_id.str.split("_").str[2].astype("float")

            main_cond = ((pd.notnull(df[weight_cols["nominal"]]))
                         & (~((df.data_type.isin(Belle_list_types.keys())) & (df.exp < 30))))

        nom_weight = HistVariable(weight_cols["nominal"], 100, None, r"Nominal Weight", "")

        conditions = {"": main_cond}
        if BtoDstarFFReweighting.daughter2_pdg_info_col in df.columns:
            assert len(abs(df[BtoDstarFFReweighting.daughter2_pdg_info_col]).unique()) == 3
            assert all(pdg in [22, 111, 211] for pdg in abs(df[BtoDstarFFReweighting.daughter2_pdg_info_col]).unique())
            gamma_cond = abs(df[BtoDstarFFReweighting.daughter2_pdg_info_col]) == 22
            pi_cond = abs(df[BtoDstarFFReweighting.daughter2_pdg_info_col]).isin([111, 211])

            assert len(df.index) == (len(df.loc[gamma_cond].index) + len(df.loc[pi_cond].index))
            conditions.update({"gamma": (gamma_cond & main_cond), "pi": (pi_cond & main_cond)})

        for cond_key, cond in conditions.items():
            name_extension = "_Dstar_to_D_{cond_key}" if cond_key else ""

            for hist_var in hist_vars:
                # Plotting variable distribution
                file_name = f"{base_name}_{decay_key}{name_extension}_{hist_var.df_label}_plot"
                fig, _ = plot_ff_variable_distribution(df=df, condition=cond, hist_var=hist_var, decay_key=decay_key,
                                                       weight_cols=weight_cols)
                rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
                plt.close(fig)

                # Plotting 2D histograms of weight vs variable
                file_name = f"{base_name}_{decay_key}{name_extension}_weight_vs_{hist_var.df_label}_plot"
                fig, _ = plot_ff_var_vs_weight(df=df, condition=cond, hist_var=hist_var, nom_weight=nom_weight,
                                               decay_key=decay_key, weight_cols=weight_cols)
                rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
                plt.close(fig)

            # Producing box plots
            file_name = f"{base_name}_{decay_key}{name_extension}_weights_box_plot"
            fig, _ = weight_box_plot(df.loc[cond], decay_key, whisker_quantile=0.0, cols=list(weight_cols.values()),
                                     plot_mean=False)
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            file_name = f"{base_name}_{decay_key}{name_extension}_weights_1to99_quantile_w_fliers_box_plot"
            fig, _ = weight_box_plot(df.loc[cond], decay_key, whisker_quantile=0.01, cols=list(weight_cols.values()),
                                     plot_fliers=True, plot_mean=False)
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            file_name = f"{base_name}_{decay_key}{name_extension}_weights_1to99_quantile_box_plot"
            fig, _ = weight_box_plot(df.loc[cond], decay_key, whisker_quantile=0.01, cols=list(weight_cols.values()))
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            file_name = f"{base_name}_{decay_key}{name_extension}_weights_5to95_quantile_box_plot"
            fig, _ = weight_box_plot(df.loc[cond], decay_key, whisker_quantile=0.05, cols=list(weight_cols.values()))
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            # Producing violin plots
            file_name = f"{base_name}_{decay_key}{name_extension}_weights_violin_plot"
            fig, _ = weight_violin_plot(df.loc[cond], decay_key, mask_quantile=0.0, cols=list(weight_cols.values()))
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            file_name = f"{base_name}_{decay_key}{name_extension}_1to99_quantile_weights_violin_plot"
            fig, _ = weight_violin_plot(df.loc[cond], decay_key, mask_quantile=0.01,
                                        cols=list(weight_cols.values()))
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            file_name = f"{base_name}_{decay_key}{name_extension}_5to95_quantile_weights_violin_plot"
            fig, _ = weight_violin_plot(df.loc[cond], decay_key, mask_quantile=0.05,
                                        cols=list(weight_cols.values()))
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

            # Plotting weight distribution
            file_name = f"{base_name}_{decay_key}{name_extension}_weights_plot"
            fig, _ = plot_ff_weight_distribution(df=df, condition=cond, decay_key=decay_key, nom_weight=nom_weight,
                                                 weight_cols=weight_cols)
            rdstar_plots.export(fig=fig, filename=file_name, target_dir=dir_path)
            plt.close(fig)

    if not os.path.islink(output_path):
        os.symlink(os.path.join(dir_path, f"{file_name}.pdf"), output_path)
