"""
Tagging Correction for Belle with FEI

Felix Metzner 2019
"""

import numpy as np
import pandas as pd


class TagCorrection:
    tagging_correction_name = "__tagging_fei_corr_factor__"
    tagging_corr_uncert_name = "__tagging_fei_corr_uncert__"

    def __init__(self, correct_only_true_b: bool = False):
        self.u4s_decay_id_col = "extraInfo__bodecayModeID__bc"
        self.tag_b_pdg_col = "daughter__bo0__cm__spPDG__bc"
        self.tag_b_mc_pdg_col = "daughter__bo0__cm__spmatchedMC__boPDG__bc__bc"

        self.correct_only_true_b = correct_only_true_b

        self.charged_tag_ids = [1001, 1010, 1020]
        self.neutral_tag_ids = [1000, 1011, 1021]
        self.charged_PDG = 521
        self.neutral_PDG = 511

        # Correction Factors from Judith Schwab's KIT Master Thesis
        self.charged_factor = 0.810
        self.charged_uncert = np.sqrt(0.013 ** 2 + 0.054 ** 2)

        self.neutral_factor = 0.853
        self.neutral_uncert = np.sqrt(0.064 ** 2 + 0.058 ** 2)

    def apply_tagging_correction(self, sample: pd.DataFrame):
        assert self.u4s_decay_id_col in sample.columns, self.u4s_decay_id_col
        assert self.tag_b_pdg_col in sample.columns, self.tag_b_pdg_col
        assert self.tag_b_mc_pdg_col in sample.columns, self.tag_b_mc_pdg_col

        assert self.tagging_correction_name not in sample.columns, self.tagging_correction_name
        assert self.tagging_corr_uncert_name not in sample.columns, self.tagging_corr_uncert_name

        neutral_mask = sample[self.u4s_decay_id_col].isin(self.neutral_tag_ids)
        charged_mask = sample[self.u4s_decay_id_col].isin(self.charged_tag_ids)
        if self.correct_only_true_b:
            neutral_mask = neutral_mask & (abs(sample[self.tag_b_mc_pdg_col]) == self.neutral_PDG)
            charged_mask = charged_mask & (abs(sample[self.tag_b_mc_pdg_col]) == self.charged_PDG)

        sample.loc[:, self.tagging_correction_name] = 1.0
        sample.loc[:, self.tagging_corr_uncert_name] = 0.0

        sample.loc[neutral_mask, self.tagging_correction_name] = self.neutral_factor
        sample.loc[neutral_mask, self.tagging_corr_uncert_name] = self.neutral_uncert

        sample.loc[charged_mask, self.tagging_correction_name] = self.charged_factor
        sample.loc[charged_mask, self.tagging_corr_uncert_name] = self.charged_uncert

        assert len(sample[self.tagging_correction_name].unique()) == 2
        assert len(sample[self.tagging_corr_uncert_name].unique()) == 2
        assert self.charged_factor in sample[self.tagging_correction_name].unique()
        assert self.neutral_factor in sample[self.tagging_correction_name].unique()
        assert self.charged_uncert in sample[self.tagging_corr_uncert_name].unique()
        assert self.neutral_uncert in sample[self.tagging_corr_uncert_name].unique()
