"""
Missing Mass Squared Resolution Correction

Felix Metzner 2020
"""

import os
import json
import logging
import probfit
import iminuit
import numpy as np
import pandas as pd
import matplotlib as mpl

from matplotlib import pyplot as plt
from matplotlib.offsetbox import AnchoredText

from recordclass import RecordClass
from typing import Any, Union, Optional, Tuple, List, Dict, NamedTuple

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "MissingMassResolutionEvaluator",
    "MissingMassResolutionCorrector"
]

missing_mass_hist_var = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=250,
    scope=(-0.5, 2.),
    var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
    unit=r"$\mathrm{GeV}^2$",
    use_log_scale=False
)

mbc_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=100,
    scope=(5.270, 5.290),
    var_name=r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
    unit="GeV",
    use_log_scale=False
)

mbc_binning = Binning(
    bins=(5.2700, 5.2750, 5.2780, 5.2805, 5.2835, 5.290),
    dimensions=1,
    scope=(5.270, 5.290),
    log_scale=False
)


class FitInputData(NamedTuple):
    mc_miss_mass: np.ndarray
    mc_weights: np.ndarray
    mc_sig_miss_mass: np.ndarray
    mc_sig_weights: np.ndarray
    data_miss_mass: np.ndarray
    data_weights: Optional[np.ndarray] = None


class FitParameters(RecordClass):
    gauss_n_guess: float
    gauss_mean_guess: float
    gauss_sigma_guess: float
    cruijiff_n_guess: float
    cruijiff_mean_guess: float
    cruijiff_sigma_l_guess: float
    cruijiff_sigma_r_guess: float
    cruijiff_alpha_l_guess: float
    cruijiff_alpha_r_guess: float
    sig_cruijiff_n_guess: float
    sig_cruijiff_mean_guess: float
    sig_cruijiff_sigma_l_guess: float
    sig_cruijiff_sigma_r_guess: float
    sig_cruijiff_alpha_l_guess: float
    sig_cruijiff_alpha_r_guess: float

    @property
    def parameter_names(self) -> Tuple[str]:
        return tuple(self.__annotations__.keys())

    @property
    def as_dict(self) -> Dict[str, float]:
        return {name: getattr(self, name) for name in self.parameter_names}


class CruijiffFitResult(NamedTuple):
    n: float
    mean: float
    sigma_l: float
    sigma_r: float
    alpha_l: float
    alpha_r: float


class GaussFitResult(NamedTuple):
    n: float
    mean: float
    sigma: float


class FitOutput(NamedTuple):
    fit_model: probfit.BinnedLH
    minuit_instance_before_fit: iminuit.Minuit
    minuit_instance_after_fit: iminuit.Minuit
    gauss: Optional[GaussFitResult] = None
    cruijiff: Optional[CruijiffFitResult] = None


class MissingMassResolutionEvaluator:
    default_is_signal_col = "isSignalAcceptMissingNeutrino"
    default_result_file = "missing_mass_squared_resolution_eval_result"  # type: str

    fit_plot_base_name = "missing_mass_resolution_fit_plot"
    default_fig_size = (6, 4)  # type: Tuple[Union[int, float], Union[int, float]]

    parameter_name_dict = {  # type: Dict[str, str]
        "N_g": r"$N_\mathrm{Gauss}$",
        "m_g": r"$m_\mathrm{Gauss}$",
        "sigma_g": r"$\sigma_\mathrm{Gauss}$",
        "N_c": r"$N_\mathrm{Cruijiff}$",
        "m_c": r"$m_\mathrm{Cruijiff}$",
        "sigma_c_L": r"$\sigma^\mathrm{L}_\mathrm{Cruijiff}$",
        "sigma_c_R": r"$\sigma^\mathrm{R}_\mathrm{Cruijiff}$",
        "alpha_c_L": r"$\alpha^\mathrm{L}_\mathrm{Cruijiff}$",
        "alpha_c_R": r"$\alpha^\mathrm{R}_\mathrm{Cruijiff}$",
    }

    type_string_mapping = {  # type: Dict[str, str]
        "mc": "on MC",
        "data": "on Data",
        "signal": "on Signal MC",
    }

    def __init__(
            self,
            weight_column: str,
            results_path=Union[str, os.fspath],
            missing_mass: Optional[HistVariable] = None,
            secondary_variable: Tuple[Optional[HistVariable], Optional[Binning]] = (None, None),
            is_signal_column: Optional[str] = None,
            number_of_bins: int = 50,
            normalize: bool = False,
            result_file_name: Optional[str] = None,
            produce_plots: bool = True,
            fig_size: Optional[Tuple[Union[int, float], Union[int, float]]] = None
    ) -> None:
        self._weight_column = weight_column
        self._missing_mass = self._initialize_missing_mass(missing_mass=missing_mass)  # type: HistVariable
        self._secondary_variable, self._secondary_binning = self._initialize_secondary_var(secondary_variable)
        self._is_signal_column = self.default_is_signal_col if is_signal_column is None else is_signal_column
        self._number_of_bins = number_of_bins
        self._normalize = normalize
        self._result_file = self.default_result_file if result_file_name is None else result_file_name
        self._results_path = os.fspath(results_path)
        self._produce_plots = produce_plots
        self._fig_size = self.default_fig_size if fig_size is None else fig_size

    def evaluate_resolution(self, mc_df: pd.DataFrame, data_df: pd.DataFrame) -> Dict[str, os.fspath]:
        self._check_columns(mc_cols=list(mc_df.columns), data_cols=list(data_df.columns))

        mc_results = {}  # type: Dict[int, float]
        data_results = {}  # type: Dict[int, float]
        output_path_dict = {}  # type: Dict[str, os.fspath]

        new_out = self.produce_secondary_variable_plot(df=mc_df, dry_run=False)
        assert all(new_k not in output_path_dict for new_k in new_out.keys())
        output_path_dict.update(new_out)

        for bin_index, secondary_edges in enumerate(self.secondary_bin_edge_pairs):
            assert bin_index not in mc_results
            assert bin_index not in data_results

            miss_mass_mc_cond = self.get_missing_mass_condition(df=mc_df)
            miss_mass_data_cond = self.get_missing_mass_condition(df=data_df)
            is_sig_mc_cond = self.get_signal_condition(df=mc_df)
            secondary_mc_cond = self.get_secondary_condition(df=mc_df, secondary_edges=secondary_edges)
            secondary_data_cond = self.get_secondary_condition(df=data_df, secondary_edges=secondary_edges)

            mc_cond = miss_mass_mc_cond & secondary_mc_cond
            data_cond = miss_mass_data_cond & secondary_data_cond

            mc_miss_mass = mc_df.loc[mc_cond, self.missing_mass_col].values  # type: np.ndarray
            mc_weights = mc_df.loc[mc_cond, self.weight_col].values  # type: np.ndarray

            mc_sig_miss_mass = mc_df.loc[mc_cond & is_sig_mc_cond, self.missing_mass_col].values  # type: np.ndarray
            mc_sig_weights = mc_df.loc[mc_cond & is_sig_mc_cond, self.weight_col].values  # type: np.ndarray

            data_miss_mass = data_df.loc[data_cond, self.missing_mass_col].values  # type: np.ndarray
            data_weights = None  # type: Optional[np.ndarray]

            if self.normalize:  # TODO: Normalize does not work properly...
                mc_weights = mc_weights / np.sum(mc_weights)
                mc_sig_weights = mc_sig_weights / np.sum(mc_sig_weights)
                data_weights = np.ones_like(data_miss_mass) * 1. / data_miss_mass.size

            fit_data = FitInputData(
                mc_miss_mass=mc_miss_mass,
                mc_weights=mc_weights,
                mc_sig_miss_mass=mc_sig_miss_mass,
                mc_sig_weights=mc_sig_weights,
                data_miss_mass=data_miss_mass,
                data_weights=data_weights
            )

            mc_and_data_sigma, new_out = self.do_fit_routine(
                fit_data=fit_data,
                bin_index=bin_index if self.multiple_secondary_bins else None,
                dry_run=False
            )

            mc_results.update({bin_index: mc_and_data_sigma[0]})
            data_results.update({bin_index: mc_and_data_sigma[1]})
            assert all(new_k not in output_path_dict for new_k in new_out.keys())
            output_path_dict.update(new_out)

        new_out = self.write_result(mc_results=mc_results, data_results=data_results)
        assert all(new_k not in output_path_dict for new_k in new_out.keys())
        output_path_dict.update(new_out)
        return output_path_dict

    def get_output_dict(self) -> Dict[str, os.fspath]:
        output_dict = {}  # type: Dict[str, os.fspath]
        for bin_index, _ in enumerate(self.secondary_bin_edge_pairs):
            _, new_out = self.do_fit_routine(bin_index=bin_index, dry_run=True)
            assert all(new_k not in output_dict for new_k in new_out.keys())
            output_dict.update(new_out)

        new_out = self.produce_secondary_variable_plot(df=pd.DataFrame(), dry_run=True)
        assert all(new_k not in output_dict for new_k in new_out.keys())
        output_dict.update(new_out)

        res_key, res_path = self.result_file_key_path_pair
        output_dict.update({res_key: res_path})

        return output_dict

    def write_result(self, mc_results: Dict[int, float], data_results: Dict[int, float]) -> Dict[str, os.fspath]:
        result_dict = {  # type: Dict[str, Any]
            "secondary_var": self.secondary_col,
            "number_of_bins": self.number_of_secondary_bins,
            "bin_edge_pairs": self.secondary_bin_edge_pairs
        }

        sigmas = []
        for bin_index, secondary_edges in enumerate(self.secondary_bin_edge_pairs):
            assert bin_index in mc_results.keys()
            assert bin_index in data_results.keys()

            mc = mc_results[bin_index]
            data = data_results[bin_index]
            diff = abs(data - mc)

            result_dict.update({
                f"bin_{bin_index}": {
                    "index": bin_index,
                    "lower_bound": secondary_edges[0],
                    "upper_bound": secondary_edges[1],
                    "mc_sigma": mc,
                    "data_sigma": data,
                    "sigma_diff": diff
                }})

            sigmas.append(diff)

            logging.info(f"Bin {bin_index} ({secondary_edges}): "
                         f"mc={mc}, "
                         f"data={data} "
                         f"diff={diff}")

        result_dict.update({"sigmas": sigmas})

        assert len(result_dict.keys()) == self.number_of_secondary_bins + 4
        res_key, res_path = self.result_file_key_path_pair
        with open(res_path, "w") as result_fp:
            json.dump(result_dict, result_fp, indent=2)

        return {res_key: res_path}

    def do_fit_routine(
            self,
            fit_data: Optional[FitInputData] = None,
            bin_index: Optional[int] = None,
            dry_run: bool = False
    ) -> Tuple[Tuple[float, float], Dict[str, os.fspath]]:
        output_dict = {}  # type: Dict[str, os.fspath]
        if dry_run:
            sig_out = self.produce_before_after_plots(fit_type="signal", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in sig_out)
            output_dict.update(sig_out)
            mc_out = self.produce_before_after_plots(fit_type="mc", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in mc_out)
            output_dict.update(mc_out)
            d_out = self.produce_before_after_plots(fit_type="data", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in d_out)
            output_dict.update(d_out)
            return (0., 0.), output_dict
        else:
            assert fit_data is not None
            assert isinstance(fit_data, FitInputData)

        fit_parameters = FitParameters(
            gauss_n_guess=float(np.sum(fit_data.mc_sig_weights)),
            gauss_mean_guess=0.0,
            gauss_sigma_guess=0.05,
            cruijiff_n_guess=float(np.sum(fit_data.mc_weights) - np.sum(fit_data.mc_sig_weights)),
            cruijiff_mean_guess=-0.1,
            cruijiff_sigma_l_guess=0.1,
            cruijiff_sigma_r_guess=0.7,
            cruijiff_alpha_l_guess=0.7,
            cruijiff_alpha_r_guess=0.1,
            sig_cruijiff_n_guess=float(np.sum(fit_data.mc_sig_weights)),
            sig_cruijiff_mean_guess=0.0,
            sig_cruijiff_sigma_l_guess=0.03,
            sig_cruijiff_sigma_r_guess=0.06,
            sig_cruijiff_alpha_l_guess=0.15,
            sig_cruijiff_alpha_r_guess=0.15,
        )

        sig_fit_res = self.do_cruijiff_fit(  # type: FitOutput
            **{k.split("_", 1)[1]: v for k, v in fit_parameters.as_dict.items() if k.startswith("sig")},
            events=fit_data.mc_sig_miss_mass,
            event_weights=fit_data.mc_sig_weights
        )
        sig_out = self.produce_before_after_plots(fit_type="signal", fit_output=sig_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in sig_out)
        output_dict.update(sig_out)

        sig_sigma_guess = abs(sig_fit_res.cruijiff.sigma_l + sig_fit_res.cruijiff.sigma_r) / 2.
        fit_parameters.gauss_sigma_guess = sig_sigma_guess

        fit_parameters.sig_cruijiff_n_guess = sig_fit_res.cruijiff.n
        fit_parameters.sig_cruijiff_mean_guess = sig_fit_res.cruijiff.mean
        fit_parameters.sig_cruijiff_sigma_l_guess = sig_fit_res.cruijiff.sigma_l
        fit_parameters.sig_cruijiff_sigma_r_guess = sig_fit_res.cruijiff.sigma_r
        fit_parameters.sig_cruijiff_alpha_l_guess = sig_fit_res.cruijiff.alpha_l
        fit_parameters.sig_cruijiff_alpha_r_guess = sig_fit_res.cruijiff.alpha_r

        mc_fit_res = self.do_gauss_cruijiff_fit(  # type: FitOutput
            **{k: v for k, v in fit_parameters.as_dict.items() if not k.startswith("sig")},
            events=fit_data.mc_miss_mass,
            event_weights=fit_data.mc_weights
        )
        mc_out = self.produce_before_after_plots(fit_type="mc", fit_output=mc_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in mc_out)
        output_dict.update(mc_out)

        data_fit_res = self.do_gauss_cruijiff_fit(  # type: FitOutput
            **{k: v for k, v in fit_parameters.as_dict.items() if not k.startswith("sig")},
            events=fit_data.data_miss_mass,
            event_weights=fit_data.data_weights
        )
        d_out = self.produce_before_after_plots(fit_type="data", fit_output=data_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in d_out)
        output_dict.update(d_out)

        return (mc_fit_res.gauss.sigma, data_fit_res.gauss.sigma), output_dict

    def do_gauss_cruijiff_fit(
            self,
            gauss_n_guess: Union[float, int],
            gauss_mean_guess: float,
            gauss_sigma_guess: float,
            cruijiff_n_guess: Union[float, int],
            cruijiff_mean_guess: float,
            cruijiff_sigma_l_guess: float,
            cruijiff_sigma_r_guess: float,
            cruijiff_alpha_l_guess: float,
            cruijiff_alpha_r_guess: float,
            events: np.ndarray,
            event_weights: Optional[np.ndarray] = None
    ) -> FitOutput:
        if event_weights is not None:
            assert len(event_weights) == len(events), (len(event_weights), len(events))

        pdf1 = probfit.rename(probfit.gaussian, ('x', 'm_g', 'sigma_g'))
        pdf2 = probfit.rename(probfit.cruijff, ('x', 'm_c', 'sigma_c_L', 'sigma_c_R', 'alpha_c_L', 'alpha_c_R'))

        ext_pdf1 = probfit.Extended(pdf1, extname='N_g')
        ext_pdf2 = probfit.Extended(pdf2, extname='N_c')

        composite_pdf = probfit.AddPdf(ext_pdf1, ext_pdf2)

        fit_model = probfit.BinnedLH(f=composite_pdf, data=events, bins=self.bins, weights=event_weights, extended=True)
        before_fit, minuit_instance = self.create_minuit_instances(
            fit_model=fit_model,
            m_g=gauss_mean_guess,
            m_c=cruijiff_mean_guess,
            sigma_g=gauss_sigma_guess,
            sigma_c_L=cruijiff_sigma_l_guess,
            sigma_c_R=cruijiff_sigma_r_guess,
            alpha_c_L=cruijiff_alpha_l_guess,
            alpha_c_R=cruijiff_alpha_r_guess,
            N_g=gauss_n_guess,
            N_c=cruijiff_n_guess
        )

        minuit_instance.migrad()

        gauss_res = GaussFitResult(
            n=minuit_instance.values["N_g"],
            mean=minuit_instance.values["m_g"],
            sigma=minuit_instance.values["sigma_g"]
        )
        cruijiff_res = CruijiffFitResult(
            n=minuit_instance.values["N_c"],
            mean=minuit_instance.values["m_c"],
            sigma_l=minuit_instance.values["sigma_c_L"],
            sigma_r=minuit_instance.values["sigma_c_R"],
            alpha_l=minuit_instance.values["alpha_c_L"],
            alpha_r=minuit_instance.values["alpha_c_R"]
        )

        return FitOutput(
            fit_model=fit_model,
            minuit_instance_before_fit=before_fit,
            minuit_instance_after_fit=minuit_instance,
            gauss=gauss_res,
            cruijiff=cruijiff_res
        )

    def do_cruijiff_fit(
            self,
            cruijiff_n_guess: Union[float, int],
            cruijiff_mean_guess: float,
            cruijiff_sigma_l_guess: float,
            cruijiff_sigma_r_guess: float,
            cruijiff_alpha_l_guess: float,
            cruijiff_alpha_r_guess: float,
            events: np.ndarray,
            event_weights: Optional[np.ndarray] = None
    ) -> FitOutput:
        if event_weights is not None:
            assert len(event_weights) == len(events), (len(event_weights), len(events))

        pdf = probfit.rename(probfit.cruijff, ('x', 'm_c', 'sigma_c_L', 'sigma_c_R', 'alpha_c_L', 'alpha_c_R'))

        ext_pdf = probfit.Extended(pdf, extname='N_c')

        fit_model = probfit.BinnedLH(f=ext_pdf, data=events, bins=self.bins, weights=event_weights, extended=True)
        before_fit, minuit_instance = self.create_minuit_instances(
            fit_model=fit_model,
            m_c=cruijiff_mean_guess,
            sigma_c_L=cruijiff_sigma_l_guess,
            sigma_c_R=cruijiff_sigma_r_guess,
            alpha_c_L=cruijiff_alpha_l_guess,
            alpha_c_R=cruijiff_alpha_r_guess,
            N_c=cruijiff_n_guess
        )

        minuit_instance.migrad()

        cruijiff_res = CruijiffFitResult(
            n=minuit_instance.values["N_c"],
            mean=minuit_instance.values["m_c"],
            sigma_l=minuit_instance.values["sigma_c_L"],
            sigma_r=minuit_instance.values["sigma_c_R"],
            alpha_l=minuit_instance.values["alpha_c_L"],
            alpha_r=minuit_instance.values["alpha_c_R"]
        )

        return FitOutput(
            fit_model=fit_model,
            minuit_instance_before_fit=before_fit,
            minuit_instance_after_fit=minuit_instance,
            cruijiff=cruijiff_res
        )

    @staticmethod
    def create_minuit_instances(fit_model: probfit.BinnedLH, **kwargs) -> Tuple[iminuit.Minuit, iminuit.Minuit]:
        instance_1 = iminuit.Minuit(fit_model, **kwargs)
        instance_2 = iminuit.Minuit(fit_model, **kwargs)
        return instance_1, instance_2

    def produce_secondary_variable_plot(self, df: pd.DataFrame, dry_run: bool = False) -> Dict[str, os.fspath]:
        file_name = "missing_mass_resolution_correction_secondary_variable_binning"
        path = os.path.join(self.results_path, f"{file_name}.pdf")
        output_dict = {file_name: path}
        if dry_run:
            return output_dict

        shp = SimpleHistogramPlot(variable=self._secondary_variable)
        shp.add_component(
            label=self._secondary_variable.variable_name,
            data=df,
            weights=self.weight_col,
            color=plot_style.KITColors.kit_blue,
        )

        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()
        shp.plot_on(ax=ax, draw_legend=False, y_axis_scale=1.1)

        for a_bin_edge in self.secondary_bin_edges:
            ax.axvline(x=a_bin_edge, lw=2, color="black")

        plt.title(f"Binning in {self._secondary_variable.variable_name} "
                  f"for {self._missing_mass.variable_name} Resolution Correction")

        export(fig=fig, filename=file_name, target_dir=self.results_path)

        return output_dict

    def produce_before_after_plots(
            self,
            fit_type: str,
            fit_output: Optional[FitOutput] = None,
            bin_index: Optional[int] = None,
            dry_run: bool = False
    ) -> Dict[str, os.fspath]:
        if not self.produce_plots:
            return {}

        assert fit_type in self.type_string_mapping.keys()
        title_info = self.map_plot_title_info(type_string=fit_type)
        bin_info = "" if bin_index is None else f"_bin{bin_index}"
        bin_title_info = ""
        if bin_index is not None:
            sec_bin_edges = self.secondary_bin_edge_pairs[bin_index]
            edges_info = f"[{sec_bin_edges[0]:.3f}, {sec_bin_edges[1]:.3f}] {self._secondary_variable.unit}"
            bin_title_info = f" in Bin {bin_index + 1}/{self.number_of_secondary_bins} " \
                             rf" in {self._secondary_variable.variable_name} $\in$ {edges_info}"
        output_dict = {}

        b_fit_file_name = self.get_fit_plot_file_name(fitted=False, bin_info=bin_info, type_string=fit_type)
        output_dict.update({b_fit_file_name: os.path.join(self.results_path, f"{b_fit_file_name}.pdf")})

        a_fit_file_name = self.get_fit_plot_file_name(fitted=True, bin_info=bin_info, type_string=fit_type)
        output_dict.update({a_fit_file_name: os.path.join(self.results_path, f"{b_fit_file_name}.pdf")})

        residuals_file_name = self.get_residual_plot_file_name(bin_info=bin_info, type_string=fit_type)
        output_dict.update({residuals_file_name: os.path.join(self.results_path, f"{residuals_file_name}.pdf")})

        if dry_run:
            return output_dict
        else:
            assert fit_output is not None
            assert isinstance(fit_output, FitOutput)

        self.plot_probfit_status(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_before_fit,
            file_name=b_fit_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} Before Fit\n{bin_title_info}"
        )

        self.plot_probfit_status(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_after_fit,
            file_name=a_fit_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} After Fit\n{bin_title_info}"
        )

        self.plot_probfit_residuals(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_after_fit,
            file_name=residuals_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} Residuals\n{bin_title_info}"
        )

        return output_dict

    def plot_probfit_status(
            self,
            fit_model: probfit.BinnedLH,
            status: iminuit.Minuit,
            file_name: str,
            title: Optional[str] = None
    ) -> None:
        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()
        fit_model.draw(minuit=status, ax=ax, parts=True, print_par=False)
        info_text = self.get_fit_info_text(fit_model=fit_model, status=status)
        anchored_text = AnchoredText(info_text, loc="upper right", frameon=False)
        plt.gca().add_artist(anchored_text)

        ax.set_xlabel(self._missing_mass.x_label, plot_style.xlabel_pos)
        ax.set_ylabel(self.get_y_axis_label(), plot_style.ylabel_pos)

        if title is not None:
            plt.title(title)

        export(fig=fig, filename=file_name, target_dir=self.results_path)

    def plot_probfit_residuals(
            self,
            fit_model: probfit.BinnedLH,
            status: iminuit.Minuit,
            file_name: str,
            title: Optional[str] = None
    ) -> None:
        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()

        # Resetting color codes used by draw_residual method implemented in probfit
        mpl.colors.colorConverter.colors["b"] = plot_style.KITColors.kit_blue
        mpl.colors.colorConverter.colors["r"] = plot_style.KITColors.kit_red

        fit_model.draw_residual(minuit=status, ax=ax, norm=True)

        ax.set_xlabel(self._missing_mass.x_label, plot_style.xlabel_pos)
        ax.set_ylabel(r"(Fit$-$Data Points) / $\sigma_\mathrm{Fit}$", plot_style.ylabel_pos)

        if title is not None:
            plt.title(title)

        export(fig=fig, filename=file_name, target_dir=self.results_path)

    def get_fit_info_text(self, fit_model: probfit.BinnedLH, status: iminuit.Minuit) -> str:
        parameters = iminuit.util.describe(fit_model)
        assert all(p in self.parameter_name_dict.keys() for p in parameters), (parameters, self.parameter_name_dict)
        txt = ""
        for (param, value) in zip(parameters, status.args):
            txt += rf"{self.parameter_name_dict[param]} = {value:.4g}"
            if status.errors is not None:
                txt += rf"$\pm${status.errors[param]:.4g}"
            txt += '\n'
        return txt

    def get_y_axis_label(self) -> str:
        if self.normalize:
            return "Normalized in arb. units"
        else:
            return "Events / ({b:.2g}{v})".format(
                b=self.bin_width,
                v=" " + self._missing_mass.unit if self._missing_mass.unit else ""
            )

    def get_fit_plot_file_name(self, fitted: bool, bin_info: str, type_string: str) -> str:
        b_or_a = "after" if fitted else "before"
        return f"{self.fit_plot_base_name}{bin_info}_{type_string}_{b_or_a}_fit"

    def get_residual_plot_file_name(self, bin_info: str, type_string: str) -> str:
        return f"{self.fit_plot_base_name}{bin_info}_{type_string}_residuals"

    def map_plot_title_info(self, type_string: str) -> str:
        assert type_string in self.type_string_mapping.keys()
        return self.type_string_mapping[type_string]

    def get_missing_mass_condition(self, df: pd.DataFrame) -> pd.Series:
        return (df[self.missing_mass_col] >= self.missing_mass_bounds[0]) \
               & (df[self.missing_mass_col] <= self.missing_mass_bounds[1])

    def get_signal_condition(self, df: pd.DataFrame) -> pd.Series:
        return df[self.is_signal_col] == 1.

    def get_secondary_condition(self, df: pd.DataFrame, secondary_edges: Tuple[float, float]) -> pd.Series:
        assert isinstance(secondary_edges, tuple), type(secondary_edges)
        assert len(secondary_edges) == 2, (len(secondary_edges), secondary_edges)
        return (df[self.secondary_col] >= secondary_edges[0]) & (df[self.secondary_col] <= secondary_edges[1])

    @property
    def missing_mass_bounds(self) -> Tuple[float, float]:
        return self._missing_mass.scope

    @property
    def secondary_bin_edge_pairs(self) -> List[Tuple[float, float]]:
        bin_edges = self._secondary_binning.bin_edges[0]
        return [(left, right) for left, right in zip(bin_edges[:-1], bin_edges[1:])]

    @property
    def secondary_bin_edges(self) -> List[float]:
        return list(self._secondary_binning.bin_edges[0])

    @property
    def missing_mass_col(self) -> str:
        return self._missing_mass.df_label

    @property
    def secondary_col(self) -> str:
        return self._secondary_variable.df_label

    @property
    def weight_col(self) -> str:
        return self._weight_column

    @property
    def is_signal_col(self) -> str:
        return self._is_signal_column

    @property
    def bins(self) -> int:
        return self._number_of_bins

    @property
    def bin_width(self) -> float:
        window_width = self.missing_mass_bounds[1] - self.missing_mass_bounds[0]
        return window_width / self.bins

    @property
    def number_of_secondary_bins(self) -> int:
        return self._secondary_binning.num_bins[0]

    @property
    def multiple_secondary_bins(self) -> bool:
        return self.number_of_secondary_bins > 1

    @property
    def normalize(self) -> bool:
        return self._normalize

    @property
    def results_path(self) -> os.fspath:
        os.makedirs(self._results_path, exist_ok=True)
        return self._results_path

    @property
    def result_file_key_path_pair(self) -> Tuple[str, os.fspath]:
        key = self._result_file
        path = os.path.join(self.results_path, f"{self._result_file}.json")
        return key, path

    @property
    def fig_size(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self._fig_size

    @property
    def produce_plots(self) -> bool:
        return self._produce_plots

    def _check_columns(self, mc_cols: List[str], data_cols: List[str]):
        assert self.missing_mass_col in mc_cols
        assert self.missing_mass_col in data_cols

        assert self.secondary_col in mc_cols
        assert self.secondary_col in data_cols

        assert self.weight_col in mc_cols
        assert self.is_signal_col in mc_cols

    @staticmethod
    def _initialize_missing_mass(missing_mass: Optional[HistVariable]) -> HistVariable:
        if missing_mass is None:
            return missing_mass_hist_var
        else:
            return missing_mass

    @staticmethod
    def _initialize_secondary_var(
            secondary_variable: Tuple[Optional[HistVariable], Optional[Binning]]
    ) -> Tuple[HistVariable, Binning]:
        if secondary_variable[0] is None:
            var = mbc_hist_var
        else:
            var = secondary_variable[0]
        if secondary_variable[1] is None:
            binning = mbc_binning
        else:
            binning = secondary_variable[1]

        assert binning.dimensions == 1, binning.dimensions
        assert isinstance(binning.range, tuple), type(binning.range)
        assert len(binning.range) == 1, (binning.range, len(binning.range))
        assert isinstance(binning.range[0], tuple), (type(binning.range[0]), binning.range[0])

        assert var.scope == binning.range[0], (var.scope == binning.range)
        return var, binning


class ResolutionInfo(NamedTuple):
    number_of_bins: int
    sigmas: Tuple[float, ...]
    bin_edges: Optional[Tuple[Tuple[float, float], ...]] = None
    secondary_var: Optional[str] = None


class MissingMassResolutionCorrector:
    default_correction_col_tag = "corrected_resolution"

    def __init__(
            self,
            resolution_info: Union[os.fspath, str, Dict[str, Any]],
            original_column: str,
            new_column: Optional[str] = None
    ) -> None:
        self._base_resolution_info = self._init_info_dict(dict_input=resolution_info)
        self._original_col = original_column
        self._new_col = self.get_corrected_col_name(original_column) if new_column is None else new_column

        self._resolution_info = self._read_res_info(info_dict=self._base_resolution_info)  # type: ResolutionInfo

    def apply_resolution_correction(
            self,
            sample: pd.DataFrame,
            random_seed: int = 13374747,
            correct_overflows: bool = True
    ) -> None:
        assert self.original_column in sample.columns, self.original_column
        assert self.secondary_var_col in sample.columns, self.secondary_var_col

        assert self.new_column not in sample.columns, self.new_column

        np.random.RandomState(seed=random_seed)

        sample.loc[:, self.new_column] = np.nan

        for bin_index, (sigma, bin_edges) in enumerate(zip(self.sigmas, self.bin_edge_pairs)):
            if correct_overflows and (bin_index == 0 or bin_index == self.number_of_bins - 1):
                if self.number_of_bins == 1:
                    condition = pd.Series([True] * len(sample.index))
                    assert len(condition.index) == len(sample.index)
                elif bin_index == 0:
                    condition = sample[self.secondary_var_col] <= bin_edges[1]
                elif bin_index == self.number_of_bins - 1:
                    condition = sample[self.secondary_var_col] > bin_edges[0]
                else:
                    raise RuntimeError("This case should not be reached...")
            else:
                left_condition = sample[self.secondary_var_col] >= bin_edges[0]

                if bin_index == self.number_of_bins - 1:
                    # Special case for last bin
                    right_condition = sample[self.secondary_var_col] <= bin_edges[1]
                else:
                    right_condition = sample[self.secondary_var_col] < bin_edges[1]

                condition = left_condition & right_condition

            assert pd.isnull(sample.loc[condition, self.new_column]).all()
            sample.loc[condition, self.new_column] = np.random.normal(
                loc=sample.loc[condition, self.original_column].values,
                scale=sigma
            )

        if not correct_overflows:
            left_condition = sample[self.secondary_var_col] < self.bin_edge_pairs[0][0]
            right_condition = sample[self.secondary_var_col] > self.bin_edge_pairs[-1][1]
            condition = left_condition & right_condition
            assert pd.isnull(sample.loc[condition, self.new_column]).all()
            sample.loc[condition, self.new_column] = sample.loc[condition, self.original_column]

        assert not pd.isnull(sample[self.new_column]).any()

    @staticmethod
    def add_new_missing_mass_column_for_data(
            sample: pd.DataFrame,
            original_column: str,
            new_column: Optional[str] = None
    ) -> None:
        if new_column is None:
            new_column = MissingMassResolutionCorrector.get_corrected_col_name(original_column=original_column)

        assert original_column in sample.columns, original_column
        assert new_column not in sample.columns, new_column

        sample.loc[:, new_column] = sample[original_column].values
        assert new_column in sample.columns, new_column

    @property
    def original_column(self) -> str:
        return self._original_col

    @property
    def new_column(self) -> str:
        return self._new_col

    @property
    def secondary_var_col(self) -> Optional[str]:
        return self._resolution_info.secondary_var

    @property
    def bin_edge_pairs(self) -> Optional[Tuple[Tuple[float, float], ...]]:
        return self._resolution_info.bin_edges

    @property
    def sigmas(self) -> Tuple[float, ...]:
        return self._resolution_info.sigmas

    @property
    def number_of_bins(self) -> int:
        return self._resolution_info.number_of_bins

    @staticmethod
    def get_corrected_col_name(original_column: str) -> str:
        return f"{original_column}_{MissingMassResolutionCorrector.default_correction_col_tag}"

    @staticmethod
    def _read_res_info(info_dict: Dict[str, Any]) -> ResolutionInfo:
        number_of_bins = info_dict["number_of_bins"]
        sigmas = tuple(info_dict["sigmas"])
        bin_edges = tuple([tuple(entry) for entry in info_dict["bin_edge_pairs"]])

        assert len(sigmas) == number_of_bins, (number_of_bins, len(sigmas), sigmas)
        assert all(sig > 0. for sig in sigmas), sigmas
        assert len(bin_edges) == number_of_bins, (number_of_bins, len(bin_edges), bin_edges)
        assert all(isinstance(edges, tuple) for edges in bin_edges), bin_edges
        assert all(len(edges) == 2 for edges in bin_edges), bin_edges
        assert all(edges[0] < edges[1] for edges in bin_edges), bin_edges

        return ResolutionInfo(
            number_of_bins=number_of_bins,
            sigmas=sigmas,
            bin_edges=bin_edges,
            secondary_var=info_dict["secondary_var"]
        )

    @staticmethod
    def _init_info_dict(dict_input: Union[os.fspath, str, Dict[str, Any]]) -> Dict[str, Any]:
        if isinstance(dict_input, dict):
            return dict_input
        else:
            assert os.path.isfile(dict_input), dict_input
            with open(dict_input, 'r') as in_file:
                info_dict = json.load(in_file)
            return info_dict
