"""
This file contains definitions and functions for adding signal and
generator ids to monte carlo samples.
"""
import copy
import logging

import numpy as np
import pandas as pd

from typing import Union, Optional, Dict, List, Callable

logging.getLogger(__name__).addHandler(logging.NullHandler())


SigQueryDictType = Dict[str, Dict[str, Union[pd.Series, int]]]
GenQueryDictType = SigQueryDictType


__all__ = [
    "lepton_pdg_ids",
    "all_d_pdg_ids",
    "valid_forth_daughter",
    "rdstar_sig_queries",
    "sig_id_to_label",
    "decay_mode_to_sig_id",
    "decay_mode_to_latex_label",
    "add_sig_id_to_df",
    "gen_b_ids",
    "gen_id_to_label",
    "add_gen_b_id_to_df"
]


# general stuff
# =============

lepton_pdg_ids = [11, 13]  # type: List[int]
tau_pdg_id = 15  # type: int

d_pdg_ids = [411, 421]  # type: List[int]
dst_pdg_ids = [413, 423]  # type: List[int]

done_pdg_ids = [10413, 10423]  # type: List[int]
dtwostar_pdg_ids = [415, 425]  # type: List[int]
dzerostar_pdg_ids = [10411, 10421]  # type: List[int]
dprimeone_pdg_ids = [20413, 20423]  # type: List[int]

dstst_pdg_ids = done_pdg_ids + dtwostar_pdg_ids + dzerostar_pdg_ids + dprimeone_pdg_ids  # type: List[int]
all_d_pdg_ids = d_pdg_ids + dst_pdg_ids + dstst_pdg_ids  # type: List[int]

b_meson_pdg_ids = [511, 521]  # type: List[int]
valid_forth_daughter = [22, -999]  # type: List[int]

pizero_pdg_id = 111  # type: int
picharged_pdg_id = 211  # type: int


# signal ids
# ==========
def is_lepton_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the mcMatched(PDG) code of daughter 1 of the signal B meson
    matches a lepton pdg id.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    fourth_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc"

    is_lepton = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc).isin(lepton_pdg_ids)
    is_from_b = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_lepton & is_from_b & has_no_forth_daughter


def is_tau_to_lepton_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the mcMatched(PDG) code of daughter 1 of the signal B meson
    matches a lepton pdg id and if the mcMother(PDG) is a tau lepton.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    b_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__boPDG__bc__bc__bc__bc"
    fourth_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc__bc"

    is_lepton = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc).isin(lepton_pdg_ids)
    is_from_tau = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc) == tau_pdg_id
    is_from_b = abs(sample[b_pdg_var_name]).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_lepton & is_from_tau & is_from_b & has_no_forth_daughter


def is_tau_to_hadron_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the reconstructed first daughter of the B meson originates from a tau,
    but does not match a lepton PDG.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    lepton_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc"
    b_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__boPDG__bc__bc__bc__bc"
    fourth_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc__bc"

    is_not_lepton = ~abs(sample[lepton_pdg_var_name]).isin(lepton_pdg_ids)
    is_from_tau = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc) == tau_pdg_id
    is_from_b = abs(sample[b_pdg_var_name]).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_not_lepton & is_from_tau & is_from_b & has_no_forth_daughter


def b_to_dstst_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D**(->D(*)pi0)lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"
    d_daughter_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp" \
                              "mcDaughter__bo1__cm__spPDG__bc__bc__bc__bc__bc"
    dstst2_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp" \
                     "mcDaughter__bo2__cm__spPDG__bc__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_dstst = np.abs(sample[d_pdg_var_name]).isin(dstst_pdg_ids)
    is_dstst_with_pi_zero = np.abs(sample[d_daughter_pdg_var_name]) == pizero_pdg_id
    is_not_continuum = sample.isNotContinuumEvent == 1.

    if dstst2_pdg_var in sample.columns:
        dstst_has_no_second_daughter = sample[dstst2_pdg_var].isin(valid_forth_daughter)
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_zero & is_not_continuum & dstst_has_no_second_daughter
    else:
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_zero & is_not_continuum


def b_to_dstst_l_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D**(->D(*)pi+)lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"
    d_daughter_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp" \
                              "mcDaughter__bo1__cm__spPDG__bc__bc__bc__bc__bc"
    dstst2_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp" \
                     "mcDaughter__bo2__cm__spPDG__bc__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_dstst = np.abs(sample[d_pdg_var_name]).isin(dstst_pdg_ids)
    is_dstst_with_pi_charged = np.abs(sample[d_daughter_pdg_var_name]) == picharged_pdg_id
    is_not_continuum = sample.isNotContinuumEvent == 1.
    if dstst2_pdg_var in sample.columns:
        dstst_has_no_second_daughter = sample[dstst2_pdg_var].isin(valid_forth_daughter)
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_charged & is_not_continuum & dstst_has_no_second_daughter
    else:
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_charged & is_not_continuum


def b_to_d_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->Dlv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_d = np.abs(sample[d_pdg_var_name]).isin(d_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.

    return is_lepton & is_b_to_d & is_not_continuum


def b_to_dst_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D*lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_dst = np.abs(sample[d_pdg_var_name]).isin(dst_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.

    return is_lepton & is_b_to_dst & is_not_continuum


def b_to_d_tau_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->Dtv.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bo" \
                     "mcDaughter__bo0__cm__spPDG__bc__bc__bc__bc__bc"

    is_tau_to_lepton = is_tau_to_lepton_signal(sample)
    is_b_to_d = np.abs(sample[d_pdg_var_name]).isin(d_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.

    return is_tau_to_lepton & is_b_to_d & is_not_continuum


def b_to_dst_tau_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D*tv.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bo" \
                     "mcDaughter__bo0__cm__spPDG__bc__bc__bc__bc__bc"

    is_tau_to_lepton = is_tau_to_lepton_signal(sample)
    is_b_to_dst = np.abs(sample[d_pdg_var_name]).isin(dst_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.

    return is_tau_to_lepton & is_b_to_dst & is_not_continuum


def hadronic_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is not
    a lepton.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    l_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc"

    fake_lepton = ~np.abs(sample[l_pdg_var_name]).isin(lepton_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.

    return fake_lepton & is_not_continuum


def continuum_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the event is a continuum event.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    return sample.isContinuumEvent == 1.


def bbar_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if an event does not match any other checks above.
    This would be classified as BBbar background.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    not_dl = ~b_to_d_l_signal(sample)
    not_dstl = ~b_to_dst_l_signal(sample)
    not_dststl = ~b_to_dstst_l_signal(sample)
    not_dststl_bkg = ~b_to_dstst_l_bkg(sample)
    not_dtau = ~b_to_d_tau_signal(sample)
    not_dsttau = ~b_to_dst_tau_signal(sample)
    not_hadronic_bkg = ~hadronic_bkg(sample)
    not_continuum = ~continuum_bkg(sample)
    return (not_dl & not_dstl & not_dststl & not_dststl_bkg & not_dtau & not_dsttau
            & not_hadronic_bkg & not_continuum)


rdstar_sig_queries = {  # type: SigQueryDictType
    "Dtau": {"query": b_to_d_tau_signal, "id": 1},
    "DStau": {"query": b_to_dst_tau_signal, "id": 2},
    "Dl": {"query": b_to_d_l_signal, "id": 3},
    "DSl": {"query": b_to_dst_l_signal, "id": 4},
    "DSSl": {"query": b_to_dstst_l_signal, "id": 5},
    "DSSlbkg": {"query": b_to_dstst_l_bkg, "id": 6},
    "HBKG": {"query": hadronic_bkg, "id": 7},
    "GBKG": {"query": bbar_bkg, "id": 8},
    "CBKG": {"query": continuum_bkg, "id": 9}
}

sig_id_to_label = {  # type: Dict[int, str]
    1: r"$D \tau \nu$",
    2: r"$D^* \tau \nu$",
    3: r"$D \ell \nu$",
    4: r"$D^* \ell \nu$",
    5: r"$D^{**}(\rightarrow D^{(*)} \pi^0) \ell \nu$",
    6: r"$D^{**}(\rightarrow D^{(*)} \pi^\pm) \ell \nu$",
    7: r"Hadronic Bkg",
    8: r"$B\bar{B}$ Bkg",
    9: r"Continuum",
}

decay_mode_to_sig_id = {  # type: Dict[int, List[int]]
    11: [1, 3],  # r"$B^0_\mathrm{sig} \rightarrow D^- e^+$"
    12: [1, 3],  # r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+$"
    13: [1, 3],  # r"$B^-_\mathrm{sig} \rightarrow D^0 e^-$"
    14: [1, 3],  # r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^-$"
    15: [2, 4],  # r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+$"
    16: [2, 4],  # r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+$"
    17: [2, 4],  # r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^-$"
    18: [2, 4],  # r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^-$"
    61: [5],  # r"$B^0_\mathrm{sig} \rightarrow D^- e^+ \pi^0$"
    62: [5],  # r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+ \pi^0$"
    63: [5],  # r"$B^-_\mathrm{sig} \rightarrow D^0 e^- \pi^0$"
    64: [5],  # r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^- \pi^0$"
    65: [5],  # r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+ \pi^0$"
    66: [5],  # r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+ \pi^0$"
    67: [5],  # r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^- \pi^0$"
    68: [5],  # r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^- \pi^0$"
    71: [6],  # r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} e^+ \pi^-$"
    72: [6],  # r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} \mu^+ \pi^-$"
    73: [6],  # r"$B^-_\mathrm{sig} \rightarrow D^+ e^- \pi^-$"
    74: [6],  # r"$B^-_\mathrm{sig} \rightarrow D^+ \mu^- \pi^-$"
    75: [6],  # r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} e^+ \pi^-$"
    76: [6],  # r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} \mu^+ \pi^-$"
    77: [6],  # r"$B^-_\mathrm{sig} \rightarrow D^{*+} e^- \pi^-$"
    78: [6],  # r"$B^-_\mathrm{sig} \rightarrow D^{*+} \mu^- \pi^-$"
}

decay_mode_to_latex_label = {  # type: Dict[int, str]
    11: r"$B^0_\mathrm{sig} \rightarrow D^- e^+$",
    12: r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+$",
    13: r"$B^-_\mathrm{sig} \rightarrow D^0 e^-$",
    14: r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^-$",
    15: r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+$",
    16: r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+$",
    17: r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^-$",
    18: r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^-$",
    61: r"$B^0_\mathrm{sig} \rightarrow D^- e^+ \pi^0$",
    62: r"$B^0_\mathrm{sig} \rightarrow D^- \mu^+ \pi^0$",
    63: r"$B^-_\mathrm{sig} \rightarrow D^0 e^- \pi^0$",
    64: r"$B^-_\mathrm{sig} \rightarrow D^0 \mu^- \pi^0$",
    65: r"$B^0_\mathrm{sig} \rightarrow D^{*-} e^+ \pi^0$",
    66: r"$B^0_\mathrm{sig} \rightarrow D^{*-} \mu^+ \pi^0$",
    67: r"$B^-_\mathrm{sig} \rightarrow D^{*0} e^- \pi^0$",
    68: r"$B^-_\mathrm{sig} \rightarrow D^{*0} \mu^- \pi^0$",
    71: r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} e^+ \pi^-$",
    72: r"$B^0_\mathrm{sig} \rightarrow \overline{D^0} \mu^+ \pi^-$",
    73: r"$B^-_\mathrm{sig} \rightarrow D^+ e^- \pi^-$",
    74: r"$B^-_\mathrm{sig} \rightarrow D^+ \mu^- \pi^-$",
    75: r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} e^+ \pi^-$",
    76: r"$B^0_\mathrm{sig} \rightarrow \overline{D^{*0}} \mu^+ \pi^-$",
    77: r"$B^-_\mathrm{sig} \rightarrow D^{*+} e^- \pi^-$",
    78: r"$B^-_\mathrm{sig} \rightarrow D^{*+} \mu^- \pi^-$",
}


def add_sig_id_to_df_dict(
        df_dict: Dict[str, Dict[str, pd.DataFrame]],
        sig_query_dict: Optional[SigQueryDictType] = None
) -> None:
    """Adds signal ids to a dictionary of dictionary pandas.DataFrames.

    :param df_dict: Standard Dict of Dict of pd.DataFrames used in this analysis.
    :param sig_query_dict: rdstar_sig_queries Dict returned from get_rdstar_sig_queries.
        Default is mc_ids.rdstar_sig_queries.
    :return: None
    """
    if sig_query_dict is None:
        sig_query_dict = copy.deepcopy(rdstar_sig_queries)

    for data_type in df_dict.keys():
        logging.info(f"Adding signal ids for data type: {data_type}")

        add_sig_id_to_df(
            df=df_dict[data_type]["Y4S_ntuple"],
            sig_query_dict=sig_query_dict
        )


def add_sig_id_to_df(df: pd.DataFrame, sig_query_dict: Optional[SigQueryDictType] = None) -> None:
    """Adds signal ids to a pandas.DataFrame.

    :param df: pandas.DataFrame.
    :param sig_query_dict: A 'query' dict like the `rdstar_sig_queries` dict.
        Default is mc_ids.rdstar_sig_queries
    """
    if sig_query_dict is None:
        sig_query_dict = copy.deepcopy(rdstar_sig_queries)

    logging.info(f"Number of candidates: {len(df.index)}")

    sig_id_column_name = "SIG_ID"
    df.loc[:, sig_id_column_name] = np.nan

    for key, values in sig_query_dict.items():
        query_fcn = values["query"]
        query_result = query_fcn(df)

        logging.info(f"Processing query {key}...")
        logging.info(
            f"Number of matching candidates: {len(df.loc[query_result].index)}"
        )

        if not np.all(pd.isnull(df.loc[query_result, sig_id_column_name])):
            raise RuntimeWarning("Careful, you are overwriting singal ids")

        df.loc[query_result, sig_id_column_name] = values["id"]

    check = df[sig_id_column_name].isnull().sum()
    assert check == 0, f"Sig ID queries do not cover all events.\nNumber of Rows w/o ID: {check}"


# generator ids
# =============
def is_continuum_event(df: pd.DataFrame) -> pd.Series:
    return df["isContinuumEvent"] == 1.


def second_daughter_is_lepton(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"]).isin(
        lepton_pdg_ids)


def second_daughter_is_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"]) == tau_pdg_id


def second_daughter_is_tau_to_l(df: pd.DataFrame, b_index: int) -> pd.Series:
    l_pdg = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"

    return second_daughter_is_tau(df, b_index) & abs(df[l_pdg]).isin(lepton_pdg_ids)


def second_daughter_is_tau_to_had(df: pd.DataFrame, b_index: int) -> pd.Series:
    l_pdg = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"

    return second_daughter_is_tau(df, b_index) & ~(abs(df[l_pdg]).isin(lepton_pdg_ids))


def first_daughter_is_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(d_pdg_ids)


def first_daughter_is_Dst(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dst_pdg_ids)


def first_daughter_is_Done(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        done_pdg_ids)


def first_daughter_is_Dtwostar(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dtwostar_pdg_ids)


def first_daughter_is_Dprimeone(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dprimeone_pdg_ids)


def first_daughter_is_Dzerostar(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dzerostar_pdg_ids)


def first_daughter_is_any_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        all_d_pdg_ids)


def first_daughter_is_not_any_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return ~abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        all_d_pdg_ids)


def b_has_three_daughters(df: pd.DataFrame, b_index: int) -> pd.Series:
    # allow radiative photons als fourth daughter
    return df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo3__cmPDG__bc__bc__bc"].isin([22, -999])


def gen_BtoDtau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_d = first_daughter_is_D(df, b_index)
    has_tau_to_l = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_d & has_tau_to_l & has_three_daughters


def gen_BtoDsttau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dst = first_daughter_is_Dst(df, b_index)
    has_tau_to_l = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dst & has_tau_to_l & has_three_daughters


def gen_BtoDl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_d = first_daughter_is_D(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_d & has_l & has_three_daughters


def gen_BtoDstl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dst = first_daughter_is_Dst(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dst & has_l & has_three_daughters


def gen_BtoDonel(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_done = first_daughter_is_Done(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_done & has_l & has_three_daughters


def gen_BtoDtwostarl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dtwostar = first_daughter_is_Dtwostar(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dtwostar & has_l & has_three_daughters


def gen_BtoDprimeonel(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dprimeone = first_daughter_is_Dprimeone(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dprimeone & has_l & has_three_daughters


def gen_BtoDzerostarl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dzerostar = first_daughter_is_Dzerostar(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dzerostar & has_l & has_three_daughters


def gen_BtoDonetau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_done = first_daughter_is_Done(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_done & has_tau & has_three_daughters


def gen_BtoDtwostartau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dtwostar = first_daughter_is_Dtwostar(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dtwostar & has_tau & has_three_daughters


def gen_BtoDprimeonetau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dprimeone = first_daughter_is_Dprimeone(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dprimeone & has_tau & has_three_daughters


def gen_BtoDzerostartau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dzerostar = first_daughter_is_Dzerostar(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dzerostar & has_tau & has_three_daughters


def gen_hadronic_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_any_d = first_daughter_is_any_D(df, b_index)
    has_tau_to_had = second_daughter_is_tau_to_had(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_any_d & has_tau_to_had & has_three_daughters


def gen_other(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    not_gen_BtoDl = ~gen_BtoDl(df, b_index)
    not_gen_BtoDstl = ~gen_BtoDstl(df, b_index)
    not_gen_BtoDtau = ~gen_BtoDtau(df, b_index)
    not_gen_BtoDsttau = ~gen_BtoDsttau(df, b_index)
    not_gen_BtoDonel = ~gen_BtoDonel(df, b_index)
    not_gen_BtoDonetau = ~gen_BtoDonetau(df, b_index)
    not_gen_BtoDtwostarl = ~gen_BtoDtwostarl(df, b_index)
    not_gen_BtoDtwostartau = ~gen_BtoDtwostartau(df, b_index)
    not_gen_BtoDprimeonel = ~gen_BtoDprimeonel(df, b_index)
    not_gen_BtoDprimeonetau = ~gen_BtoDprimeonetau(df, b_index)
    not_gen_BtoDzerostarl = ~gen_BtoDzerostarl(df, b_index)
    not_gen_BtoDzerostartau = ~gen_BtoDzerostartau(df, b_index)
    not_gen_hadronic_tau = ~gen_hadronic_tau(df, b_index)

    return (not_continuum & not_gen_hadronic_tau &
            not_gen_BtoDl & not_gen_BtoDstl &
            not_gen_BtoDtau & not_gen_BtoDsttau &
            not_gen_BtoDonel & not_gen_BtoDonetau &
            not_gen_BtoDtwostarl & not_gen_BtoDtwostartau &
            not_gen_BtoDprimeonel & not_gen_BtoDprimeonetau &
            not_gen_BtoDzerostarl & not_gen_BtoDzerostartau)


def gen_continuum(df: pd.DataFrame, b_index: int) -> pd.Series:
    return is_continuum_event(df)


gen_b_ids = {  # type: GenQueryDictType
    "Dtau": {"query": gen_BtoDtau, "id": 1},
    "Dsttau": {"query": gen_BtoDsttau, "id": 2},
    "Dl": {"query": gen_BtoDl, "id": 3},
    "Dstl": {"query": gen_BtoDstl, "id": 4},
    "Donel": {"query": gen_BtoDonel, "id": 51},
    "Donetau": {"query": gen_BtoDonetau, "id": 55},
    "Dtwostarl": {"query": gen_BtoDtwostarl, "id": 52},
    "Dtwostartau": {"query": gen_BtoDtwostartau, "id": 56},
    "Dprimeonel": {"query": gen_BtoDprimeonel, "id": 53},
    "Dprimeonetau": {"query": gen_BtoDprimeonetau, "id": 57},
    "Dzerostarl": {"query": gen_BtoDzerostarl, "id": 54},
    "Dzerostartau": {"query": gen_BtoDzerostartau, "id": 58},
    "hadronic_tau": {"query": gen_hadronic_tau, "id": 6},
    "other": {"query": gen_other, "id": 7},
    "continuum": {"query": gen_continuum, "id": 8},
}

gen_id_to_label = {  # type: Dict[int, str]
    1: r"$D \tau \nu$",
    2: r"$D^* \tau \nu$",
    3: r"$D \ell \nu$",
    4: r"$D^* \ell \nu$",
    51: r"$D_1 \ell \nu$",
    52: r"$D_2^* \ell \nu$",
    53: r"$D'_1 \ell \nu$",
    54: r"$D_0^* \ell \nu$",
    55: r"$D_1 \tau \nu$",
    56: r"$D_2^* \tau \nu$",
    57: r"$D'_1 \tau \nu$",
    58: r"$D_0^* \tau \nu$",
    6: r"Hadronic $\tau$",
    7: r"Other",
    8: r"Continuum",
    -999: r"Inconclusive"
}


def add_gen_b_id_to_df(
        df: pd.DataFrame,
        gen_b_mdst_index: int,
        gen_query_dict: Optional[GenQueryDictType] = None
) -> None:
    """Adds the given generator B meson ids for the generator B
    with mdst index `gen_b_mdst_index` to a given data frame.

    :param df: A pandas.DataFrame instance (has to contain
    genUpsilon4S variables.)
    :param gen_b_mdst_index: Integer specifying the generator B
    mdst index. Has to be in {1, 2}.
    :param gen_query_dict: Dictionary with True/False values for
    the index. Default is mc_ids.gen_b_ids
    """
    assert gen_b_mdst_index in [1, 2]

    if gen_query_dict is None:
        gen_query_dict = copy.deepcopy(gen_b_ids)

    logging.debug(f"Number of candidates: {len(df.index)}")
    gen_b_column_name = f"GEN_B{gen_b_mdst_index}_ID"
    df.loc[:, gen_b_column_name] = np.nan

    for key, values in gen_query_dict.items():
        logging.debug(f"Processing query {key}...")
        query_fcn = values["query"]
        query_result = query_fcn(df, gen_b_mdst_index - 1)

        logging.debug(f"Number of matching candidates: {len(df.loc[query_result].index)}")

        if not np.all(pd.isnull(df.loc[query_result, gen_b_column_name])):
            raise RuntimeError("Careful, you are overwriting singal ids")

        df.loc[query_result, gen_b_column_name] = values["id"]

    check = df[gen_b_column_name].isnull().sum()

    if check > 0:
        raise ValueError(f"Gen ID queries do not cover all events.\nNumber of Rows w/o ID: {check}")


# generic functions
def apply_fcn_to_df_dict(
        df_dict: Dict[str, Dict[str, pd.DataFrame]],
        fcn: Callable,
        ntuple_key: str = "Y4S_ntuple",
        **kwargs
) -> None:
    for key, val in df_dict.items():
        logging.info(f"Applying function {fcn.__name__} to data type {key}..")
        fcn(val[ntuple_key], **kwargs)
