"""
Weights for RDStar analysis samples

Felix Metzner 2019
"""

import pandas as pd

from typing import Any, Optional, Dict

from analysistools.batch_tools.file_handler import Belle_url_mc_types, Belle_list_types, Belle_data_types
from analysistools.offline_analysis import sample_weights as sw


__all__ = [
    "scale_weight_column_name",
    "signal_mc_bfs",
    "signal_mc_evts",
    "add_scaling_weights"
]


BFSDType = Dict[str, Dict[str, Any]]
EvtNumDictType = Dict[str, Dict[str, int]]

scale_weight_column_name = sw.scale_weight_column_name  # type: str

sig_mc_bfs = {  # type: BFSDType
    # B0
    "DSME": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- e+ nu_e", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DME": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- e+ nu_e", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    "BZD1E": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- e+ nu_e", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "BZD0sE": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- e+ nu_e", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BZDp1E": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- e+ nu_e", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BZD2sE": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- e+ nu_e", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "DSMMu": {"bf": 0.05110, "charged_B": False, "decay_descr": "B0 -> D*- mu+ nu_mu", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DMMu": {"bf": 0.02140, "charged_B": False, "decay_descr": "B0 -> D- mu+ nu_mu", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    "BZD1Mu": {"bf": 0.00422, "charged_B": False, "decay_descr": "B0 -> D_1- mu+ nu_mu", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "BZD0sMu": {"bf": 0.00362, "charged_B": False, "decay_descr": "B0 -> D_0*- mu+ nu_mu", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BZDp1Mu": {"bf": 0.00401, "charged_B": False, "decay_descr": "B0 -> D'_1- mu+ nu_mu", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BZD2sMu": {"bf": 0.00347, "charged_B": False, "decay_descr": "B0 -> D_2*- mu+ nu_mu", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "DSMTau": {"bf": 0.0125416, "charged_B": False, "decay_descr": "B0 -> D*- tau+ nu_tau", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DMTau": {"bf": 0.0063687, "charged_B": False, "decay_descr": "B0 -> D- tau+ nu_tau", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    # "BZD1Tau": {"bf": 0.0013, "charged_B": False, "decay_descr": "B0 -> D_1- tau+ nu_tau", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},  # Not considered
    # "BZD0sTau": {"bf": 0.0013, "charged_B": False, "decay_descr": "B0 -> D_0*- tau+ nu_tau", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},  # Not considered
    # "BZDp1Tau": {"bf": 0.0020, "charged_B": False, "decay_descr": "B0 -> D'_1- tau+ nu_tau", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},  # Not considered
    # "BZD2sTau": {"bf": 0.0020, "charged_B": False, "decay_descr": "B0 -> D_2*- tau+ nu_tau", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},  # Not considered
    "GapD1Mixed": {"bf": 0.00564, "charged_B": False, "decay_descr": "B0 -> D_1- (-> D pi pi) l+ nu_l", "evtgen_info": "PHOTOS ISGW2;"},
    "GapDetaMixed": {"bf": 0.00868, "charged_B": False, "decay_descr": "B0 -> D(*)- eta l+ nu_l", "evtgen_info": "PHSP;"},
    "GapDpipiMixed": {"bf": 0.00588, "charged_B": False, "decay_descr": "B0 -> D(*) l+ nu_l", "evtgen_info": "PHSP;"},
    # B+
    "DSZE": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 e+ nu_e", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DZE": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 e+ nu_e", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    "BMD1E": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 e+ nu_e", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "BMD0sE": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 e+ nu_e", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BMDp1E": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 e+ nu_e", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BMD2sE": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 e+ nu_e", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "DSZMu": {"bf": 0.05490, "charged_B": True, "decay_descr": "B+ -> anti-D*0 mu+ nu_mu", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DZMu": {"bf": 0.02310, "charged_B": True, "decay_descr": "B+ -> anti-D0 mu+ nu_mu", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    "BMD1Mu": {"bf": 0.00454, "charged_B": True,  "decay_descr": "B+ -> anti-D_10 mu+ nu_mu", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "BMD0sMu": {"bf": 0.00389, "charged_B": True,  "decay_descr": "B+ -> anti-D_0*0 mu+ nu_mu", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BMDp1Mu": {"bf": 0.00431, "charged_B": True,  "decay_descr": "B+ -> anti-D'_10 mu+ nu_mu", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},
    "BMD2sMu": {"bf": 0.00373, "charged_B": True,  "decay_descr": "B+ -> anti-D_2*0 mu+ nu_mu", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},
    "DSZTau": {"bf": 0.0134976, "charged_B": True, "decay_descr": "B+ -> anti-D*0 tau+ nu_tau", "evtgen_info": "HQET3 0.912 1.205 1.15 1.404 0.854; #FF values taken from HFLAV 2016 (1612.07233)"},
    "DZTau": {"bf": 0.0068527, "charged_B": True, "decay_descr": "B+ -> anti-D0 tau+ nu_tau", "evtgen_info": "HQET3 1.0541 1.128 1.0; #FF values taken from HFLAV 2016 (1612.07233)"},
    # "BMD1Tau": {"bf": 0.0013, "charged_B": True, "decay_descr": "B+ -> anti-D_10 tau+ nu_tau", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},  # Not considered
    # "BMD0sTau": {"bf": 0.0013, "charged_B": True, "decay_descr": "B+ -> anti-D_0*0 tau+ nu_tau", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},  # Not considered
    # "BMDp1Tau": {"bf": 0.0020, "charged_B": True, "decay_descr": "B+ -> anti-D'_10 tau+ nu_tau", "evtgen_info": "LLSW 0.68 -0.2 0.3;"},  # Not considered
    # "BMD2sTau": {"bf": 0.0020, "charged_B": True, "decay_descr": "B+ -> anti-D_2*0 tau+ nu_tau", "evtgen_info": "LLSW 0.71 -1.6 -0.5 2.9;"},  # Not considered
    "GapD1Charged": {"bf": 0.00607, "charged_B": True, "decay_descr": "B+ -> anti-D_10 (-> D pi pi) l+ nu_l", "evtgen_info": "PHOTOS ISGW2;"},
    "GapDetaCharged": {"bf": 0.00804, "charged_B": True, "decay_descr": "B+ -> anti-D(*)0 eta l+ nu_l", "evtgen_info": "PHSP;"},
    "GapDpipiCharged": {"bf": 0.00632, "charged_B": True, "decay_descr": "B+ -> D(*) pi pi l+ nu_l", "evtgen_info": "PHSP;"},
}

# Adding branching fraction infos also for test and efficiency samples.
# For the efficiency samples of the B -> D** decays these are not correct, as we have separate
# efficiency samples for the charged and neutral pion cases, but this does not affect the
# efficiency calculation performed in the EfficiencyEvaluationTask!
Dss_Test_Info = {f"{k}Test": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_Eff_Info = {f"{k}Eff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_ChargedPiEff_Info = {f"{k}ChargedPiEff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
sig_mc_bfs.update(Dss_Test_Info)
sig_mc_bfs.update(Dss_Eff_Info)
sig_mc_bfs.update(Dss_ChargedPiEff_Info)
signal_mc_bfs = sig_mc_bfs

#  Number of simulated events in private MC samples
signal_mc_evts = {  # type: EvtNumDictType
    # B0
    "DSME": {"prod": 2500000, "meta": 2500000, "summary": 2498381},
    "DME": {"prod": 2500000, "meta": 2500000, "summary": 2498259},
    "BZD1E": {"prod": 72999988, "meta": 72955090, "summary": 72970488},  # ratio .9995, 
    "BZD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72972981},  # ratio .9996, ok
    "BZDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72971014},  # ratio .9996, ok
    "BZD2sE": {"prod": 72899988, "meta": 72999988, "summary": 72972825},  # ratio .9996, 
    "BZD1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499020},
    "BZD0sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "BZDp1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499004},
    "BZD2sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499058},
    "BZD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498775},
    "BZD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498849},
    "BZDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2498816},
    "BZD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2498877},
    "BZD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499099},
    "BZDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "BZD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499205},
    "BZD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499172},
    "DSMMu": {"prod": 2500000, "meta": 2500000, "summary": 2495652},
    "DMMu": {"prod": 2500000, "meta": 2500000, "summary": 2496318},
    "BZD1Mu": {"prod": 72999988, "meta": 72928778, "summary": 72935520},  # ratio .9991, ok
    "BZD0sMu": {"prod": 72999988, "meta": 72999988, "summary": 72945960},  # ratio .9992, ok
    "BZDp1Mu": {"prod": 72899988, "meta": 72999988, "summary": 72937478},  # ratio .9991,ok
    "BZD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72885254},  # ratio .9984, ok
    "BZD1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497746},
    "BZD0sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498095},
    "BZDp1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497821},
    "BZD2sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498115},
    "BZD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497335},
    "BZD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497482},
    "BZDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497318},
    "BZD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2497486},
    "BZD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497982},
    "BZDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498097},
    "BZD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498273},
    "BZD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498428},
    "DSMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495647},
    "DMTau": {"prod": 2500000, "meta": 2500000, "summary": 2495568},
    # "BZD1Tau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZD0sTau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZDp1Tau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZD2sTau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZD1TauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZD0sTauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZDp1TauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BZD2sTauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    "GapD1Mixed": {"prod": 9999998, "meta": 9999998, "summary": 9990965},
    "GapDetaMixed": {"prod": 12059997, "meta": 12059997, "summary": 12053835},
    "GapDpipiMixed": {"prod": 9479999, "meta": 9479999, "summary": 9477142},
    # B+
    "DSZE": {"prod": 2500000, "meta": 2500000, "summary": 2499051},
    "DZE": {"prod": 2500000, "meta": 2500000, "summary": 2498851},
    "BMD1E": {"prod": 72999988, "meta": 72999988, "summary": 72972451},  # ratio .9996, ok
    "BMD0sE": {"prod": 72999988, "meta": 72999988, "summary": 72976112},  # ratio .996, ok
    "BMDp1E": {"prod": 72999988, "meta": 72999988, "summary": 72973658},  # ratio .9996, ok
    "BMD2sE": {"prod": 72999988, "meta": 72986680, "summary": 72976436},  # ratio .9996, ok
    "BMD1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499020},
    "BMD0sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499081},
    "BMDp1ETest": {"prod": 2500000, "meta": 2500000, "summary": 2499004},
    "BMD2sETest": {"prod": 2500000, "meta": 2500000, "summary": 2499195},
    "BMD1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499285},
    "BMD0sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499248},
    "BMDp1EEff": {"prod": 2500000, "meta": 2500000, "summary": 2499208},
    "BMD2sEEff": {"prod": 2500000, "meta": 2500000, "summary": 2499272},
    "BMD1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498942},
    "BMDp1EChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499014},
    "BMD2sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499063},
    "BMD0sEChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2499165},
    "DSZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497271},
    "DZMu": {"prod": 2500000, "meta": 2500000, "summary": 2497540},
    "BMD1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72941370},  # ratio .992, ok
    "BMD0sMu": {"prod": 72899988, "meta": 72999988, "summary": 72951794},  # ratio .993, ok
    "BMDp1Mu": {"prod": 72999988, "meta": 72999988, "summary": 72942764},  # ratio .9992, ok
    "BMD2sMu": {"prod": 72999988, "meta": 72999988, "summary": 72950054},  # ratio .9995, ok
    "BMD1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497746},
    "BMD0sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498095},
    "BMDp1MuTest": {"prod": 2500000, "meta": 2500000, "summary": 2497821},
    "BMD2sMuTest": {"prod": 2500000, "meta": 2500000, "summary": 2498308},
    "BMD1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498342},
    "BMD0sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498401},
    "BMDp1MuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498384},
    "BMD2sMuEff": {"prod": 2500000, "meta": 2500000, "summary": 2498407},
    "BMD1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497809},
    "BMDp1MuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2497867},
    "BMD2sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498191},
    "BMD0sMuChargedPiEff": {"prod": 2500000, "meta": 2500000, "summary": 2498300},
    "DSZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497354},
    "DZTau": {"prod": 2500000, "meta": 2500000, "summary": 2497181},
    # "BMD1Tau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMD0sTau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMDp1Tau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMD2sTau": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMD1TauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMD0sTauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMDp1TauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    # "BMD2sTauTest": {"prod": 0, "meta": 0, "summary": 0},  # Not considered
    "GapD1Charged": {"prod": 9999998, "meta": 9928198, "summary": 9993029},  # TODO: Missing events in file /group/belle2/users/welsch/ntuples/b_sl_gap/gsim/mdst/GapD1Charged/evtgen_exp_25_GapD1Charged-0.mdst
    "GapDetaCharged": {"prod": 12759999, "meta": 12759999, "summary": 12755786},
    "GapDpipiCharged": {"prod": 10039998, "meta": 10039998, "summary": 10037734},
}

# Checking consistency of the dictionaries signal_mc_evts and signal_mc_bfs:
assert set(signal_mc_evts.keys()) == set(signal_mc_bfs.keys())


def add_scaling_weights(
        df: pd.DataFrame,
        data_type: str,
        generic_stream_dict: Optional[Dict[str, int]] = None
) -> None:
    if data_type in Belle_url_mc_types:
        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=generic_stream_dict[data_type] if generic_stream_dict is not None else 1.
        )
    elif data_type in Belle_list_types:
        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=sw.FIXED_SPECIAL_WEIGHTS[data_type]
        )
    elif data_type in signal_mc_evts:
        n_events_source = "prod"
        if data_type.startswith("B"):
            n_events_source = "meta"
        applied_weight = sw.add_private_mc_weight(
            df=df,
            produced_events=signal_mc_evts[data_type][n_events_source],
            charged_B=signal_mc_bfs[data_type]["charged_B"],
            branching_fraction=signal_mc_bfs[data_type]["bf"]
        )
    elif data_type.startswith("B") and (data_type.endswith("Test") or data_type.endswith("Eff")):
        # No weighting for additional D** sample:
        applied_weight = 1.0
        df.loc[:, sw.scale_weight_column_name] = applied_weight
    elif data_type in Belle_data_types:
        # Also adding no scaling weight for actual data.
        applied_weight = 1.0
        df.loc[:, sw.scale_weight_column_name] = applied_weight
    else:
        raise RuntimeError(f"Data type {data_type} is not covered by add_scaling_weights!")

    assert applied_weight > 0.

    print(f"Applied scaling factor for {data_type:>10}: {applied_weight:2.4f}")
