"""
Definition of Signal and Sideband Selections

This file covers the definitions of
    - Selection cuts
    - Best candidate selection
for
    - the main selection
    - the q2 sideband
    - the mbc sideband
    - the selection for the form factor study
    - the dedicated D** selection  # TODO!


Felix Metzner 2019
"""

from typing import List, Tuple, Union, NamedTuple
from collections import namedtuple

SelectionInfo = namedtuple(
    typename="SelectionInfo",
    field_names=["selection_tag", "selection", "target_path", "tmp_target_path", "output_key"]
)


class BCSVariableInfo(NamedTuple):
    """
    Definition of a information set used for the BCS:
    Contains:
      - bcs_variable: variable name of variable to be used for bcs
                      or "random" if selection shall be done random
                      or None e.g. if now secondary bcs variable shall be used or no BCS shall be applied at all.
      - absolute_value: Boolean indicating if absolute value of variable should be taken for the bcs
      - max: Boolean whether to select the candidate based on the maximum (True) or minimum (False)
    """
    bcs_variable: Union[None, str]
    absolute_value: bool = False
    max: bool = True


# Definition of my first version of the selection:
class FirstSelection:
    selection_name = "first_selection"

    # Default Cuts
    fei_prob_cut = "daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc > 0.001"
    q_square_cut = "qSquaredReco > 4."
    e_extra_cut = "ROE_eextra__boCleanROE__bc < 3."
    d_mass_cut = "abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.35"
    tag_mbc_cut = "daughter__bo0__cm__spMbc__bc > 5.27"
    tag_delta_e_cut = "-0.05 < daughter__bo0__cm__spdeltaE__bc < 0.1"
    cs_mva_cut = "cs_mva_output > 0.5"
    lepton_p_cut = "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2"
    b_vertex_fit_cut = "daughter__bo1__cm__spchiProb__bc > 0.0"

    # Possible Cuts; Not applied
    lepton_p_star_cut = "pStarL < 2"  # TODO: Maybe use this instead of lepton_p_cut above?

    # Sideband Cut: q Squared Sideband
    q_square_anti_cut = "qSquaredReco <= 4."

    # Sideband Cut: Mbc Sideband
    tag_mbc_anti_cut = "daughter__bo0__cm__spMbc__bc <= 5.27"

    cut_label_dict = {
        fei_prob_cut: r"$\mathcal{P}_\mathrm{FEI} > 0.001$",
        # daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc > 0.001
        q_square_cut: r"$\mathrm{q}^2 > 4.0\,$GeV",  # qSquaredReco > 4.
        e_extra_cut: r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra} < 3.0\,$GeV",  # ROE_eextra__boCleanROE__bc < 3.
        d_mass_cut: r"$|\Delta \mathrm{M}_{D^{(*)}}| < 0.35\,$GeV",
        # abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.35
        tag_mbc_cut: r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag} > 5.27\,$GeV",  # daughter__bo0__cm__spMbc__bc > 5.27
        tag_delta_e_cut: r"$\Delta\mathrm{E}_\mathrm{tag}\;\in\;(-0.05, 0.1)\,$GeV",
        # -0.05 < daughter__bo0__cm__spdeltaE__bc < 0.1
        cs_mva_cut: r"$\mathcal{P}_\mathrm{CS} > 0.5$",  # cs_mva_output > 0.5
        lepton_p_cut: r"$\mathrm{p}_\ell < 2\,$GeV",  # daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2
        b_vertex_fit_cut: r"$\mathcal{P}_\mathrm{Vertex} > 0.0$",  # daughter__bo1__cm__spchiProb__bc > 0.0
        lepton_p_star_cut: r"$\mathrm{p}_\ell^\ast < 2.0\,$GeV",  # pStarL < 2
        q_square_anti_cut: r"$\mathrm{q}^2 \leq 4.0\,$GeV",  # qSquaredReco <= 4.
        tag_mbc_anti_cut: r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag} \leq 5.27\,$GeV",
        # daughter__bo0__cm__spMbc__bc <= 5.27
    }

    # Definition of Best Candidate Selection
    @staticmethod
    def get_bcs_info():
        primary_bcs_var = BCSVariableInfo(
            bcs_variable="ROE_neextra__boCleanROE__bc",
            absolute_value=False,
            max=False
        )
        secondary_bcs_var = BCSVariableInfo(
            bcs_variable="daughter__bo0__cm__spdeltaE__bc",
            absolute_value=True,
            max=False
        )
        return primary_bcs_var, secondary_bcs_var

    # Selection Definitions
    @staticmethod
    def get_selection(cs_tag: str) -> Tuple[str, List[str], List[str], bool]:
        # Main Selection
        selection_tag = f"{FirstSelection.selection_name}_with_CS_{cs_tag}"
        cut_list = [
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.tag_delta_e_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut
        ]
        check_cuts = []
        check_bcs = False
        yield selection_tag, cut_list, check_cuts, check_bcs

        sb_check_cut_list = [
            FirstSelection.e_extra_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut
        ]

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        mbc_sb_check_cuts = [c for c in sb_check_cut_list if c != FirstSelection.tag_mbc_cut]
        mbc_sb_check_bcs = True
        yield mbc_sb_tag, mbc_sb_cut_list, mbc_sb_check_cuts, mbc_sb_check_bcs

        # q squared Sideband Selection
        q2_sb_tag = f"{selection_tag}_q2_sideband"
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        q2_sbc_check_cuts = [c for c in sb_check_cut_list if c != FirstSelection.q_square_cut]
        q2_sb_check_bcs = True
        yield q2_sb_tag, q2_sb_cut_list, q2_sbc_check_cuts, q2_sb_check_bcs

        # Form Factor Study Selection (full q squared range)
        full_q2_tag = f"{selection_tag}_full_q2"
        full_q2_cut_list = [
            c for c in cut_list if (
                    c != FirstSelection.q_square_cut
                    and c != FirstSelection.fei_prob_cut
                    and c != FirstSelection.lepton_p_cut
                    and c != FirstSelection.lepton_p_star_cut
                    and c != FirstSelection.tag_delta_e_cut
            )
        ]
        full_q2_check_cuts = []
        full_q2_check_bcs = False
        yield full_q2_tag, full_q2_cut_list, full_q2_check_cuts, full_q2_check_bcs


# This interface class should be used to access all defined selections...
class Selection:
    @staticmethod
    def _get_selection_class(selection_name: str):
        if selection_name == FirstSelection.selection_name:
            return FirstSelection
        else:
            raise ValueError(f"Selection with name {selection_name} is not available!")

    @staticmethod
    def _get_selection_as_dict(selection_name, cs_tag):
        selection_class = Selection._get_selection_class(selection_name=selection_name)

        return {
            infos[0]: {
                "cuts": infos[1],
                "primary_bcs_var": selection_class.get_bcs_info()[0],
                "secondary_bcs_var": selection_class.get_bcs_info()[1],
                "check_cuts": infos[2],
                "check_bcs": infos[3]
            }
            for infos in selection_class.get_selection(cs_tag=cs_tag)
        }

    @staticmethod
    def get_selection_infos(selection_name, cs_tag):
        return Selection._get_selection_as_dict(
            selection_name=selection_name,
            cs_tag=cs_tag
        )

    @staticmethod
    def get_selection_cuts(selection_name, selection_tag, cs_tag):
        cuts = Selection._get_selection_as_dict(
            selection_name=selection_name,
            cs_tag=cs_tag
        )[selection_tag]["cuts"]

        check_cuts = Selection._get_selection_as_dict(
            selection_name=selection_name,
            cs_tag=cs_tag
        )[selection_tag]["check_cuts"]

        initial_cuts = [c for c in cuts if c not in check_cuts]

        return cuts, initial_cuts, check_cuts

    @staticmethod
    def get_selection_bcs(selection_name, selection_tag, cs_tag):
        primary = Selection._get_selection_as_dict(selection_name, cs_tag=cs_tag)[selection_tag]["primary_bcs_var"]
        secondary = Selection._get_selection_as_dict(selection_name, cs_tag=cs_tag)[selection_tag]["secondary_bcs_var"]
        check = Selection._get_selection_as_dict(selection_name, cs_tag=cs_tag)[selection_tag]["check_bcs"]

        return primary, secondary, check

    @staticmethod
    def get_selection_tags(selection_name: str, cs_tag: str) -> List[str]:
        selection_class = Selection._get_selection_class(selection_name=selection_name)
        return [k[0] for k in selection_class.get_selection(cs_tag=cs_tag)]

    @staticmethod
    def get_main_selection_tag(selection_name, cs_tag):
        selection_class = Selection._get_selection_class(selection_name=selection_name)
        # Getting first element returned by the generator get_selection.
        # This first item is supposed to be the main selection definition of the analysis.
        return next(selection_class.get_selection(cs_tag=cs_tag))[0]

    @staticmethod
    def print_selection_info(selection_name, cs_tag):
        selection_class = Selection._get_selection_class(selection_name=selection_name)
        print(f"Selection name: {selection_class.selection_name}\n")

        print(
            "All defined selections:\n\t- "
            + "\n\t- ".join(Selection.get_selection_tags(selection_name, cs_tag=cs_tag)) + "\n"
        )

        print(f"Main selection tag: {Selection.get_main_selection_tag(selection_name, cs_tag=cs_tag)}\n")

        for key, value in Selection._get_selection_as_dict(selection_name=selection_name, cs_tag=cs_tag).items():
            cut_string_list = [f"  {c}" if c not in value["check_cuts"] else f"C {c}" for c in value["cuts"]]
            print(f"{key}:\n\t" + "\n\t".join(cut_string_list) + "\n")

    @staticmethod
    def get_selection_cut_label_dict(selection_name):
        selection_class = Selection._get_selection_class(selection_name=selection_name)
        return selection_class.cut_label_dict


if __name__ == "__main__":
    Selection.print_selection_info(selection_name=FirstSelection.selection_name, cs_tag="test")
