"""
Continuum Suppression Tasks

Felix Metzner 2019
"""

import os
import yaml
import numpy as np
import pandas as pd
import b2luigi as luigi

from PyFastBDT import FastBDT
from luigi.util import inherits
from sklearn.model_selection import train_test_split

from typing import Optional, List, Tuple, Dict, Generator

from analysistools.dataframe_tools import hdf5_utils

from rdstar.offline_analysis.tasks import utilities
from rdstar import continuum_suppression as cs_tools
from rdstar.offline_analysis.validation import cs_validation
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask


__all__ = [
    "CSDataPreparationTask",
    "CSTrainTask",
    "CSTestTask"
]


@inherits(DataCopyTask)
class CSDataPreparationTask(utilities.BaseOfflineTask):
    cs_data_tag = luigi.Parameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")

        self.cs_base_tuple_type = "Y4S_ntuple"

        self.cs_sig_data_types = [
            "DME", "DMMu", "DMTau", "DZE", "DZMu", "DZTau",
            "DSME", "DSMMu", "DSMTau", "DSZE", "DSZMu", "DSZTau",
            "BMD0sETest", "BMD0sMuTest", "BMD1ETest", "BMD1MuTest",
            "BMD2sETest", "BMD2sMuTest", "BMDp1ETest", "BMDp1MuTest",
            "BZD0sETest", "BZD0sMuTest", "BZD1ETest", "BZD1MuTest",
            "BZD2sETest", "BZD2sMuTest", "BZDp1ETest", "BZDp1MuTest"
        ]
        self.cs_sig_stream = -999

        self.cs_bkg_data_types = ["UDS", "Charm"]
        self.cs_bkg_stream = 0

        self.cs_signal_query = "isSignalAcceptMissingNeutrino == 1"

        self.cs_vars = list(cs_tools.cs_variable_dict.keys())
        self.cs_target_var = "isNotContinuumEvent"
        self.signal_vars = [
            "m2RecoilSignalSide",
            "daughter__bo1__cm__spuseRestFrame__bodaughter__bo1__cm__spp__bc__bc__bc",
            "isSignalAcceptMissingNeutrino"
        ]

        self.vars = self.cs_vars + [self.cs_target_var] + self.signal_vars

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_key_pair(self, data_type: str, extension: Optional[str] = None) -> Tuple[str, str]:
        assert data_type in self.cs_sig_data_types + self.cs_bkg_data_types

        if data_type in self.cs_sig_data_types:
            hdf5_key = f"{data_type}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_{DataPreparationTask.base_file_tag}"
        elif extension is None and data_type in self.cs_bkg_data_types:
            hdf5_key = f"{data_type}_{self.cs_bkg_stream}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_{DataPreparationTask.base_file_tag}"
        else:
            assert extension in ["test", "train"]
            hdf5_key = f"{data_type}_{self.cs_bkg_stream}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_forCS_{extension}"

        return file_key, hdf5_key

    def get_data_frames(self, data_types: List[str], extension: Optional[str] = None) -> List[pd.DataFrame]:
        dfs = []
        for file_key, hdf5_key in [self.get_key_pair(d, extension) for d in data_types]:
            dfs.append(pd.read_hdf(path_or_buf=self.input_dict[file_key], key=hdf5_key)[self.vars])
        return dfs

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for data_type in self.cs_bkg_data_types:
            yield self.clone(
                DataPreparationMergerTask,
                data_type=data_type,
                stream=self.cs_bkg_stream,
                tuple_type=self.cs_base_tuple_type,
                n_subsets=utilities.GlobalParameters().n_subsets_generic
            )

        for data_type in self.cs_sig_data_types:
            yield self.clone(
                DataPreparationMergerTask,
                data_type=data_type,
                stream=self.cs_sig_stream,
                tuple_type=self.cs_base_tuple_type
            )

    def cs_preparation_process(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        cs_train_data_file_name = os.path.join(self.cs_dir, f"cs_train_data_{self.cs_data_tag}.h5")
        cs_test_data_file_name = os.path.join(self.cs_dir, f"cs_test_data_{self.cs_data_tag}.h5")
        tmp_cs_train_data_file_name = os.path.join(self.cs_dir, f"tmp_cs_train_data_{self.cs_data_tag}.h5")
        tmp_cs_test_data_file_name = os.path.join(self.cs_dir, f"tmp_cs_test_data_{self.cs_data_tag}.h5")

        if not dry_run:
            os.makedirs(self.cs_dir, exist_ok=True)

            bkg_test_dfs = self.get_data_frames(self.cs_bkg_data_types, 'test')
            bkg_train_dfs = self.get_data_frames(self.cs_bkg_data_types, 'train')

            sig_dfs = self.get_data_frames(self.cs_sig_data_types)

            sig_df = pd.concat([df.query(self.cs_signal_query) for df in sig_dfs], ignore_index=True, sort=False)

            print(f"Number of considered Signal events: {len(sig_df.index)}")
            sig_test_df, sig_train_df = train_test_split(sig_df, train_size=0.5)

            train_df = pd.concat(bkg_train_dfs + [sig_train_df], ignore_index=True, sort=False)
            test_df = pd.concat(bkg_test_dfs + [sig_test_df], ignore_index=True, sort=False)

            train_df.to_hdf(path_or_buf=tmp_cs_train_data_file_name, key="cs_train")
            test_df.to_hdf(path_or_buf=tmp_cs_test_data_file_name, key="cs_test")

            os.rename(src=tmp_cs_train_data_file_name, dst=cs_train_data_file_name)
            os.rename(src=tmp_cs_test_data_file_name, dst=cs_test_data_file_name)

        return {
            f"cs_train_{self.cs_data_tag}": cs_train_data_file_name,
            f"cs_test_{self.cs_data_tag}": cs_test_data_file_name,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.cs_preparation_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.cs_preparation_process(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict


@inherits(DataCopyTask)
class CSTrainTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)
    cs_fastbdt_settings = luigi.DictParameter(positional=False, hashed=True, significant=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")

        self.cs_vars = list(cs_tools.cs_variable_dict.keys())  # TODO: provide cs_vars via luigi param to change them!
        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]

        self.cs_data_tag = offline_config["cs_config"]["cs_data_tag"]

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_sample(self) -> pd.DataFrame:
        train_data_path = self.input_dict[f"cs_train_{self.cs_data_tag}"]
        return hdf5_utils.load_hdf5(hdf5_file_path=train_data_path, hdf5_key="cs_train")

    def run_cs_training(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        cs_output_weight_file = os.path.join(self.cs_dir, f"cs_classifier_weights_{self.cs_tag}.xml")
        cs_var_list_file = os.path.join(self.cs_dir, f"cs_mva_vars_{self.cs_tag}.pkl")
        tmp_cs_output_weight_file = os.path.join(self.cs_dir, f"tmp_cs_classifier_weights_{self.cs_tag}.xml")
        tmp_cs_var_list_file = os.path.join(self.cs_dir, f"tmp_cs_mva_vars_{self.cs_tag}.pkl")

        if not dry_run:
            cs_train_data = self.get_sample()

            cs_train_data.loc[:, "__cs_weights__"] = np.ones(len(cs_train_data.index))

            cs = cs_tools.ContinuumSuppression(
                cs_variables=self.cs_vars,
                target_variable=self.cs_target_var,
                weight='__cs_weights__'
            )

            if not self.cs_fastbdt_settings:
                fast_bdt_settings = {
                    "nTrees": 200,
                }
            else:
                fast_bdt_settings = self.cs_fastbdt_settings

            cs.add_classifier(classifier=FastBDT.Classifier(**fast_bdt_settings))
            cs.train(cs_train_data)

            cs.save_classifier_weights(weight_file_path=tmp_cs_output_weight_file, overwrite=True)

            cs_tools.ContinuumSuppression.dump_list_of_variables(
                variable_list=self.cs_vars,
                pickle_file_path=tmp_cs_var_list_file
            )

            os.rename(src=tmp_cs_output_weight_file, dst=cs_output_weight_file)
            os.rename(src=tmp_cs_var_list_file, dst=cs_var_list_file)

        return {
            "cs_output_weight_file": cs_output_weight_file,
            "cs_var_list_file": cs_var_list_file,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_cs_training()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_cs_training(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CSDataPreparationTask, cs_data_tag=self.cs_data_tag)


# TODO: Implement CS Test Tasks which evaluates several CS Trainings and picks the best.
@inherits(CSTrainTask)
class CSTestTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")

        self.cs_data_tag = offline_config["cs_config"]["cs_data_tag"]
        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

        self.cs_train_data_path = self.input_dict[f"cs_train_{self.cs_data_tag}"]
        self.cs_test_data_path = self.input_dict[f"cs_test_{self.cs_data_tag}"]

    def run_cs_evaluation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_key = f"cs_validation_plot"
        output_path = os.path.join(self.cs_dir, f"cs_{self.cs_tag}_{self.cs_data_tag}_validation_plot.pdf")

        if not dry_run:
            roc_auc_on_test = cs_validation.run_cs_validation(
                path_train_df=self.cs_train_data_path,
                path_test_df=self.cs_test_data_path,
                path_cs_var_list=self.input_dict["cs_var_list_file"],
                path_cs_bdt_weights=self.input_dict["cs_output_weight_file"],
                target_file_path=output_path,
                cs_target_var=self.cs_target_var
            )

            # TODO: Use ROC AUC for comparison.
            print(f"ROC AUC for CS tag {self.cs_tag}  with CS data {self.cs_data_tag} is {roc_auc_on_test}")

        return {output_key: output_path}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_cs_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_cs_evaluation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CSTrainTask)
        yield self.clone(CSDataPreparationTask, cs_data_tag=self.cs_data_tag)
