"""
Data Preparation Tasks

    - Removing replaced events from generic MC
    - MC reweighting for D, D* and D** modes
    - Applying pID corrections
    - Applying tracking corrections
    - Applying K-short corrections
    - ...

Felix Metzner 2019
"""

import os
import yaml
import json
import tempfile
import numpy as np
import pandas as pd
import b2luigi as luigi

from typing import Union, Optional, Dict, List, Tuple, Generator

from collections import defaultdict
from luigi.util import inherits, requires
from more_itertools import split_into, random_permutation

from analysistools.batch_tools.file_handler import Belle_list_types

from rdstar.offline_analysis import mc_ids
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.logistics_tasks import DataInitialConversionTask

from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.corrections.qed_correction import get_coulomb_corr_factor
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection

from rdstar.mc_reweighting.BtoDlnu.BtoDlnu_weights import BtoDFFReweighting
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import BtoDststFFReweighting, DststType


__all__ = [
    "DataPreparationMergerTask",
    "DataPreparationSplitterTask",
    "DataPreparationTask"
]


PathType = Union[str, os.fspath]


@requires(DataInitialConversionTask)
class DataPreparationSplitterTask(utilities.BaseOfflineTask):
    n_subsets = luigi.IntParameter(positional=False, default=1)

    n_entries_before_split_key = "n_entries_before_split"

    @property
    def resources(self) -> Dict[str, int]:
        if self.data_type in ["Charged", "Charm", "UDS"]:
            return {"offline_memory": 8}
        elif self.data_type in ["Mixed"]:
            return {"offline_memory": 4}
        else:
            return {"offline_memory": 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.n_subsets, int)
        assert self.n_subsets > 0

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.file_tag = "base_split"

        self.input_dict = {k: v.path for k, v in self.input().items()}

    @staticmethod
    def get_output_key(file_name: str, index: int) -> str:
        return f"{file_name}_base_split_{index}"

    def get_target_path(self, index: int) -> os.fspath:
        return os.path.join(self.base_dir, f'{self.file_name}_{self.file_tag}_{index}.h5')

    def get_temp_target_path(self, index: int) -> os.fspath:
        return os.path.join(self.base_dir, f'temp_{self.file_name}_{self.file_tag}_{index}.h5')

    def split_process(self, dry_run: bool = False) -> Dict[str, PathType]:
        source = self.input_dict[self.file_name]
        output_dict = {}

        assert isinstance(self.n_subsets, int)

        for index in range(self.n_subsets):
            target = self.get_target_path(index=index)
            output_dict[self.get_output_key(file_name=self.file_name, index=index)] = target

        if not dry_run:
            if self.n_subsets == 1:
                assert len(output_dict) == 1
                target = list(output_dict.values())[0]
                os.symlink(source, target)
            else:
                self.split_sample(source_path=source, n_sub_samples=self.n_subsets)

        return output_dict

    def split_sample(self, source_path: PathType, n_sub_samples: int) -> None:
        assert n_sub_samples > 1
        df = pd.read_hdf(source_path)
        assert isinstance(df, pd.DataFrame)
        initial_entries = len(df.index)

        index_pairs = self.get_index_pairs(sample=df, n_sub_samples=n_sub_samples)
        splitted_entries = 0

        for sub_sample_index, index_pair in enumerate(index_pairs):
            splitted_entries += len(df.loc[index_pair[0]:index_pair[1]].index)
            df.loc[index_pair[0]:index_pair[1]].to_hdf(
                path_or_buf=self.get_temp_target_path(index=sub_sample_index),
                key=self.file_name
            )

            with pd.HDFStore(path=self.get_temp_target_path(index=sub_sample_index)) as hdf5store:
                hdf5store.append(key=self.n_entries_before_split_key, value=pd.Series(initial_entries))

            os.rename(
                src=self.get_temp_target_path(index=sub_sample_index),
                dst=self.get_target_path(index=sub_sample_index)
            )

        assert initial_entries == splitted_entries

    @staticmethod
    def get_index_pairs(sample: pd.DataFrame, n_sub_samples: int) -> List[Tuple[int, int]]:
        n_entries = len(sample.index)
        assert sample.index[0] == 0
        assert sample.index[-1] == n_entries - 1

        sub_sample_size = int(np.floor(n_entries / n_sub_samples))
        index_pairs = []

        for n in range(n_sub_samples):
            lower_index = 0
            if n > 0:
                lower_index = index_pairs[n - 1][1] + 1

            if n < n_sub_samples - 1:
                upper_index = (n + 1) * sub_sample_size
                new_index = upper_index
                while sample.iloc[upper_index - 1]["unique_event_id"] == sample.iloc[new_index]["unique_event_id"]:
                    new_index += 1
                upper_index = new_index
            else:
                upper_index = sample.index[-1]
                assert upper_index == n_entries - 1

            index_pairs.append((lower_index, upper_index))

        return index_pairs

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.split_process()

    def output(self) -> Dict[str, PathType]:
        output_dict = {k: luigi.LocalTarget(os.fspath(t)) for k, t in self.split_process(dry_run=True).items()}
        return output_dict


@inherits(DataPreparationSplitterTask)
class DataPreparationTask(utilities.BaseOfflineTask):
    subset_index = luigi.IntParameter(positional=False)

    base_file_tag = "prepared"

    scale_corr_column = "__scaling_corr__"

    n_entries_after_split_key = "n_entries_after_split"
    n_entries_key = "n_entries"
    n_entries_before_key = "n_entries_before"
    n_cs_sample_key = "n_cs_sample"

    @property
    def resources(self) -> Dict[str, int]:
        if self.data_type in ["Charged", "Charm", "UDS"]:
            return {"offline_memory": 8}
        elif self.data_type in ["Mixed"]:
            return {"offline_memory": 4}
        else:
            return {"offline_memory": 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.file_tag = self.base_file_tag

        self.removed_tag = "removed_from_generic"

        if isinstance(offline_config["cs_config"], dict):
            self.cs_config = offline_config["cs_config"]
        else:
            self.cs_config = None

        self.pid_weight_info_target = os.path.join(
            self.base_dir,
            f"{self.file_name}_pid_weight_info_split_{self.subset_index}.h5"
        )
        self.pid_weight_info_output_key = f"{self.file_name}_pid_weight_info_split_{self.subset_index}"

        self.ff_d_weight_info_target = os.path.join(
            self.base_dir,
            f"{self.file_name}_ff_d_weight_info_split_{self.subset_index}.h5"
        )
        self.ff_d_weight_info_output_key = f"{self.file_name}_ff_d_weight_info_split_{self.subset_index}"

        self.ff_ds_weight_info_target = os.path.join(
            self.base_dir,
            f"{self.file_name}_ff_ds_weight_info_split_{self.subset_index}.h5"
        )
        self.ff_ds_weight_info_output_key = f"{self.file_name}_ff_ds_weight_info_split_{self.subset_index}"

        self.ff_dss_weight_info_target = os.path.join(
            self.base_dir,
            f"{self.file_name}_ff_dss_weight_info_split_{self.subset_index}.h5"
        )
        self.ff_dss_weight_info_output_key = f"{self.file_name}_ff_dss_weight_info_split_{self.subset_index}"

        self.ff_ds_gammas_path = os.path.join(os.path.dirname(self.working_dir), "dst_reweighting_gamma_integrals.json")

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        self.input_key = DataPreparationSplitterTask.get_output_key(file_name=self.file_name, index=self.subset_index)
        self.input_file = self.input_dict[self.input_key]

    def get_precomputed_gamma_integrals(self) -> Optional[Dict[str, Dict[int, float]]]:
        if os.path.exists(self.ff_ds_gammas_path):
            with open(self.ff_ds_gammas_path, 'r') as in_file:
                json_gamma_ints = json.load(in_file)

            return self.convert_gamma_int_dict_keys(json_gamma_ints)
        else:
            return None

    @staticmethod
    def convert_gamma_int_dict_keys(gamma_dict: Dict[str, Dict[str, float]]) -> Dict[str, Dict[int, float]]:
        new_gamma_dict = {}
        for k, v in gamma_dict.items():
            new_gamma_dict[k] = {int(key): value for key, value in gamma_dict[k].items()}

        return new_gamma_dict

    def write_computed_ff_ds_gammas(self, gamma_dicts: Dict[str, Dict[int, float]]) -> None:
        if os.path.exists(self.ff_ds_gammas_path):
            with open(self.ff_ds_gammas_path, 'r') as in_file:
                json_gamma_ints = json.load(in_file)
                gamma_ints = self.convert_gamma_int_dict_keys(json_gamma_ints)
            if not all(
                    ks in gamma_ints
                    and all(k in gamma_ints[ks] and v == gamma_ints[ks][k] for k, v in vs.items())
                    for ks, vs in gamma_dicts.items()
            ):
                raise ValueError(
                    f"The redundantly computed gamma integral values do not agree with the available ones:\n"
                    f"Available:\n{gamma_ints}\nComputed:\n{gamma_dicts}"
                )
        else:
            output_data = json.dumps(gamma_dicts, indent=2)
            with tempfile.NamedTemporaryFile(mode="w", dir=os.path.dirname(self.ff_ds_gammas_path)) as out_file:
                out_file.write(output_data)
                out_file.flush()
                os.fsync(out_file.fileno())
                if not os.path.exists(self.ff_ds_gammas_path):
                    os.rename(out_file.name, self.ff_ds_gammas_path)
                    open(out_file.name, 'w').close()

    def prepare_data_frame(
            self,
            df: pd.DataFrame,
            target: PathType,
            cs_train_target: Optional[PathType],
            cs_test_target: Optional[PathType],
            remove_target: Optional[PathType],
            n_entries_before_split: Optional[int]
    ) -> None:
        assert (cs_train_target is not None and cs_test_target is not None) \
               or (cs_train_target is None and cs_test_target is None)

        if self.n_subsets > 1:
            assert n_entries_before_split is not None

        n_entries_after_split = len(df.index)

        # Add unique candidate IDs
        df.loc[:, "temp_index"] = df.index
        df.loc[:, "unique_candidate_id"] = df.groupby("unique_event_id")["temp_index"].rank()
        df.drop(columns=["temp_index"], inplace=True)

        # Add Sig- and Gen-IDs
        mc_ids.add_sig_id_to_df(df=df)
        mc_ids.add_gen_b_id_to_df(df=df, gen_b_mdst_index=1)
        mc_ids.add_gen_b_id_to_df(df=df, gen_b_mdst_index=2)

        # Remove decay types from generic MC, which are covered in other dedicated samples
        if remove_target is not None:
            d_starstar_gen_ids = [51, 52, 53, 54]
            do_remove_cond = (df["GEN_B1_ID"].isin(d_starstar_gen_ids)) | (df["GEN_B2_ID"].isin(d_starstar_gen_ids))

            df_remove = df.loc[do_remove_cond]
            df_keep = df.loc[~do_remove_cond]

            assert len(df.index) == (len(df_keep.index) + len(df_remove.index))

            df_remove.to_hdf(path_or_buf=remove_target, key=self.file_name)
        else:
            df_keep = df

        # PID Correction
        my_pid_correction = ParticleIDCorrection(number_of_variations=20)
        my_pid_correction.apply_pid_correction(
            sample=df_keep,
            apply_for_lep_fakes=True,
            apply_for_had_fakes=True,
            info_output_path=self.pid_weight_info_target
        )

        wcs = [c for c in df_keep.columns if c.startswith("__weight_pid_")]
        assert not pd.isnull(df_keep[wcs]).any().any()

        my_tracking_correction = TrackingEffCorrector(consider_pi0_for_sys=True)
        my_tracking_correction.apply_tracking_correction(sample=df_keep)

        wcs = [c for c in df_keep.columns if c.startswith("__weight_slow_pi_")]
        assert not pd.isnull(df_keep[wcs]).values.any()

        my_k_short_correction = KShortEffCorrector()
        my_k_short_correction.apply_k_short_correction(sample=df_keep)

        wcs = [c for c in df_keep.columns if c.startswith("__weight_k_short_")]
        assert not pd.isnull(df_keep[wcs]).values.any()

        # MC Reweighting
        # B -> D l nu
        btodlnu_reweighter = BtoDFFReweighting(info_target_path=self.ff_d_weight_info_target)
        btodlnu_reweighter.add_weights(df_keep)
        btodlnu_reweighter.release()

        # B -> D* l nu
        btodstlnu_reweighter = BtoDstarFFReweighting(
            num_cpu=4,
            cached_gammas=self.get_precomputed_gamma_integrals(),
            info_target_path=self.ff_ds_weight_info_target
        )
        btodstlnu_reweighter.add_weights(df_keep)
        self.write_computed_ff_ds_gammas(btodstlnu_reweighter.get_gamma_integrals())
        btodstlnu_reweighter.release()

        # B -> D** l nu
        if self.data_type not in Belle_list_types:
            btodststlnu_reweighter = BtoDststFFReweighting(info_target_path=self.ff_dss_weight_info_target)
            for dstst_type in DststType:
                btodststlnu_reweighter.add_weights(df_keep, dstst_type=dstst_type)
            btodststlnu_reweighter.release()

        # Tagging Correction:
        fei_tag_corrector = TagCorrection(correct_only_true_b=False)
        fei_tag_corrector.apply_tagging_correction(sample=df_keep)

        # QED Coulomb correction:
        get_coulomb_corr_factor(
            df=df_keep,
            only_leptonic_taus=False,
            d_mesons=None,
            on_reco_sig_side=False
        )

        # Split up for CS Training and testing.
        if cs_train_target is not None:
            sample_size = int(np.ceil(int(self.cs_config["cs_sample_size"]) / self.n_subsets))

            unique_event_groups = random_permutation(df_keep.groupby("unique_event_id", group_keys=False))
            cs_train, cs_test, keep = split_into(unique_event_groups, [sample_size, sample_size, None])

            cs_train_ids = pd.concat([group["unique_event_id"] for _, group in cs_train], axis=0, sort=False).index
            cs_test_ids = pd.concat([group["unique_event_id"] for _, group in cs_test], axis=0, sort=False).index
            keep_ids = pd.concat([group["unique_event_id"] for _, group in keep], axis=0, sort=False).index

            df_cs_train = df_keep.loc[cs_train_ids]
            df_cs_test = df_keep.loc[cs_test_ids]
            df_keep_final = df_keep.loc[keep_ids]

            assert len(df_keep.index) == (len(df_cs_train.index) + len(df_cs_test.index) + len(df_keep_final.index))

            n_cs_sample = 2 * sample_size
            n_entries_before = len(df_keep["unique_event_id"].unique())
            scaling_correction = 1. * n_cs_sample / n_entries_before

            df_cs_train.to_hdf(path_or_buf=cs_train_target, key=self.file_name)
            df_cs_test.to_hdf(path_or_buf=cs_test_target, key=self.file_name)
        else:
            n_cs_sample = 0
            n_entries_before = 0
            df_keep_final = df_keep
            scaling_correction = 0.

        df_keep_final.loc[:, self.scale_corr_column] = scaling_correction

        n_entries = len(df_keep_final.index)
        df_keep_final.to_hdf(path_or_buf=target, key=self.file_name)

        if self.n_subsets > 1:
            with pd.HDFStore(path=target) as hdf5store:
                hdf5store.append(key=self.n_entries_key, value=pd.Series(n_entries))
                hdf5store.append(
                    key=DataPreparationSplitterTask.n_entries_before_split_key,
                    value=pd.Series(n_entries_before_split)
                )
                hdf5store.append(key=self.n_entries_after_split_key, value=pd.Series(n_entries_after_split))
                if cs_train_target is not None:
                    hdf5store.append(key=self.n_entries_before_key, value=pd.Series(n_entries_before))
                    hdf5store.append(key=self.n_cs_sample_key, value=pd.Series(n_cs_sample))

    def preparation_process(self, dry_run: bool = False) -> List[Tuple[PathType, PathType, PathType]]:
        source = self.input_file
        target = os.path.join(self.base_dir, f"{self.file_name}_{self.file_tag}_split_{self.subset_index}.h5")
        temp_target = os.path.join(self.base_dir, f"tmp_{self.file_name}_{self.file_tag}_split_{self.subset_index}.h5")
        output_key = f"{self.file_name}_{self.file_tag}_split_{self.subset_index}"

        source_target_list = [(source, target, output_key)]
        cs_train_target = None
        cs_test_target = None
        remove_target = None

        if self.data_type in ["Charm", "UDS"] and self.cs_config is not None:
            cs_train_target = os.path.join(self.base_dir, f"{self.file_name}_forCS_train_split_{self.subset_index}.h5")
            cs_train_output_key = f"{self.file_name}_forCS_train_split_{self.subset_index}"
            source_target_list.append((source, cs_train_target, cs_train_output_key))

            cs_test_target = os.path.join(self.base_dir, f"{self.file_name}_forCS_test_split_{self.subset_index}.h5")
            cs_test_output_key = f"{self.file_name}_forCS_test_split_{self.subset_index}"
            source_target_list.append((source, cs_test_target, cs_test_output_key))

        # TODO: Remove in all but dedicated sample or all but newly simulated samples or as is?
        if self.data_type in ["Charged", "Mixed"]:
            remove_target = os.path.join(
                self.base_dir,
                f"{self.file_name}_{self.removed_tag}_split_{self.subset_index}.h5"
            )
            remove_output_key = f"{self.file_name}_{self.removed_tag}_split_{self.subset_index}"

            source_target_list.append((source, remove_target, remove_output_key))

        if self.data_type not in rdstar_samples.recorded_data_samples:
            if self.pid_weight_info_target is not None:
                source_target_list.append((source, self.pid_weight_info_target, self.pid_weight_info_output_key))
            if self.ff_d_weight_info_target is not None:
                source_target_list.append((source, self.ff_d_weight_info_target, self.ff_d_weight_info_output_key))
            if self.ff_ds_weight_info_target is not None:
                source_target_list.append((source, self.ff_ds_weight_info_target, self.ff_ds_weight_info_output_key))
            if self.ff_dss_weight_info_target is not None and self.data_type not in Belle_list_types:
                source_target_list.append((source, self.ff_dss_weight_info_target, self.ff_dss_weight_info_output_key))

        if not dry_run:
            if self.data_type in rdstar_samples.recorded_data_samples:
                os.symlink(source, target)
            else:
                df = pd.read_hdf(path_or_buf=source, key=self.file_name)
                assert isinstance(df, pd.DataFrame)
                n_entries_before_split = None
                if self.n_subsets > 1:
                    n_entries_before_split = pd.read_hdf(
                        path_or_buf=source,
                        key=DataPreparationSplitterTask.n_entries_before_split_key
                    )[0]

                self.prepare_data_frame(
                    df=df,
                    target=temp_target,
                    cs_train_target=cs_train_target,
                    cs_test_target=cs_test_target,
                    remove_target=remove_target,
                    n_entries_before_split=n_entries_before_split
                )
                os.rename(src=temp_target, dst=target)

        return source_target_list

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.preparation_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for _, target_path, output_key in self.preparation_process(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(DataPreparationSplitterTask)


@inherits(DataPreparationSplitterTask)
class DataPreparationMergerTask(utilities.BaseOfflineTask):

    @property
    def resources(self) -> Dict[str, int]:
        if self.data_type in ["Charged", "Charm", "UDS"]:
            return {"offline_memory": 8}
        elif self.data_type in ["Mixed"]:
            return {"offline_memory": 4}
        else:
            return {"offline_memory": 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        if isinstance(offline_config["cs_config"], dict):
            self.cs_config = offline_config["cs_config"]
        else:
            self.cs_config = None

        self.has_cs_sample = self.data_type in ["UDS", "Charm"] and self.cs_config is not None

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_output_key(self, tag: str) -> str:
        return f"{self.file_name}_{tag}"

    def get_output_path(self, tag: str) -> PathType:
        return os.path.join(self.base_dir, f"{self.file_name}_{tag}.h5")

    def get_temp_output_path(self, tag: str) -> PathType:
        return os.path.join(self.base_dir, f"temp_{self.file_name}_{tag}.h5")

    def get_file_type_dict(self) -> Dict[str, Dict[int, PathType]]:
        file_type_dict = defaultdict(dict)

        for key, path in self.input_dict.items():
            name_and_tag, index_str = key.split("_split_")
            assert index_str.isdigit()
            index = int(index_str)
            tag = name_and_tag.split(f"{self.file_name}_")[1]
            file_type_dict[tag][index] = path

        assert DataPreparationTask.base_file_tag in file_type_dict.keys()
        assert all(len(d) == len(file_type_dict[list(file_type_dict.keys())[0]]) for d in file_type_dict.values())

        return file_type_dict

    def merge_split_files(self, file_type_dict: Dict[str, Dict[int, PathType]]) -> None:
        for tag, files in file_type_dict.items():
            df = pd.DataFrame()
            n_entries = []
            n_entries_before_split = []
            n_entries_after_split = []
            n_cs_sample = []
            n_before = []

            hdf_key = self.file_name
            if "weight_info" in tag:
                hdf_key = "weight_info"

            for index in range(len(files)):
                df_part = pd.read_hdf(files[index], key=hdf_key)

                if tag == DataPreparationTask.base_file_tag:
                    n_entries.append(pd.read_hdf(files[index], key=DataPreparationTask.n_entries_key)[0])
                    n_entries_before_split.append(pd.read_hdf(
                        path_or_buf=files[index],
                        key=DataPreparationSplitterTask.n_entries_before_split_key
                    )[0])
                    n_entries_after_split.append(pd.read_hdf(
                        path_or_buf=files[index],
                        key=DataPreparationTask.n_entries_after_split_key
                    )[0])

                if tag == DataPreparationTask.base_file_tag and self.has_cs_sample:
                    n_cs_sample.append(pd.read_hdf(files[index], key=DataPreparationTask.n_cs_sample_key)[0])
                    n_before.append(pd.read_hdf(files[index], key=DataPreparationTask.n_entries_before_key)[0])

                df = df.append(df_part, sort=False)

            if tag == DataPreparationTask.base_file_tag:
                assert len(set(n_entries_before_split)) == 1
                assert sum(n_entries_after_split) == n_entries_before_split[0]
                assert sum(n_entries) == len(df.index)

            # Update scaling correction for UDS and Charm
            if tag == DataPreparationTask.base_file_tag and self.has_cs_sample:
                new_scaling_correction = sum(n_cs_sample) * 1. / sum(n_before)
                df.loc[:, DataPreparationTask.scale_corr_column] = new_scaling_correction

            df.to_hdf(path_or_buf=self.get_temp_output_path(tag=tag), key=hdf_key)
            os.rename(src=self.get_temp_output_path(tag=tag), dst=self.get_output_path(tag=tag))

    def link_files(self, file_type_dict: Dict[str, Dict[int, PathType]]) -> None:
        for tag, files in file_type_dict.items():
            assert len(files) == 1
            os.symlink(files[0], self.get_output_path(tag=tag))

    def merge_process(self, dry_run: bool = False) -> Dict[str, PathType]:
        file_type_dict = self.get_file_type_dict()

        output_dict = {self.get_output_key(tag): self.get_output_path(tag) for tag in file_type_dict.keys()}

        if not dry_run:
            if self.n_subsets > 1:
                self.merge_split_files(file_type_dict=file_type_dict)
            else:
                self.link_files(file_type_dict=file_type_dict)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.merge_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {
            key: luigi.LocalTarget(os.fspath(target)) for key, target in self.merge_process(dry_run=True).items()
        }
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for subset_index in range(self.n_subsets):
            yield self.clone(DataPreparationTask, subset_index=subset_index)
