"""
The objective of the tasks provided here are:
    - the determination of the difference in resolution of the observable missing mass squared on data and MC
    - and the application of a smearing based on this difference on the MC data sets.
The Resolution difference is obtained from the q^2 sideband.

Felix Metzner 2019
"""

import os
import pandas as pd
import multiprocessing
import b2luigi as luigi

from collections import abc
from functools import partial
from typing import Optional, Tuple, List, Dict, Generator

from luigi.util import inherits

from rdstar.offline_analysis import rdstar_samples

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.utilities import InputDictType as IDType
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightAnalysisTask

from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.sideband_validation import plot_data_mc_comparison, SidebandInfo, \
    reset_systematics_in_sideband_info

from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.corrections.resolution_correction import MissingMassResolutionEvaluator, \
    MissingMassResolutionCorrector

from templatefitter.plotter.histogram_variable import HistVariable


__all__ = [
    "ResolutionEvaluationTask",
    "ResolutionCorrectionTask",
    "ResolutionCorrectionValidationTask"
]


@inherits(WeightAnalysisTask)
class ResolutionEvaluationTask(utilities.BaseOfflineTask):
    base_dir_name = "ResolutionEvaluation"  # type: str
    resolution_eval_result = "missing_mass_squared_resolution_eval_result.json"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name)  # type: os.fspath

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"  # type: str
        self.sideband_key = "q2_sideband"  # type: str

        self.mc_sample_keys = rdstar_samples.main_sample_keys  # type: List[str]
        self.data_sample_key = "DataY4S"  # type: str

        self.full_weight_column_name = "__weight_final_wFEI_corr__"  # TODO: Remove once provided by weight finalization

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}  # type: IDType

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"
        mc_paths = [  # type: List[os.fspath]
            p for mc_key in self.mc_sample_keys for k, p in self.input_dict.items()
            if (k.startswith(f"{mc_key}_{WeightFinalizationTask.appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p for k, p in self.input_dict.items()
            if (k.startswith(f"{self.data_sample_key}_{WeightFinalizationTask.appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        return mc_sample, data_sample

    def add_fei_correction(self, mc_sample: pd.DataFrame) -> None:
        assert self.full_weight_column_name not in mc_sample.columns
        assert WeightFinalizationTask.final_weight_col in mc_sample.columns
        assert TagCorrection.tagging_correction_name in mc_sample.columns

        product_cols = [
            WeightFinalizationTask.final_weight_col,
            TagCorrection.tagging_correction_name
        ]
        mc_sample.loc[:, self.full_weight_column_name] = mc_sample[product_cols].product(axis=1)

    def evaluate_resolution(self, dry_run: bool = False) -> Optional[Dict[str, os.fspath]]:
        resolution_evaluator = MissingMassResolutionEvaluator(
            weight_column=self.full_weight_column_name,
            results_path=self.base_path
        )

        if dry_run:
            return resolution_evaluator.get_output_dict()

        mc_sample, data_sample = self.get_mc_and_data_sample()
        self.add_fei_correction(mc_sample=mc_sample)
        resolution_evaluator.evaluate_resolution(mc_df=mc_sample, data_df=data_sample)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.evaluate_resolution(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.evaluate_resolution(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield WeightFinalizationTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations
            )


@inherits(WeightFinalizationTask)
class ResolutionCorrectionTask(utilities.BaseOfflineTask):
    appendix = "withMMSQResCorr"
    original_mmsq_col = "m2RecoilSignalSide"
    new_mmsq_col = MissingMassResolutionCorrector.get_corrected_col_name(original_column=original_mmsq_col)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}  # type: IDType

    def get_input_files(self) -> Dict[str, os.fspath]:
        key_path_dict = {
            key: path for key, path in self.input_dict.items()
            if WeightFinalizationTask.appendix in key and os.path.splitext(path)[1] == ".h5"
        }
        assert all(key.startswith(self.combo_name) for key in key_path_dict.keys()), key_path_dict
        return key_path_dict

    def get_resolution_correction_info_path(self) -> os.fspath:
        return self.input_dict[MissingMassResolutionEvaluator.default_result_file]

    def apply_resolution_correction_for_key(
            self,
            key_path_tuple: Tuple[str, os.fspath],
            dry_run: bool = False
    ) -> Tuple[str, os.fspath]:
        input_key, input_path = key_path_tuple

        dir_path, file_name = os.path.split(input_path)
        output_name_1, output_name_2 = file_name.split(f"_{WeightFinalizationTask.appendix}", 1)
        output_name = f"{output_name_1}_{WeightFinalizationTask.appendix}_{self.appendix}_{output_name_2}"
        output_path = os.path.join(dir_path, output_name)
        temp_output_path = os.path.join(dir_path, f"temp_{output_name}")
        output_key_1, output_key_2 = input_key.split(f"_{WeightFinalizationTask.appendix}", 1)
        output_key = f"{output_key_1}_{WeightFinalizationTask.appendix}_{self.appendix}_{output_key_2}"

        if dry_run:
            return output_key, output_path

        assert os.path.isfile(input_path), key_path_tuple
        sample = pd.read_hdf(input_path)
        assert isinstance(sample, pd.DataFrame)

        if input_key.startswith("Data"):
            MissingMassResolutionCorrector.add_new_missing_mass_column_for_data(
                sample=sample,
                original_column=self.original_mmsq_col,
                new_column=self.new_mmsq_col
            )
        else:
            res_corrector = MissingMassResolutionCorrector(
                resolution_info=self.get_resolution_correction_info_path(),
                original_column=self.original_mmsq_col,
                new_column=self.new_mmsq_col
            )
            res_corrector.apply_resolution_correction(sample=sample, correct_overflows=True)

        sample.to_hdf(temp_output_path, key=output_key)
        os.rename(temp_output_path, output_path)
        return output_key, output_path

    def apply_resolution_correction(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        with multiprocessing.Pool(processes=10) as pool:
            output_tuples = pool.map(
                partial(ResolutionCorrectionTask.apply_resolution_correction_for_key, self, dry_run=dry_run),
                self.get_input_files().items()
            )

        return {output_key:  output_path for output_key, output_path in output_tuples}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.apply_resolution_correction(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.apply_resolution_correction(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(WeightFinalizationTask)
        yield ResolutionEvaluationTask(
            working_dir=self.working_dir,
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
            config_combinations=self.config_combinations
        )


@inherits(ResolutionEvaluationTask)
class ResolutionCorrectionValidationTask(utilities.BaseOfflineTask):
    base_dir_name = "ResolutionEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.before_col = ResolutionCorrectionTask.original_mmsq_col
        self.after_col = ResolutionCorrectionTask.new_mmsq_col

        self.sideband_key = "q2_sideband"  # type: str

        self.key_prefix = f"{WeightFinalizationTask.appendix}_{ResolutionCorrectionTask.appendix}_" \
                          f"{self.selection_name}_with_CS_{self.cs_tag}_"  # type: str

        self.full_weight_column_name = "__weight_final_wFEI_corr__"  # TODO: Remove once provided by weight finalization

        self.figure_size = (12., 4.)  # type: Tuple[float, float]

        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name)  # type: os.fspath
        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}  # type: IDType

    def get_hist_variables(self) -> List[Tuple[HistVariable, bool]]:
        return [(HistVariable(
            df_label=var,
            n_bins=250,
            scope=(-0.5, 2.),
            var_name=r"$\mathrm{M}_\mathrm{miss}^2$",
            unit=r"$\mathrm{GeV}^2$"
        ), var == self.after_col) for var in [self.before_col, self.after_col]]

    def get_sideband_infos(self) -> List[Tuple[SidebandInfo, bool]]:
        title_end = "Resolution Correction"
        q2_sideband_title = r"$\mathrm{q}^2$ Sideband"
        q2_sideband_text = r"$\mathrm{q}^2 < 4.0$ GeV$^2$"
        sideband_infos = [(SidebandInfo(
            key="q2_sideband",
            hist_var=var,
            title=q2_sideband_title + f"{var.variable_name} {'after' if is_corrected else 'before'} {title_end}",
            sb_text=q2_sideband_text,
            systematics=None
        ), is_corrected) for var, is_corrected in self.get_hist_variables()]

        sideband_infos_with_full_sys = [(reset_systematics_in_sideband_info(
            sideband_info=sbi,
            new_systematics=full_systematics_cols,
            sys_name_tag=full_systematics_name), is_corrected) for sbi, is_corrected in sideband_infos]

        return sideband_infos + sideband_infos_with_full_sys

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"
        mc_paths = [  # type: List[os.fspath]
            p for mc_key in self.mc_sample_keys for k, p in self.input_dict.items()
            if (k == f"{mc_key}_{full_key}" and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p for k, p in self.input_dict.items()
            if (k == f"{self.data_sample_key}_{full_key}_" and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        return mc_sample, data_sample

    def add_fei_correction(self, mc_sample: pd.DataFrame) -> None:
        assert self.full_weight_column_name not in mc_sample.columns
        assert WeightFinalizationTask.final_weight_col in mc_sample.columns
        assert TagCorrection.tagging_correction_name in mc_sample.columns

        product_cols = [
            WeightFinalizationTask.final_weight_col,
            TagCorrection.tagging_correction_name
        ]
        mc_sample.loc[:, self.full_weight_column_name] = mc_sample[product_cols].product(axis=1)

    def do_resolution_correction_evaluation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        mc_sample, data_sample = self.get_mc_and_data_sample()
        self.add_fei_correction(mc_sample=mc_sample)

        output_dict = {}

        for sideband_info, is_corrected in self.get_sideband_infos():
            for style in sideband_info.styles:
                new_output = plot_data_mc_comparison(
                    mc_df=mc_sample,
                    data_df=data_sample,
                    weight_col=self.full_weight_column_name,
                    sideband_info=sideband_info,
                    dec_modes=None,
                    dec_titles=None,
                    export_path=self.base_path,
                    style=style,
                    ratio_type="vs_uncert",
                    gof_check_method="toys",
                    fig_size=self.figure_size,
                    normalize_mc_to_data=True,
                    adaptive_binning=True,
                    dry_run=dry_run
                )
                output_dict.update(new_output)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_resolution_correction_evaluation(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_resolution_correction_evaluation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations
            )
