"""
Tasks producing the Results of the Analysis

Felix Metzner 2019
"""

import os
import json
import stat
import tempfile
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits

from typing import Union, Tuple, List, Dict, Generator

from analysistools.statistics import limits as rdstar_limits

from rdstar.offline_analysis import mc_ids
from rdstar.offline_analysis.selection import Selection
from rdstar.offline_analysis import rdstar_sample_weights

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.weight_tasks import WeightAnalysisTask
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask


__all__ = [
    "EfficiencyEvaluationTask"
]


@inherits(WeightAnalysisTask)
class EfficiencyEvaluationTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, "Efficiencies")

        self.eff_output_key = "efficiencies"
        self.eff_target_path = os.path.join(self.base_path, "efficiency_dict.json")

        self.relevant_combos = ["SignalEff", "DssSignalEff", "DssPiSignalEff", "DssPiPlusSignalEff"]

        self.sig_id_col = "SIG_ID"
        self.decay_mode_id_col = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_relevant_data_types(self) -> List[str]:
        data_combo_dict = SampleCombinerTask.get_data_combo_dict()
        assert all(k in data_combo_dict for k in self.relevant_combos)
        relevant_data_types = [v for type_key in self.relevant_combos for v in data_combo_dict[type_key]]

        assert all(k in rdstar_sample_weights.signal_mc_evts.keys() for k in relevant_data_types)

        return relevant_data_types

    def get_efficiency_samples(self) -> List[Tuple[str, os.fspath]]:
        main_selection_tag = Selection.get_main_selection_tag(selection_name=self.selection_name, cs_tag=self.cs_tag)
        sample_key_path_list = [
            (k.split("_", 1)[0], v) for k, v in self.input_dict.items()
            if k.endswith(main_selection_tag) and v.endswith(".h5")
        ]
        return sample_key_path_list

    @staticmethod
    def calculate_efficiency_values(
            df: pd.DataFrame,
            n_before: int,
            dictionary,
            mask=None,
            name=None
    ) -> None:
        assert (mask is None) == (name is None)
        assert "before" in dictionary.keys()

        after_name = f"{name}_after" if name is not None else "after"
        eff_name = f"{name}_eff" if name is not None else "eff"
        up_name = f"{name}_up" if name is not None else "up"
        low_name = f"{name}_low" if name is not None else "low"

        assert after_name not in dictionary.keys()
        assert eff_name not in dictionary.keys()
        assert up_name not in dictionary.keys()
        assert low_name not in dictionary.keys()

        if mask is None:
            n_after = len(df.index)
        else:
            n_after = len(df.loc[mask].index)
        efficiency = 1. * n_after / n_before
        upper, lower = rdstar_limits.binomial_limit(n=n_before, k=n_after, sigma=1)

        dictionary.update({
            after_name: n_after,
            eff_name: efficiency,
            up_name: upper,
            low_name: lower
        })

    @staticmethod
    def get_sig_id(data_type: str) -> int:
        if data_type in SampleCombinerTask.get_data_combo_dict()["DssSignalEff"]:
            sig_type = "DSSl"
        elif data_type in SampleCombinerTask.get_data_combo_dict()["DssPiSignalEff"]:
            sig_type = "DSSl"
        elif data_type in SampleCombinerTask.get_data_combo_dict()["DssPiPlusSignalEff"]:
            sig_type = "DSSlbkg"
        elif data_type in SampleCombinerTask.get_data_combo_dict()["SignalEff"]:
            if data_type.startswith("DS"):
                if data_type.endswith("Tau"):
                    sig_type = "DStau"
                else:
                    sig_type = "DSl"
            else:
                if data_type.endswith("Tau"):
                    sig_type = "Dtau"
                else:
                    sig_type = "Dl"
        else:
            raise KeyError(f"Trying to calculate signal efficiency for unexpected data type {data_type}...")

        return mc_ids.rdstar_sig_queries[sig_type]["id"]

    def evaluate_efficiencies(self) -> Dict[str, Dict[str, Union[float, int]]]:
        efficiency_dict = dict()

        for data_type, df_path in self.get_efficiency_samples():
            df = pd.read_hdf(df_path)
            assert isinstance(df, pd.DataFrame)
            n_before = rdstar_sample_weights.signal_mc_evts[data_type]["meta"]
            sig_id = self.get_sig_id(data_type=data_type)

            efficiency_dict[data_type] = {
                "before": n_before,
                "sig_id": sig_id
            }

            self.calculate_efficiency_values(df=df, n_before=n_before, dictionary=efficiency_dict[data_type])

            sig_mask = df[self.sig_id_col] == sig_id
            self.calculate_efficiency_values(
                df=df,
                n_before=n_before,
                dictionary=efficiency_dict[data_type],
                mask=sig_mask,
                name="sig"
            )

            dss_mask = df[self.decay_mode_id_col].isin([61, 62, 63, 64, 65, 66, 67, 68])
            self.calculate_efficiency_values(
                df=df,
                n_before=n_before,
                dictionary=efficiency_dict[data_type],
                mask=(sig_mask & dss_mask),
                name="sig_dstst"
            )

            dss_charged_mask = df[self.decay_mode_id_col].isin([71, 72, 73, 74, 75, 76, 77, 78])
            self.calculate_efficiency_values(
                df=df,
                n_before=n_before,
                dictionary=efficiency_dict[data_type],
                mask=(sig_mask & dss_charged_mask),
                name="sig_dstst_charged"
            )

        return efficiency_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)

        eff_dict = self.evaluate_efficiencies()
        assert len(eff_dict) > 0

        os.makedirs(self.base_path, exist_ok=True)

        output_data = json.dumps(eff_dict, indent=2)
        with tempfile.NamedTemporaryFile(mode="w", dir=self.base_path) as out_file:
            out_file.write(output_data)
            out_file.flush()
            os.fsync(out_file.fileno())
            os.rename(out_file.name, self.eff_target_path)
            os.chmod(path=self.eff_target_path, mode=(stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH))
            open(out_file.name, 'w').close()

    def output(self) -> Dict[str, os.fspath]:
        return {self.eff_output_key: luigi.LocalTarget(self.eff_target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        used_data_types = []
        for _, _, relevant_configs in self.config_combinations:
            for params in relevant_configs:
                if params["data_type"] in self.get_relevant_data_types():
                    used_data_types.append(params["data_type"])
                    yield DataSelectionTask(selection_name=self.selection_name, **params)
        if len(used_data_types) < 1:
            raise RuntimeError(f"No data for efficiency evaluation available!\n"
                               f"Please ensure that any of the following data types are processed:"
                               f"\n {self.get_relevant_data_types()}")
