"""
Signal Selection Tasks

    - Application of Continuum Suppression
    - Best candidate selection
    - Signal selection cuts
    - Creation of side band samples


Felix Metzner 2019
"""

import os
import json
import copy
import yaml
import warnings
import numpy as np
import pandas as pd
import b2luigi as luigi

from collections import defaultdict
from luigi.util import inherits, requires
from typing import Optional, Union, List, Tuple, Dict, Any, Generator

from pandas.core.common import SettingWithCopyWarning

from analysistools.dataframe_tools import hdf5_utils

from rdstar.offline_analysis import rdstar_samples
from rdstar import continuum_suppression as cs_tools
from rdstar.offline_analysis import offline_observables
from rdstar.offline_analysis import rdstar_sample_weights
from rdstar.offline_analysis.rdstar_plots import decay_mode_id_dict, u4s_decay_mode_id_dict

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.cs_tasks import CSTrainTask, CSTestTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationSplitterTask
from rdstar.offline_analysis.selection import Selection, SelectionInfo, BCSVariableInfo
from rdstar.offline_analysis.mc_ids import rdstar_sig_queries, gen_b_ids, decay_mode_to_sig_id
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask


__all__ = [
    "DataReductionTask",
    "DataSelectionTask"
]


@inherits(CSTrainTask)
@inherits(DataPreparationSplitterTask)
class DataReductionTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        if isinstance(offline_config["cs_config"], dict):
            self.cs_offline_config_path = self.offline_config_path
            cs_config = offline_config["cs_config"]
        else:
            configs_dir = utilities.get_config_directory()
            self.cs_offline_config_path = os.path.join(configs_dir, offline_config["cs_config"])

            if not os.path.isfile(self.cs_offline_config_path):
                raise RuntimeError(f"Encountered invalid offline_config path for cs training:\n"
                                   f"\t{self.cs_offline_config_path}")

            with open(self.cs_offline_config_path) as cs_f:
                cs_offline_config = yaml.load(cs_f)
                cs_config = cs_offline_config["cs_config"]

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.input_key = f"{self.file_name}_{DataPreparationTask.base_file_tag}"

        self.target_path = os.path.join(self.base_dir, f"{self.file_name}_with_CS_{self.cs_tag}.h5")
        self.tmp_target_path = os.path.join(self.base_dir, f"tmp_{self.file_name}_with_CS_{self.cs_tag}.h5")
        self.output_key = f"{self.file_name}_with_CS_{self.cs_tag}"

        self.cs_column_name = cs_config["cs_column_name"]
        self.cs_cut_value = cs_config["cs_cut_value"] if "cs_cut_value" in cs_config else None

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def apply_cs_mva(self, df: pd.DataFrame) -> None:
        cs_classifier = cs_tools.ContinuumSuppression.load_classifier(
            weight_file_path=self.input_dict["cs_output_weight_file"],
            var_file_path=self.input_dict["cs_var_list_file"]
        )
        print("  Applying CS classifier to data frame...")

        cs_classifier.apply(data=df, output_column=self.cs_column_name)

        if self.cs_cut_value is not None:
            print("  Applying cut on CS output to data frame...")
            df.query(f"{self.cs_column_name} > {self.cs_cut_value}", inplace=True)

    @staticmethod
    def discard_unneeded_columns(df: pd.DataFrame) -> None:
        discard_cols = [c for c in offline_observables.observables_to_discard if c in df.columns]
        df.drop(discard_cols, axis=1, inplace=True)

    def add_scaling_weights(self, df: pd.DataFrame) -> None:
        rdstar_sample_weights.add_scaling_weights(
            df=df,
            data_type=self.data_type,
            generic_stream_dict=None
        )

    def data_reduction(self) -> None:
        df = hdf5_utils.load_hdf5(hdf5_file_path=self.input_dict[self.input_key], hdf5_key=self.file_name)

        self.apply_cs_mva(df=df)
        self.discard_unneeded_columns(df=df)
        self.add_scaling_weights(df=df)

        df.to_hdf(path_or_buf=self.tmp_target_path, key=self.file_name)
        os.rename(src=self.tmp_target_path, dst=self.target_path)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.data_reduction()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {self.output_key: luigi.LocalTarget(self.target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CSTrainTask, offline_config_path=self.cs_offline_config_path)
        yield self.clone(CSTestTask, offline_config_path=self.cs_offline_config_path)
        yield self.clone(DataPreparationMergerTask)


@requires(DataReductionTask)
class DataSelectionTask(utilities.BaseOfflineTask):
    selection_name = luigi.Parameter(positional=False)

    evaluation_info_prefix = "selection_info"

    # Column name information:
    u_eid_col = "unique_event_id"
    is_sig_col = "isSignalAcceptMissingNeutrino"
    is_b_sig_col = "daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc"
    decay_mode_id_col = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        with open(self.offline_config_path) as f:
            offline_config = yaml.load(f)

        self.working_dir = offline_config["working_dir"]
        self.base_dir = os.path.join(self.working_dir, self.data_type)

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.bcs_random_seed = 1337

        self.file_name = f'{self.data_type}{self.stream_tag}_{self.tuple_type}'
        self.input_key = f"{self.file_name}_with_CS_{self.cs_tag}"

        self.input_dict = {k: v.path for k, v in self.input().items()}

    @staticmethod
    def apply_cuts(df: pd.DataFrame, cut_list: List[str]) -> pd.DataFrame:
        cut_string = " and ".join(cut_list)
        return df.query(cut_string)

    def apply_bcs(
            self,
            df: pd.DataFrame,
            primary_var: BCSVariableInfo,
            secondary_var: BCSVariableInfo
    ) -> pd.DataFrame:
        assert isinstance(primary_var, BCSVariableInfo), (primary_var, type(primary_var))
        assert isinstance(secondary_var, BCSVariableInfo), (secondary_var, type(secondary_var))

        if len(df.index) == 0:
            return df

        if primary_var.bcs_variable is None:
            return df

        if primary_var.bcs_variable == "random":
            return df.sample(frac=1.0, random_state=self.bcs_random_seed).groupby(self.u_eid_col).head(1).sort_index()

        prim_var_name = primary_var.bcs_variable
        if primary_var.absolute_value:
            prim_var_name = f"bcs_abs_{primary_var.bcs_variable}"
            assert prim_var_name not in df.columns

            # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
            with warnings.catch_warnings(record=True) as warning:
                warnings.simplefilter("always")
                # Actual command:
                df.loc[:, prim_var_name] = np.abs(df[primary_var.bcs_variable])
                for w in warning:
                    assert issubclass(w.category, SettingWithCopyWarning)
                    assert "A value is trying to be set on a copy" in str(w.message)

        if secondary_var.bcs_variable is not None:
            # Use secondary bcs var if duplicates for best primary bcs variable exist
            sec_var_name = secondary_var.bcs_variable
            if secondary_var.absolute_value:
                sec_var_name = f"bcs_abs_{secondary_var.bcs_variable}"
                assert sec_var_name not in df.columns

                # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
                with warnings.catch_warnings(record=True) as warning:
                    warnings.simplefilter("always")
                    # Actual command:
                    df.loc[:, sec_var_name] = np.abs(df[secondary_var.bcs_variable])
                    for w in warning:
                        assert issubclass(w.category, SettingWithCopyWarning)
                        assert "A value is trying to be set on a copy" in str(w.message)

            if primary_var.max:
                selected_ids = df.sample(
                    frac=1.,
                    random_state=self.bcs_random_seed
                ).sort_values(
                    by=[sec_var_name],
                    ascending=True if not secondary_var.max else False
                ).groupby(self.u_eid_col)[prim_var_name].idxmax()

                if secondary_var.absolute_value:
                    del df[sec_var_name]
            else:
                selected_ids = df.sample(
                    frac=1.,
                    random_state=self.bcs_random_seed
                ).sort_values(
                    by=[sec_var_name],
                    ascending=True if not secondary_var.max else False
                ).groupby(self.u_eid_col)[prim_var_name].idxmin()

                if secondary_var.absolute_value:
                    del df[sec_var_name]
        else:
            # Don't use a secondary bcs variable, instead choose candidate by random
            # if duplicates for best primary bcs variable exist
            if primary_var.max:
                selected_ids = df.sample(
                    frac=1.,
                    random_state=self.bcs_random_seed
                ).groupby(self.u_eid_col)[prim_var_name].idxmax()
            else:
                selected_ids = df.sample(
                    frac=1.,
                    random_state=self.bcs_random_seed
                ).groupby(self.u_eid_col)[prim_var_name].idxmin()

        if primary_var.absolute_value:
            del df[prim_var_name]

        return df.loc[selected_ids]

    def apply_selection(
            self,
            df: pd.DataFrame,
            selection_info: SelectionInfo,
            evaluate_selection: bool = False
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, Dict[str, Any]]]:
        if not evaluate_selection:
            selected_df = self.apply_cuts(df=df, cut_list=selection_info.selection["cuts"])
            selected_df = self.apply_bcs(
                df=selected_df,
                primary_var=selection_info.selection["primary_bcs_var"],
                secondary_var=selection_info.selection["secondary_bcs_var"]
            )
            return selected_df
        else:
            evaluation_info = {}
            evaluation_info.update({"0_initial_no_cuts": self.get_evaluation_info(df=df)})
            applied_cuts = []
            selected_df = copy.copy(df)
            for num, cut in enumerate(selection_info.selection["cuts"]):
                applied_cuts.append(cut)
                selected_df = self.apply_cuts(df=selected_df, cut_list=applied_cuts)
                evaluation_info.update({
                    f"{num + 1}_after_cut_{num + 1}": self.get_evaluation_info(df=selected_df, cuts=applied_cuts)
                })

            selected_df = self.apply_bcs(
                df=selected_df,
                primary_var=selection_info.selection["primary_bcs_var"],
                secondary_var=selection_info.selection["secondary_bcs_var"]
            )
            bcs = (selection_info.selection['primary_bcs_var'], selection_info.selection['secondary_bcs_var'])
            assert isinstance(bcs[0], BCSVariableInfo), type(bcs[0])
            assert isinstance(bcs[1], BCSVariableInfo), type(bcs[1])

            index = len(applied_cuts) + 1
            evaluation_info.update({
                f"{index}_after_bcs": self.get_evaluation_info(df=selected_df, cuts=applied_cuts, bcs=bcs)
            })

            return selected_df, evaluation_info

    def add_vs_comparisons(
            self,
            info_dict: Dict[str, Any],
            df: pd.DataFrame,
            info_name: str,
            col: str,
            cond_col: str,
            cond_val: int
    ) -> None:
        cond = df[cond_col] == cond_val
        assert f"{info_name}_{cond_val}" not in info_dict, f"{info_name}_{cond_val}"
        assert f"u_{info_name}_{cond_val}" not in info_dict, f"u_{info_name}_{cond_val}"

        vs_info = df.loc[cond, col].value_counts()
        info_dict.update({
            f"{info_name}_{cond_val}": {k: v for k, v in zip(vs_info.index, vs_info)}
        })
        u_vs_info = df.loc[cond].groupby(self.u_eid_col)[col].first().value_counts()
        info_dict.update({
            f"u_{info_name}_{cond_val}": {k: v for k, v in zip(u_vs_info.index, u_vs_info)}
        })

    def add_cut_and_bcs_validation_info(self, info_dict: Dict[str, Any], df: pd.DataFrame) -> None:
        assert isinstance(df, pd.DataFrame)
        res = defaultdict(dict)
        is_sig_strict = df[self.is_sig_col] == 1
        is_b_sig_strict = df[self.is_b_sig_col] == 1

        for decay_id in decay_mode_id_dict():
            is_rec = df[self.decay_mode_id_col] == decay_id
            is_sig = df["SIG_ID"].isin(decay_mode_to_sig_id[decay_id])

            if len(decay_mode_to_sig_id[decay_id]) == 2:
                is_sig_wn = df["SIG_ID"].isin(decay_mode_to_sig_id[decay_id][:1])
                is_norm = df["SIG_ID"].isin(decay_mode_to_sig_id[decay_id][1:])
            else:
                is_sig_wn = is_sig
                is_norm = False

            sig = len(df.loc[is_rec & is_sig].index)
            res["cut_info_sig"].update({decay_id: sig})
            sig_wn = len(df.loc[is_rec & is_sig_wn].index)
            res["cut_info_wn_sig"].update({decay_id: sig_wn})
            norm = len(df.loc[is_rec & is_norm].index)
            res["cut_info_wn_norm"].update({decay_id: norm})
            sig_u = len(df.loc[is_rec & is_sig].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_sig"].update({decay_id: sig_u})
            sig_wn_u = len(df.loc[is_rec & is_sig_wn].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_sig"].update({decay_id: sig_wn_u})
            norm_u = len(df.loc[is_rec & is_norm].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_norm"].update({decay_id: norm_u})

            strict_sig = len(df.loc[is_rec & is_sig_strict].index)
            res["cut_info_strict_sig"].update({decay_id: strict_sig})
            strict_b_sig = len(df.loc[is_rec & is_b_sig_strict].index)
            res["cut_info_strict_b_sig"].update({decay_id: strict_b_sig})
            strict_sig_u = len(df.loc[is_rec & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_strict_sig"].update({decay_id: strict_sig_u})
            strict_b_sig_u = len(df.loc[is_rec & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_strict_b_sig"].update({decay_id: strict_b_sig_u})

            v_strict_sig = len(df.loc[is_rec & is_sig & is_sig_strict].index)
            res["cut_info_v_strict_sig"].update({decay_id: v_strict_sig})
            v_strict_b_sig = len(df.loc[is_rec & is_sig & is_b_sig_strict].index)
            res["cut_info_v_strict_b_sig"].update({decay_id: v_strict_b_sig})
            v_strict_sig_wn = len(df.loc[is_rec & is_sig_wn & is_sig_strict].index)
            res["cut_info_wn_v_strict_sig"].update({decay_id: v_strict_sig_wn})
            v_strict_b_sig_wn = len(df.loc[is_rec & is_sig_wn & is_b_sig_strict].index)
            res["cut_info_wn_v_strict_b_sig"].update({decay_id: v_strict_b_sig_wn})
            v_strict_norm = len(df.loc[is_rec & is_norm & is_sig_strict].index)
            res["cut_info_wn_v_strict_norm"].update({decay_id: v_strict_norm})
            v_strict_b_norm = len(df.loc[is_rec & is_norm & is_b_sig_strict].index)
            res["cut_info_wn_v_strict_b_norm"].update({decay_id: v_strict_b_norm})
            v_strict_sig_u = len(df.loc[is_rec & is_sig & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_v_strict_sig"].update({decay_id: v_strict_sig_u})
            v_stri_b_sig_u = len(df.loc[is_rec & is_sig & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_v_strict_b_sig"].update({decay_id: v_stri_b_sig_u})
            v_strict_sig_wn_u = len(df.loc[is_rec & is_sig_wn & is_sig_strict].drop_duplicates(
                subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_sig"].update({decay_id: v_strict_sig_wn_u})
            v_strict_b_sig_wn_u = len(df.loc[is_rec & is_sig_wn & is_b_sig_strict].drop_duplicates(
                subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_b_sig"].update({decay_id: v_strict_b_sig_wn_u})
            v_strict_norm_u = len(df.loc[is_rec & is_norm & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_norm"].update({decay_id: v_strict_norm_u})
            v_st_b_norm_u = len(df.loc[is_rec & is_norm & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_b_norm"].update({decay_id: v_st_b_norm_u})

            bkg = len(df.loc[is_rec & ~is_sig].index)
            res["bcs_info_bkg"].update({decay_id: bkg})
            bkg_wn = len(df.loc[is_rec & ~is_sig_wn & ~is_norm].index)
            res["bcs_info_wn_bkg"].update({decay_id: bkg_wn})

            strict_bkg = len(df.loc[is_rec & ~is_sig_strict].index)
            res["bcs_info_strict_bkg"].update({decay_id: strict_bkg})
            strict_b_bkg = len(df.loc[is_rec & ~is_b_sig_strict].index)
            res["bcs_info_strict_b_bkg"].update({decay_id: strict_b_bkg})

            v_strict_bkg = len(df.loc[is_rec & ~(is_sig & is_sig_strict)].index)
            res["bcs_info_v_strict_bkg"].update({decay_id: v_strict_bkg})
            v_strict_b_bkg = len(df.loc[is_rec & ~(is_sig & is_b_sig_strict)].index)
            res["bcs_info_v_strict_b_bkg"].update({decay_id: v_strict_b_bkg})
            v_strict_bkg_wn = len(df.loc[is_rec & ~(is_sig_wn & is_sig_strict) & ~(is_norm & is_sig_strict)].index)
            res["bcs_info_wn_v_strict_bkg"].update({decay_id: v_strict_bkg_wn})
            v_stric_b_bkg_wn = len(df.loc[is_rec & ~(is_sig_wn & is_b_sig_strict) & ~(is_norm & is_b_sig_strict)].index)
            res["bcs_info_wn_v_strict_b_bkg"].update({decay_id: v_stric_b_bkg_wn})

        assert all(k not in info_dict for k in res)
        info_dict.update(res)

    def add_event_multiplicity_info(self, info_dict: Dict[str, Any], df: pd.DataFrame) -> None:
        em_name = "event_multiplicities"
        em_per_dt_name = "event_multiplicities_per_data_type"
        assert em_name not in info_dict.keys(), em_name
        assert em_per_dt_name not in info_dict.keys(), em_per_dt_name

        # Event multiplicity; get histogram by applying value_counts twice:
        multiplicity_hist = df[self.u_eid_col].value_counts().value_counts()
        info_dict.update({em_name: {k: v for k, v in zip(multiplicity_hist.index, multiplicity_hist)}})
        info_dict.update({em_per_dt_name: {str(self.data_type): info_dict[em_name]}})

        em_per_decay_id_name = "event_multiplicities_per_decay_id"
        assert em_per_decay_id_name not in info_dict.keys(), em_per_decay_id_name
        em_per_decay_id_is_sig_name = "event_multiplicities_is_sig_per_decay_id"
        assert em_per_decay_id_is_sig_name not in info_dict.keys(), em_per_decay_id_is_sig_name
        em_per_decay_id_is_sig_strict_name = "event_multiplicities_is_sig_strict_per_decay_id"
        assert em_per_decay_id_is_sig_strict_name not in info_dict.keys(), em_per_decay_id_is_sig_strict_name
        em_per_decay_id_is_b_sig_strict_name = "event_multiplicities_is_sig_strict_b_per_decay_id"
        assert em_per_decay_id_is_b_sig_strict_name not in info_dict.keys(), em_per_decay_id_is_b_sig_strict_name
        em_per_decay_id_dict = {}
        em_p_did_is_sig_dict = {}
        em_p_did_is_sig_strict_dict = {}
        em_p_did_is_b_sig_stri_dict = {}
        for decay_id in decay_mode_id_dict():
            is_decay = df[self.decay_mode_id_col] == decay_id
            d_id_em_hist = df.loc[is_decay, self.u_eid_col].value_counts().value_counts()
            em_per_decay_id_dict.update({decay_id: {k: v for k, v in zip(d_id_em_hist.index, d_id_em_hist)}})

            if self.data_type not in rdstar_samples.recorded_data_samples:
                is_sig = df["SIG_ID"].isin(decay_mode_to_sig_id[decay_id])
                is_sig_strict = df[self.is_sig_col] == 1
                is_b_sig_strict = df[self.is_b_sig_col] == 1

                sig_em_h = df.loc[is_decay & is_sig, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_sig_dict.update({decay_id: {k: v for k, v in zip(sig_em_h.index, sig_em_h)}})
                strict_em_h = df.loc[is_decay & is_sig & is_sig_strict, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_sig_strict_dict.update({decay_id: {k: v for k, v in zip(strict_em_h.index, strict_em_h)}})
                stric_b_em_h = df.loc[is_decay & is_sig & is_b_sig_strict, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_b_sig_stri_dict.update({decay_id: {k: v for k, v in zip(stric_b_em_h.index, stric_b_em_h)}})

        info_dict.update({em_per_decay_id_name: em_per_decay_id_dict})
        info_dict.update({em_per_decay_id_is_sig_name: em_p_did_is_sig_dict})
        info_dict.update({em_per_decay_id_is_sig_strict_name: em_p_did_is_sig_strict_dict})
        info_dict.update({em_per_decay_id_is_b_sig_strict_name: em_p_did_is_b_sig_stri_dict})

        if self.data_type not in rdstar_samples.recorded_data_samples:
            em_per_sig_id_name = "event_multiplicities_per_sig_id"
            assert em_per_sig_id_name not in info_dict.keys(), em_per_sig_id_name
            em_per_sig_id_dict = {}
            for sig_id_info in rdstar_sig_queries.values():
                sig_id = sig_id_info["id"]
                is_sig = df.SIG_ID == sig_id
                sig_id_em_hist = df.loc[is_sig, self.u_eid_col].value_counts().value_counts()
                em_per_sig_id_dict.update({sig_id: {k: v for k, v in zip(sig_id_em_hist.index, sig_id_em_hist)}})
            info_dict.update({em_per_sig_id_name: em_per_sig_id_dict})

    def get_evaluation_info(
            self,
            df: pd.DataFrame,
            cuts: Optional[List[str]] = None,
            bcs: Optional[Tuple[BCSVariableInfo, BCSVariableInfo]] = None
    ) -> Dict[str, Any]:
        info = {"bcs": bcs, "cuts": copy.copy(cuts)}

        tag_decay_mode_id = "daughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc"
        u4s_decay_mode_id = "extraInfo__bodecayModeID__bc"

        # Event multiplicity
        self.add_event_multiplicity_info(info_dict=info, df=df)

        info.update({"n_events": len(df[self.u_eid_col].unique())})
        info.update({"n_candidates": len(df.index)})
        # Number of candidates per data type
        info.update({"data_type_counts": {str(self.data_type): len(df.index)}})
        # Number of events per data type
        info.update({"data_type_counts_unique": {str(self.data_type): len(df[self.u_eid_col].unique())}})

        # Reconstruction modes
        decay_modes = df[self.decay_mode_id_col].value_counts()
        info.update({"events_per_decay_mode_id": {k: v for k, v in zip(decay_modes.index, decay_modes)}})
        info.update({"decay_id_vs_data_type": {str(self.data_type): info["events_per_decay_mode_id"]}})
        u_decay_modes = df.groupby(self.u_eid_col)[self.decay_mode_id_col].first().value_counts()
        info.update({"unique_events_per_decay_mode_id": {k: v for k, v in zip(u_decay_modes.index, u_decay_modes)}})
        info.update({"decay_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_decay_mode_id"]}})
        tag_modes = df[tag_decay_mode_id].value_counts()
        info.update({"events_per_tag_mode_id": {k: v for k, v in zip(tag_modes.index, tag_modes)}})
        info.update({"tag_id_vs_data_type": {str(self.data_type): info["events_per_tag_mode_id"]}})
        u_tag_modes = df.groupby(self.u_eid_col)[tag_decay_mode_id].first().value_counts()
        info.update({"unique_events_per_tag_mode_id": {k: v for k, v in zip(u_tag_modes.index, u_tag_modes)}})
        info.update({"tag_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_tag_mode_id"]}})
        u4s_modes = df[u4s_decay_mode_id].value_counts()
        info.update({"events_per_u4s_mode_id": {k: v for k, v in zip(u4s_modes.index, u4s_modes)}})
        info.update({"u4s_id_vs_data_type": {str(self.data_type): info["events_per_u4s_mode_id"]}})
        u_u4s_modes = df.groupby(self.u_eid_col)[u4s_decay_mode_id].first().value_counts()
        info.update({"unique_events_per_u4s_mode_id": {k: v for k, v in zip(u_u4s_modes.index, u_u4s_modes)}})
        info.update({"u4s_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_u4s_mode_id"]}})

        for decay_id in decay_mode_id_dict():
            self.add_vs_comparisons(info_dict=info, df=df, info_name="tag_id_vs_decay_id", col=tag_decay_mode_id,
                                    cond_col=self.decay_mode_id_col, cond_val=decay_id)

            self.add_vs_comparisons(info_dict=info, df=df, info_name="u4s_id_vs_decay_id", col=u4s_decay_mode_id,
                                    cond_col=self.decay_mode_id_col, cond_val=decay_id)

        for u4s_id in u4s_decay_mode_id_dict():
            self.add_vs_comparisons(info_dict=info, df=df, info_name="tag_id_vs_u4s_id", col=tag_decay_mode_id,
                                    cond_col=u4s_decay_mode_id, cond_val=u4s_id)

        if self.data_type in rdstar_samples.recorded_data_samples:
            return info

        # MC-Dependent information from here on!

        self.add_cut_and_bcs_validation_info(info_dict=info, df=df)

        is_signal_strict = df.isSignalAcceptMissingNeutrino == 1
        info.update({"n_signal_strict": len(df.loc[is_signal_strict].index)})
        info.update({"n_signal_unique_strict": len(df.loc[is_signal_strict, self.u_eid_col].unique())})

        decay_modes = df.loc[is_signal_strict, self.decay_mode_id_col].value_counts()
        info.update({"is_signal_events_per_decay_mode_id": {k: v for k, v in zip(decay_modes.index, decay_modes)}})
        info.update(
            {"is_signal_decay_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_decay_mode_id"]}})
        u_decay_modes = df.loc[is_signal_strict].groupby(
            self.u_eid_col
        )[self.decay_mode_id_col].first().value_counts()
        info.update(
            {"is_signal_unique_events_per_decay_mode_id": {k: v for k, v in zip(u_decay_modes.index, u_decay_modes)}})
        info.update({"is_signal_decay_id_vs_data_type_unique": {
            str(self.data_type): info["is_signal_unique_events_per_decay_mode_id"]}})
        tag_modes = df.loc[is_signal_strict, tag_decay_mode_id].value_counts()
        info.update({"is_signal_events_per_tag_mode_id": {k: v for k, v in zip(tag_modes.index, tag_modes)}})
        info.update({"is_signal_tag_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_tag_mode_id"]}})
        u_tag_modes = df.loc[is_signal_strict].groupby(self.u_eid_col)[tag_decay_mode_id].first().value_counts()
        info.update({"is_signal_unique_events_per_tag_mode_id": {k: v for k, v in zip(u_tag_modes.index, u_tag_modes)}})
        info.update({"is_signal_tag_id_vs_data_type_unique": {
            str(self.data_type): info["is_signal_unique_events_per_tag_mode_id"]}})
        u4s_modes = df.loc[is_signal_strict, u4s_decay_mode_id].value_counts()
        info.update({"is_signal_events_per_u4s_mode_id": {k: v for k, v in zip(u4s_modes.index, u4s_modes)}})
        info.update({"is_signal_u4s_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_u4s_mode_id"]}})
        u_u4s_modes = df.loc[is_signal_strict].groupby(self.u_eid_col)[u4s_decay_mode_id].first().value_counts()
        info.update({"is_signal_unique_events_per_u4s_mode_id": {k: v for k, v in zip(u_u4s_modes.index, u_u4s_modes)}})
        info.update({"is_signal_u4s_id_vs_data_type_unique": {
            str(self.data_type): info["is_signal_unique_events_per_u4s_mode_id"]}})

        # Number of candidates per sig_id
        sig_id_counts = df["SIG_ID"].value_counts()
        info.update({"sig_id_counts": {k: v for k, v in zip(sig_id_counts.index, sig_id_counts)}})
        u_sig_id_counts = df.groupby(self.u_eid_col)["SIG_ID"].first().value_counts()
        info.update({"sig_id_counts_unique": {k: v for k, v in zip(u_sig_id_counts.index, u_sig_id_counts)}})

        mdst_index = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        mdst_index_tau = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        l_mother_pdg = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc"
        is_continuum = df["isContinuumEvent"] == 1

        tmp_gen_id_col = "tmp_gen_id"
        assert tmp_gen_id_col not in df.columns

        conditions = [
            (df[mdst_index] == 1),
            (df[mdst_index] == 2),
            ((df[mdst_index_tau] == 1) & (abs(df[l_mother_pdg]) == 15)),
            ((df[mdst_index_tau] == 2) & (abs(df[l_mother_pdg]) == 15)),
            (is_continuum & (df["GEN_B1_ID"] == df["GEN_B2_ID"]))
        ]

        continuum_check_series = df.loc[is_continuum, "GEN_B1_ID"] == df.loc[is_continuum, "GEN_B2_ID"]
        assert isinstance(continuum_check_series, pd.Series)
        assert continuum_check_series.all()
        assert all(pd.DataFrame({c: d for c, d in enumerate(conditions)}).sum(axis=1) <= 1), "Conditions are not unique"

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, tmp_gen_id_col] = np.select(
                condlist=conditions,
                choicelist=[df["GEN_B1_ID"], df["GEN_B2_ID"], df["GEN_B1_ID"], df["GEN_B2_ID"], df["GEN_B1_ID"]],
                default=-999
            )
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        incon_gen_1_id = df.loc[df[tmp_gen_id_col] == -999, "GEN_B1_ID"].value_counts()
        incon_gen_2_id = df.loc[df[tmp_gen_id_col] == -999, "GEN_B2_ID"].value_counts()
        info.update({
            "gen_b1_id_for_inconclusive": {k: v for k, v in zip(incon_gen_1_id.index, incon_gen_1_id)},
            "gen_b2_id_for_inconclusive": {k: v for k, v in zip(incon_gen_2_id.index, incon_gen_2_id)}
        })

        gen_id_counts = df[tmp_gen_id_col].value_counts()
        info.update({"gen_id_counts": {k: v for k, v in zip(gen_id_counts.index, gen_id_counts)}})
        info.update({"gen_id_per_data_type": {str(self.data_type): info["gen_id_counts"]}})
        u_gen_id_counts = df.groupby(self.u_eid_col)[tmp_gen_id_col].first().value_counts()
        info.update({"gen_id_counts_unique": {k: v for k, v in zip(u_gen_id_counts.index, u_gen_id_counts)}})
        info.update({"gen_id_per_data_type_unique": {str(self.data_type): info["gen_id_counts_unique"]}})
        s_gen_id_counts = df.loc[is_signal_strict, tmp_gen_id_col].value_counts()
        info.update({"gen_id_counts_strict": {k: v for k, v in zip(s_gen_id_counts.index, s_gen_id_counts)}})
        us_gen_id_counts = df.loc[is_signal_strict].groupby(self.u_eid_col)[tmp_gen_id_col].first().value_counts()
        info.update({"gen_id_counts_unique_strict": {k: v for k, v in zip(us_gen_id_counts.index, us_gen_id_counts)}})

        for sig_id, sig_id_info in rdstar_sig_queries.items():
            is_sig = df.SIG_ID == sig_id_info["id"]

            info.update({f"n_signal_{sig_id}": len(df.loc[is_sig].index)})
            info.update({f"n_signal_unique_{sig_id}": len(df.loc[is_sig, self.u_eid_col].unique())})

            info.update({
                f"data_type_per_sig_id_{sig_id_info['id']}": {str(self.data_type): len(df.loc[is_sig].index)}
            })
            num_unique_per_sig_id = len(df.loc[is_sig, self.u_eid_col].unique())
            info.update({
                f"u_data_type_per_sig_id_{sig_id_info['id']}": {str(self.data_type): num_unique_per_sig_id}
            })

            self.add_vs_comparisons(info_dict=info, df=df, info_name="gen_id_per_sig_id", col=tmp_gen_id_col,
                                    cond_col="SIG_ID", cond_val=sig_id_info["id"])

            self.add_vs_comparisons(info_dict=info, df=df, info_name="decay_id_vs_sig_id", col=self.decay_mode_id_col,
                                    cond_col="SIG_ID", cond_val=sig_id_info["id"])
            self.add_vs_comparisons(info_dict=info, df=df, info_name="tag_id_vs_sig_id", col=tag_decay_mode_id,
                                    cond_col="SIG_ID", cond_val=sig_id_info["id"])
            self.add_vs_comparisons(info_dict=info, df=df, info_name="u4s_id_vs_sig_id", col=u4s_decay_mode_id,
                                    cond_col="SIG_ID", cond_val=sig_id_info["id"])

        for gen_id, gen_id_info in gen_b_ids.items():
            self.add_vs_comparisons(info_dict=info, df=df, info_name="decay_id_vs_gen_id", col=self.decay_mode_id_col,
                                    cond_col=tmp_gen_id_col, cond_val=gen_id_info["id"])
            self.add_vs_comparisons(info_dict=info, df=df, info_name="tag_id_vs_gen_id", col=tag_decay_mode_id,
                                    cond_col=tmp_gen_id_col, cond_val=gen_id_info["id"])
            self.add_vs_comparisons(info_dict=info, df=df, info_name="u4s_id_vs_gen_id", col=u4s_decay_mode_id,
                                    cond_col=tmp_gen_id_col, cond_val=gen_id_info["id"])

        del df[tmp_gen_id_col]

        return info

    def get_target_paths(self, selection_tag: str) -> Tuple[os.fspath, os.fspath]:
        target_path = os.path.join(self.base_dir, f"{self.file_name}_{selection_tag}.h5")
        tmp_target_path = os.path.join(self.base_dir, f"tmp_{self.file_name}_{selection_tag}.h5")
        return target_path, tmp_target_path

    def get_evaluation_paths(self, target_path: os.fspath, selection_tag: str) -> Tuple[str, os.fspath, os.fspath]:
        dir_name, base_name = os.path.split(os.path.splitext(target_path)[0])
        tmp_path = os.path.join(dir_name, f"tmp_{self.evaluation_info_prefix}_{base_name}.json")
        path = os.path.join(dir_name, f"{self.evaluation_info_prefix}_{base_name}.json")
        key = f"{self.file_name}_{self.evaluation_info_prefix}_{selection_tag}"
        return key, path, tmp_path

    def get_selection_infos(self) -> Generator[SelectionInfo, None, None]:
        for selection_tag, selection in Selection.get_selection_infos(self.selection_name, self.cs_tag).items():
            target_path, tmp_target_path = self.get_target_paths(selection_tag=selection_tag)
            output_key = f"{self.file_name}_{selection_tag}"

            yield SelectionInfo(
                selection_tag=selection_tag,
                selection=selection,
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=output_key
            )

    @staticmethod
    def reset_selection(
            selection_dict: Dict[str, Any],
            new_cuts: Optional[List[str]] = None,
            reset_bcs: Union[Tuple, bool] = False
    ) -> Dict[str, Any]:
        selection = copy.copy(selection_dict)
        assert isinstance(reset_bcs, bool) or isinstance(reset_bcs, tuple), (reset_bcs, type(reset_bcs))
        if new_cuts is not None:
            selection["check_cuts"] = new_cuts
        if reset_bcs is True:
            selection["primary_bcs_var"] = BCSVariableInfo(None)
            selection["secondary_bcs_var"] = BCSVariableInfo(None)
        if isinstance(reset_bcs, tuple):
            assert len(reset_bcs) == 2, len(reset_bcs)
            assert all(isinstance(r_bcs, BCSVariableInfo) for r_bcs in reset_bcs), reset_bcs
            selection["primary_bcs_var"] = reset_bcs[0]
            selection["secondary_bcs_var"] = reset_bcs[1]
        selection["check_cuts"] = []
        selection["check_bcs"] = False
        return selection

    @staticmethod
    def get_bcs_check_extension(random: bool = False) -> str:
        if random:
            return "_random_bcs"
        return "_no_bcs"

    @staticmethod
    def get_cut_check_extension(
            cut_number: int,
            no_bcs: bool = False,
            random_bcs: bool = False
    ) -> str:
        # Check that not both no_bcs and random_bcs are True at the same time:
        assert isinstance(cut_number, int) and cut_number >= 0, cut_number
        if (no_bcs is random_bcs) and no_bcs:
            raise ValueError("The arguments no_bcs and random_bcs must not be both True at the same time!")
        if no_bcs:
            return f"_without_cut_{cut_number}{DataSelectionTask.get_bcs_check_extension(False)}"
        elif random_bcs:
            return f"_without_cut_{cut_number}{DataSelectionTask.get_bcs_check_extension(True)}"
        else:
            return f"_without_cut_{cut_number}"

    @staticmethod
    def get_possible_subsample_extensions(input_dict: Dict[str, Any]) -> List[str]:
        extensions = []
        checking = True
        index = 0
        while checking:
            checking = False
            if index == 0:
                ext = DataSelectionTask.get_bcs_check_extension(random=False)
                if any(ext in key for key in input_dict):
                    extensions.append(ext)
                ext = DataSelectionTask.get_bcs_check_extension(random=True)
                if any(ext in key for key in input_dict):
                    extensions.append(ext)
            ext = DataSelectionTask.get_cut_check_extension(cut_number=index)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            ext = DataSelectionTask.get_cut_check_extension(cut_number=index, no_bcs=True)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            ext = DataSelectionTask.get_cut_check_extension(cut_number=index, random_bcs=True)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            index += 1
        return extensions

    def get_subselection_info(self, selection_info: SelectionInfo) -> Generator[SelectionInfo, None, None]:
        if not selection_info.selection["check_cuts"] and not selection_info.selection["check_bcs"]:
            return

        if selection_info.selection["check_bcs"]:
            subselection_tag = f"{selection_info.selection_tag}{self.get_bcs_check_extension(random=False)}"
            target_path, tmp_target_path = self.get_target_paths(selection_tag=subselection_tag)
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, reset_bcs=True),
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=f"{self.file_name}_{subselection_tag}"
            )
            subselection_tag = f"{selection_info.selection_tag}{self.get_bcs_check_extension(random=True)}"
            target_path, tmp_target_path = self.get_target_paths(selection_tag=subselection_tag)
            bcs_tuple_rand = (BCSVariableInfo("random"), BCSVariableInfo(None))
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, reset_bcs=bcs_tuple_rand),
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=f"{self.file_name}_{subselection_tag}"
            )

        if not selection_info.selection["check_cuts"]:
            return

        for cut_num, cut in enumerate(selection_info.selection["check_cuts"]):
            other_cuts = [c for c in selection_info.selection["cuts"] if c is not cut]
            subselection_tag = f"{selection_info.selection_tag}{self.get_cut_check_extension(cut_num)}"
            target_path, tmp_target_path = self.get_target_paths(selection_tag=subselection_tag)
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, new_cuts=other_cuts, reset_bcs=False),
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=f"{self.file_name}_{subselection_tag}"
            )
            subselection_tag = f"{selection_info.selection_tag}{self.get_cut_check_extension(cut_num, no_bcs=True)}"
            target_path, tmp_target_path = self.get_target_paths(selection_tag=subselection_tag)
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, new_cuts=other_cuts, reset_bcs=True),
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=f"{self.file_name}_{subselection_tag}"
            )
            subselection_tag = f"{selection_info.selection_tag}{self.get_cut_check_extension(cut_num, random_bcs=True)}"
            target_path, tmp_target_path = self.get_target_paths(selection_tag=subselection_tag)
            bcs_tuple_rand = (BCSVariableInfo("random"), BCSVariableInfo(None))
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, new_cuts=other_cuts, reset_bcs=bcs_tuple_rand),
                target_path=target_path,
                tmp_target_path=tmp_target_path,
                output_key=f"{self.file_name}_{subselection_tag}"
            )

    def data_selection(self) -> None:
        df = hdf5_utils.load_hdf5(hdf5_file_path=self.input_dict[self.input_key], hdf5_key=self.file_name)

        for selection_info in self.get_selection_infos():
            selected_df, eval_info = self.apply_selection(df=df, selection_info=selection_info, evaluate_selection=True)

            selected_df.to_hdf(path_or_buf=selection_info.tmp_target_path, key=self.file_name)
            os.rename(src=selection_info.tmp_target_path, dst=selection_info.target_path)

            _, info_path, tmp_info_path = self.get_evaluation_paths(
                target_path=selection_info.target_path,
                selection_tag=selection_info.selection_tag
            )
            with open(tmp_info_path, 'w') as out_file:
                json.dump(eval_info, out_file, indent=2)
            os.rename(src=tmp_info_path, dst=info_path)

            for subselection_info in self.get_subselection_info(selection_info=selection_info):
                selected_df = self.apply_selection(df=df, selection_info=subselection_info)
                selected_df.to_hdf(path_or_buf=subselection_info.tmp_target_path, key=self.file_name)
                os.rename(src=subselection_info.tmp_target_path, dst=subselection_info.target_path)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.data_selection()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {info.output_key: luigi.LocalTarget(info.target_path) for info in self.get_selection_infos()}
        for selection_info in self.get_selection_infos():
            for sub_selection in self.get_subselection_info(selection_info=selection_info):
                assert sub_selection.output_key not in output_dict, sub_selection.output_key
                assert sub_selection.target_path not in output_dict.values(), sub_selection.target_path
                output_dict.update({sub_selection.output_key: luigi.LocalTarget(sub_selection.target_path)})
            info_key, info_path, _ = self.get_evaluation_paths(
                target_path=selection_info.target_path,
                selection_tag=selection_info.selection_tag
            )
            assert info_key not in output_dict, info_key
            assert info_path not in output_dict.values(), info_path
            output_dict.update({info_key: luigi.LocalTarget(info_path)})

        return output_dict
