"""
Tasks performing the Sideband Validation

Felix Metzner 2019
"""

import os
import copy
import pandas as pd
import multiprocessing
import b2luigi as luigi

from functools import partial
from luigi.util import inherits
from collections import abc, defaultdict

from typing import List, Tuple, Dict, Generator

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.validation.sideband_validation import plot_data_mc_comparison, sideband_info_collection, \
    SidebandInfo

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightAnalysisTask


__all__ = [
    "SidebandEvaluationTask"
]


@inherits(WeightAnalysisTask)
class SidebandEvaluationTask(utilities.BaseOfflineTask):
    base_dir_name = "SidebandValidation"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(os.fspath(self.working_dir), self.base_dir_name)

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.mc_sample_keys = rdstar_samples.main_sample_keys
        self.data_sample_key = "DataY4S"

        self.full_weight_column_name = "__weight_final_wFEI_corr__"  # TODO: Remove once provided by weight finalization

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"

        self.sideband_infos = copy.copy(sideband_info_collection)

        self.input_appendix = f"{WeightFinalizationTask.appendix}_{ResolutionCorrectionTask.appendix}"
        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_mc_and_data_sample(self, selection_key: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{selection_key}"
        mc_paths = [  # type: List[os.fspath]
            p for mc_key in self.mc_sample_keys for k, p in self.input_dict.items()
            if (k.startswith(f"{mc_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p for k, p in self.input_dict.items()
            if (k.startswith(f"{self.data_sample_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame)

        return mc_sample, data_sample

    def get_sorted_sideband_infos(self) -> Dict[str, List[SidebandInfo]]:
        sorted_sideband_infos = defaultdict(list)
        for sideband_info in self.sideband_infos:
            sorted_sideband_infos[sideband_info.key].append(sideband_info)

        return sorted_sideband_infos

    def produce_sideband_plot(self, input_tuple: Tuple[str, SidebandInfo, str], dry_run: bool) -> Dict[str, os.fspath]:
        sample_key, sideband_info, style = input_tuple
        assert sideband_info.key == sample_key, (sideband_info.key, sample_key)

        mc_df, data_df = (None, None)
        if not dry_run:
            mc_df, data_df = self.get_mc_and_data_sample(selection_key=sample_key)

            assert self.full_weight_column_name not in mc_df.columns
            assert WeightFinalizationTask.final_weight_col in mc_df.columns
            assert TagCorrection.tagging_correction_name in mc_df.columns

            product_cols = [
                WeightFinalizationTask.final_weight_col,
                TagCorrection.tagging_correction_name
            ]
            mc_df.loc[:, self.full_weight_column_name] = mc_df[product_cols].product(axis=1)

        plot_output_dict = plot_data_mc_comparison(
            mc_df=mc_df,
            data_df=data_df,
            dry_run=dry_run,
            weight_col=self.full_weight_column_name,
            sideband_info=sideband_info,
            style=style,
            normalize_mc_to_data=True,
            ratio_type="vs_uncert",
            gof_check_method="toys",
            adaptive_binning=True,
            export_path=self.base_path
        )

        if not dry_run:
            del mc_df
            del data_df

        return plot_output_dict

    def do_sideband_validation(self, dry_run: bool = False) -> Dict[str, os.fspath]:
        output_dict = {}

        sb_plot_info_tuples = [
            (sample_key, info, style)
            for sample_key, sb_info_list in self.get_sorted_sideband_infos().items()
            for info in sb_info_list
            for style in info.styles
        ]

        with multiprocessing.Pool(processes=1 if dry_run else 10) as pool:
            eval_outputs = pool.map(
                partial(SidebandEvaluationTask.produce_sideband_plot, self, dry_run=dry_run),
                sb_plot_info_tuples
            )

        for eval_output in eval_outputs:
            assert all(k not in output_dict.keys() for k in eval_output.keys()),\
                [k for k in eval_output.keys() if k in output_dict.keys()]
            assert all(v not in output_dict.values() for v in eval_output.values())
            output_dict.update(eval_output)

        assert len(output_dict) % len(self.sideband_infos) == 0, (len(output_dict), len(self.sideband_infos))
        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_sideband_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_sideband_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations
            )
