"""
Utility Functions for Analysis Tasks

Felix Metzner 2019
"""

import os
import json
import rdstar
import socket
import inspect

import numpy as np
import pandas as pd
import b2luigi as luigi

from typing import Any, Union, Optional, List, Dict, NamedTuple


__all__ = [
    "TaskType",
    "InputDictType",
    "GlobalParameters",
    "BaseOfflineTask",
    "get_config_directory",
    "host_check",
    "load_dict_from_json",
    "to_float_if_float",
    "fix_json_dict",
    "decompose_event_id",
    "recreate_event_id",
    "UniqueEventIDColumnInfos",
    "add_unique_candidate_id"
]

TaskType = luigi.Task
InputDictType = Dict[str, os.fspath]


class GlobalParameters(luigi.Config):
    work_host_list = luigi.ListParameter(default=[])
    copy_host_list = luigi.ListParameter(default=[])

    n_subsets_mixed = luigi.IntParameter(default=1)
    n_subsets_generic = luigi.IntParameter(default=1)


class BaseOfflineTask(luigi.Task):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_task_file_dir(self) -> os.fspath:
        assert self.working_dir is not None
        return os.fspath(self.working_dir)

    def get_log_file_dir(self) -> os.fspath:
        assert self.get_task_file_dir() is not None
        return self.get_task_file_dir()


def get_config_directory() -> os.fspath:
    return os.path.join(os.path.dirname(inspect.getfile(rdstar)), "configs")


def host_check(hosts: List[str], message: Optional[str] = None) -> bool:
    if not hosts:
        return True

    if message is None:
        message = "The task must be run on a suitable machine...\n"
        message += f"Switch to any of the following hosts:\n {hosts}"

    if not any(host in socket.gethostname() for host in hosts):
        raise RuntimeError(message)

    return True


def load_dict_from_json(path: os.fspath) -> Dict[Any, Any]:
    with open(path, 'r') as in_file:
        eval_dict = json.load(in_file)
    return fix_json_dict(eval_dict)


def to_float_if_float(input_string: str) -> Union[float, str]:
    try:
        num = float(input_string)
    except ValueError:
        return input_string
    return num


def fix_json_dict(json_dict: Dict[str, Any]) -> Dict[Union[float, str], Any]:
    if not isinstance(json_dict, dict):
        return json_dict
    return {to_float_if_float(k): fix_json_dict(v) for k, v in json_dict.items()}


class UniqueEventIDColumnInfos(NamedTuple):
    event_id_col: str = "unique_event_id"
    data_type_col: str = "temp_data_type_info"
    exp_col: str = "temp_exp_info"
    run_col: str = "temp_run_info"
    evt_col: str = "temp_evt_info"
    split_str: str = "_"
    split_n: int = 3


def decompose_event_id(df: pd.DataFrame, remove_original_column: bool = False) -> None:
    col_info = UniqueEventIDColumnInfos()
    num_cols = len(df.columns)

    assert col_info.event_id_col in df.columns, col_info.data_type_col

    assert col_info.data_type_col not in df.columns, col_info.data_type_col
    assert col_info.exp_col not in df.columns, col_info.exp_col
    assert col_info.run_col not in df.columns, col_info.run_col
    assert col_info.evt_col not in df.columns, col_info.evt_col

    split_id = df.loc[:, col_info.event_id_col].str.rsplit(pat=col_info.split_str, n=col_info.split_n, expand=True)
    assert not pd.isnull(split_id).any().any()

    df.loc[:, col_info.data_type_col] = split_id[0].astype('category')
    df.loc[:, col_info.exp_col] = split_id[1].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.run_col] = split_id[2].str.split(pat=".", expand=True)[0].astype("int")
    df.loc[:, col_info.evt_col] = split_id[3].str.split(pat=".", expand=True)[0].astype("int")

    if remove_original_column:
        del df[col_info.event_id_col]
        assert len(df.columns) == num_cols + 3, (len(df.columns), num_cols)
    else:
        assert len(df.columns) == num_cols + 4, (len(df.columns), num_cols)


def recreate_event_id(df: pd.DataFrame, remove_original_columns: bool = False) -> None:
    col_info = UniqueEventIDColumnInfos()

    assert col_info.data_type_col in df.columns, col_info.data_type_col
    assert col_info.exp_col in df.columns, col_info.exp_col
    assert col_info.run_col in df.columns, col_info.run_col
    assert col_info.evt_col in df.columns, col_info.evt_col

    assert col_info.event_id_col not in df.columns, col_info.data_type_col

    info_cols = [col_info.data_type_col]
    int_info_cols = [col_info.exp_col, col_info.run_col, col_info.evt_col]
    original_info_cols = info_cols + int_info_cols
    converted_cols = []
    for col in int_info_cols:
        converted_col_name = f"temp_{col}"
        assert converted_col_name not in df.columns
        df.loc[:, converted_col_name] = df[col].astype(float).astype(str)
        converted_cols.append(converted_col_name)
        info_cols.append(converted_col_name)

    df.loc[:, col_info.event_id_col] = pd.Series(df.loc[:, info_cols].values.tolist()).str.join(col_info.split_str)
    assert not pd.isnull(df[col_info.event_id_col]).any()

    for column in converted_cols:
        del df[column]
    if remove_original_columns:
        for column in original_info_cols:
            del df[column]


def add_unique_candidate_id(
        df: pd.DataFrame,
        id_column_name: str,
        event_id_name: str,
        candidate_id_name: str,
        check_if_unique: bool = True
) -> np.ndarray:
    assert event_id_name in df.columns, event_id_name
    assert candidate_id_name in df.columns, candidate_id_name
    assert id_column_name not in df.columns, id_column_name

    cand_id_col_name = candidate_id_name
    if df[candidate_id_name].dtype == np.float or df[candidate_id_name].dtype == np.int:
        temp_cand_id_as_str_name = f"temp_{candidate_id_name}"
        assert temp_cand_id_as_str_name not in df.columns, temp_cand_id_as_str_name
        df.loc[:, temp_cand_id_as_str_name] = df[candidate_id_name].astype(str)
        cand_id_col_name = temp_cand_id_as_str_name

    df.loc[:, id_column_name] = pd.Series(df.loc[:, [event_id_name, cand_id_col_name]].values.tolist()).str.join("_")
    assert not pd.isnull(df[id_column_name]).any()

    assert id_column_name in df.columns, id_column_name
    if check_if_unique:
        assert len(df.index) == len(df[id_column_name].unique()), (len(df.index), len(df[id_column_name].unique()))

    if cand_id_col_name != candidate_id_name:
        del df[cand_id_col_name]

    return df[id_column_name].values
