"""
Analysis Flow Validation
    - Cut Flow Validation
    - BCS Validation

Felix Metzner 2019
"""

import numpy as np
import pandas as pd

from rdstar.offline_analysis.selection import Selection
from rdstar.offline_analysis.rdstar_plots import decay_mode_id_dict

# BCS Validation
bcs_info_key_parts = [
    "bcs_info_",
    "bcs_info_strict_",
    "bcs_info_v_strict_",
    "bcs_info_strict_b_",
    "bcs_info_v_strict_b_",
    # With Normalisation modes
    "bcs_info_wn_",
    "bcs_info_wn_v_strict_",
    "bcs_info_wn_v_strict_b_"
]


def create_bcs_table(info_dict, info_key_parts):
    # counts after bcs
    nom_info_dict = get_sub_dict_by_index(info_dict=info_dict, index=len(info_dict) - 1)
    # counts before bcs
    denom_info_dict = get_sub_dict_by_index(info_dict=info_dict, index=len(info_dict) - 2)

    output = {}
    for truth in info_key_parts:
        norm_nom, norm_denom = (None, None)
        has_norm = False
        if "_wn_" in truth:
            has_norm = True

        sig_nom = nom_info_dict[f"{truth}sig"]
        sig_denom = denom_info_dict[f"{truth}sig"]
        bkg_nom = nom_info_dict[f"{truth}bkg"]
        bkg_denom = denom_info_dict[f"{truth}bkg"]
        if has_norm:
            norm_nom = nom_info_dict[f"{truth}norm"]
            norm_denom = denom_info_dict[f"{truth}norm"]

        base_dict = {
            "Sig before": sig_denom,
            "Sig after": sig_nom,
        }
        if has_norm:
            base_dict.update({
                "Norm before": norm_denom,
                "Norm after": norm_nom,
            })
        base_dict.update({
            "Bkg before": bkg_denom,
            "Bkg after": bkg_nom,
        })

        df = pd.DataFrame(base_dict)

        df.loc[:, "Sig Eff"] = df["Sig after"] / df["Sig before"]
        if has_norm:
            df.loc[:, "Norm Eff"] = df["Norm after"] / df["Norm before"]
            df.loc[~np.isfinite(df["Norm Eff"]), "Norm Eff"] = np.nan
        df.loc[:, "Bkg Eff"] = df["Bkg after"] / df["Bkg before"]

        # For norm: set all zeroes to NaNs to be able to remove irrelevant row:
        if has_norm:
            for row in [c for c in df.columns if c.startswith("Norm ")]:
                df.loc[df[row] == 0., row] = np.nan

        df = df.T
        df = df.rename(columns={k: rf"{k}_{v}" for k, v in decay_mode_id_dict().items()})

        df.fillna("---", inplace=True)

        output.update({truth: df})
    return output


# Cut Flow Validation
cut_flow_info_keys = [
    "cut_info_sig",
    "cut_info_strict_sig",
    "cut_info_strict_b_sig",
    "cut_info_v_strict_sig",
    "cut_info_v_strict_b_sig",
    # With Normalisation modes
    "cut_info_wn_sig",
    "cut_info_wn_v_strict_sig",
    "cut_info_wn_v_strict_b_sig"
]


def get_cut_flow_table_col_name(index, cut_list, selection_name):
    if index == 0:
        return "No Cuts"
    if index == len(cut_list) + 1:
        return "After BCS"
    assert index - 1 < len(cut_list)
    return Selection.get_selection_cut_label_dict(selection_name=selection_name)[cut_list[index - 1]]


def create_cut_flow_table(info_dict, info_key, selection_name):
    assert all(k.split("_", 1)[0].isdigit() for k in info_dict.keys())
    assert all(isinstance(v, dict) for v in info_dict.values())
    assert all("cuts" in v.keys() for v in info_dict.values())

    bkg_info_key = info_key.replace("cut", "bcs").replace("sig", "bkg")
    assert all(bkg_info_key in info_dict[k].keys() for k in info_dict.keys())

    norm_info_key = None
    has_norm = False
    if "_wn_" in info_key:
        norm_info_key = info_key.replace("sig", "norm")
        assert all(norm_info_key in info_dict[k].keys() for k in info_dict.keys())
        has_norm = True

    cut_list = get_sub_dict_by_index(info_dict=info_dict, index=len(info_dict) - 1)["cuts"]
    assert len(cut_list) == len(info_dict.keys()) - 2

    table_dict = {
        (i, "1_sig"): info_dict[k][info_key]
        for i, k in enumerate(info_dict.keys())
    }
    sub_index = 2
    if has_norm:
        sub_index = 3
        table_dict.update({
            (i, "2_norm"): info_dict[k][norm_info_key]
            for i, k in enumerate(info_dict.keys())
        })
    table_dict.update({
        (i, f"{sub_index}_bkg"): info_dict[k][bkg_info_key]
        for i, k in enumerate(info_dict.keys())
    })

    df = pd.DataFrame.from_dict(table_dict, orient="index")

    # Calculate Efficiencies for each Cut and create Multicolumn Data Frame
    for c in df.columns:
        df[c, "Counts"] = df[c]
        df[c, "Eff"] = df[c, "Counts"] * 100. / df[c, "Counts"].shift(3 if has_norm else 2)
        del df[c]

    # Calculate Total Efficiencies
    total_row_name = "Total Eff"
    fc = list(df.columns)[1]
    df = df.T
    df.loc[fc::2, pd.IndexSlice[total_row_name, "1_sig"]] = (
            df.loc[::2, pd.IndexSlice[len(info_dict.keys()) - 1, "1_sig"]]
            * 100. / df.loc[::2, pd.IndexSlice[0, "1_sig"]]).values
    if has_norm:
        df.loc[fc::2, pd.IndexSlice[total_row_name, "2_norm"]] = (
                df.loc[::2, pd.IndexSlice[len(info_dict.keys()) - 1, "2_norm"]]
                * 100. / df.loc[::2, pd.IndexSlice[0, "2_norm"]]).values
        df.loc[~np.isfinite(df[(total_row_name, "2_norm")]), (total_row_name, "2_norm")] = np.nan
    df.loc[fc::2, pd.IndexSlice[total_row_name, f"{sub_index}_bkg"]] = (
            df.loc[::2, pd.IndexSlice[len(info_dict.keys()) - 1, f"{sub_index}_bkg"]]
            * 100. / df.loc[::2, pd.IndexSlice[0, f"{sub_index}_bkg"]]).values

    # For norm: set all zeroes to NaNs to be able to remove irrelevant row:
    if has_norm:
        for row in df.columns.levels[0]:
            df.loc[df[(row, "2_norm")] == 0.0, (row, "2_norm")] = np.nan
    df = df.T

    df.fillna("---", inplace=True)
    df.columns = pd.MultiIndex.from_tuples(df.columns)

    # Renaming:
    df = df.rename(columns={k: rf"{k}_{v}" for k, v in decay_mode_id_dict().items()})
    df.columns.set_levels(
        [r"\multicolumn{1}{c}{Counts}", r"\multicolumn{1}{c}{Eff in \si{\percent}}"],
        level=1,
        inplace=True
    )

    df = df.T
    df.columns.set_levels(
        [get_cut_flow_table_col_name(index=i, cut_list=cut_list, selection_name=selection_name)
         if i != total_row_name else i
         for i in df.columns.levels[0]],
        level=0,
        inplace=True
    )
    index_l2_names = ["Signal", "Bkg"]
    if has_norm:
        index_l2_names.insert(1, "Norm")
    df.columns.set_levels(index_l2_names, level=1, inplace=True)
    df = df.T

    df.index.names = [r"\multicolumn{1}{c}{Cuts}", None]

    return df


# Utility
def get_sub_dict_by_index(info_dict, index):
    keys = [k for k in info_dict.keys() if k.startswith(str(index))]
    assert len(keys) == 1, (len(keys), keys)
    key = keys[0]
    return info_dict[key]
