import os
import pickle

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from typing import Union
from sklearn.metrics import roc_curve, roc_auc_score

import rdstar.continuum_suppression as cs_tools

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos

set_matplotlibrc_params()


def load_dfs(train_df, test_df):
    return pd.read_hdf(train_df), pd.read_hdf(test_df)


def load_cs_vars(path):
    with open(path, "rb") as f:
        cs_vars = pickle.load(f)
    return cs_vars


def get_sig_bkg(df):
    return df.query("isNotContinuumEvent==1"), df.query("isNotContinuumEvent==0")


def plot_cs_vars(df, cs_vars, target_dir, title=None, base_name=None):
    sig, bkg = get_sig_bkg(df)

    new_file_paths = []

    for var in cs_vars:
        hist_var = cs_tools.cs_variable_dict[var]

        name = f"{base_name}_" if base_name else ""
        file_name = f"{name}cs_var_distribution_{hist_var.df_label}.pdf"
        file_path = os.path.join(target_dir, file_name)

        shp = SimpleHistogramPlot(hist_var)
        shp.add_component(label="Background", data=bkg, hist_type="stepfilled", color=KITColors.kit_red)
        shp.add_component(label="Signal", data=sig, hist_type="stepfilled", color=KITColors.kit_blue, alpha=0.8)

        fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
        shp.plot_on(ax, normed=True, y_axis_scale=1.1)

        if title is not None:
            ax.set_title(title)

        plt.savefig(file_path, bbox_inches="tight")
        plt.close(fig)
        new_file_paths.append(file_path)

    return new_file_paths


def plot_cs_output(train_df, test_df, cs_output_label, target_dir, appendix=None, title=None, base_name=None):
    name = f"{base_name}_" if base_name else ""
    file_name = f"{name}cs_output{'_' + appendix if appendix is not None else ''}.pdf"
    file_path = os.path.join(target_dir, file_name)

    train_sig, train_bkg = get_sig_bkg(train_df)
    test_sig, test_bkg = get_sig_bkg(test_df)

    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)

    bins = np.linspace(0, 1.0, 26)
    mids = (bins[:-1] + bins[1:]) / 2

    ax.hist(train_sig[cs_output_label], bins=bins, density=True,
            color=KITColors.kit_blue, alpha=0.8, histtype="stepfilled", lw=0.5, edgecolor=KITColors.kit_black,
            label="Signal (Train)")
    ax.hist(train_bkg[cs_output_label], bins=bins, density=True,
            color=KITColors.kit_red, alpha=0.8, histtype="stepfilled", lw=0.5, edgecolor=KITColors.kit_black,
            label="Background (Train)")

    test_sig_bc, _ = np.histogram(test_sig[cs_output_label], bins=bins, density=True)
    test_bkg_bc, _ = np.histogram(test_bkg[cs_output_label], bins=bins, density=True)

    ax.plot(mids, test_sig_bc, ls="", marker=".", color=KITColors.kit_blue, markeredgewidth=0.5,
            markeredgecolor=KITColors.kit_black, label="Signal (Test)")
    ax.plot(mids, test_bkg_bc, ls="", marker=".", color=KITColors.kit_red, markeredgewidth=0.5,
            markeredgecolor=KITColors.kit_black, label="Background (Test)")

    if title is not None:
        ax.set_title(title)

    ax.set_xlabel("Continuum Suppression Output", **xlabel_pos)
    ax.set_ylabel("Normalized in arb. units", **ylabel_pos)
    ax.legend(loc="best")

    plt.savefig(file_path, bbox_inches="tight")
    plt.close(fig)

    return file_path


def plot_roc_curve(
        df,
        cs_output_label,
        target_dir,
        cs_target_var="isNotContinuumEvent",
        appendix=None,
        title=None,
        base_name=None,
        working_point=0.5
):
    name = f"{base_name}_" if base_name else ""
    file_name = f"{name}roc_curve{'_' + appendix if appendix is not None else ''}"
    file_path = os.path.join(target_dir, f"{file_name}.pdf")

    y_true = df[cs_target_var]
    y_predict = df[cs_output_label]

    selected = y_predict > working_point

    sig_eff = np.sum(selected & (y_true == 1)) / np.sum(y_true == 1)
    bkg_rej = np.sum(selected & (y_true == 0)) / np.sum(y_true == 0)

    fpr, tpr, thr = roc_curve(y_true, y_predict)
    roc_auc = roc_auc_score(y_true, y_predict)

    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
    ax.set_xlim([0.0, 1.05])
    ax.set_ylim([0.0, 1.05])

    ax.plot(1 - fpr, tpr, lw=1.5, color=KITColors().kit_blue, label="ROC Curve")
    ax.plot(1 - bkg_rej, sig_eff, ls="", marker="o", color=KITColors().kit_red,
            label=r"Working Point, $\mathcal{P}_\mathrm{CS}>0.5$")

    ax.axhline(sig_eff, 0, 1 - bkg_rej, ls="--", color=KITColors().kit_red)
    ax.axvline(1 - bkg_rej, 0, sig_eff, ls="--", color=KITColors().kit_red)

    info_text = f"ROC Integral = {roc_auc:.3f}\n" \
        f"Signal Efficiency = {100 * sig_eff:.1f}%\n" \
        f"Background Rejection = {100 * (1 - bkg_rej):.1f}%"

    ax.text(x=0.05, y=0.5, s=info_text)

    if title is not None:
        ax.set_title(title)

    ax.set_xlabel("Background Suppression", xlabel_pos)
    ax.set_ylabel("Signal Efficiency", ylabel_pos)
    plt.legend()

    export(fig, file_name, target_dir)
    plt.close(fig)

    return file_path, roc_auc


def run_cs_validation(
        path_train_df: Union[str, os.fspath],
        path_test_df: Union[str, os.fspath],
        path_cs_var_list: Union[str, os.fspath],
        path_cs_bdt_weights: Union[str, os.fspath],
        target_file_path: Union[str, os.fspath],
        cs_target_var: str = "isNotContinuumEvent",
):
    target_path, target_base_name = os.path.split(target_file_path)
    base_name = target_base_name.replace(".pdf", "")
    os.makedirs(target_path, exist_ok=True)

    train_df, test_df = load_dfs(path_train_df, path_test_df)  # type: pd.DataFrame
    cs_vars = load_cs_vars(path_cs_var_list)

    cs_classifier = cs_tools.ContinuumSuppression.load_classifier(
        weight_file_path=path_cs_bdt_weights,
        var_file_path=path_cs_var_list
    )

    cs_output_label = "cs_output"
    for df in [train_df, test_df]:
        cs_classifier.apply(data=df, output_column=cs_output_label)

    # Input variables
    plot_cs_vars(train_df, cs_vars, target_dir=target_path, base_name=base_name)

    # ROC curves
    _, test_roc_auc = plot_roc_curve(
        df=test_df,
        cs_output_label=cs_output_label,
        target_dir=target_path,
        cs_target_var=cs_target_var,
        base_name=base_name
    )
    plot_roc_curve(
        df=train_df,
        cs_output_label=cs_output_label,
        target_dir=target_path,
        cs_target_var=cs_target_var,
        appendix="on_train_data",
        base_name=base_name
    )

    # CS output distribution
    cs_output_path = plot_cs_output(train_df, test_df, cs_output_label, target_dir=target_path, base_name=base_name)

    if target_base_name:
        os.symlink(cs_output_path, target_file_path)

    return test_roc_auc


if __name__ == "__main__":
    src_dir = "/storage/9/welsch/CS_Trainings"
    cwd = os.getcwd()
    my_plot_dir = os.path.join(cwd, "plots")

    run_cs_validation(
        path_train_df=os.path.join(src_dir, "cs_train_data_CS_dataset1.h5"),
        path_test_df=os.path.join(src_dir, "cs_test_data_CS_dataset1.h5"),
        path_cs_var_list=os.path.join(src_dir, "cs_mva_vars_CS1.pkl"),
        path_cs_bdt_weights=os.path.join(src_dir, "cs_classifier_weights_CS1.xml"),
        target_file_path=my_plot_dir
    )
