"""
This script creates validation plots of the D and D* meson reconstruction.
"""

import os
import glob
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors, xlabel_pos
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import *

logging.getLogger(__name__).addHandler(logging.NullHandler())

set_matplotlibrc_params()

D_id_to_label = {
    31: r'$D^+ \rightarrow K^- \pi^+ \pi^+$',
    32: r'$D^+ \rightarrow K^- \pi^+ \pi^+ \pi^0$',
    33: r'$D^+ \rightarrow K^- \pi^+ \pi^+ \pi^+ \pi^-$',
    34: r'$D^+ \rightarrow K_S^0 \pi^+$',
    35: r'$D^+ \rightarrow K_S^0 \pi^+ \pi^0$',
    36: r'$D^+ \rightarrow K_S^0 \pi^+ \pi^+ \pi^-$',
    37: r'$D^+ \rightarrow K_S^0 K^+$',
    38: r'$D^+ \rightarrow K^+ K^- \pi+$',
    41: r'$D^0 \rightarrow K^- \pi^+$',
    42: r'$D^0 \rightarrow K^- \pi^+ \pi^0$',
    43: r'$D^0 \rightarrow K^- \pi^+ \pi^+ \pi^-$',
    44: r'$D^0 \rightarrow K^- \pi^+ \pi^+ \pi^- \pi^0$',
    45: r'$D^0 \rightarrow K_S^0 \pi^0$',
    46: r'$D^0 \rightarrow K_S^0 \pi^+ \pi^-$',
    47: r'$D^0 \rightarrow K_S^0 \pi^+ \pi^- \pi^0$',
    48: r'$D^0 \rightarrow K^- K^+$',
}

Dst_id_to_label = {
    21: r'$D^{*0} \rightarrow D^0 \pi^0$',
    22: r'$D^{*0} \rightarrow D^0 \gamma$',
    23: r'$D^{*+} \rightarrow D^0 \pi^+$',
    24: r'$D^{*+} \rightarrow D^+ \pi^0$',
}

B_PDGs = [511, 521]
D_PDGs = [411, 421]
Dst_PDGs = [413, 423]
Dstst_PDGs = [10421, 10411, 10413, 10423, 20413, 20423, 415, 425]
lepton_tau_pdg = [11, 13, 15]

B_mother_pdg_condition = " or ".join(
    [f"abs(mcMother__boPDG__bc) == {p}" for p in B_PDGs]
)
B_grandmother_pdg_condition = " or ".join(
    [f"abs(mcMother__bomcMother__boPDG__bc__bc) == {p}" for p in B_PDGs]
)
B_grandgrandmother_pdg_condition = " or ".join(
    [f"abs(mcMother__bomcMother__bomcMother__boPDG__bc__bc__bc) == {p}" for p in B_PDGs]
)

is_d_meson = " or ".join(
    [f"abs(matchedMC__boPDG__bc) == {pdg}" for pdg in D_PDGs]
)
is_dst_meson = " or ".join(
    [f"abs(matchedMC__boPDG__bc) == {pdg}" for pdg in Dst_PDGs]
)

mother_sl_condition = " or ".join(
    [f"abs(mcMother__bomcDaughter__bo1__cm__spPDG__bc__bc) == {p}" for p in lepton_tau_pdg]
)
B_mother_sl = f"({B_mother_pdg_condition}) and ({mother_sl_condition})"

grandmother_sl_condition = " or ".join(
    [f"abs(mcMother__bomcMother__bomcDaughter__bo1__cm__spPDG__bc__bc__bc) == {p}" for p in lepton_tau_pdg]
)
B_grandmother_sl = f"({B_grandmother_pdg_condition}) and {grandmother_sl_condition}"

grandgrandmother_sl_condition = " or ".join(
    [f"abs(mcMother__bomcMother__bomcMother__bomcDaughter__bo1__cm__spPDG__bc__bc__bc__bc) == {p}" for p in
     lepton_tau_pdg]
)
B_grandgrandmother_sl = f"({B_grandgrandmother_pdg_condition}) and {grandgrandmother_sl_condition}"

is_continuum = "isContinuumEvent == 1"

d_hvars = [
    HistVariable('E', n_bins=30, var_name=r"$E_D$", unit="GeV"),
    HistVariable('abs__bodM__bc', n_bins=30, scope=(0, 0.2), var_name=r"$\left| \Delta_{M_D} \right|$", unit="GeV"),
    HistVariable('abs__boSigMBF__bc', n_bins=30, scope=(0, 10), var_name=r"$\left| \mathrm{sig(MBF)} \right|$",
                 unit=""),
    HistVariable('InvM', n_bins=30, scope=(1.6, 2.15), var_name=r"$Inv(M_D)$", unit="GeV"),
    HistVariable('useCMSFrame__bop__bc', n_bins=30, var_name=r"$ p^\ast_D$", unit="GeV"),
    HistVariable('useCMSFrame__bopt__bc', n_bins=30, var_name=r"$ p^\ast_{t,D}$", unit="GeV"),

]

dst_hvars = [
    HistVariable('abs__bodM__bc', n_bins=20, scope=(0, 0.2), var_name=r"$\left| \Delta_{M_D} \right|$", unit="GeV"),
    HistVariable('abs__boSigMBF__bc', n_bins=20, scope=(0, 10), var_name=r"$\left| \mathrm{sig(MBF)} \right|$",
                 unit=""),
    HistVariable('abs__bomassDifferenceSignificance__bo0__bc__bc', n_bins=20, scope=(0, 10),
                 var_name=r"$\left| \mathrm{sig(M_{D^\ast} - M_D)} \right|$", unit=""),
    HistVariable('daughterMotherDiffOf__bo0__cm__spM__bc', n_bins=20, scope=(0.115, 0.18),
                 var_name=r"$ \mathrm{M_{D^\ast} - M_D} $", unit="GeV"),
    HistVariable('InvM', n_bins=20, scope=(1.80, 2.2), var_name=r"$Inv(M_D)$", unit="GeV"),
    HistVariable('useCMSFrame__bop__bc', n_bins=20, var_name=r"$ p^\ast_{D^\ast}$", unit="GeV"),
    HistVariable('useCMSFrame__bopt__bc', n_bins=20, var_name=r"$ p^\ast_{t,D^{\ast}}$", unit="GeV"),
    HistVariable('useLabFrame__bodaughterAngleInBetween__bo0__cm__sp1__bc__bc', scope=(0, np.pi), n_bins=20,
                 var_name=r"$\measuredangle$ between first and second daughter", unit=""),
    HistVariable('daughter__bo1__cm__spp__bc', n_bins=20, scope=(0, 0.5), var_name=r"$p_{(\pi^0, \gamma)}$",
                 unit="GeV"),
    HistVariable('daughter__bo1__cm__spE__bc', n_bins=20, scope=(0, 0.5), var_name=r"$E_{(\pi^0, \gamma)}$",
                 unit="GeV"),
    HistVariable('daughterInvariantMass__bo0__cm__sp1__bc', n_bins=20, scope=(1.82, 2.18),
                 var_name=r"Inv. Mass of first and second daughter", unit="GeV"),
]


def sig_vs_bkg_plot(ax, df, hvar, title=""):
    """Plots 'isSignal==1' vs 'isSignal==0' for the given HistVariable
    on the given axis.

    :param ax: A matplotlib.axis instance.
    :param df: A pandas.DataFrame instance.
    :param hvar: A templatefitter.plotter.histogram_variable.HistVariable instance.
    :param title: Optional axis title. will be plotted on the top right corner.
    """
    logging.info(f"Plotting {hvar} for channel {title}")
    d_bkg = df.query("isSignal == 0")
    d_sig = df.query("isSignal == 1")

    shp = StackedHistogramPlot(hvar)
    shp.add_component("Background", d_bkg, weights=d_bkg.__weight__, color=KITColors.light_grey)
    shp.add_component("Signal", d_sig, weights=d_sig.__weight__, color=KITColors.kit_blue)

    shp.plot_on(ax, y_label="Candidates")
    ax.set_title(title, loc="right")


def split_df_into_D_components(df):
    """Split a DataFrame with D meson candidates into different
    components. The components are continuum, combinatorics,
    bbbar bkg and b->D(*/**)sl

    :param df: D meson candidate pandas.DataFrame.
    :return: Tuple of four pandas.DataFrames.
    """
    cont = df.query(f"{is_continuum}")
    not_cont = df.query(f"not({is_continuum})")

    comb = not_cont.query(f"not({is_d_meson})")
    not_comb = not_cont.query(f"{is_d_meson}")

    bkg = not_comb.query(f"not({B_mother_sl} or {B_grandmother_sl} or {B_grandgrandmother_sl})")
    bdsl = not_comb.query(f"{B_mother_sl} or {B_grandmother_sl} or {B_grandgrandmother_sl}")

    assert (len(cont.index) + len(comb.index) + len(bkg.index) + len(bdsl.index)) == len(df.index)

    return cont, comb, bkg, bdsl


def split_df_into_Dst_components(df):
    """Split a DataFrame with D* meson candidates into different
    components. The components are continuum, combinatorics,
    bbbar bkg and b->D(*/**)sl

    :param df: D* meson candidate pandas.DataFrame.
    :return: Tuple of four pandas.DataFrames.
    """
    cont = df.query(f"{is_continuum}")
    not_cont = df.query(f"not({is_continuum})")

    comb = not_cont.query(f"not({is_dst_meson})")
    not_comb = not_cont.query(f"{is_dst_meson}")

    bkg = not_comb.query(f"not({B_mother_sl} or {B_grandmother_sl})")
    bdsl = not_comb.query(f"{B_mother_sl} or {B_grandmother_sl}")

    assert (len(cont.index) + len(comb.index) + len(bkg.index) + len(bdsl.index)) == len(df.index)

    return cont, comb, bkg, bdsl


def D_component_plot(ax, df, hvar, title):
    """Plots normalized histograms of the given HistVariable for different
    D meson candidate types.

    :param ax: A matplotlib.axis instance.
    :param df: A pandas.DataFrame instance.
    :param hvar: A templatefitter.plotter.histogram_variable.HistVariable instance.
    :param title: Optional axis title. will be plotted on the top right corner.
    """
    logging.info(f"Plotting {hvar.df_label} for channel {title}")

    d_cont, d_comb, d_bkg, d_bdsl = split_df_into_D_components(df)

    shp = SimpleHistogramPlot(hvar)
    shp.add_component(r"Cont. Background", d_cont, hist_type="step", color=KITColors.kit_yellow)
    shp.add_component(r"Comb. Background", d_comb, hist_type="step", color=KITColors.kit_green)
    shp.add_component(r"$B\bar{B}$ Background", d_bkg, hist_type="step", color=KITColors.kit_red)
    shp.add_component(r"$B\rightarrow D^{(\ast/\ast\ast)}\ell\nu$", d_bdsl, hist_type="step", color=KITColors.kit_blue)
    shp.plot_on(ax, y_label="Candidates", normed=True)
    ax.set_title(title, loc="right")


def Dst_component_plot(ax, df, hvar, title):
    """Plots normalized histograms of the given HistVariable for different
    D* meson candidate types.

    :param ax: A matplotlib.axis instance.
    :param df: A pandas.DataFrame instance.
    :param hvar: A templatefitter.plotter.histogram_variable.HistVariable instance.
    :param title: Optional axis title. will be plotted on the top right corner.
    """
    logging.info(f"Plotting {hvar.df_label} for channel {title}")

    dst_cont, dst_comb, dst_bkg, dst_bdsl = split_df_into_Dst_components(df)

    shp = SimpleHistogramPlot(hvar)
    shp.add_component(r"Cont. Background", dst_cont, hist_type="step", color=KITColors.kit_yellow)
    shp.add_component(r"Comb. Background", dst_comb, hist_type="step", color=KITColors.kit_green)
    shp.add_component(r"$B\bar{B}$ Background", dst_bkg, hist_type="step", color=KITColors.kit_red)
    shp.add_component(r"$B\rightarrow D^{(\ast/\ast\ast)}\ell\nu$", dst_bdsl, hist_type="step",
                      color=KITColors.kit_blue)
    shp.plot_on(ax, y_label="Candidates", normed=True)
    ax.set_title(title, loc="right")


def plot_per_decay_mode(df, hvars, id_to_label_dict, plot_fcn, out_dir, plot_id):
    for decay_mode_id, label in id_to_label_dict.items():
        df_per_decay_mode = df.query(f"extraInfo__bodecayModeID__bc == {decay_mode_id}")

        if len(df_per_decay_mode.index) == 0:
            continue

        for hvar in hvars:
            fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=200)
            plot_fcn(ax, df_per_decay_mode, hvar, label)

            ylim = ax.get_ylim()
            ax.set_ylim((ylim[0], ylim[1] * 1.3))

            fig.savefig(os.path.join(out_dir, f"{decay_mode_id}_{hvar.df_label}_{plot_id}.pdf"), bbox_inches="tight")
            fig.savefig(os.path.join(out_dir, f"{decay_mode_id}_{hvar.df_label}_{plot_id}.png"), bbox_inches="tight")
            plt.close(fig)


def plot_whole_df(df, hvars, plot_fcn, out_dir, label, plot_id):
    for hvar in hvars:
        fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=200)
        plot_fcn(ax, df, hvar, label)

        ylim = ax.get_ylim()
        ax.set_ylim((ylim[0], ylim[1] * 1.3))

        fig.savefig(os.path.join(out_dir, f"{hvar.df_label}_{plot_id}.pdf"), bbox_inches="tight")
        fig.savefig(os.path.join(out_dir, f"{hvar.df_label}_{plot_id}.png"), bbox_inches="tight")
        plt.close(fig)


def sig_vs_bkg_dmID_bar_plot(df, id_to_label_dict, plot_id, out_dir, log=False):
    """Creates a bar plot of the decay mode id for the signal and background
    candidates of a D candidate DataFrame.

    :param df: D candidate pandas.DataFrame.
    :param id_to_label_dict: Dictionary with mappings from dmID to a latex label.
    :param plot_id: A identifier use in the plot name (for saving the figure).
    :param out_dir: Output directory.
    :param log: Whether to use log scale for the x axis or not.
    :return: None
    """
    id_counts_bkg = df.query("isSignal==0").groupby("extraInfo__bodecayModeID__bc")[
        "extraInfo__bodecayModeID__bc"].count()
    id_counts_sig = df.query("isSignal==1").groupby("extraInfo__bodecayModeID__bc")[
        "extraInfo__bodecayModeID__bc"].count()

    xtick_labels = [id_to_label_dict[dec_id] for dec_id in list(id_counts_sig.index)]

    index = np.arange(0, len(xtick_labels) * 2, 2)

    fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=200)

    ax.barh(index, id_counts_bkg, height=1.5, color=KITColors.light_grey, label="Background")
    ax.barh(index, width=id_counts_sig, left=id_counts_bkg, height=1.5, color=KITColors.kit_blue, label="Signal")

    if log:
        ax.set_xscale("log", nonposx='clip')
        ax.set_xlabel("Candidates", **xlabel_pos)
    else:
        ax.set_xlabel("Candidates", **xlabel_pos, labelpad=15)

    ax.legend(loc=0)
    plt.yticks(index, xtick_labels)

    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_{plot_id}.pdf"), bbox_inches="tight")
    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_{plot_id}.png"), bbox_inches="tight")
    plt.close(fig)


def d_meson_dmId_bar_plot(d_df, out_dir, log=False):
    """Creates a bar plot of the decay mode id for the different
    components of a D candidate DataFrame.

    :param d_df: D candidate pandas.DataFrame.
    :param out_dir: Output directory.
    :param log: Whether to use log scale for the x axis or not.
    :return: None
    """
    d_cont, d_comb, d_bkg, d_bdsl = split_df_into_D_components(d_df)

    id_counts_cont = d_cont.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_comb = d_comb.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_bkg = d_bkg.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_bdsl = d_bdsl.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()

    xtick_labels = [D_id_to_label[dec_id] for dec_id in list(id_counts_cont.index)]
    index = np.arange(0, len(xtick_labels) * 2, 2)

    fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=200)

    ax.barh(index, width=id_counts_cont, height=1.5, color=KITColors.kit_yellow, label="Cont. Background")
    ax.barh(index, width=id_counts_comb, left=id_counts_cont, height=1.5, color=KITColors.kit_green,
            label="Comb. Background")
    ax.barh(index, width=id_counts_bkg, left=id_counts_cont + id_counts_comb, height=1.5, color=KITColors.kit_red,
            label=r"$B\bar{B}$ Background")
    ax.barh(index, width=id_counts_bdsl, left=id_counts_cont + id_counts_comb + id_counts_bkg, height=1.5,
            color=KITColors.kit_blue, label=r"$B\rightarrow D^{(\ast/\ast\ast)}\ell\nu$")

    if log:
        ax.set_xscale("log", nonposx='clip')
        ax.set_xlabel("Candidates", **xlabel_pos)
    else:
        ax.set_xlabel("Candidates", **xlabel_pos, labelpad=15)

    ax.legend(bbox_to_anchor=(1, 1))
    plt.yticks(index, xtick_labels)

    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_D.pdf"), bbox_inches="tight")
    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_D.png"), bbox_inches="tight")
    plt.close(fig)


def dst_meson_dmId_bar_plot(dst_df, out_dir, log=False):
    """Creates a bar plot of the decay mode id for the different
    components of a D* candidate DataFrame.

    :param dst_df: D* candidate pandas.DataFrame.
    :param out_dir: Output directory.
    :param log: Whether to use log scale for the x axis or not.
    :return: None
    """

    dst_cont, dst_comb, dst_bkg, dst_bdsl = split_df_into_Dst_components(dst_df)

    id_counts_cont = dst_cont.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_comb = dst_comb.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_bkg = dst_bkg.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()
    id_counts_bdsl = dst_bdsl.groupby("extraInfo__bodecayModeID__bc")["extraInfo__bodecayModeID__bc"].count()

    xtick_labels = [Dst_id_to_label[dec_id] for dec_id in list(id_counts_cont.index)]
    index = np.arange(0, len(xtick_labels) * 2, 2)

    fig, ax = plt.subplots(1, 1, figsize=(5, 5), dpi=200)

    ax.barh(index, width=id_counts_cont, height=1.5, color=KITColors.kit_yellow, label="Cont. Background")
    ax.barh(index, width=id_counts_comb, left=id_counts_cont, height=1.5, color=KITColors.kit_green,
            label="Comb. Background")
    ax.barh(index, width=id_counts_bkg, left=id_counts_cont + id_counts_comb, height=1.5, color=KITColors.kit_red,
            label=r"$B\bar{B}$ Background")
    ax.barh(index, width=id_counts_bdsl, left=id_counts_cont + id_counts_comb + id_counts_bkg, height=1.5,
            color=KITColors.kit_blue, label=r"$B\rightarrow D^{(\ast/\ast\ast)}\ell\nu$")

    if log:
        ax.set_xscale("log", nonposx='clip')
        ax.set_xlabel("Candidates", **xlabel_pos)
    else:
        ax.set_xlabel("Candidates", **xlabel_pos, labelpad=15)

    ax.legend(bbox_to_anchor=(1, 1))
    plt.yticks(index, xtick_labels)

    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_Dst.pdf"), bbox_inches="tight")
    fig.savefig(os.path.join(out_dir, f"dmID_sigvsbkg_Dst.png"), bbox_inches="tight")
    plt.close(fig)


def d_meson_validation(source_path, output_dir):
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    print("=========================")
    print("D meson validation plots:")
    print("=========================")

    dirs = os.listdir(source_path)
    dirs.remove("Test2")

    data = dict()

    for ddir in dirs:
        data[ddir] = dict()

        flist = [f for f in glob.glob(os.path.join(g_source_path, ddir, "*.hdf")) if "gen" not in f]

        for f in flist:
            dtype = os.path.basename(f).split("_")[2]
            data[ddir][dtype] = pd.read_hdf(f)

    dzero_df = pd.concat([df["dzero"] for df in data.values()], sort=False)
    dcharged_df = pd.concat([df["dcharged"] for df in data.values()], sort=False)
    dstzero_df = pd.concat([df["dstzero"] for df in data.values()], sort=False)
    dstcharged_df = pd.concat([df["dstcharged"] for df in data.values()], sort=False)
    D_df = pd.concat([dzero_df, dcharged_df], sort=False)
    Dst_df = pd.concat([dstcharged_df, dstzero_df], sort=False)

    #
    # print(f"Plots will be saved in {output_dir}")
    # print(f"Loading Dzero ntuple from: {fdzero}")
    # print(f"Loading Dzero ntuple from: {fdcharged}")
    # D_df = pd.concat((load_root_file(fdzero), load_root_file(fdcharged)))
    #
    # print(f"Loading Dzero ntuple from: {fdstzero}")
    # print(f"Loading Dzero ntuple from: {fdstcharged}")
    # Dst_df = pd.concat((load_root_file(fdstzero), load_root_file(fdstcharged)))

    print("Creating plots for all D modes...")
    plot_whole_df(D_df, d_hvars, sig_vs_bkg_plot, output_dir, r"$D$ mesons", "D_sig_vs_bkg")
    plot_whole_df(D_df, d_hvars, D_component_plot, output_dir, r"$D$ mesons", "D_components")
    print("Creating plots for individual D modes...")
    plot_per_decay_mode(D_df, d_hvars, D_id_to_label, sig_vs_bkg_plot, output_dir, "sig_vs_bkg")
    plot_per_decay_mode(D_df, d_hvars, D_id_to_label, D_component_plot, output_dir, "components")
    print("Creating decay mode id plots for the D mesons")
    sig_vs_bkg_dmID_bar_plot(D_df, D_id_to_label, "D_mesons", output_dir)
    d_meson_dmId_bar_plot(D_df, output_dir)

    print("Creating plots for all D* modes...")
    plot_whole_df(Dst_df, dst_hvars, sig_vs_bkg_plot, output_dir, r"$D^\ast$ mesons", "Dst_sig_vs_bkg")
    plot_whole_df(Dst_df, dst_hvars, Dst_component_plot, output_dir, r"$D^\ast$ mesons", "Dst_components")
    print("Creating plots for individual D* modes...")
    plot_per_decay_mode(Dst_df, dst_hvars, Dst_id_to_label, sig_vs_bkg_plot, output_dir, "Dst_sig_vs_bkg")
    plot_per_decay_mode(Dst_df, dst_hvars, Dst_id_to_label, Dst_component_plot, output_dir, "Dst_components")
    print("Creating decay mode id plots for the D* mesons")
    sig_vs_bkg_dmID_bar_plot(Dst_df, Dst_id_to_label, "Dst_mesons", output_dir)
    dst_meson_dmId_bar_plot(Dst_df, output_dir)


if __name__ == "__main__":
    g_source_path = '/ceph/welsch/rdstar_data/DMeson_study_loose_cuts_prod_24thFeb2019'
    d_meson_validation(
        g_source_path,
        "/home/welsch/dmeson_validation_plots"
    )
