"""
Sideband Validation
    - Sideband Plot Definitions in sideband_info_collection
    - Plotting Function

Felix Metzner 2019
"""

import os
import copy
import pandas as pd

from matplotlib import pyplot as plt
from typing import NamedTuple, Optional, Union, List, Tuple, Dict

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import DataMCHistogramPlot

from rdstar.offline_analysis.mc_ids import sig_id_to_label, rdstar_sig_queries
from rdstar.offline_analysis.rdstar_plots import get_rdstar_color_dict, plot_chi_squared_toys
from rdstar.offline_analysis.systematics import SysColsInfo, get_systematics, full_systematics_cols, \
    full_systematics_name


__all__ = [
    "SidebandInfo",
    "hvar_neextra_fine",
    "hvar_eextra",
    "hvar_cs_mva_output",
    "hvar_loose_cs_mva_output",
    "hvar_pStarL",
    "hvar_missing_mass_q2sb",
    "hvar_mbc",
    "hvar_tag_delta_e",
    "dss_dec_modes",
    "rdstar_dec_modes",
    "dss_dec_titles",
    "rdstar_dec_titles",
    "plot_data_mc_comparison",
    "sideband_info_collection",
    "reset_systematics_in_sideband_info"
]


DecayModeDictType = Dict[str, List[int]]
DecayTitleDictType = Dict[str, str]


class SidebandInfo(NamedTuple):
    key: str
    hist_var: HistVariable
    title: str
    sb_text: str
    styles: Tuple[str, str] = ("summed", "stacked")
    text_h_pos: Union[float, Tuple[float, float]] = 0.66
    text_v_pos: Optional[Union[float, Tuple[float, float]]] = (0.88, 0.55)
    legend_loc: Optional[Union[int, str, Tuple[Union[int, str], Union[int, str]]]] = None
    legend_cols: Optional[Union[int, Tuple[int, int]]] = None
    y_scale: Optional[Union[float, Tuple[float, float]]] = None
    split_chi_info_text: Union[bool, Tuple[bool, bool]] = True
    systematics: Optional[List[SysColsInfo]] = None
    sys_name_tag: Optional[str] = None


# TODO
# hvar_missing_mass = HistVariable("m2RecoilSignalSide", 25, (-2.5, 8), r"$M_\mathrm{miss}^2$", r"$\mathrm{GeV}^2$")
# hvar_missing_mass_lower_bin = HistVariable("m2RecoilSignalSide", 15, (-2.5, 8), r"$M_\mathrm{miss}^2$", r"$\mathrm{GeV}^2$")
# hvar_missing_mass_ge_1p5 =  HistVariable("m2RecoilSignalSide", 25, (1.5, 8), r"$M_\mathrm{miss}^2$", r"$\mathrm{GeV}^2$")


# hvar_mbc = HistVariable("daughter__bo0__cm__spMbc__bc", 25, (5.22, 5.27), r"$M_\mathrm{bc}$", "GeV")
# hvar_mbc_lower_bin = HistVariable("daughter__bo0__cm__spMbc__bc", 15, (5.22, 5.27), r"$M_\mathrm{bc}$", "GeV")


# hvar_diff_dpid = HistVariable(
#     "daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
#     30, (0.1, 1.0), r"$M(D^{(\ast)}\pi^0) - M(D^{(\ast)})$", r"$\mathrm{GeV}$"
# )

# hvar_missing_mom_cms = HistVariable("missingMomentumInCMSFrame", 30, (0, 1.5), r"$p_\mathrm{miss}^\ast$", "GeV")
# TODO


hvar_neextra_fine = HistVariable("ROE_neextra__boCleanROE__bc", 40, (0, 2.5),
                                 r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra}$", r"$\mathrm{GeV}$")

hvar_eextra = HistVariable("ROE_eextra__boCleanROE__bc", 20, (0, 3), r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
                           r"$\mathrm{GeV}$")

hvar_cs_mva_output = HistVariable("cs_mva_output", 40, (0.5, 1), r"$\mathcal{P}_\mathrm{CS}$")
hvar_loose_cs_mva_output = HistVariable("cs_mva_output", 40, (0.2, 1), r"$\mathcal{P}_\mathrm{CS}$")

hvar_pStarL = HistVariable("pStarL", 25, (0.1, 2.2), r"$\mathrm{p}_\ell^\ast$", r"$\mathrm{GeV}$")
# hvar_pStarL = HistVariable("pStarL", 30, (0.1, 2.2), r"$\mathrm{p}_\ell^\ast$", r"$\mathrm{GeV}$")

hvar_missing_mass_q2sb = HistVariable("m2RecoilSignalSide", 25, (-2., 4.), r"$\mathrm{M}_\mathrm{miss}^2$",
                                      r"$\mathrm{GeV}^2$")

hvar_mbc = HistVariable("daughter__bo0__cm__spMbc__bc", 25, (5.27, 5.29), r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
                        "GeV")

hvar_tag_delta_e = HistVariable("daughter__bo0__cm__spdeltaE__bc", 30, (-.05, 0.1),
                                r"$\Delta\mathrm{E}^\mathrm{tag}$", "GeV")

# q^2 Sideband
q2_sideband_title = r"$\mathrm{q}^2$ Sideband"
q2_sideband_text = r"$\mathrm{q}^2 < 4.0$ GeV$^2$"

_q2_sideband_info_collection = [  # type: List[SidebandInfo]
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS",
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS",
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        text_v_pos=(0.88, 0.62),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=hvar_neextra_fine,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        text_v_pos=(0.88, 0.62),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=hvar_eextra,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS\n" + r"No cut on $\mathrm{E}^\mathrm{ECL}_\mathrm{extra}$",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_cs_mva_output,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=hvar_cs_mva_output,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=hvar_cs_mva_output,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_2_random_bcs",
        hist_var=hvar_loose_cs_mva_output,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS\n" + r"Loose cut on $\mathcal{P}_\mathrm{CS}$",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_2_no_bcs",
        hist_var=hvar_loose_cs_mva_output,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS\n" + r"Loose cut on $\mathcal{P}_\mathrm{CS}$",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_pStarL,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.78),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3",
        hist_var=hvar_pStarL,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\n" + r"No cut on $\mathrm{p}_\ell$",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.0),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3_random_bcs",
        hist_var=hvar_pStarL,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS\n" + r"No cut on $\mathrm{p}_\ell$",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.4, 2.4),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3_no_bcs",
        hist_var=hvar_pStarL,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS\n" + r"No cut on $\mathrm{p}_\ell$",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.4, 2.6),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_missing_mass_q2sb,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=hvar_missing_mass_q2sb,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=hvar_missing_mass_q2sb,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_mbc,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        text_v_pos=(0.97, 0.78),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1",
        hist_var=hvar_mbc,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\n" + r"No cut on $\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
        text_v_pos=(0.97, 0.78),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_random_bcs",
        hist_var=hvar_mbc,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS\n" + r"No cut on $\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
        text_v_pos=(0.97, 0.8),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_no_bcs",
        hist_var=hvar_mbc,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS\n" + r"No cut on $\mathrm{M}_\mathrm{bc}^\mathrm{tag}$",
        text_v_pos=(0.97, 0.8),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=hvar_tag_delta_e,
        title=q2_sideband_title,
        sb_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=hvar_tag_delta_e,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nNo BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=hvar_tag_delta_e,
        title=q2_sideband_title,
        sb_text=q2_sideband_text + "\nRandom BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    )
]

# Tag M_bc Sideband
mbc_sideband_title = r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag}$ Sideband"
mbc_sideband_text = r"$\mathrm{M}_\mathrm{bc}^\mathrm{tag} \leq 5.27\,$GeV"

_tag_mbc_sideband_info_collection = [
    SidebandInfo(
        key="mbc_sideband",
        hist_var=hvar_neextra_fine,
        title=mbc_sideband_title,
        sb_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    ),
    SidebandInfo(
        key="mbc_sideband_no_bcs",
        hist_var=hvar_neextra_fine,
        title=mbc_sideband_title,
        sb_text=mbc_sideband_text + "\nNo BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    ),
    SidebandInfo(
        key="mbc_sideband_random_bcs",
        hist_var=hvar_neextra_fine,
        title=mbc_sideband_title,
        sb_text=mbc_sideband_text + "\nRandom BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None
    )
]

_sideband_info_collection = _q2_sideband_info_collection + _tag_mbc_sideband_info_collection

dss_dec_modes = {  # type: DecayModeDictType
    "DZPi": [63, 64],
    "DMPi": [61, 62],
    "DSZPi": [67, 68],
    "DSMPi": [65, 66],
}

rdstar_dec_modes = {  # type: DecayModeDictType
    "DZ": [13, 14],
    "DM": [11, 12],
    "DSZ": [17, 18],
    "DSM": [15, 16],
}

dss_dec_titles = {  # type: DecayTitleDictType
    "DZPi": r"$B^- \rightarrow D^0 \pi^0 \ell^-$",
    "DMPi": r"$B^0 \rightarrow D^- \pi^0 \ell^+$",
    "DSZPi": r"$B^- \rightarrow D^{*0} \pi^0 \ell^-$",
    "DSMPi": r"$B^0 \rightarrow D^{*-} \pi^0 \ell^+$",
}

rdstar_dec_titles = {  # type: DecayTitleDictType
    "DZ": r"$B^- \rightarrow D^0 \ell^-$",
    "DM": r"$B^0 \rightarrow D^- \ell^+$",
    "DSZ": r"$B^- \rightarrow D^{*0} \ell^-$",
    "DSM": r"$B^0 \rightarrow D^{*-} \ell^+$",
    "all_modes": "All Reconstruction Modes",
}


def plot_data_mc_comparison(
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        weight_col: str,
        sideband_info: SidebandInfo,
        dec_modes: Optional[DecayModeDictType] = None,
        dec_titles: Optional[DecayTitleDictType] = None,
        lumi_label: bool = True,
        draw_legend: bool = True,
        legend_inside: bool = True,
        export_path: Optional[Union[str, os.fspath]] = None,
        style: str = "summed",
        ratio_type: str = "normal",
        fig_size: Tuple[float, float] = (5.5, 5.),
        normalize_mc_to_data: bool = True,
        gof_check_method: Optional[str] = None,
        adaptive_binning: bool = False,
        dry_run: bool = False,
        show_plots: bool = False
) -> Optional[Dict[str, os.fspath]]:
    if show_plots:
        assert not dry_run

    info_text_pos = get_info_text_position(sideband_info=sideband_info, style=style, draw_legend=draw_legend)
    legend_loc, legend_cols = get_legend_info(sideband_info=sideband_info, style=style)
    y_scale = get_y_scale(sideband_info=sideband_info, style=style)
    split_chi_info_text = get_chi_text_split_info(sideband_info=sideband_info, style=style)

    if dec_modes is None:
        dec_modes = rdstar_dec_modes
    if dec_titles is None:
        dec_titles = rdstar_dec_titles

    systematics_cols = None
    include_sys = False
    if sideband_info.systematics is not None and not dry_run:
        systematics_cols = get_systematics(df=mc_df, sys_cols_infos=sideband_info.systematics)
        if systematics_cols:
            include_sys = True

    tag_info = f"{sideband_info.key}_" if sideband_info.key is not None else ""
    sys_tag = f"{sideband_info.sys_name_tag}_" if sideband_info.sys_name_tag is not None else ""
    file_name_base = f"data_vs_mc_{style}_{sideband_info.hist_var.df_label}_{tag_info}{sys_tag}"

    plot_order = ["CBKG", "GBKG", "HBKG", "DSSlbkg", "DSSl", "DSl", "Dl", "DStau", "Dtau"]

    all_mode_ids = [d_id for d_ids in dec_modes.values() for d_id in d_ids]
    output_dict = {}

    mode_infos = [(k, v) for k, v in dec_modes.items()]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        file_name = f"{file_name_base}{mode_name}"
        if export_path is not None:
            output_path = os.path.join(export_path, f"{file_name}.pdf")
            assert file_name not in output_dict.keys(), file_name
            output_dict.update({file_name: output_path})

        if dry_run:
            continue

        dmc_hp = DataMCHistogramPlot(sideband_info.hist_var)

        dmc_hp.add_data_component("Sideband Data", data_df.loc[
            data_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)])

        for comp in plot_order:
            reco_id = rdstar_sig_queries[comp]["id"]
            dmc_hp.add_mc_component(
                sig_id_to_label[reco_id],
                mc_df.loc[mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)].query(
                    f"SIG_ID == {reco_id}"),
                weights=weight_col,
                systematics=systematics_cols,
                color=get_rdstar_color_dict()[reco_id]
            )

        fig, ax = DataMCHistogramPlot.create_hist_ratio_figure(fig_size=fig_size)

        chi2, dof, p_val, toy_output = dmc_hp.plot_on(
            ax1=ax[0], ax2=ax[1],
            style=style,
            ratio_type=ratio_type,
            draw_legend=draw_legend,
            legend_inside=legend_inside,
            legend_cols=legend_cols,
            legend_loc=legend_loc,
            y_scale=y_scale,
            normalize_to_data=normalize_mc_to_data,
            include_sys=include_sys,
            gof_check_method=gof_check_method,
            adaptive_binning=adaptive_binning
        )

        ax[0].set_title(dec_titles[mode_name], loc="right")
        if sideband_info.title is not None:
            use_title = sideband_info.title
            if ax[0].get_ylim()[1] > 0.85e4:
                padding = " " * 9
                use_title = padding + sideband_info.title
            ax[0].set_title(use_title, loc="left")

        info_text = r"$\int \mathcal{L} \,dt=711\,\mathrm{fb}^{-1}$" if lumi_label else ""
        if sideband_info.sb_text is not None:
            info_text += "\n" + sideband_info.sb_text
        if chi2 is not None:
            info_text += "\n" + r"$p$-Value = " + f"{p_val:.2f}"
            if split_chi_info_text:
                if chi2 >= 1000:
                    info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.0f}/{int(dof)}"
                else:
                    info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.1f}/{int(dof)}"
                info_text += "\n" + r"$\qquad\quad\,\,$ = " + f"{chi2 / dof:.2f}"
            else:
                if chi2 >= 1000:
                    info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.0f}/{int(dof)} = {chi2 / dof:.2f}"
                else:
                    info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.1f}/{int(dof)} = {chi2 / dof:.2f}"

        ax[0].text(
            x=info_text_pos[0],
            y=info_text_pos[1],
            s=info_text,
            transform=ax[0].transAxes,
            va="top",
            linespacing=1.5
        )

        fig.canvas.draw()
        x_ticks = ax[1].get_xticklabels()
        if len(x_ticks[0].get_text().split("{")[1].split("}")[0]) > 4:
            new_x_ticks = x_ticks[1::2]
            if all(tick.get_text().split("{")[1].split("}")[0][-1] == "0" for tick in new_x_ticks):
                for tick in x_ticks:
                    original_text = tick.get_text()
                    original_value = original_text.split("{")[1].split("}")[0]
                    tick.set_text(original_text.replace(original_value, original_value[:-1]))
            ax[1].set_xticklabels(x_ticks)
            plt.setp(x_ticks[::2], visible=False)
            plt.setp(new_x_ticks, visible=True)

        if export_path is not None:
            export(fig, file_name, export_path)

        if not show_plots:
            plt.close(fig)

        if toy_output is not None:
            _, bin_edges, toys = toy_output
            plot_chi_squared_toys(
                toys=toys,
                bin_edges=bin_edges,
                ndf=dof,
                observed_chi2=chi2,
                full_output_path=None if export_path is None else os.path.join(export_path, file_name),
                tag="chi2_distribution"
            )

    if export_path is not None:
        return output_dict


def get_info_text_position(
        sideband_info: SidebandInfo,
        style: str,
        draw_legend: bool
) -> Tuple[Union[float, int], Union[float, int]]:
    info_text_pos = [0.66, 0.88]  # type: List[float, float]
    if isinstance(sideband_info.text_h_pos, tuple):
        assert all(isinstance(element, (int, float)) for element in sideband_info.text_h_pos)
        info_text_pos[0] = sideband_info.text_h_pos[0] if style == "summed" else sideband_info.text_h_pos[1]
    else:
        assert isinstance(sideband_info.text_h_pos, (int, float))
        info_text_pos[0] = sideband_info.text_h_pos
    if sideband_info.text_v_pos is not None:
        if isinstance(sideband_info.text_v_pos, tuple):
            assert all(isinstance(element, (int, float)) for element in sideband_info.text_v_pos)
            info_text_pos[1] = sideband_info.text_v_pos[0] if style == "summed" else sideband_info.text_v_pos[1]
        else:
            assert isinstance(sideband_info.text_v_pos, (int, float))
            info_text_pos[1] = sideband_info.text_v_pos

    if info_text_pos[0] == 0.66 and info_text_pos[1] == 0.55:
        if style == "stacked" and not draw_legend:
            info_text_pos[1] = info_text_pos[1] + 0.3

    return info_text_pos[0], info_text_pos[1]


def get_legend_info(sideband_info: SidebandInfo, style: str) -> Tuple[Optional[int], Optional[int]]:
    legend_loc, legend_cols = (None, None)
    if sideband_info.legend_loc is not None:
        if isinstance(sideband_info.legend_loc, tuple):
            legend_loc = sideband_info.legend_loc[0] if style == "summed" else sideband_info.legend_loc[1]
        else:
            legend_loc = sideband_info.legend_loc
    if sideband_info.legend_cols is not None:
        if isinstance(sideband_info.legend_cols, tuple):
            legend_cols = sideband_info.legend_cols[0] if style == "summed" else sideband_info.legend_cols[1]
        else:
            legend_cols = sideband_info.legend_cols
    return legend_loc, legend_cols


def get_y_scale(sideband_info: SidebandInfo, style: str) -> float:
    y_scale = 1.1
    if sideband_info.y_scale is not None:
        if isinstance(sideband_info.y_scale, tuple):
            y_scale = 1.1 * sideband_info.y_scale[0] if style == "summed" else sideband_info.y_scale[1]
        else:
            y_scale = sideband_info.y_scale
    return y_scale


def get_chi_text_split_info(sideband_info: SidebandInfo, style: str) -> Union[bool, Tuple[bool, bool]]:
    if isinstance(sideband_info.split_chi_info_text, tuple):
        return sideband_info.split_chi_info_text[0] if style == "summed" else sideband_info.split_chi_info_text[1]
    else:
        return sideband_info.split_chi_info_text


def reset_systematics_in_sideband_info(
        sideband_info: SidebandInfo,
        new_systematics: Optional[List[SysColsInfo]],
        sys_name_tag: Optional[str]
) -> SidebandInfo:
    assert isinstance(sideband_info, SidebandInfo), type(sideband_info)
    return SidebandInfo(
        key=copy.copy(sideband_info.key),
        hist_var=copy.copy(sideband_info.hist_var),
        title=copy.copy(sideband_info.title),
        sb_text=copy.copy(sideband_info.sb_text),
        styles=copy.copy(sideband_info.styles),
        text_h_pos=copy.copy(sideband_info.text_h_pos),
        text_v_pos=copy.copy(sideband_info.text_v_pos),
        legend_loc=copy.copy(sideband_info.legend_loc),
        legend_cols=copy.copy(sideband_info.legend_cols),
        y_scale=copy.copy(sideband_info.y_scale),
        split_chi_info_text=copy.copy(sideband_info.split_chi_info_text),
        systematics=copy.copy(new_systematics),
        sys_name_tag=sys_name_tag
    )


def create_sideband_infos_with_systematics(
        sideband_info_list: List[SidebandInfo],
        systematics_collection: List[SysColsInfo],
        sys_name_tag: str
) -> List[SidebandInfo]:
    assert isinstance(systematics_collection, list), type(systematics_collection)
    assert all(isinstance(sys, SysColsInfo) for sys in systematics_collection)
    return [
        reset_systematics_in_sideband_info(
            sideband_info=sb_info,
            new_systematics=systematics_collection,
            sys_name_tag=sys_name_tag
        ) for sb_info in sideband_info_list
    ]


_sideband_info_collection_with_full_sys = create_sideband_infos_with_systematics(
    sideband_info_list=_sideband_info_collection,
    systematics_collection=full_systematics_cols,
    sys_name_tag=full_systematics_name
)

sideband_info_collection = _sideband_info_collection + _sideband_info_collection_with_full_sys
