"""
Definition of convenience functions for the analysis-specific decay channel reconstruction.

it originates from:
Felix Metzner 2018
"""

from analysistools.processing import reco_utilities as reco_utils
from analysistools.processing import roe_masks_utils
from analysistools.processing import decay_channels_utils
from analysistools.processing import best_cand_selection_utils as bcs

from rdstar.online_analysis import decay_channels
from rdstar.online_analysis import final_state_particles as fsp
from rdstar.online_analysis import roe_masks
from rdstar.online_analysis import skim_cuts
from rdstar.online_analysis import best_cand_selection


def reconstruct_Kshorts(path, config):
    for channel in decay_channels.KS0_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii']
        )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.KS0_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='reco'
    )


def reconstruct_pi0(path, config):
    for channel in decay_channels.pi0_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii']
        )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.pi0_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='reco'
    )


def reconstruct_tau_for_gen(path, config):
    if config['use_mc_info'] is False:
        return

    for channel in decay_channels.tau_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii']
        )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.tau_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='all'
    )


def reconstruct_D(path, config):
    for channel in decay_channels.D_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii'],
        )

    particle_lists, gen_particle_lists = reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.D_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='all'
    )

    bcs.apply_bcs_from_dict(path, best_cand_selection.D_meson_by_dM_bcs, particle_lists)


def reconstruct_Dstar(path, config):
    for channel in decay_channels.Dstar_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii']
        )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.Dstar_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='all'
    )


def reconstruct_D2star_for_gen(path, config):
    if config['use_mc_info'] is False:
        return

    for channel in decay_channels.D2star_decay_channels:
        reco_utils.fillParticleListsFromMC([(f"{channel}_gen", '')], path=path, addDaughters=True)

    # for channel in decay_channels.D2star_decay_channels:
    #     reco_utils.reconstruct_decay_channel(
    #         path=path,
    #         decay_label=channel,
    #         decay_chl_dict=decay_channels.decay_channel_dict,
    #         use_mc_info=config['use_mc_info'],
    #         use_b2bii=config['use_b2bii']
    #     )

    # for channel in decay_channels.D2star_decay_channels:
    #     reco_utils.findMCDecay(
    #         list_name=f"{channel}_gen",
    #         decay=decay_channels.decay_channel_dict['decay_label_to_chain'][channel].format(suf='', neutrino=''),
    #         writeOut=True,
    #         path=path
    #     )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.D2star_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='all'
    )


def reconstruct_sig_B(path, config):
    sigB_channels = {
        **decay_channels.B_decay_channels,
        **decay_channels.B_Dlpi0_decay_channels,
        **decay_channels.B_Dlpim_decay_channels,
        **decay_channels.Btau_decay_channels,
        **decay_channels.B_D2star_decay_channels,
    }
    for channel in sigB_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii'],
            add_neutrinos=True
        )

    reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=sigB_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='sig'
    )


def reconstruct_upsilon4s(path, config):
    for channel in decay_channels.U4S_decay_channels:
        reco_utils.reconstruct_decay_channel(
            path=path,
            decay_label=channel,
            decay_chl_dict=decay_channels.decay_channel_dict,
            use_mc_info=config['use_mc_info'],
            use_b2bii=config['use_b2bii'],
        )

    particle_lists, gen_particle_lists = reco_utils.merge_particle_lists(
        path=path,
        merge_channel_dict=decay_channels.U4S_decay_channels,
        decay_chl_dict=decay_channels.decay_channel_dict,
        use_mc_info=config['use_mc_info'],
        label='all'
    )

    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_fei_probability_bcs, particle_lists)
    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_vertex_fit_probability_bcs, particle_lists)
    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_roe_eextra_bcs, particle_lists)
    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_roe_neextra_bcs, particle_lists)
    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_deltaE_bcs, particle_lists)
    bcs.apply_bcs_from_dict(path, best_cand_selection.upsilon4s_with_DMeson_deltaM_bcs, particle_lists)


def print_decay_summary(
        fsp_collection=None,
        decay_channel_dict=None,
        list_of_skims=None,
        bcs_dict=None,
        roe_mask_list=None
):
    if fsp_collection:
        print("Final State Particles Selection Criteria:")
        print(fsp_collection)

    if decay_channels:
        print("\n\nDecay Channels Reconstruction Criteria:")
        decay_channels_utils.print_decay_channels_table(decay_channel_dict)

    if list_of_skims:
        print("\n\nPreselection and Skim Criteria:")
        skim_cuts.print_skim_cuts(list_of_skims)

    if bcs_dict:
        print("\n\nBest Candidate Selection Criteria:")
        bcs.print_bcs_settings(bcs_dict)

    if roe_mask_list:
        b2bii = True
        print(f"\n\nAll Defined ROE Masks for use_b2bii = {b2bii}:")
        roe_masks_utils.print_roe_masks(roe_mask_list, use_b2bii=b2bii)


if __name__ == '__main__':
    print_decay_summary(
        fsp_collection=fsp.fsp_collection,
        decay_channel_dict=decay_channels.decay_channel_dict,
        list_of_skims=skim_cuts.list_of_skims,
        bcs_dict=best_cand_selection.my_bcs_settings,
        roe_mask_list=roe_masks.roe_mask_list
    )
