"""
Steering file for the n-tuple file generation for the B -> D(*) tau nu analysis.
Felix Metzner 2018
"""

import basf2
import b2biiConversion

from modularAnalysis import *

from analysistools.processing import path_utilities
from analysistools.processing import reco_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import observables, decay_utilities, skim_cuts, decay_channels

input_files, input_sequences, output_dict, config = path_utilities.get_commandline_args()

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()

b2biiConversion.convertBelleMdstToBelleIIMdst(
    inputBelleMDSTFile=input_files,
    applyHadronBJSkim=True,
    useBelleDBServer=os.environ['BELLE_POSTGRES_SERVER'],
    path=main_path,
    entrySequences=input_sequences
)

setAnalysisConfigParams({'mcMatchingVersion': 'Belle'}, main_path)


# Get Final State Particles
main_path.add_module('CurlerCloneMarker')
decay_utilities.fsp.fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config['use_mc_info'])
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.pre_skim_cuts)

decay_utilities.reconstruct_tau_for_gen(main_path, config)


# Build intermediate D mesons
decay_utilities.reconstruct_D(main_path, config)
decay_utilities.reconstruct_Dstar(main_path, config)
decay_utilities.reconstruct_D2star_for_gen(main_path, config)


# Build signal side B meson
decay_utilities.reconstruct_sig_B(main_path, config)
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.post_signal_skim_cuts)


# Get tag side B meson via FEI
reco_utilities.reconstruct_tag_B(
    path=main_path,
    config=config,
    b_tag_chl_dict=decay_channels.BTag_decay_channels,
    decay_chl_dict=decay_channels.decay_channel_dict
)
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.post_tag_skim_cuts)


# Build Upsilon(4S)
decay_utilities.reconstruct_upsilon4s(main_path, config)
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.post_reco_skim_cuts)


##########
# Output #
##########

# Get Decay Tree info if MC info is available
if config["use_mc_info"] is True:
    main_path.add_module(
        'ParticleMCDecayString',
        fileName=(output_dict['Y4S_decay_trees']),
        listName='Upsilon(4S):all'
    )


up4s_coll = observables.get_upsilon4s_collection(config)
main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['Y4S_ntuple'],
    particleList='Upsilon(4S):all',
    variables=up4s_coll.as_list()
)


if config["use_mc_info"] is True:
    # Move this VariablesToNtuple call in front of the last skim, if all
    # possible Upsilon(4S) candidates shall be written out...
    ups4s_gen_coll = observables.get_gen_level_upsilon4s_vars(config)
    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['Y4S_generator_level_reco_ntuple'],
        particleList='Upsilon(4S):all_gen',
        variables=ups4s_gen_coll.as_list()
    )

    b_gen_coll = observables.get_gen_level_b_collection(config)
    b_direct_gen_coll = observables.get_direct_gen_level_b_collection(config)
    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['B_minus_generator_info_ntuple'],
        particleList='B-:sig_gen',
        variables=b_gen_coll.as_list()
    )
    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['B_zero_generator_info_ntuple'],
        particleList='B0:sig_gen',
        variables=b_gen_coll.as_list()
    )

    fillParticleListsFromMC([('B-:mc', '')], addDaughters=True, path=main_path)
    fillParticleListsFromMC([('B0:mc', '')], addDaughters=True, path=main_path)
    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['B_minus_generator_direct_info_ntuple'],
        particleList='B-:mc',
        variables=b_direct_gen_coll.as_list()
    )
    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['B_zero_generator_direct_info_ntuple'],
        particleList='B0:mc',
        variables=b_direct_gen_coll.as_list()
    )

basf2.print_path(main_path)
basf2.process(main_path)
print(statistics)
