"""
Steering file for testing the n-tuple file generation for the B -> D(*) tau nu analysis.

Contains the reconstruction analog to reconstruction.py but produces additional output.

Felix Metzner 2018
"""

import basf2
import b2biiConversion

from modularAnalysis import *

from analysistools.processing import path_utilities
from analysistools.processing import reco_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import observables, decay_utilities, skim_cuts, decay_channels

input_files, input_sequences, output_dict, config = path_utilities.get_commandline_args()

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()

b2biiConversion.convertBelleMdstToBelleIIMdst(
    inputBelleMDSTFile=input_files,
    applyHadronBJSkim=True,
    useBelleDBServer=os.environ['BELLE_POSTGRES_SERVER'],
    path=main_path,
    entrySequences=input_sequences
)

setAnalysisConfigParams({'mcMatchingVersion': 'Belle'}, main_path)

main_path.add_module('CurlerCloneMarker')
decay_utilities.fsp.fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config['use_mc_info'])

#fillParticleListsFromMC([('D0:gen', '')], addDaughters=True, path=main_path)
fillParticleListsFromMC([('anti-nu_e:gen', '')], addDaughters=True, path=main_path)

#copyLists(
#    outputListName='e-:all',
#    inputListNames=['e-:fsp','e-:corr'],
#    path=main_path
#)

# reconstruct D
reconstructDecay(
    decayString='D0:reco -> K-:fsp pi+:fsp pi0:fsp',
    cut='useCMSFrame(p) < 3.0 and abs(dM) < 0.05 and useCMSFrame(daughter(2, p)) > 0.2',
    dmID=1,
    path=main_path
)
matchMCTruth("D0:reco", path=main_path)

# reconstruct D*
reconstructDecay(
    decayString='D*0:reco -> D0:reco pi0:fsp',
    cut="useCMSFrame(p) < 3.0 and daughterMotherDiffOf(0, M) < 0.16",
    dmID=1,
    path=main_path
)
matchMCTruth("D*0:reco", path=main_path)

# reconstruct B with corrected and uncorrected electrons
reconstructDecay(
    decayString='B-:uncorr -> D*0:reco e-:fsp anti-nu_e:gen',
    cut="1.0 < M < 6.0 and R2 < 0.6 and daughter(1, eIDBelle) > 0.6 and daughter(1, p) > 0.3",
    dmID=1,
    path=main_path
)
matchMCTruth("B-:uncorr", path=main_path)

reconstructDecay(
    decayString='B-:corr -> D*0:reco e-:corr anti-nu_e:gen',
    cut="1.0 < M < 6.0 and R2 < 0.6 and daughter(1, eIDBelle) > 0.6 and daughter(1, p) > 0.3",
    dmID=1,
    path=main_path
)
matchMCTruth("B-:corr", path=main_path)

# get tag side B-Meson
fillParticleListsFromMC([('B+:tag', '')], addDaughters=True, path=main_path)


# recostruct Upsilon(4S) corrected and uncorrected
reconstructDecay(
    decayString='Upsilon(4S):corr -> B+:tag B-:corr',
    cut="",
    dmID=1,
    path=main_path
)
matchMCTruth("Upsilon(4S):corr", path=main_path)

reconstructDecay(
    decayString='Upsilon(4S):uncorr -> B+:tag B-:uncorr',
    cut="",
    dmID=1,
    path=main_path
)
matchMCTruth("Upsilon(4S):uncorr", path=main_path)

b_gen_var_list = [
    "expNum",
    "runNum",
    "evtNum",
    'PDG',
    'extraInfo(decayModeID)',
    'Mbc',
    'M',
    'InvM',
    'matchedMC(E)',
    'matchedMC(px)',
    'matchedMC(py)',
    'matchedMC(pz)',
    'matchedMC(M)',
    'matchedMC(mdstIndex)',
    'matchedMC(InvM)',
    'E',
    'px', 'py', 'pz',
    'deltaE',
    'chiProb',
    'charge',
    'decayAngle(0)',
    'decayAngle(1)',
    'cosThetaBetweenParticleAndTrueB',
    'daughterInvM(0, 2)',
    'formula(daughterInvM(0, 2) - daughter(0, InvM))',
    'daughter(2, p)',
    'daughter(2, theta)',
    'useCMSFrame(daughter(0, p))',
    'useCMSFrame(daughter(1, p))',
    'useCMSFrame(daughter(2, p))',
    'useCMSFrame(daughterAngle(0, 1))',
    'useCMSFrame(daughterAngleInBetween(0, 1))',
    'useCMSFrame(daughterDiffOfPhi(0, 1))',
    'useCMSFrame(daughterAngle(0, 2))',
    'useCMSFrame(daughterAngleInBetween(0, 2))',
    'useCMSFrame(daughterDiffOfPhi(0, 2))',
    'useRestFrame(daughter(0, p))',
    'useRestFrame(daughter(1, p))',
    'useRestFrame(daughter(2, p))',
    'daughter(1, PDG)',
    'daughter(1, E)',
    'daughter(1, px)',
    'daughter(1, py)',
    'daughter(1, pz)',
    'daughter(1, matchedMC(E))',
    'daughter(1, matchedMC(px))',
    'daughter(1, matchedMC(py))',
    'daughter(1, matchedMC(pz))',
    'daughter(1, mdstIndex)',
    'daughter(2, PDG)',
    'daughter(2, E)',
    'daughter(2, px)',
    'daughter(2, py)',
    'daughter(2, pz)',
    'daughter(2, matchedMC(E))',
    'daughter(2, matchedMC(px))',
    'daughter(2, matchedMC(py))',
    'daughter(2, matchedMC(pz))',
    'daughter(1, extraInfo(fsrCorrected))',
    'daughter(1, extraInfo(hasFSRCorrection))',
    'wasFSRCorrected'
]

b_gen_uncorr_var_list = [
        'daughter(1, extraInfo(fsrCorrectedE))',
        'daughter(1, extraInfo(fsrCorrectedPx))',
        'daughter(1, extraInfo(fsrCorrectedPy))',
        'daughter(1, extraInfo(fsrCorrectedPz))',
        'daughter(1, extraInfo(fsrCorrectedP))',
        'daughter(1, extraInfo(fsrGammaE))',
        'daughter(1, extraInfo(fsrGammaPx))',
        'daughter(1, extraInfo(fsrGammaPy))',
        'daughter(1, extraInfo(fsrGammaPz))',
        'daughter(1, extraInfo(fsrGammaP))',
        'daughter(1, extraInfo(fsrCorrected))'
]

y4s_var_list = [
    "expNum",
    "runNum",
    "evtNum",
    'PDG',
    'extraInfo(decayModeID)',
    'Mbc',
    'M',
    'InvM',
    'matchedMC(E)',
    'matchedMC(px)',
    'matchedMC(py)',
    'matchedMC(pz)',
    'matchedMC(M)',
    'matchedMC(mdstIndex)',
    'matchedMC(InvM)',
    'E',
    'px', 'py', 'pz',
    'deltaE',
    'chiProb',
    'charge',
    'decayAngle(0)',
    'decayAngle(1)',
    'cosThetaBetweenParticleAndTrueB',
    'daughterInvM(0, 2)',
    'formula(daughterInvM(0, 2) - daughter(0, InvM))',
    'daughter(2, p)',
    'daughter(2, theta)',
    'useCMSFrame(daughter(0, p))',
    'useCMSFrame(daughter(1, p))',
    'useCMSFrame(daughter(2, p))',
    'useCMSFrame(daughterAngle(0, 1))',
    'useCMSFrame(daughterAngleInBetween(0, 1))',
    'useCMSFrame(daughterDiffOfPhi(0, 1))',
    'useCMSFrame(daughterAngle(0, 2))',
    'useCMSFrame(daughterAngleInBetween(0, 2))',
    'useCMSFrame(daughterDiffOfPhi(0, 2))',
    'useRestFrame(daughter(0, p))',
    'useRestFrame(daughter(1, p))',
    'useRestFrame(daughter(2, p))',
    'daughter(1, PDG)',
    'daughter(1, E)',
    'daughter(1, px)',
    'daughter(1, py)',
    'daughter(1, pz)',
    'daughter(1, matchedMC(E))',
    'daughter(1, matchedMC(px))',
    'daughter(1, matchedMC(py))',
    'daughter(1, matchedMC(pz))',
    'daughter(2, PDG)',
    'daughter(2, E)',
    'daughter(2, px)',
    'daughter(2, py)',
    'daughter(2, pz)',
    'daughter(2, matchedMC(E))',
    'daughter(2, matchedMC(px))',
    'daughter(2, matchedMC(py))',
    'daughter(2, matchedMC(pz))',
    'missingMomentum',
    'missingMomentumInCMSFrame',
    'matchedMC(missingMomentum)',
    'matchedMC(missingMomentumInCMSFrame)',
    'wasFSRCorrected'
]

y4s_uncorr_var_list = [
    'daughter(1,daughter(1, extraInfo(fsrCorrected)))',
    'daughter(1,daughter(1, extraInfo(hasFSRCorrection)))'
]

mdst_variables = [
    'mdstIndex',    # B
    'daughter(0,mdstIndex)',    # D0*
    'daughter(1,mdstIndex)',    # e
    'daughter(2,mdstIndex)',    # nu
    'daughter(0,daughter(0,mdstIndex))',    # D0
    'daughter(0,daughter(1,mdstIndex))',    # pi0_1
    'daughter(0,daughter(1,daughter(0,mdstIndex)))', # gamma from pi0_1
    'daughter(0,daughter(1,daughter(1,mdstIndex)))', # gamma from pi0_1
    'daughter(0,daughter(2,mdstIndex))', # not existing
    'daughter(0,daughter(0,daughter(0,mdstIndex)))',    # K
    'daughter(0,daughter(0,daughter(1,mdstIndex)))',    # pi+
    'daughter(0,daughter(0,daughter(2,mdstIndex)))',    # pi0_2
    'daughter(0,daughter(0,daughter(2,daughter(0,mdstIndex))))',    # gamma from pi0_2
    'daughter(0,daughter(0,daughter(2,daughter(1,mdstIndex))))',    # gamma from pi0_2
]

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['B_minus_uncorrected_info_ntuple'],
    particleList='B-:uncorr',
    variables=b_gen_var_list+b_gen_uncorr_var_list
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['B_minus_corrected_info_ntuple'],
    particleList='B-:corr',
    variables=b_gen_var_list
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['Y4S_uncorrected_info_ntuple'],
    particleList='Upsilon(4S):uncorr',
    variables=y4s_var_list+y4s_uncorr_var_list
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['Y4S_corrected_info_ntuple'],
    particleList='Upsilon(4S):corr',
    variables=y4s_var_list
)

main_path.add_module(
    'VariablesToNtuple',
    fileName=output_dict['mdst_info_ntuple'],
    particleList='B-:corr',
    variables=mdst_variables
)



basf2.print_path(main_path)
basf2.process(main_path)
print(statistics)
