"""
Steering file for testing the final state radiation correction.

Felix Metzner 2018
"""

import basf2
import b2biiConversion

from modularAnalysis import *

from analysistools.processing import path_utilities

import analysistools.processing.final_state_particles_utils as fsp


def create_fsr_test_path(
        input_files,
        input_sequences,
        output_dict,
        config,
        method,
        max_phi_angle,
        max_theta_angle,
        max_energy,
        momentum_factor
):
    path_utilities.setup_environment_vars()
    path_utilities.setup_databases(config)
    path_utilities.fei_backward_compatibility()

    main_path = basf2.create_path()

    b2biiConversion.convertBelleMdstToBelleIIMdst(
        inputBelleMDSTFile=input_files,
        applyHadronBJSkim=True,
        useBelleDBServer=os.environ['BELLE_POSTGRES_SERVER'],
        path=main_path,
        entrySequences=input_sequences
    )

    setAnalysisConfigParams({'mcMatchingVersion': 'Belle'}, main_path)

    fsp_collection = fsp.FinalStateParticleCollection()

    photon = fsp.FinalStateParticle(
        label='gamma',
        particlelist='gamma:fsp',
        origin=['gamma:mdst', 'gamma:v0mdst'],
        cuts=[
            'goodBelleGamma == 1'
        ]
    )

    fsp_collection.add_fsps(photon=photon)

    electron = fsp.FinalStateParticle(
        label='e',
        particlelist='e-:fsp',
        origin='reco',
        cuts=[
            'dr < 2',
            'abs(dz) < 4',
            'eIDBelle > 0.1',
            'useCMSFrame(pt) > 0.1'
        ],
        electron_fsr_correction=fsp.FSRCorrectionSettings(
            fsr_gamma_list_name="gamma:fsp",
            angle_threshold=max_phi_angle,
            theta_threshold=max_theta_angle,
            energy_threshold=max_energy,
            momentum_factor=momentum_factor,
            method=method
        )
    )

    fsp_collection.add_fsps(electron=electron)

    fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config['use_mc_info'])

    ############
    # e and mu #
    ############

    lepton_vars = [
        'PDG',
        'matchedMC(PDG)',
        'mcMother(PDG)',
        'mcMother(mcMother(PDG))',
        'mcMother(mcDaughter(0,PDG))',
        'mcDaughter(0,PDG)',
        'isSignal',
        'M', 'dM', 'InvM',
        'daughterInvariantMass(0, 1)',
        'E',
        'Q', 'dQ',
        'useCMSFrame(p)',
        'useCMSFrame(pt)',
        'p', 'pt', 'px', 'py', 'pz',
        'eIDBelle',
        'muIDBelle',
        'muIDBelleQuality',
        'atcPIDBelle(2,3)',
        'extraInfo(ksnbStandard)',
        'extraInfo(goodKs)',
        'dr', 'dz',
        'pValue',
        'nCDCHits',
        'nSVDHits',
        'daughter(0, pValue)',
        'daughter(0, nCDCHits)',
        'daughter(0, nSVDHits)',
        'daughter(0, dr)',
        'daughter(0, dz)',
        'matchedMC(E)',
        'matchedMC(p)',
        'matchedMC(px)',
        'matchedMC(py)',
        'matchedMC(pz)',
        'extraInfo(fsrCorrectedE)',
        'extraInfo(fsrCorrectedPx)',
        'extraInfo(fsrCorrectedPy)',
        'extraInfo(fsrCorrectedPz)',
        'extraInfo(fsrCorrectedP)',
        'extraInfo(fsrGammaE)',
        'extraInfo(fsrGammaPx)',
        'extraInfo(fsrGammaPy)',
        'extraInfo(fsrGammaPz)',
        'extraInfo(fsrGammaP)',
        'extraInfo(fsrCorrected)'
    ]

    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['e_ntuple'],
        particleList='e-:fsp',
        variables=lepton_vars
    )

    #########
    # Gamma #
    #########

    gamma_vars = [
        'PDG',
        'matchedMC(PDG)',
        'mcMother(PDG)',
        'mcMother(mcMother(PDG))',
        'mcMother(mcMother(mcMother(PDG)))',
        'mcMother(mcMother(mcMother(mcMother(PDG))))',
        'mcMother(mcDaughter(0,PDG))',
        'isSignal',
        'goodBelleGamma',
        'clusterE9E25',
        'clusterNHits',
        'clusterHighestE',
        'M', 'dM', 'InvM',
        'abs(SigMBF)',
        'E',
        'Q', 'dQ',
        'useCMSFrame(p)',
        'useCMSFrame(pt)',
        'p', 'pt', 'px', 'py', 'pz',
        'eIDBelle',
        'muIDBelle',
        'muIDBelleQuality',
        'atcPIDBelle(2,3)',
        'extraInfo(ksnbStandard)',
        'extraInfo(goodKs)',
        'dr', 'dz',
        'matchedMC(E)',
        'matchedMC(px)',
        'matchedMC(py)',
        'matchedMC(pz)',
        'matchedMC(p)',
        'extraInfo(fsrCorrected)'
    ]

    main_path.add_module(
        'VariablesToNtuple',
        fileName=output_dict['gamma_ntuple'],
        particleList='gamma:fsp',
        variables=gamma_vars
    )

    return main_path


if __name__ == "__main__":
    main_path = create_fsr_test_path()

    basf2.print_path(main_path)
    basf2.process(main_path)
    print(statistics)
